// Demonstration code of USB I/O on PIC 18F2455 (and siblings) -
// turn LED on/off and echo a buffer back to host.
//
// Copyright (C) 2005 Alexander Enzmann
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//

#include <pic18fregs.h>
#include <stdio.h>
#include <usart.h>
#include "evadefs.h"
#include "usb.h"
#include "programmer.h"

// Define configuration registers (fuses)
#if defined(pic18f2550) || defined(pic18f2455) || defined(pic18f4550) || defined(pic18f4455)
#if CLK_SPD == 48
code char at 0x300000 CONFIG1L = 0x24; // USB, /2 post (48MHz), /5 pre (20 MHz)
code char at 0x300001 CONFIG1H = 0x0e; // IESO=0, FCMEN=0, HS-PLL
#elif CLK_SPD == 24
code char at 0x300000 CONFIG1L = 0x34; // USB, /4 post (24MHz), /5 pre (20 MHz)
code char at 0x300001 CONFIG1H = 0x0e; // IESO=0, FCMEN=0, HS-PLL
#elif CLK_SPD == 20
code char at 0x300000 CONFIG1L = 0x24; // USB, (20MHz), /5 pre (20 MHz)
code char at 0x300001 CONFIG1H = 0x0c; // IESO=0, FCMEN=0, HS-PLL
#elif CLK_SPD == 16
code char at 0x300000 CONFIG1L = 0x3c; // USB, /6 post (24MHz), /5 pre (20 MHz)
code char at 0x300001 CONFIG1H = 0x0e; // IESO=0, FCMEN=0, HS-PLL
#elif CLK_SPD == 10
code char at 0x300000 CONFIG1L = 0x2c; // USB, /2 post (10 MHz), /5 pre (20 MHz)
code char at 0x300001 CONFIG1H = 0x0c; // IESO=0, FCMEN=0, HS
#elif CLK_SPD == 5
code char at 0x300000 CONFIG1L = 0x3c; // USB, /4 post (5 MHz), /5 pre (20 MHz)
code char at 0x300001 CONFIG1H = 0x0c; // IESO=0, FCMEN=0, HS
#else
#error "Invalid clock value - set to 48, 24, 16, 10, 5"
#endif
code char at 0x300002 CONFIG2L = 0x20; // Brown out off, PWRT On
code char at 0x300003 CONFIG2H = 0x00; // WDT off
code char at 0x300004 CONFIG3L = 0xff; // Unused configuration bits
code char at 0x300005 CONFIG3H = 0x01; // No MCLR, PORTB digital, CCP2 - RC1
code char at 0x300006 CONFIG4L = 0x80; // ICD off, ext off, LVP off, stk ovr off
code char at 0x300007 CONFIG4H = 0xff; // Unused configuration bits
code char at 0x300008 CONFIG5L = 0xff; // No code read protection
code char at 0x300009 CONFIG5H = 0xff; // No data/boot read protection
code char at 0x30000A CONFIG6L = 0xff; // No code write protection
code char at 0x30000B CONFIG6H = 0xff; // No data/boot/table protection
code char at 0x30000C CONFIG7L = 0xff; // No table read protection
code char at 0x30000D CONFIG7H = 0xff; // No boot table protection
#endif

// #pragma udata stack1 stack
#pragma stack 0x200 0xff

static unsigned char getRepCtr;

// If we get a packet via the serial port, this flag will
// be set.  It must be cleared before further packet
// processing can occur.
volatile unsigned char serialPacketFlag;
volatile unsigned char USBPacketFlag;
volatile unsigned char packetOutReadyFlag;

// HID feature buffer
volatile byte HIDBufferReady; // Set to 1 when an entire buffer has arrived

void high_isr(void) shadowregs interrupt 1
{
	byte oldINTCON = INTCON;
	// This is unexpected - something has gone wrong
	// Disable interrupts and report what happened
	INTCONbits.GIE = 0;

#if 0
	// Show state of interrupts
	printf("RCON: 0x, INTCON: 0x, INTCON2, INTCON3\r\n", RCON, INTCON, INTCON2, INTCON3);
	printf("PIR1, PIR2: 0x\r\n", PIR1, PIR2);
	printf("PIE1, PIE2: 0x\r\n", PIE1, PIE2);
	printf("IPR1, IPR2: 0x\r\n", IPR1, IPR2);

	_asm
	reset
	_endasm;
#endif

	INTCON = oldINTCON;
}

byte rxChar;

void low_isr(void) shadowregs interrupt 2
{
    // If the timer associated with the servo timed out, then
    // ask the servo code to handle it.
	if (INTCONbits.TMR0IE && INTCONbits.TMR0IF)
	{
        TMR0L = 0x00;           // Reset Timer0 count
		INTCONbits.TMR0IF = 0;
        VBoostISR();
    }
	else if (PIR1bits.RCIF)
	{
		do {
			// Got a byte through the serial port
			rxChar = RCREG;

			// Debug: echo it
			// TXREG = rxChar;

			if (serialPacketFlag)
			{
				// Already have a complete packet - if we take this
				// character then we will overwrite it before the
				// packet has been processed.
			}
			else
			{
#if 0
				serialPacketFlag = addByteToPacket(rxChar);
#else
				// Low clock speeds (< 10MHz) can't seem to keep up
				// with 115 KBaud serial input - try recoding in asm
				// to improve speed.
				serialPacketFlag = asmAddByteToPacket(rxChar);
#endif
			}

			// Clear RX flag
			PIR1bits.RCIF = 0;
		} while (PIR1bits.RCIF);
	}
    else if (PIE1bits.TMR2IE && PIR1bits.TMR2IF)
    {
        // Timer 2 interrupt.
		PIR1bits.TMR2IF  = 0; // Ensure overflow flag is clear
    }
	else if (PIR2bits.USBIF)
	{
		// TBD: Determine which USB interrupt/error got thrown
		// LED1 = 1;
#if DEBUG_PRINT
		if (debugPrintFlag)
			printf("USBIF: %x, %x\r\n", UIR, UEIR);
#endif
		UIR = 0x00;  // Clear USB interrupt status register
		UEIR = 0x00; // Clear USB error interrupt status register

		// Clear interrupt flag
		PIR2bits.USBIF = 0;
	}
	else
	{
		// Clear any other interrupts?
		// LED1 = 1;
	}
}

// Allocate buffers in RAM for storage of bytes that have either just
// come in from the SIE or are waiting to go out to the SIE.
char txBuffer[HID_INPUT_REPORT_BYTES];
char rxBuffer[HID_OUTPUT_REPORT_BYTES];

// If we got some bytes from the host, then echo them back.
static void USBEcho(void)
{
    byte rxCnt, i;

    // Find out if an Output report has been received from the host.
    rxCnt = HIDRxReport(rxBuffer, HID_OUTPUT_REPORT_BYTES);

	// If no bytes in, then nothing to do
	if (rxCnt == 0)
		return;

    // Copy input bytes to the output buffer
	for (i=0;i<HID_OUTPUT_REPORT_BYTES;i++)
        txBuffer[i] = rxBuffer[i];            

    // As long as the SIE is owned by the processor, we let USB tasks continue.
	while (ep1Bi.Stat & UOWN)
        ProcessUSBTransactions(); 

    // The report will be sent in the next interrupt IN transfer.
    HIDTxReport(txBuffer, HID_INPUT_REPORT_BYTES);
}

byte mFlashResult;
// Initialization for a SET_FEATURE request.  This routine will be
// invoked during the setup stage and is used to set up the buffer
// for receiving data from the host
int SetupFeatureReport(byte reportID)
{
	if (reportID == 0)
	{
		inPtr = &PACKET;
		return 1;
	}
    else
        return 0;
}

byte val0, val1, val2;

byte
ProcessFeatureReport()
{
	// Execute firmware update routines
	byte cmd, cmdOpt, DLen, UAddr, HAddr, LAddr;
	byte *buf = &PACKET;
	byte *dataBuf = &DATA_BUF;

	serialPacketFlag = 0; // Clear flag
	USBPacketFlag = 0;

	cmd    = buf[0];
	DLen   = buf[1];
	LAddr  = buf[2];
	HAddr  = buf[3];
	UAddr  = buf[4];
	cmdOpt = buf[5];
#if DEBUG_PRINT
	//if (debugPrintFlag)
	// printf("Cmd: %d, buf: %d %d %d %d\r\n", cmd, buf[0], buf[1], buf[2], buf[3]);
    // printf("Cmd: %d, opt: %d\r\n", cmd, cmdOpt);
#endif
	if (cmd == 0)
	{
		return HID_FEATURE_REPORT_BYTES;
	}
	else if (cmd == ECHO_CMD)
	{
		byte sval = buf[2];

		LED1 = (sval & 0x01) ? 1 : 0;
		LED2 = (sval & 0x02) ? 1 : 0;
		LED3 = (sval & 0x04) ? 1 : 0;

		return DLen;
	}
	else if (cmd == IREAD_CMD)
	{
		// Read a chunk of internal memory
		readFlash(UAddr, HAddr, LAddr, dataBuf, cmdOpt);

		return HID_FEATURE_REPORT_BYTES;
	}
	else if (cmd == IERASE_CMD)
	{
		// Erase the portion of internal memory that is used for updatable firmware
		eraseFlash(UAddr, HAddr, LAddr, cmdOpt);

		return HID_FEATURE_REPORT_BYTES;
	}
	else if (cmd == IWRITE_CMD)
	{
		// Write new firmware to internal memory
		// Packet format: <cmd><DLEN><ADDRL><ADDRH><ADDRU><CMDOPT><... data bytes ...>
		updateFlash(UAddr, HAddr, LAddr, dataBuf, cmdOpt);

		return HID_FEATURE_REPORT_BYTES;
	}
	else if (cmd == IRUN_CMD)
	{
		// Run a routine in internal memory
		// Packet format: <cmd><DLEN><ADDRL><ADDRH><ADDRU><CMDOPT>
		executeFlash(UAddr, HAddr, LAddr, cmdOpt);
		return HID_FEATURE_REPORT_BYTES;
	}
	else if (cmd == IDEBUG_PRINT_CMD)
	{
		// The flag is stored in the fourth byte of the packet
		// <cmd><DLEN><N/A><Flag><N/A><CMDOPT>
		debugPrintFlag = HAddr;
		return HID_FEATURE_REPORT_BYTES;
	}
	else
	{
		// This is a command for the programmer to execute - pass it
		// along.
#if DEBUG_PRINT
	//if (debugPrintFlag)
    // printf("Ex: %x\r\n", cmd);
#endif
#if 1
		_asm
		movff _INTCON, POSTDEC1
		bcf _INTCON, 7
		push
		movlw LOW(returnFromFlash)
		movwf _TOSL
		movlw HIGH(returnFromFlash)
		movwf _TOSH
		movlw UPPER(returnFromFlash)
		movwf _TOSU
		movf _PREINC1, W
		movwf _INTCON
		; Call routine at address to 0x004000
		clrf _PCLATU
		movlw 0x40
		movwf _PCLATH
		clrf _PCL
	returnFromFlash:
		movwf _mFlashResult
		_endasm;
#else
	FlashResult = ProcessCmdBuffer()();
#endif
#if DEBUG_PRINT
	// if (debugPrintFlag)
    // printf("Res: %x\r\n", mFlashResult);
#endif

		return mFlashResult;
	}
}

// Post processing for a SET_FEATURE request.  After all the data has
// been delivered from host to device, this will be invoked to perform
// application specific processing.
int SetFeatureReport(byte reportID)
{
#if 0 && DEBUG_PRINT
	if (debugPrintFlag)
	// printf("SetFeatureReport(%d): %d\r\n", (int)reportID, wCount);
	printf("SetF(%d): %d\r\n", (int)reportID, wCount);
#endif

	if (reportID == 0)
	{
		// SET_FEATURE relies on control transfers, so this routine will
		// be called enough times to fill the entire feature buffer.
		// We only want to process the buffer after it is full.
		if (wCount == HID_FEATURE_REPORT_BYTES)
		{
			byte result;
#if 0
			// Firmware locks up if we process the packet after leaving the
			// ProcessControlTransfer() function in usb.c.  TBD: Figure out
			// how to get this code to work.
			USBPacketFlag = 1;
			packetOutReadyFlag = 1;
			return 1;
#else
			result = ProcessFeatureReport();
			packetOutReadyFlag = 1;
#endif
			return result;
		}
		else
		{
			// Haven't read the entire packet yet.  Keep waiting...
			return 1;
		}
	}
    else
        return 0;
}

// Handle a feature report request on the control pipe
int GetFeatureReport(byte reportID)
{
#if 0 && DEBUG_PRINT
	if (debugPrintFlag)
    // printf("GetFeatureReport: %x\r\n", reportID);
    printf("GetF: %x\r\n", reportID);
#endif

	if (reportID == 0)
	{
		// Make sure there is a packet ready to send
		if (!packetOutReadyFlag)
		{
#if 0
			byte *buf = &PACKET;
			byte cmd = buf[0];

			outPtr = buf;

			packetOutReadyFlag = 0;
			return HID_FEATURE_REPORT_BYTES;
#else
        outPtr = &HIDTxBuffer;

        // The number of bytes in the report (from usb.h).
        return HID_INPUT_REPORT_BYTES;
#endif
		}
		else
		{
			byte *buf = &PACKET;
			byte cmd = buf[0];

			outPtr = buf;

			packetOutReadyFlag = 0;

#if 1
			wCount = HID_FEATURE_REPORT_BYTES;
			return 1;
#else
			if (cmd == ECHO_CMD)
			{
				wCount = setEchoBytes();

				return 1;
			}
			else
			{
				wCount = HID_FEATURE_REPORT_BYTES;
				return 1;
			}
#endif
		}
	}
    else
        return 0;
}

// Handle control out.  This might be an alternate way of processing
// an output report, so all that's needed is to point the output
// pointer to the output buffer
// Initialization for a SET_REPORT request.  This routine will be
// invoked during the setup stage and is used to set up the buffer
// for receiving data from the host
int SetupOutputReport(byte reportID)
{
    if (reportID == 0)
    {
        // When the report arrives in the data stage, the data will be  
        // stored in HIDFeatureBuffer
        inPtr = (byte*)&HIDRxBuffer;

        return 1;
    }
    else
        return 0;
}

// Post processing for a SET_REPORT request.  After all the data has
// been delivered from host to device, this will be invoked to perform
// application specific processing.
int SetOutputReport(byte reportID)
{
#if DEBUG_PRINT
	//if (debugPrintFlag)
    //printf("SetOutputReport(0x%hx)\r\n", reportID);
#endif
    // Currently only handling report 0, ignore any others.
    if (reportID != 0)
        return 0;

    // TBD: do something.  Not currently implemented because the output
    // report is being handled by an interrupt endpoint.
    return 1;
}

// Handle a control input report
int GetInputReport(byte reportID)
{
#if DEBUG_PRINT
	if (debugPrintFlag)
    printf("GetInputReport: 0x%uhx\r\n", reportID);
#endif
    if (reportID == 0)
    {
        // Send back the contents of the HID report
        // TBD: provide useful information...
        outPtr = &HIDTxBuffer;

        // The number of bytes in the report (from usb.h).
        wCount = HID_INPUT_REPORT_BYTES;

        return 1;
    }
    else
        return 0;
}

// Central processing loop.  Whenever the firmware isn't busy servicing
// the USB, we will get control here to do other processing.
void ApplicationTasks(void)
{
    // Note: put tasks here that have to be performed when
    // the circuit is powered.

    // Check to see if we have external power.
    // selfPowered = SELF_POWERED;

    // Note: Put tasks here that only make sense if we are attached
    // to the host.
	if (RCSTAbits.OERR)
	{
		byte readBad = RCREG;
		RCSTAbits.CREN = 0;   // Disable USART to clear overrun
		RCSTAbits.OERR = 0;
		RCSTAbits.CREN = 1;   // Reenable USART
	}
	else if (RCSTAbits.FERR)
	{
		byte readBad = RCREG;
		RCSTAbits.CREN = 0;
		RCSTAbits.FERR = 0;
		RCSTAbits.CREN = 1;
	}
	else if (serialPacketFlag)
	{
		byte res;

		// Tell PC we don't want any characters yet
#if USE_RTS_CTS
		CTS = 1;
#endif

		// Process the command packet
#if 1
		res = ProcessFeatureReport();
		packetOutReadyFlag = 1;
#else
		if (COMMAND == ECHO_CMD)
			res = setEchoBytes();
		else
			res = ProcessFeatureReport();
#endif

		// Done processing, now ok to get characters
#if USE_RTS_CTS
		CTS = 0;
#endif

		if (res)
		{
			// Echo back the packet
			serialResponse(res+2); // res = length of packet w/o command and length bytes
		}

		serialPacketFlag = 0; // Clear flag
	}
	else if (USBPacketFlag)
	{
		ProcessFeatureReport();
	}

    // User Application USB tasks
    if ((deviceState < CONFIGURED) || (UCONbits.SUSPND == 1))
    {
        return;
    }

	// Process USB: Echo back any bytes that have come in.
    // USBEcho();
}

void InitUSB()
{
    UCON = 0;
    UIE = 0;

    // Initialize USB
    UCFG = 0x14; // Enable pullup resistors; full speed mode
	UCONbits.USBEN = 0; // Disable USB module till we are ready

    deviceState = DETACHED;
    remoteWakeup = 0x00;
    selfPowered = 0x00;
    currentConfiguration = 0x00;

	debugPrintFlag = 1;
}

// Entry point of the firmware
void main(void)
{
    // Disable interrupts
	INTCON  = 0x00;
	INTCON2 = 0x00;
	INTCON3 = 0x00;

	// Start with all interrupts at low priority
	IPR1 = 0x00;
	IPR2 = 0x00;

	// Clear interrupt flags
	PIR1 = 0x00;
	PIR2 = 0x00;

	// Start with all peripheral interrupts disabled
	PIE1 = 0x00;
	PIE2 = 0x00;

	// Check to see if this is a Power On Reset (POR) or if something
	// bad happened.  If STKPTR > 1 then what likely happened is that
	// we executed code all the way through address space (can happen
	// if programming functions called before the code is loaded into
	// the upper addresses (empty memory will be all NOP).
	if (!RCONbits.POR)
	{
		// This is a POR - set the flag so we can tell if it isn't a
		// POR if we start executing this code again.
		RCONbits.POR = 1;
		RCONbits.BOR = 1;
	}
	else if (STKPTR > 1)
	{
		// Bad.  Need to fully reset the chip.  If we kept on going, then
		// at some point the stack would overflow.
		_asm
			reset
		_endasm; 
	}
	else
	{
		// TBD: handle other reset possibilities.
	}

    // Set all I/O pins to digital
    ADCON1 |= 0x0F;

	InitUSB();

	serialPacketFlag = 0; // No report yet.
	packetOutReadyFlag = 0;

    getRepCtr = 0; // Debug - track # of calls to SET_FEATURE...

    // Call user initialization function
    UserInit();

    while(1)
    {
		if (USB_PWR)
		{
			// Ensure USB module is available
			EnableUSBModule();

			ProcessUSBTransactions();

			if (UCONbits.PKTDIS)
			{
				// For some reason packet processing is disabled...
				// TBD: Figure out why this is happening.
				UCONbits.PKTDIS = 0;
				// InitUSB();
			}
		}

        // Application specific tasks
        ApplicationTasks();
    }
}
