// Firmware for USB based PIC programmer.  This file provides
// the routines that connect the USB framework to the upgradable
// routines for actually performing the programming operations.
//
// Copyright (C) 2005 Alexander Enzmann
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//
#include "programmer.h"
#include <stdio.h>
#include <usart.h>

#define ALL_C_CODE 0
#define DISABLE_PWM 0 // Debug: Check electrical characteristics without PWM active

byte programmerBusy; // Is the programmer actively doing something?

// Buffer to hold characters as they come in from the USART
byte *serialPacketPtr;

// Registers for I/O and programming variables (bank 0)
byte at 0x22 PCADDR0;     // Low byte of current ICSP address
byte at 0x23 PCADDR1;     // High byte of current ICSP address
byte at 0x24 PCADDR2;     // Upper byte of current ICSP address
byte at 0x25 ICTR;        // Temp counter used for ICSP and Serial I/O
byte at 0x26 TEMP;        // Temp value
byte at 0x27 DLYTEMP;     // Temp counter used in delay routine
byte at 0x28 PWMCtr;    // Number of passes to wait before checking PWM
byte at 0x29 SFlags;      // Escape flag during RS232 input
byte at 0x2a trxChar;     // Last value read from RS232
byte at 0x2b packetCtr;   // Temp counter used for packet IO
byte at 0x2c STXCount;    // # of start of packet tokens in a row
byte at 0x2d TMPSTATUS;   // Save STATUS during interrupt
byte at 0x2e TMPW;        // Save W during interrupt
byte at 0x2f TMPPCLATH;   // Save PCLATH during interrupt
byte at 0x30 CHECKSUM;    // Checksum for command packet
byte at 0x31 CMDTEMP;     // Temp ICSP command value (for block programming)
byte at 0x32 PROGTEMP;    // Temp ICSP begin programming command value
byte at 0x33 ENDPROGTEMP; // Temp ICSP end programming command value
byte at 0x34 RETRYCTR;    // # of times write attempted at one location
byte at 0x35 PICDATA0;    // Low byte returned from target PIC
byte at 0x36 PICDATA1;    // High byte returned from target PIC
byte at 0x37 PICDATA2;    // Upper byte returned from target PIC
byte at 0x38 analogVal;
byte at 0x39 PWMVal;
// Storage for command packet:
//    <cmd><data len><addrL><addrH><addrU><cmdOPT><...data...>
//    <cmd><data len><data0><data1><data2>
//    <cmd><data len><ICSP cmd><data high byte><data low byte>
//
byte at 0x3a PACKET;      // Start of packet
byte at 0x3a COMMAND;     // Command value is first entry in the packet
byte at 0x3b DATALEN;     // Length of data to follow
byte at 0x3c DATABUF0;    // Byte 0 of generic data buffer
byte at 0x3c ADDRL;       // Low byte of target address
byte at 0x3c ICSP_CMD;    // command value to send to PIC
byte at 0x3d DATABUF1;    // Byte 1 of generic data buffer
byte at 0x3d ADDRH;       // High byte of target address
byte at 0x3d DATA0;       // Low byte of data to send to PIC
byte at 0x3e DATABUF2;    // Byte 2 of generic data buffer
byte at 0x3e ADDRU;       // Upper byte of target address
byte at 0x3e DATA1;       // High byte of data to send to PIC
byte at 0x3f CMDOPT;      // Optional value (e.g., block length)
byte at 0x40 DATA_BUF[DATA_BUF_SIZE];// Up to 64 bytes of data to send/receive

//
// Address 0x40..0x7f are used for data payload.
//

// Info for voltage boost circuit
// Capacitor charge equations:
//     Q = C * V * (1 - e ^ (-t / (R * C)))
//     I = (V / R) * e ^ (-t / (R * C))
//     Qmax = C * V.
// Capacitor discharge equations:
//     V = Q / C
//     V = V0 * e ^ (-t / (R * C))
//     I = (V0 / R) * e ^ (-t / (R * C))
// Resistor divider equations (Vt -- R1 -- R2 -- Gnd):
//     Vt     - Target boosted voltage
//     V0     - Voltage measured between R1 and R2.
//     R1, R2 - Values of resistors in voltage divider
//     R2 = (V0 * R1) / (Vt - V0)
//     V0 = (Vt * R2) / (R1 + R2)
//     Vt = V0 * (R1 + R2) / R2
// For a target voltage of 20V and resistors of 47K and 10K,
// the measured voltage will be Vt = 4.7 * (47 + 10) / 10,
// or V = 3.51V.  The current draw for R1=47K, R2=10K would
// be 20V / 57K = 0.35 mA.
//
// Ripple in the output should be low (target of 1 mA), so selection
// of the inductor (based on Coiltronics app note) should be:
//     L = V1 * dT / dI
// Where V1 = 4.7 (on), (20 - 4.7) = 15.3 (off); D = 0.25
// For various values of T (1/f), max ripple current 10 mA:
//     f (KHz)    T (uSec)    L (uH)
//     47        21.3        2,662
//     94        10.6        1,325
//    188         5.3          663
//    375         2.7          331
//
// Peak current is Ip = V / Lf (Average current Ia = D * Ip).
// For this table, V = 4.7 volts, D = 0.75
//     f (KHz)    L (uH)      Ip (mA)   Ia (mA)
//     47        33         3030      2273
//     94        33         1515      1136
//    188        33          758       568
//    375        33          379       284
//     47       100         1010       758
//     94       100          505       379
//    188       100          253       189
//    375       100          126        95
//     47      1000          101        76
//     94      1000           50        38
//    188      1000           25        19
//    375      1000           13        10
//
// Condition for continuous conduction:
//     Vs / (1 - D^2)R > (D * T * Vs) / (2 * L), or
//     f > R * D * (1 - D)^2 / 2 * L
// or
//     L = (Vout - Vin + Vd)(1 - D) / f
//
void VBoostISR(void)
{
    unsigned char boostVal;

    PIE1bits.TMR2IE  = 0; // Disable Timer 2 interrupts until done

    // Read the analog value
    boostVal = getVoltage();

    // A/D version.  Compare the value we read with the desired
    // value and adjust the duty cycle accordingly.
    if (boostVal < VBOOST_LOW_HOLD)
    {
        // Only increase duty cycle if it is less than max
        if (PWMVal < MAX_DUTY_CYCLE)
        {
            if (boostVal < VBOOST_LOW)
            {
                PWMVal += 2;
            }
            else
            {
                PWMVal += 1;
            }
        }
    }
    else if (boostVal > VBOOST_HIGH_HOLD)
    {
        // Only reduce the duty cycle if it is greater than 0
        if (PWMVal > 0)
        {
            if (boostVal > VBOOST_HIGH)
			{
                PWMVal -= 4;
			}
            else
			{
                PWMVal -= 1;
			}
        }
    }

    // Set duty cycle register, shifting based on value of PR2 -
    // higher frequencies (lower PR2) require smaller timeout
    // values for the same duty cycle.
#if (PR2_VAL == 0xff)
        // Use all 8 bits of PWMVal in CCPR1L
        CCPR1L = PWMVal;
        CCP1CONbits.DC1B1 = 0;
        CCP1CONbits.DC1B0 = 0;
#elif (PR2_VAL == 0x7f)
        // Use 7 bits of PWMVal in CCPR1L, use bottom bit of PWMVal
        // as high bit of duty cycle part of CCP1CON
        CCPR1L = PWMVal >> 1;
        CCP1CONbits.DC1B1 = (PWMVal & 0x01) ? 1 : 0;
        CCP1CONbits.DC1B0 = 0;
#elif (PR2_VAL == 0x3F)
        CCPR1L = PWMVal >> 2;
        CCP1CONbits.DC1B1 = (PWMVal & 0x02) ? 1 : 0;
        CCP1CONbits.DC1B0 = (PWMVal & 0x01) ? 1 : 0;
#elif (PR2_VAL == 0x1F)
        CCPR1L = PWMVal >> 3;
        CCP1CONbits.DC1B1 = (PWMVal & 0x04) ? 1 : 0;
        CCP1CONbits.DC1B0 = (PWMVal & 0x02) ? 1 : 0;
#elif (PR2_VAL == 0x0F)
        CCPR1L = PWMVal >> 4;
        CCP1CONbits.DC1B1 = (PWMVal & 0x08) ? 1 : 0;
        CCP1CONbits.DC1B0 = (PWMVal & 0x04) ? 1 : 0;
#elif (PR2_VAL == 0x07)
        CCPR1L = PWMVal >> 5;
        CCP1CONbits.DC1B1 = (PWMVal & 0x10) ? 1 : 0;
        CCP1CONbits.DC1B0 = (PWMVal & 0x08) ? 1 : 0;
#else
#error Bad value for PR2 - set to one of: 0xff, 0x7f, 0x3f, 0x1f, 0x0f.
	// Unknown frequency - don't do any shifting.
    CCPR1L = PWMVal;
#endif
}

// Note: rxChar is passed in W so it doesn't appear
// to be used when compiling this function.
byte asmAddByteToPacket(byte rxChar) wparam _naked
{
	// Low clock speeds (< 10MHz) can't seem to keep up
	// with 115 KBaud serial input - try recoding in asm
	// to improve speed.
_asm
    banksel _trxChar
    movwf _trxChar    // Move rxChar to trxChar

	// See if this is the start of a packet.  We need to
	// see two STX bytes in a row.
	movlw 0x02
	cpfslt _STXCount
	bra checkDLEFlag      // Already have 2 STX's
	// Compare what we just got to STX
	movlw STX
	cpfseq _trxChar
	bra resetPacket       // Didn't just get an STX
	incf _STXCount, F
	retlw 0x00

checkDLEFlag:
	btfss _SFlags, 0
	bra checkDLE
	// Already got a DLE.  Clear the flag
	// and add this byte to the packet.
	clrf _SFlags
	bra addByteToPacket   // Process escaped byte

checkDLE:
	movlw DLE
	cpfseq _trxChar
	bra checkETX          // Not a DLE
	bsf _SFlags, 0
	retlw 0x00

checkETX:
	movlw ETX
	cpfseq _trxChar
	bra checkSTX          // Not an ETX

	// Packet is complete - can process it now.  Make sure all counters
	// and flags are clear for the next packet.
	clrf _packetCtr
	clrf _STXCount
	clrf _SFlags
	retlw 0x01

checkSTX:
	movlw STX
	cpfseq _trxChar
	bra addByteToPacket
	bra resetPacket        // Unexpected STX in middle of packet

// This is a byte we need to 
addByteToPacket:
	// Check length of packet
	movlw 0x50
	cpfslt _packetCtr
	bra resetPacket     // Counter exceeded packet storage length

	// Save indirection registers
	// movff _FSR0L, _SERIALTMP0
	// movff _FSR0H, _SERIALTMP1

	// Add data to packet and check for overflow
	clrf _FSR0H
	movlw _PACKET
	addwf _packetCtr, W  // Point to byte at PACKET[packetCtr]
	movwf _FSR0L
	movff _trxChar, _INDF0 // Put the byte in the packet

	// Restore indirection registers
	// movff _SERIALTMP0, _FSR0L
	// movff _SERIALTMP1, _FSR0H

	// Increment pointer into packet
	incf _packetCtr, F

	// Not done with packet yet
	retlw 0x00

// An error occured in the packet.  Clear all counters and
// start processing as if we hadn't seen any data to this point
resetPacket:
	clrf _packetCtr
	clrf _STXCount
	clrf _SFlags
	retlw 0x00
_endasm;
}

void resetSerialPacketRx()
{
	packetCtr = 0;
	serialPacketPtr = &PACKET;

	STXCount = 0; // No Start-of-Packet bytes yet
	SFlags = 0;
}

// Got a byte from the serial port.  Returns 1 if an entire packet has
// been assembled.  0 otherwise (even on error).
byte addByteToPacket(byte inData)
{
#if 0
	if (STXCount == 0)
	{
		if (inData == STX)
			STXCount++;
		else
			resetSerialPacketRx();
	}
	else if (STXCount == 1)
	{
		if (inData == STX)
			STXCount++;
		else
			resetSerialPacketRx();			
	}
#else
	if (STXCount < 2)
	{
		if (inData == STX)
			STXCount++;
		else
			resetSerialPacketRx();			
	}
#endif
	else if (SFlags & 0x01)
	{
		// Read an escaped character
        *serialPacketPtr++ = inData;
        packetCtr++;
		SFlags &= 0xfe; // Clear the DLE flag
	}
	else if (inData == DLE)
	{
		// Read the next byte, no matter what the value
		SFlags |= 0x01; // Set the DLE flag
	}
	else if (inData == ETX)
	{
		// End of packet - can process it now.
		resetSerialPacketRx();
		return 1;
	}
	else if (inData == STX)
	{
		resetSerialPacketRx();
	}
	else
	{
		// Add data to the buffer
        *serialPacketPtr++ = inData;
        packetCtr++;
	}

	if (packetCtr == 80)
	{
		// Buffer overflow.  Reset Rx packet
		resetSerialPacketRx();
	}

	return 0;
}


// SDCC wants a user-defined function for putchar(c) that it can reuse for
// all the printf() style functions.  The macro PUTCHAR is defined in stdio.h.
PUTCHAR(c)
{
    __stream_putchar(stdout, c);
}


byte RxRS232()
{
    byte rxByte;
 
    _asm clrwdt _endasm;

    // Check for Overrun Error
    if (RCSTA & 0x02)
    {
        // OERR - Overrun Error, should be able to clear by clearing CREN
        RCSTA &= 0xef;
        RCSTA |= 0x10;
    }


    // Wait until RCIF (bit 5) is clear in PIR
    while (!usart_drdy())
        ;

    // Valid data is now in RCREG
    rxByte = usart_getc();

    return rxByte;
}

byte TxRS232(unsigned char txData)
{
#if !USE_RTS_CTS
	putchar(txData);
#else
	int ctr;
    _asm clrwdt _endasm;

	for (ctr=0;ctr<1000;ctr++)
	{
		if (!RTS)
		{
		    putchar(txData);
			return 1;
		}
	}
#endif

	// Timeout when trying to send the character.
	return 0;
}

static void InitializeUSART()
{
    SERIAL_TX_MASK = 0; // Set TX pin as an output
    SERIAL_RX_MASK = 1; // Set RX pin as an input
    RCSTA   = 0x90; // Enable serial port, enable receiver

    // Baud rate formula for BRG16=1, BRGH=1: Baud Rate = Fosc/(4 (n + 1)),
    // or n = (Fosc / (4 * BaudRate)) - 1
    // At 48 MHz, for 115.2K Baud:
    //     SPBRGH:SPBRG = n = Fosc / (4 * 115200) - 1 = 103.17
    TXSTA   = 0x24; // Asynch, TSR empty, BRGH=1
    BAUDCON = 0x08; // BRG16=1
    SPBRGH  = 0x00; // At 16 MHz, SPBRGH=0, SPBRG=52 gives 115.2K Baud
#if CLK_SPD == 48
	// X = ((48000000 / 115200) / 4) - 1 = 103.16
	// Calculated Baud Rate: 48000000 / (4 * (X + 1)) = 116504
	// Error: (115200 - 116504) / 115200 = 1.1%
    SPBRG   = 103;  // For 48 MHz clock
#elif CLK_SPD == 24
	// X = ((24000000 / 115200) / 4) - 1 = 51.08
	// Calculated Baud Rate: 24000000 / (4 * (X + 1)) = 115384
	// Error: (115200 - 117647) / 115200 = 0.16%
    SPBRG   = 51;
#elif CLK_SPD == 20 // For 20 MHz clock
	// X = ((20000000 / 115200) / 4) - 1 = 43 (42.4)
	// Calculated Baud Rate: 20000000 / (4 * (X + 1)) = 116279
	// Error: (115200 - 116279) / 115200 = 0.9%
    SPBRG   = 43;
#elif CLK_SPD == 16 // For 16 MHz clock
	// X = ((16000000 / 115200) / 4) - 1 = 33.72
	// Calculated Baud Rate: 16000000 / (4 * (X + 1)) = 114285.7
	// Error: (115200 - 114285.7) / 115200 = 0.8%
    SPBRG   = 34;
#elif CLK_SPD == 10 // For 10 MHz clock
	// X = ((10000000 / 115200) / 4) - 1 = 21 (20.7)
	// Calculated Baud Rate: 10000000 / (4 * (X + 1)) = 113636
	// Error: (115200 - 114285.7) / 115200 = 1.4%
    SPBRG   = 21;
#elif CLK_SPD == 7 // For 6.67 MHz clock
	// X = ((10000000 / 115200) / 4) - 1 =  ()
	// Calculated Baud Rate: 10000000 / (4 * (X + 1)) = 113636
	// Error: (115200 - 114285.7) / 115200 = 1.4%
    SPBRG   = 14;
#elif CLK_SPD == 5 // For 5 MHz clock
	// X = ((5000000 / 115200) / 4) - 1 = 10 (9.85)
	// Calculated Baud Rate: 10000000 / (4 * (X + 1)) = 113636
	// Error: (115200 - 113636) / 115200 = 1.4%
    SPBRG   = 10;
#else
#error "Unsupported processor speed in serial I/O"
#endif

#if 1
	// Enable low-priority interrupts for Rx of bytes
	PIE1bits.RCIE = 1; // Enable Rx interrupts
#else
	// Disable interrupts for Rx
	PIE1bits.RCIE = 0;
#endif
	IPR1bits.RCIP = 0; // Low priority

	// Disable interrupts for Tx
	PIE1bits.TXIE = 0; // Disable Tx interrupts

    // Assign standard input and output to the USART
    stdin = STREAM_USART;
    stdout = STREAM_USART;

	// Initialize serial reception of command packets
	resetSerialPacketRx();

	// Allow PC to send characters
	CTS_MASK = 0; // CTS is an output
	RTS_MASK = 1; // RTS is an input
	CTS = 0;      // Ready to receive characters
	RTS = 0;

#if DEBUG_PRINT
	if (debugPrintFlag)
    printf("Eva Startup\r\n");
#endif
}

// Send the response packet via the serial port
void serialResponse(byte packetLen)
{
	byte ctr, check, *buf;

	// Start checksum at 0
	check = 0;

    TxRS232(STX);
    TxRS232(STX);

	buf = &PACKET;
	for (ctr=0;ctr<packetLen;ctr++)
	{
		byte b = *buf++;
		if (b == STX)
			TxRS232(DLE);
		else if (b == ETX)
			TxRS232(DLE);
		else if (b == DLE)
			TxRS232(DLE);
		TxRS232(b);
		check += b;
	}

	// End of packet
	if ((check == STX) || (check == ETX) || (check == DLE))
		TxRS232(DLE);
	TxRS232(check);
	TxRS232(ETX);
}

// Entry point for user initialization
void UserInit(void)
{
    programmerBusy = 0;

    // Assignment to ensure reserved isn't removed through optimization
    DATA_BUF[0] = 0;

    InitializeUSART();

    // Configure analog
    ADCON0 = 0x00; // A/D off, AN0 is selected channel
    ADCON1 = 0x0d; // Volt ref Vdd/Vss, AN0/AN1 are analog
    ADCON2 = 0x22; // Left justified, Fosc/32, 8 Tad
    ADCON0bits.ADON = 1; // enable the A/D module

    // Configure Timer0 - want a couple of hundred interrupts
    // per second to check the analog value.  With a prescale
    // value of 1:128, counting to 256, and using the internal
    // instruction cycle, Timer0 will overflow every 2.73 msec,
    // giving 366.2 interrupts per second (with 48 MHz clock).
    // T0CON = 0x47; // Timer off, 8-bit timer, use CLK0, 1:256 prescale, 
    T0CON = 0x46; // Timer off, 8-bit timer, use CLK0, 1:128 prescale, 

    // Set AN0/AN1 as input, rest of PORTA output
    VBOOST_MASK  = 1; // Analog measurement of boost converter output
    EXT_PWR_MASK = 1; // Analog measurement of external power voltage

    // Configure pin used to detect attachment to USB port
    // Check USB_PWR pin to see if the device is self-powered.
    // TBD: check self power every USB GET_STATUS
    selfPowered = 0;  // SELF_POWERED;
    USB_PWR_MASK = 1; // Digital input
    USB_PWR = 0;

    // Set LED1-LED3 as outputs
    LED1_MASK = 0; LED1 = 0;
    LED2_MASK = 0; LED2 = 0;
    LED3_MASK = 0; LED3 = 0;

	// Control pins for voltage (Vdd, two Vpp lines) are outputs.
	// Make sure they start off (Vdd/Vpp turned off).
    VDD_MASK = 0;
    VDD = 0;
    VPP1_MASK = 0;
    VPP1 = 0;
    VPP2_MASK = 0;
    VPP2 = 0;

	// Set control line for programming select line to output and
	// default value (off = 16F/18F series, on = dsPIC).
	PRG_SEL_MASK = 0;
	PRG_SEL = 0;

#if DISABLE_PWM
#else
    // Configure PWM
    //    Duty cycle = (PR2 + 1) * 4 * Tosc * (TMR2 Prescale)
    //    PWM Freq   = 1 / (Duty cycle)
    // Assuming the system clock is 48 MHz (Tosc = 1/48 uSec):
    // For PR2 = 0x7f, 4 * Tosc = 0.083 uSec;
    //    Duty cycle = 128 * 0.083 uSec = 10.67 uSec
    //    Frequency  = 93.750 KHz
    // For PR2 = 0xff, 4 * Tosc = 0.083 uSec;
    //    Duty cycle = 256 * 0.083 uSec = 21.33 uSec
    //    Frequency  = 46.875 KHz
    // Frequency versus PR2 value:
    //     0xff -  46.875 KHz
    //     0x7f -  93.75 KHz
    //     0x3f - 187.5 KHz
    //     0x1f - 375 KHz
    PR2 = PR2_VAL;

    PIR1bits.TMR2IF = 0; // Clear Timer 2 interrupt flag
    IPR1bits.TMR2IP = 0; // Low priority for Timer 2 interrupt
    // T2CON = 0x0c;       // 1:1 prescale, 1:2 postscale, T2 on
    // T2CON = 0x0d;       // 1:4 prescale, 1:2 postscale, T2 on
    // T2CON = 0x24;       // 1:1 prescale, 1:4 postscale, T2 on
    T2CON = 0x7c;          // 1:1 prescale, 1:16 postscale, T2 on

    // Enable PWM outputs after a new PWM cycle has started
    while (!PIR1bits.TMR2IF)
        ;

    // Configure PWM2 to drive voltage booster
    CCP1CON = 0x0c; // Single PWM output, LSBs set to 0, PWM mode

    PWMVal = DEFAULT_DUTY_CYCLE;
    CCPR1L = DEFAULT_DUTY_CYCLE;  // Test duty cycle
    EnablePWM1();
#endif

    // Turn off comparator voltage reference module
    CVRCON = 0x00;
    CMCON = 0x17; // C1 inverted, module off

#if !DISABLE_PWM
    // Enable interrupts.
    RCONbits.IPEN   = 1;   // Enable priority levels on interrupts
    INTCONbits.GIEL = 1;   // Enable low-priority interrupts
    INTCONbits.GIEH = 1;   // Enable high-priority interrupts

    // Enable interrupts on Timer 0
    INTCON2bits.TMR0IP = 0; // Low priority for Timer 0 interrupts
    INTCONbits.TMR0IE = 1;  // Enable Timer0 interrupts
    TMR0L = 0x00;           // Reset Timer0 count
    T0CONbits.TMR0ON = 1;   // Turn on Timer 0

    // Enable Timer 2 interrupts
    // PIE1bits.TMR2IE = 1;
#endif
}

// Read an analog value on currently selected A/D pin
byte getVoltage()
{
    // TBD: Wait acquisition time?

    // Start conversion
    ADCON0bits.GO = 1;

    // Wait until conversion complete
    while (ADCON0bits.GO)
        ;

    // Return the top 8 bits of the result
    analogVal = ADRESH;

    return ADRESH;
}

// Read len bytes from program memory and put them into buf.  Note: no
// range checking performed by this routine - make sure buf is big
// enough to hold the result.
byte readFlash(byte UAddr, byte HAddr, byte LAddr, byte *buf, byte len)
{
    byte ctr;

#if DEBUG_PRINT
	if (debugPrintFlag)
       printf("Read %d at: %x %x %x\r\n", len, UAddr, HAddr, LAddr);
#endif

    TBLPTRU = UAddr;
    TBLPTRH = HAddr;
    TBLPTRL = LAddr;

    for (ctr=len;ctr>0;--ctr)
    {
        _asm
            TBLRD*+
        _endasm;

        *buf++ = TABLAT;
   }

#if DEBUG_PRINT
	if (debugPrintFlag)
       printf("Read done\r\n", UAddr, HAddr, LAddr);
#endif

    return 1;
}

#define ERASE_BLOCK_SIZE 64
// Erase a section of program memory.  Note that the 18F2455 performs erases
// in 64 byte blocks - the lowest 6 bits of LAddr will be ignored.
byte eraseFlash(byte UAddr, byte HAddr, byte LAddr, byte len)
{
    int numBlocks;

    // Only allow erases above 16K boundary...
    if ((UAddr == 0) && (HAddr < 64))
        return 0;

    // Mask off bottom 6 bits of LAddr
    LAddr &= 0xC0;

    numBlocks = (len + ERASE_BLOCK_SIZE) / ERASE_BLOCK_SIZE;
#if DEBUG_PRINT
	//if (debugPrintFlag)
    // printf("Erase at: %x %x %x, blocks: %d\r\n", UAddr, HAddr, LAddr, numBlocks);
#endif
    for (;numBlocks>0;numBlocks--)
    {
        TBLPTRU = UAddr;
        TBLPTRH = HAddr;
        TBLPTRL = LAddr;
        // Erase the block
        EECON1bits.EEPGD = 1;    // Point to Flash program memory
        EECON1bits.CFGS  = 0;    // Access Flash program memory
        EECON1bits.WREN  = 1;    // Enable write to memory
        EECON1bits.FREE  = 1;    // Enable Row Erase operation
        INTCONbits.GIE   = 0;    // Disable Interrupts
        EECON2           = 0x55; // Step 1 of required sequence
        EECON2           = 0xAA; // Step 2 of required sequence
        EECON1bits.WR    = 1;    // Start erase (CPU stall)
        INTCONbits.GIE   = 1;    // Enable interrupts

		// Increment block address
        LAddr += 64;
        if (LAddr == 0)
        {
            // Rollover
            HAddr++;
            if (HAddr == 0)
                UAddr++;
        }
    }
    return 1;
}

#define WRITE_BLOCK_SIZE 32
// Write a buffer to program memory using self-write.  Note: it is possible
// to overwrite portions of base firmware, resulting in crash/unusable chip.
byte writeFlash(byte UAddr, byte HAddr, byte LAddr, byte *buf, byte len)
{
    byte ctr;
    int numBlocks;

    // Only allow writes above 16K boundary...
    if ((UAddr == 0) && (HAddr < 64))
        return 0;

    // Mask off bottom 6 bits of LAddr
    LAddr &= 0xC0;

    numBlocks = (len + WRITE_BLOCK_SIZE - 1) / WRITE_BLOCK_SIZE;
#if DEBUG_PRINT
	//if (debugPrintFlag)
    // printf("Write at: %x %x %x, blocks: %d\r\n", UAddr, HAddr, LAddr, numBlocks);
#endif
    for (;numBlocks>0;numBlocks--)
    {
        // Copy a chunk of buf into table write buffer.
        TBLPTRU = UAddr;
        TBLPTRH = HAddr;
        TBLPTRL = LAddr;
        for (ctr=WRITE_BLOCK_SIZE;ctr>0;ctr--)
        {
#if 0 && DEBUG_PRINT
	if (debugPrintFlag)
		printf("%x ", *buf);
#endif
            TABLAT = *buf++;
            _asm
                tblwt*+
            _endasm;
        }

#if 0
        // Move table pointer back into the 32 byte block before the write
        // TBD: This causes bad compile...
        _asm
            tblrd*-
        endasm;
#else
        TBLPTRU = UAddr;
        TBLPTRH = HAddr;
        TBLPTRL = LAddr;
#endif

        // Write the buffer
        EECON1bits.EEPGD = 1;    // Point to Flash program memory
        EECON1bits.CFGS  = 0;    // Access Flash program memory
        EECON1bits.WREN  = 1;    // Enable write to memory
        INTCONbits.GIE   = 0;    // Disable Interrupts
        EECON2           = 0x55; // Step 1 of required sequence
        EECON2           = 0xAA; // Step 2 of required sequence
        EECON1bits.WR    = 1;    // Start erase (CPU stall)
        INTCONbits.GIE   = 1;    // Enable interrupts
        EECON1bits.WREN  = 0;

        // Increment block address
        LAddr += WRITE_BLOCK_SIZE;
        if (LAddr == 0)
        {
            // Rollover
            HAddr++;
            if (HAddr == 0)
                UAddr++;
        }
    }
#if 0 && DEBUG_PRINT
	if (debugPrintFlag)
	{
    buf -= 6;
    printf("Last:");
    printf(" %x", *buf++);
    printf(" %x", *buf++);
    printf(" %x", *buf++);
    printf(" %x", *buf++);
    printf(" %x", *buf++);
    printf(" %x", *buf++);
    printf("\r\n");
	}
#endif

    return 1;
}

// Clear and write a buffer into program memory.  This routine assumes
// that the new buffer will fit into memory and there is nothing within
// a 64 byte block of the new code.  The size of a write buffer must
// be less than the size of an erase block (and also must be an even
// divisor of the erase block size).
byte updateFlash(byte UAddr, byte HAddr, byte LAddr, byte *buf, byte len)
{
    byte ctr, oldGIE;
    int numEraseBlocks, numWriteBlocks;


#if 0 && DEBUG_PRINT
	if (debugPrintFlag)
       printf("Update Flash at: %x %x %x\r\n", UAddr, HAddr, LAddr);
#endif

    numEraseBlocks = (len + ERASE_BLOCK_SIZE - 1) / ERASE_BLOCK_SIZE;
    numWriteBlocks = (ERASE_BLOCK_SIZE + WRITE_BLOCK_SIZE - 1) / WRITE_BLOCK_SIZE;

	oldGIE = INTCON;
    for (;numEraseBlocks>0;numEraseBlocks--)
    {
        // Erase the block of memory
        TBLPTRU = UAddr;
        TBLPTRH = HAddr;
        TBLPTRL = LAddr;

#if 1
		_asm
		bsf _EECON1, 7      // Point to Flash program memory
		bcf _EECON1, 6      // Access Flash program memory
		bsf _EECON1, 2      // Enable write to memory
		bsf _EECON1, 4      // Enable Row Erase operation
		bcf _INTCON, 7      // Disable Interrupts
		movlw 0x55          // Step 1 of required sequence
		movwf _EECON2
		movlw 0xaa          // Step 2 of required sequence
		movwf _EECON2
		bsf _EECON1, 1      // Start erase (CPU stall)
		_endasm;
#else
        // Erase the block
        EECON1bits.EEPGD = 1;      // Point to Flash program memory
        EECON1bits.CFGS  = 0;      // Access Flash program memory
        EECON1bits.WREN  = 1;      // Enable write to memory
        EECON1bits.FREE  = 1;      // Enable Row Erase operation
        INTCONbits.GIE   = 0;      // Disable Interrupts
		EECON2           = 0x55;   // Step 1 of required sequence
		EECON2           = 0xAA;   // Step 2 of required sequence
#if 0
		_asm
		bsf _EECON1, 1      // Start erase (CPU stall)
        _endasm;
#else
		EECON1bits.WR    = 1;      // Start erase (CPU stall)
#endif
#endif
        INTCON  = oldGIE;  // Enable interrupts (if they were enabled before)

        // Fill the block with code
        for (;numWriteBlocks>0;numWriteBlocks--)
        {
            // Copy a chunk of buf into table write buffer.
            TBLPTRU = UAddr;
            TBLPTRH = HAddr;
            TBLPTRL = LAddr;
            for (ctr=WRITE_BLOCK_SIZE;ctr>0;ctr--)
            {
                TABLAT = *buf++;
                _asm
                    tblwt*+
                _endasm;
            }

            // Write the buffer
#if 1
			_asm
			tblrd*-        // Back up address into the block we are writing
			movlw 0x55     // Step 1 of required sequence
			movwf _EECON2
			movlw 0xaa     // Step 2 of required sequence
			movwf _EECON2
			bsf _EECON1, 1 // Start erase (CPU stall)
			_endasm;
#else
			// Reset address to start of block to move (it was updated
			// by the tblwt*+ command).
			TBLPTRU = UAddr;
			TBLPTRH = HAddr;
			TBLPTRL = LAddr;

            EECON2           = 0x55;   // Step 1 of required sequence
            EECON2           = 0xAA;   // Step 2 of required sequence
#if 1
			_asm
			bsf _EECON1, 1      // Start erase (CPU stall)
            _endasm;
#else
			EECON1bits.WR    = 1;      // Start erase (CPU stall)
#endif
#endif
            INTCON = oldGIE; // Enable interrupts

            // Increment block address
            LAddr += WRITE_BLOCK_SIZE;
            if (LAddr == 0)
            {
                // Rollover
                HAddr++;
                if (HAddr == 0)
                    UAddr++;
            }
        }
#if 1
		_asm
		bsf _EECON1, 2        // Disable write to memory
		_endasm;
#else
        EECON1bits.WREN  = 0; // Disable write to memory
#endif
    }

#if 0 && DEBUG_PRINT
	if (debugPrintFlag)
       printf("Update Complete at: %x %x %x\r\n", UAddr, HAddr, LAddr);
#endif
    return 1;
}

volatile byte UFlashAddr, HFlashAddr, LFlashAddr, FlashResult;

// Placeholder for default entry point in high memory.  The arguments
// are the address of a RAM buffer and it's length.  The contents of
// the buffer must be one of the forms understood by the programming
// code:
//    <cmd><data len><addrL><addrH><addrU><cmdOPT><...data...>
//    <cmd><data len><data0><data1><data2>
//    <cmd><data len><ICSP cmd><data high byte><data low byte>
//
byte executeFlash(byte UAddr, byte HAddr, byte LAddr, byte arg)
{
    UFlashAddr = UAddr;
    HFlashAddr = HAddr;
    LFlashAddr = LAddr;

#if 0 && DEBUG_PRINT
 	if (debugPrintFlag)
   printf("Execute at: %x %x %x\r\n", UAddr, HAddr, LAddr);
#endif
#if DEBUG_PRINT
    // if (debugPrintFlag)
   // printf("Res:");
#endif
#if 1
    _asm
    movff _INTCON, POSTDEC1
    bcf _INTCON, 7              // Disable interrupts while building the return address
    push                        // Make a space on the stack for the return address
    movlw LOW(returnFromFlash)
    movwf _TOSL
    movlw HIGH(returnFromFlash)
    movwf _TOSH
    movlw UPPER(returnFromFlash)
    movwf _TOSU
    movf _PREINC1, W            // Get the old state of interrupts and restore it
    movwf _INTCON
    movff _UFlashAddr, _PCLATU // Start executing at the address
    movff _HFlashAddr, _PCLATH
    movf _LFlashAddr, W
    movwf _PCL
returnFromFlash:                // Called routine will return here
    movwf _FlashResult
    _endasm;
#endif

#if DEBUG_PRINT
	// if (debugPrintFlag)
    // printf(" %x\r\n", FlashResult);
#endif

    return FlashResult;
}

byte ProcessCmdBuffer()
{
    byte res;

#if DEBUG_PRINT
	// if (debugPrintFlag)
    // printf("Ex2: %x\r\n", COMMAND);
#endif
    if (COMMAND == 0)
        return 0;

    res = executeFlash(0x00, 0x40, 0x00, 0x00);

    return res;
}
