// Firmware framework for USB I/O on PIC 18F2455 (and siblings)
// Copyright (C) 2005 Alexander Enzmann
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
//
#ifndef PROGRAMMER_H
#define PROGRAMMER_H

#include "evadefs.h"
#include "usb.h"

extern byte dutyCycle1; // Duty cycle for CCP1

extern void VBoostISR(void);
extern byte getVoltage();

// Using a voltage divider with a 47K and 10K resistor, the measured voltage
// will be 10/(10+47) times the actual voltage.  For a target voltage of 20V,
// we want to measure 0.175 * 20 = 3.5 V at the A/D converter. The supply
// voltage ranges from 4.0 - 5.0 (TBD: ensure more stable voltage when
// under USB power alone).  The target value we want will be in the range
// Target = 256 * 3.5 / 5 to 256 * 3.5 / 4 = 179 to 224 [0xb3, 0xe0]
//                                Measured        Boost (max)  Output
#define VBOOST_LOW 0xb3        // 2.4 - 3.43 V     3.2         11 - 16 V
#define VBOOST_LOW_HOLD 0xc0   //
#define VBOOST_TARGET 0xc8     //
#define VBOOST_HIGH_HOLD 0xd0  //
#define VBOOST_HIGH 0xe0       // 2.8 - 4.1 V      5.3         19 - 27 V

// Enable RC2/CCP1 by setting pin as an output
#define EnablePWM1() TRISCbits.TRISC2 = 0
// Disable RC2/CCP1 by setting pin as an input
#define DisablePWM1() TRISCbits.TRISC2 = 1

// Set the duty cycle.  The value 0x20 will give about a 25% duty cycle.
// (The duty cycle value is scaled according to the value of PR2 so the
// firmware doesn't have to make different calculations if the clock
// speed changes.  The boundary conditions for the duty cycle should
// be adjusted based on the current capacity of the chosen inductor,
// the desired voltage, and the speed at which the voltage is generated.
// Calculations from datasheet:
//   Period = (PR2 + 1) * 4 * TOsc * TMR2 prescale
//   (Use 1:1 prescale to maximize PWM frequency)
//   TOsc: 48 MHz: 0.0208 uSec, 24 MHz: 0.0104, 16 MHz: 0.0625 uSec
//   Period: 48 MHz: , 24 MHz: , 16 MHz: 64 uSec
//   PWM Frequency (in KHz) versus PR2 value with 1:1 prescale:
//     PR2    PWM @ 48 MHz   PWM @ 24 MHz    PWM @ 16 MHz
//     0xff     46.875         23.4375         15.625
//     0x7f     93.75          46.875          31.25
//     0x3f    187.5           93.75           62.5
//     0x1f    375            187.5           125
//     0x0f    750            375             250
#if CLK_SPD == 48
// #define PR2_VAL 0x7f // 93.75 KHz
#define PR2_VAL 0x3f // 187.5 KHz
#elif (CLK_SPD == 24) || (CLK_SPD == 20)
// #define PR2_VAL 0x3f // 93.75 KHz
#define PR2_VAL 0x1f // 187.5 KHz (24MHz clock); 156.3 KHz (20MHz clock)
#elif CLK_SPD == 16
#define PR2_VAL 0x1f
#else
#define PR2_VAL 0x0f
#endif
#define DEFAULT_DUTY_CYCLE 0x20
// #define MAX_DUTY_CYCLE 0xcf
#define MAX_DUTY_CYCLE 0xa0

// Should we use hardware flow control?
#define USE_RTS_CTS 1

byte TxRS232(unsigned char txData); // Returns 1 if char sent, 0 if failed (timeout)
byte RxRS232();
byte addByteToPacket(byte inData);
byte asmAddByteToPacket(byte rxChar) wparam _naked;
void resetSerialPacketRx();

void UserInit(void);
extern unsigned word at 0x20 BLKCTRX;
extern unsigned word at 0x22 BLKCTRY;

extern byte at 0x38 analogVal;   // Last measured analog value
extern byte at 0x39 PWMVal;      // Duty cycle of PWM generator

// Every packet coming in will be placed in memory starting at
// PACKET.  The first six bytes are command information and
// small data parameters.  DATA_BUF holds any additional
// information (buffer to write, etc.) needed for the command.
extern byte at 0x3a PACKET;      // Start of packet
extern byte at 0x3a COMMAND;     // Command value is first entry in the packet
extern byte at 0x3b DATALEN;     // Length of data to follow
extern byte at 0x3c DATABUF0;    // Byte 0 of generic data buffer
extern byte at 0x3c ADDRL;       // Low byte of target address
extern byte at 0x3c ICSP_CMD;    // command value to send to PIC
extern byte at 0x3d DATABUF1;    // Byte 1 of generic data buffer
extern byte at 0x3d ADDRH;       // High byte of target address
extern byte at 0x3d DATA0;       // Low byte of data to send to PIC
extern byte at 0x3e DATABUF2;    // Byte 2 of generic data buffer
extern byte at 0x3e ADDRU;       // Upper byte of target address
extern byte at 0x3e DATA1;       // High byte of data to send to PIC
extern byte at 0x3f CMDOPT;      // Optional value (e.g., block length)
#define DATA_BUF_SIZE (HID_FEATURE_REPORT_BYTES - (0x3f - 0x3a))
#if (DATA_BUF_SIZE < 64)
#error "HID Feature Report size is too small to accomodate a 64 byte buffer."
#endif
extern byte at 0x40 DATA_BUF[DATA_BUF_SIZE]; // Up to 64 bytes of data to send/receive
// Note that DATA_BUF is larger than 64 in order to pad out to the
// full length of HID_FEATURE_REPORT_BYTES (which is 80 bytes).



// Incoming data format:
//
//    <STX><STX><DATA><CHKSUM><ETX>
//              |     |
//     _________|     |____________________________
//    |                                            |
//    <COMMAND><DLEN><ADDRL><ADDRH><ADDRU><DATA>...
//
//
// Definitions:
//
//     STX     -  Start of packet indicator
//     ETX     -  End of packet indicator
//     DATA    -  General data up to 64 bytes
//     CHKSUM  -  The 8-bit two's compliment sum of DLEN & DATA
//     COMMAND -  Base command
//     DLEN    -  Length of data buffer
//     ADDR    -  Address up to 24 bits
//     DATA    -  Data buffer (if any)
#define STX 0x55
#define ETX 0x04
#define DLE 0x05

// After an entire command buffer has been received, it is sent
// to the programmer to deal with it.  Note the flag programmerBusy
// below - this routine should not be called if the programmer
// is already in the middle of an operation.  A 1 will be
// returned on success, a 0 if it failed.
byte ProcessCmdBuffer();

// Fill a response packet that gives the result of processing
// the last command.
byte GetResponseBuffer(byte *buf, int len);

// Send a response packet via USART
void serialResponse(byte packetLen);

// The flag programmerBusy is set to 1 if programmer is in the middle
// of something and should not be interrupted.  This will typically be
// when it is actively reading/writing the target chip.
extern byte programmerBusy;

byte readFlash(byte UAddr, byte HAddr, byte LAddr, byte *buf, byte len);
byte eraseFlash(byte UAddr, byte HAddr, byte LAddr, byte len);
byte writeFlash(byte UAddr, byte HAddr, byte LAddr, byte *buf, byte len);
byte updateFlash(byte UAddr, byte HAddr, byte LAddr, byte *buf, byte len);

// Invoke a routine in ROM at a given location.  Pass in the
byte executeFlash(byte UAddr, byte HAddr, byte LAddr, byte arg);


#endif // PROGRAMMER_H
