#include "StdAfx.h"
#include "PIC/HexFileIO.h"
#include <assert.h>

const int HexFileIO::WORDS_PER_HEX_DUMP = 8; // dump this many words per hex record

const int HexFileIO::PROG_BUF_SIZE   = 65536; // max size of program memory
const int HexFileIO::DATA_BUF_SIZE   =  1024; // max size of EEDATA memory
const int HexFileIO::ID_BUFFER_SIZE  =     4; // max size of ID memory
const int HexFileIO::NR_CONFIG_WORDS =     7; // max count of configuration words

const int HexFileIO::FORMAT_UNKNOWN = -1;      // hex formats
const int HexFileIO::FORMAT_INHX8M  =  0;
const int HexFileIO::FORMAT_INHX16  =  1;
const int HexFileIO::FORMAT_INHX32  =  2;

const int HexFileIO::ERROR_EOF = -1;      // unexpected EOF
const int HexFileIO::ERROR_DEX = -2;      // hex digit required
const int HexFileIO::ERROR_CEX = -3;      // missing ':'
const int HexFileIO::ERROR_CHK = -4;      // checksum error
const int HexFileIO::ERROR_UNK = -5;      // Undefined error
const int HexFileIO::ERROR_IGN = 1;       // warning that some records were ignored

// FLAGS for each word in the program buffer
const int HexFileIO::BUFFER_FLAG_USED      =  1;  // LOADED or READ<>3FFF
const int HexFileIO::BUFFER_FLAG_PRG_ERROR =  2;  // PROGRAM or VERIFY error
const int HexFileIO::BUFFER_FLAG_VFY_ERROR =  4;
const int HexFileIO::BUFFER_FLAG_DEAD      =  8;  // Failed repeatedly (R.I.P.)
const int HexFileIO::BUFFER_FLAG_SPECIAL   = 16;  // special location, like osc calib word

HexFileIO::HexFileIO(void)
{
	readFailure = 0;
}

HexFileIO::~HexFileIO(void)
{
}

char HexFileIO::lineChar(int &line, int &column)
{
	// Can't get a character from an empty file
	if (lines.size() == 0)
		return (char)EOF;

	// Skip any empty lines
	while (line < (int)lines.size() && strlen(lines[line]) == 0)
		line++;
	// If everything was empty till the end of the file then return EOF
	if (line >= (int)lines.size())
		return (char)EOF;

	// Grab the next available character
	char *tline = lines[line];
	char res = lines[line][column++];

	// Move to next character
	if (column >= (int)strlen(lines[line]))
	{
		column = 0;
		line++;
	}

	return res;
}

byte HexFileIO::ReadDigitFromFile(int &line, int &column)
{
	char c;

	if (readFailure != 0)
		return 0;

	if ((c = lineChar(line, column)) == (char)EOF)
	{
		readFailure = ERROR_EOF;    // unexpected EOF
		return 0;
	}

	// Convert char to lower case
	c = _totlower(c);

	c -= (c > _T('9')) ? (char)(_T('a') - 10) : _T('0');
	if (c < 0 || c > 0xF )
		// hex digit expected
		readFailure = ERROR_DEX;

	return (byte)c;
}

byte HexFileIO::ReadHexByteFromFile(int &line, int &column,	int *checksum)
{
	byte b;

	b = ReadDigitFromFile(line, column);
	b = (byte)((b << 4) + ReadDigitFromFile(line, column));
	*checksum += b;

	return b;
}

ushort HexFileIO::ReadHexWordFromFile(int &line, int &column, int *checksum)
{
	ushort w;

	w = ReadHexByteFromFile(line, column, checksum);
	w = (ushort)((w << 8) + ReadHexByteFromFile(line, column, checksum));

	return w;
}

ushort HexFileIO::ReadHexWord32FromFile(int &line, int &column, int *checksum)
{
	uint w;

	w = ReadHexWordFromFile(line, column, checksum);
	w = (uint)((w << 16) + ReadHexWordFromFile(line, column, checksum));

	return w;
}

// swap byte macro
ushort HexFileIO::SWAB(ushort w)
{
	return (ushort)(((w)>>8) + (((w)&0xFF) << 8));
}

int HexFileIO::IdentifyFileType()
{
	int b;
	int line = 0, col = 1, check = 0;

	// get number of bytes in record
	b = ReadHexByteFromFile(line, col, &check);

	// See if the record length matches the actual length
	b = 2 * b + 11;
	int len = (int)strlen(lines[0]);
	TCHAR lastCh = lines[0][strlen(lines[0])-1];
	if (len == b)
	{
		// Either INHX8M or INHX32.  Most INHX32 files seem to have an extended
		// linear address segment record in the first line (BB = 04).  Look for
		// it.
		if (lines[0][7] == '0' && lines[0][8] == '4')
			return FORMAT_INHX32;
		else
			return FORMAT_INHX8M;
	}
	else
		return FORMAT_INHX16;
}

// Load a hex file for a Microchip PIC processor.
int HexFileIO::Load(LPCTSTR fileName, PICHexFile *hexFile)
{
	uint wordSize = hexFile->Processor->wordSize;
	uint configWordsSize = hexFile->Processor->configWordsSize;
	uint programWordsSize = hexFile->Processor->programWordsSize;
	uint EEPROMDataSize = hexFile->Processor->EEPROMDataSize;
	uint userIDWordsSize = hexFile->Processor->userIDWordsSize;
	uint deviceIDSize = hexFile->Processor->deviceIDSize;

	uint EEPROMDataAddr, configAddr, configWordsAddr;
	uint userIDWordsAddr, deviceIDAddr;
	switch (wordSize)
	{
	case 14:
		configAddr = PICDevice::PIC14_CONFIG_ADDRESS;
		configWordsAddr = PICDevice::PIC14_FUSE_ADDRESS;
		EEPROMDataAddr = PICDevice::PIC14_EEPROM_ADDRESS;
		userIDWordsAddr = PICDevice::PIC14_USER_ADDRESS;
		deviceIDAddr = PICDevice::PIC14_DEVID_ADDRESS;
		break;
	case 16:
		configAddr = PICDevice::PIC16_CONFIG_ADDRESS;
		configWordsAddr = PICDevice::PIC16_FUSE_ADDRESS;
		EEPROMDataAddr = PICDevice::PIC16_EEPROM_ADDRESS;
		userIDWordsAddr = PICDevice::PIC16_USER_ADDRESS;
		deviceIDAddr = PICDevice::PIC16_DEVID_ADDRESS;
		break;
	case 24:
		configAddr = PICDevice::PIC24_CONFIG_ADDRESS;
		configWordsAddr = PICDevice::PIC24_FUSE_ADDRESS;
		EEPROMDataAddr = PICDevice::PIC24_EEPROM_ADDRESS;
		userIDWordsAddr = PICDevice::PIC24_USER_ADDRESS;
		deviceIDAddr = PICDevice::PIC24_DEVID_ADDRESS;
		break;
	default:
		return ERROR_UNK;
	}

	LoadFileContents(fileName);
	if (lines.size() == 0)
	{
		// No contents or missing file.
		return -1;
	}

	// Start at first line, first column
	int line = 0;
	int column = 0;

	uint extendedAddr = 0;	// High word (bits 31:16)

	// TBD: Clear contents of buffers?

	// Set up counters, etc.
	hexFile->programWordsBase = programWordsSize;
	hexFile->programWordsCeil = 0;
	hexFile->EEPROMDataBase = EEPROMDataSize;
	hexFile->EEPROMDataCeil = 0;

	int fileFormat = IdentifyFileType();
	int readFailure = 0;

	uint address, iIgnoredAtAddress;
	int recordType, nwords, i, ign = 0;
	int iIgnoredInRecord;
	ushort w;
	int checksum;

	recordType = 0;
	while (recordType != 1)
	{
		// Every line must start with a ':'
		if ( lineChar(line, column) != ':' )
			// expected ':'
			return ERROR_CEX;

		checksum = 0;
		nwords = ReadHexByteFromFile(line, column, &checksum);
		if (readFailure != 0) return readFailure;
		address =ReadHexWordFromFile(line, column, &checksum);
		if (readFailure != 0) return readFailure;

		// Add extended address to address (Note: this must come after the division
		// by two above or the extended address will be uncorrectly modified.)
		address += extendedAddr;

		recordType = ReadHexByteFromFile(line, column, &checksum);
		if (readFailure != 0) return readFailure;

		if (recordType == 1)
		{
			// End of file record
			break;
		}
		else if (recordType == 4)
		{
			extendedAddr = ReadHexWordFromFile(line, column, &checksum);
			extendedAddr <<= 16;
			// TBD: look at checksum before continuing.
			line++;
			column = 0;
			continue;
		}
		else if (recordType != 0)
		{
			// Not a data record - just skip it
			line++;
			column = 0;
			continue;
		}

		// A data record of some sort.  Parse the address and see where it belongs.
		iIgnoredInRecord=0; iIgnoredAtAddress=address;
		for ( i=0; i<nwords; ++i, ++address )
		{
			if (fileFormat == FORMAT_INHX16)
				w = ReadHexWordFromFile(line, column, &checksum);
			else
				w = ReadHexByteFromFile(line, column, &checksum);
			if (readFailure != 0) return readFailure;

			uint wordAddr;
			if (fileFormat == FORMAT_INHX32)
				wordAddr = address / 2;
			if (fileFormat == FORMAT_INHX16)
				wordAddr = address;
			else
				wordAddr = address / 2;
			bool nonCodeAddress = false;
#if 1
			// Debug: move executive program down to regular code space.
			if (wordSize == 24)
			{
				if (wordAddr >= userIDWordsAddr && wordAddr < userIDWordsAddr + userIDWordsSize)
					wordAddr -= userIDWordsAddr;
			}
#endif
			if ((address < 0) ||
				(wordAddr >= configAddr) ||
				(wordSize == 16 && wordAddr >= userIDWordsAddr) ||
				(wordSize == 24 && wordAddr >= userIDWordsAddr) ||
				(wordSize == 24 && wordAddr >= EEPROMDataAddr))
				nonCodeAddress = true;
			if (nonCodeAddress)
			{
				// The data is configuration memory space->  Either it is one of
				// the config/ID words or is EEPROM data->
				if ((wordAddr >= configWordsAddr)
					&& ((wordAddr-configWordsAddr) < configWordsSize))
				{
					if (fileFormat == FORMAT_INHX16)
						hexFile->configWords[address-configWordsAddr] = w;
					else
					{
						byte *buf = (byte *)hexFile->configWords;
						buf[address - 2 * configWordsAddr] = (byte)w;
					}
				}
				else if ((wordAddr >= userIDWordsAddr)
					&& ((wordAddr-userIDWordsAddr) < userIDWordsSize))
				{
					if (fileFormat == FORMAT_INHX16)
						hexFile->userIDWords[address-userIDWordsAddr] = w;
					else
					{
						byte *buf = (byte *)hexFile->userIDWords;
						buf[address - 2 * userIDWordsAddr] = (byte)w;
					}
				}
				else if ((wordAddr >= deviceIDAddr)
					&& ((wordAddr-deviceIDAddr) < deviceIDSize))
				{
					if (fileFormat == FORMAT_INHX16)
						hexFile->deviceID[address-deviceIDAddr] = w;
					else
					{
						byte *buf = (byte *)hexFile->deviceID;
						buf[address - 2 * deviceIDAddr] = (byte)w;
					}
				}
				else if ((wordAddr>=EEPROMDataAddr)
					&& ((wordAddr-EEPROMDataAddr) < EEPROMDataSize))
				{
					if (fileFormat == FORMAT_INHX16)
						hexFile->EEPROMData[address-EEPROMDataAddr] = w;
					else
					{
						byte *buf = (byte *)hexFile->EEPROMData;
						buf[address - 2 * EEPROMDataAddr] = (byte)w;
						// hexFile->EEPROMData[address - 2 * EEPROMDataAddr] = w;
					}
					// LOADED this byte
					// SetMemoryFlags(address, BUFFER_FLAG_USED);
					if (hexFile->EEPROMDataCeil < wordAddr )
						hexFile->EEPROMDataCeil = wordAddr;
					if (hexFile->EEPROMDataBase > wordAddr )
						hexFile->EEPROMDataBase = wordAddr;
				}
				else
				{
					ign = 1;                    /* else ignore */
					++iIgnoredInRecord;
				}
			}
			else if (wordAddr < 0 || wordAddr >= programWordsSize)
			{
				// Address is outside the allowed range for this processor
				ign = 1;
				++iIgnoredInRecord;
			}
			else
			{
				// The data is within the program address range
				if (fileFormat == FORMAT_INHX16)
					hexFile->programWords[address] = w;
				else
				{
					byte *buf = (byte *)hexFile->programWords;
				    if (wordSize == 24 && address >= 2 * userIDWordsAddr)
						buf[address-2*userIDWordsAddr] = (byte)w;
					else
						buf[address] = (byte)w;
				}

				hexFile->programWordsUsed[wordAddr] = true;
				// Update the range of addresses used.
				if (hexFile->programWordsCeil < wordAddr )
					hexFile->programWordsCeil = wordAddr;
				if (hexFile->programWordsBase > wordAddr )
					hexFile->programWordsBase = wordAddr;
			}
		} // end for < all WORDs in the hex record >
		if (iIgnoredInRecord > 0)
		{
			// TBD: Warn that words were ignored
			/*
			sprintf(szErrorMsg,
				"LoadHex: Ignored %d word%c at address 0x%04X.",
				(int)iIgnoredInRecord,
				(char) (iIgnoredInRecord ==1 ) ? ' ' : 's',
				(int)iIgnoredAtAddress );
			APPL_ShowMsg( APPL_CALLER_PIC_HEX,0, szErrorMsg);
			*/
		}
		// get checksum
		ReadHexByteFromFile(line, column, &checksum);
		if (readFailure != 0)
			return readFailure;

		// discard end-of-line (ARE: Lines already parsed out, no EOL remaining
		// lineChar(line, column);

		if ((checksum & 0xFF) != 0)
			// checksum error
			return ERROR_CHK;
	} // end while

	return ((ign != 0) ? ERROR_IGN : 0);
}

#define CHBUFLEN 256

void HexFileIO::LoadFileContents(LPCTSTR fileName)
{
	FILE *fp = _tfopen(fileName, _T("rt"));
	if (fp == NULL)
		return;

	// Clear out any old contents
	int numLines = (int)lines.size();
	for (int lineIndex=0;lineIndex<numLines;lineIndex++)
	{
		if (lines[lineIndex] != NULL)
			delete lines[lineIndex];
		lines[lineIndex] = NULL;
	}
	lines.clear();

	char chBuf[CHBUFLEN];
	while (fgets(chBuf, CHBUFLEN-2*sizeof(char), fp) != NULL)
	{
		size_t lineLen = strlen(chBuf);
		if (chBuf[lineLen-1] == '\n')
		{
			// Strip off terminating EOL character
			chBuf[lineLen-1] = 0;
		}
		char *lineBuf = new char[lineLen];
		strcpy(lineBuf, chBuf);
		lines.push_back(lineBuf);
		char *lastLine = lines[lines.size()-1];
		char firstChar = lastLine[0];
	}
	fclose(fp);
}

// Write a block of memory in HEX file format.  Note: this routine assumes
// that integer values are stored little-endian.
void HexFileIO::writeMemory(FILE *fp, int wordSize, int startAddr, int bufSize, byte *memBuf)
{
	int checksum;
	uint addr, addrU, endAddr;
	uint addrIndex;

	addrU = -1;
	endAddr = startAddr + bufSize;
	addr = startAddr;
	addrIndex = 0; // values start at the beginning of the memory buffer
	while (addr < endAddr)
	{
		// See if there has been a change in the upper part of the address.  If
		// so, then we need to write an extended address to the hex file.
		uint taddrU = ((addr & 0xffff0000) >> 16); // Mask off bottom 16 bits
		if (addrU != taddrU)
		{
			// Change in upper part of address.  Need to output an extended
			// address.
			addrU = taddrU;
			checksum = 0x02 + 0x04 + (addrU & 0xff) + ((addrU >> 8) & 0xff);
			fprintf(fp, ":02000004%04X%02X\n", addrU, ((-checksum) & 0xff));
		}

		// See if this line of values is blank
		bool emptyFlag = true;
		int lineLen = (endAddr - addr) < 16 ? (endAddr - addr) : 16;
		for (int k=0;k<lineLen && emptyFlag;k++)
		{
			if (memBuf[addrIndex+k] != 0xff)
				emptyFlag = false;
			// If this memory buffer is for 1 byte values (but stored in
			// an array of 2 byte values) then skip the high byte.
			if (wordSize == 1)
				k++;
		}
		if (!emptyFlag)
		{
			// There is a non-empty value in this group of bytes.  Dump
			// the entire line out.
			fprintf(fp, ":%02X%04X00", lineLen, (addr & 0xffff));
			checksum = (lineLen & 0xff) + (((addr & 0xff00) >> 8) & 0xff) + (addr & 0xff);
			for (int wIndex=0;wIndex<lineLen;wIndex++) {
				byte byteVal = memBuf[addrIndex++];
				checksum += byteVal;
				addr++;
				fprintf(fp, "%02X", byteVal);
			}
			fprintf(fp, "%02X\n", ((-checksum) & 0xff));
		}
		else
		{
			addrIndex += lineLen;
			addr += lineLen;
		}
	}
}

int HexFileIO::Save(LPCTSTR fileName, PICHexFile *hexFile)
{
	FILE *fp = _tfopen(fileName, _T("wt"));
	if (fp == NULL)
		return -1;

	const PICDevice *proc = hexFile->Processor;
	

	// Write code memory
	writeMemory(fp, 2, 0, proc->programWordsSize*2, (byte *)hexFile->programWords);
	// Write data memory
	writeMemory(fp, 1, proc->EEPROMDataAddr, proc->EEPROMDataSize*2, (byte *)hexFile->EEPROMData);
	// Write config memory
	writeMemory(fp, 2, proc->configAddr, proc->configWordsSize*2, (byte *)hexFile->configWords);
	// Write User ID memory
	writeMemory(fp, 2, proc->userIDWordsAddr, proc->userIDWordsSize*2, (byte *)hexFile->userIDWords);

	// Write the end of file record
	fprintf(fp, ":00000001FF\n");

	// Done writing, close the file
	fclose(fp);

	return 0;
}
