#pragma once

#include <vector>
#include "PIC/PICHexFile.h"

#if !defined(byte)
#define byte unsigned char
#endif
#if !defined(ushort)
#define ushort unsigned short
#endif

/// <summary>
/// PICHexFileIO exists for the sole purpose of translating between the .hex
/// files and the binary representation used in PIC microcontrollers.  The class
/// PICHexFile holds all of the program, data, and configuration words that are
/// loaded to/from the PIC.  The methods in PICHexFileIO are used to either populate
/// the contents of the PICHexFile class from a .hex file or to dump the contents
/// of a PICHexFile class into a .hex file.  The supported formats for .hex files
/// are: INHX8M, INHX16, INHX32.
/// </summary>
class HexFileIO
{
public:
	HexFileIO(void);
	~HexFileIO(void);

	// Contents of the file
	std::vector<LPSTR> lines;

	// From 
	// Intel HEX Format (.HEX)
	//   Code of this format is often downloaded from a PC to a development system
	//   and run from RAM. Alternatively, the hex file can be converted to a binary file
	//   and programmed into an EPROM. This format produces one 8-bit HEX file with a low
	//   byte, high byte combination. Each data record begins with a 9 character prefix
	//   and ends with a 2 character checksum. Each record has the following format: 
	//		:BBAAAATTHHHH....HHHCC
	//
	//	 where 
	//		BB 
	//			is a two digit hexadecimal byte count representing the number of data
	//          bytes that will appear on the line; 
	//		AAAA 
	//			is a four digit hexadecimal address representing the starting address
	//          of the data record; 
	//		TT 
	//			is a two digit record type: 
	//				00 - Data record 
	//				01 - End of file record 
	//				02 - Extended segment address record 
	//				03 - Start segment address record 
	//				04 - Extended linear address record 
	//				05 - Start linear address record 
	//		HH 
	//			is a two digit hexadecimal data word, presented in low byte, high byte
	//			combinations; 
	//		CC 
	//			is a two digit hexadecimal checksum that is the two's compliment of the
	//			sum of all preceding bytes in the record including the prefix; (sum of
	//			all bytes + checksum = 00). 
	// INTEL 32-bit HEX Format INHX32 (Intel Extended HEX Format) (.HEX)
	//   The extended 32-bit address HEX format is similar to the Hex 8 format
	//   described above, except that the Intel extended linear address record is
	//   output also to establish the upper 16 bits of the data address.
	//
	//	 Each data record begins with 9 charcter prefix and ends with a 2 character
	//   checksum. Each record has the following format: 
	//
	//		:BBAAAATTHHHH....HHHCC
	//
	//	 where 
	//		BB 
	//			is a two digit hexadecimal byte count representing the number of data
	//			bytes that will appear on the line; 
	//		AAAA 
	//			is a four digit hexadecimal address representing the starting address
	//			of the data record; 
	//		TT 
	//			is a two digit record type: 
	//				00 - Data record 
	//				01 - End of file record 
	//				02 - Extended segment address record 
	//				03 - Start segment address record 
	//				04 - Extended linear address record 
	//				05 - Start linear address record 
	//		HH 
	//			is a two digit hexadecimal data word, presented in low byte, high byte
	//			combinations; 
	//		CC 
	//			is a two digit hexadecimal checksum that is the two's compliment of
	//			the sum of all preceding bytes in the record including the prefix; 
	//
	// PIC 8-bit HEX Format INHX8M (.OBJ)
	//   This format produces one 8-bit hexadecimal file with a low-byte/high-byte
	//   combination. Since each address can only contain 8 bits in this format,
	//   all addresses are doubled, low byte first. This format are useful for
	//   transferring PIC series object code to the third party EPROM programmers. 
	//
	// PIC 16-bit HEX Format INHX16
	//   The 16-bit word format is basically the same as the 8-bit word format. The
	//   difference between this format and INHX8M is the word length and the
	//   high/low byte order. INHX8M has 8-bit words (two hexadecimal digits) with
	//   the low byte first, rather than 16-bit words (four hexadecimal digits) with
	//   the high byte first. The byte count (BB), however, is based on 16-bit words. 
	//
	static const int WORDS_PER_HEX_DUMP; // = 8; // dump this many words per hex record

	static const int PROG_BUF_SIZE  ; // = 65536; // max size of program memory
	static const int DATA_BUF_SIZE  ; // =  1024; // max size of EEDATA memory
	static const int ID_BUFFER_SIZE ; // =     4; // max size of ID memory
	static const int NR_CONFIG_WORDS; // =     7; // max count of configuration words

	static const int FORMAT_UNKNOWN; // = -1;      // hex formats
	static const int FORMAT_INHX8M ; // =  0;
	static const int FORMAT_INHX16 ; // =  1;
	static const int FORMAT_INHX32 ; // =  2;

	static const int ERROR_EOF; // = -1;      // unexpected EOF
	static const int ERROR_DEX; // = -2;      // hex digit required
	static const int ERROR_CEX; // = -3;      // missing ':'
	static const int ERROR_CHK; // = -4;      // checksum error
	static const int ERROR_UNK; // = -5;      // Unknown error
	static const int ERROR_IGN; // = 1;       // warning that some records were ignored

	// FLAGS for each word in the program buffer
	static const int BUFFER_FLAG_USED     ; // =  1;  // LOADED or READ<>3FFF
	static const int BUFFER_FLAG_PRG_ERROR; // =  2;  // PROGRAM or VERIFY error
	static const int BUFFER_FLAG_VFY_ERROR; // =  4;
	static const int BUFFER_FLAG_DEAD     ; // =  8;  // Failed repeatedly (R.I.P.)
	static const int BUFFER_FLAG_SPECIAL  ; // = 16;  // special location, like osc calib word

	// Code for End of File (EOF)
#if !defined(EOF)
	static const int EOF = -1;
#endif

	// Loads a Microchip hex file.  Words between 0 and programWordsMax-1 are stored
	// in the program memory buffer; words between dbase and dbase+dsize-1
	// are stored in the EEDATA buffer; 4 ID words at ibase and above are
	// stored in the ID buffer;
	// and the word(s) at cbase is stored in PIC_HEX_config_word[0] and maybe [1].
	// Returns 0 for success, -ve for failure and +ve if any records were
	// out-of-range.  On exit the gobal variable PIC_HEX_format can be used to
	// determine the style (INHX8M or INHX16) loaded.
	int Load(LPCTSTR fileName, PICHexFile *hexFile);

	// Save a hex file
	int Save(LPCTSTR fileName, PICHexFile *hexFile);

private:
	// Members used during file I/O.
	int readFailure;

	char lineChar(int &line, int &column);
	byte ReadDigitFromFile(int &line, int &column);
	byte ReadHexByteFromFile(int &line, int &column, int *checksum);
	ushort ReadHexWordFromFile(int &line, int &column, int *checksum);
	ushort ReadHexWord32FromFile(int &line, int &column, int *checksum);

	void writeMemory(FILE *fp, int wordSize, int startAddr, int bufSize, byte *memBuf);

	// swap byte macro
	static ushort SWAB(ushort w);

	int IdentifyFileType();
	void LoadFileContents(LPCTSTR fileName);
};
