#include "StdAfx.h"
#include <assert.h>
#include "PIC12F6XXProgram.h"

PIC12F6XXProgram PIC12F6XXProgram::m_Driver;

// Get the singleton programmer instance for this class of chips
const PICProgram *PIC12F6XXProgram::GetDriver()
{
	return &m_Driver;
}

int PIC12F6XXProgram::ReadCode(const PICDevice *device, CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	uint addr, i;
	if (statusIF) statusIF->InitProgress(_T("Reading Program Memory"));

	// Reset PIC.  After this is executed we will be at address 0 and can
	// increment to get to the address we actually want to read.
	PICOpen(port, device->devVppMask);

	// Read code memory in CMD_BLOCKSIZE byte chunks.
	byte outBuf[128], inBuf[128];
	uint block, resLen;
	uint blockSize = 32; // CMD_BLOCKSIZE; // TBD: USB is failing for 64 bytes (intermittently)
	for (addr=0,block=0;addr<device->programWordsSize;block+=blockSize)
	{
		outBuf[0] = 0;
		outBuf[1] = 0;
		outBuf[2] = 0;
		outBuf[3] = blockSize;
		i = PICProgram::SendPacket(port, PICProgram::EVA_PIC14RB_CMD,  4,
			outBuf, 128, &resLen, inBuf, 4*blockSize);
		if (i)
		{
			PICClose(port);
			return addr+1;
		}

		// Move the block response into the result array
		ushort *valBuf = (ushort *)&inBuf[6]; // Location of first value in response buffer
		for (uint i0=0;i0<blockSize;i0+=2)
		{
			// The programmer sends back the data LSB first so we can just pull
			// the values right out of the buffer.
			hexFile->programWords[addr++] = *valBuf++;
		}

		if (statusIF && !statusIF->UpdateProgress(addr, device->programWordsSize))
		{
			PICClose(port);
			return 1;
		}
	}

	PICClose(port);

	return 0;
}

int PIC12F6XXProgram::ReadConfig(const PICDevice *device, CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	uint addr, wordVal, i, j;
	if (statusIF) statusIF->InitProgress(_T("Reading Config"));

	// Reset PIC
	PICOpen(port, device->devVppMask);

	// Enter configuration memory
	i = TxCmd(port, EVA_PIC14R_CMD, PIC14_LOADCFG | 0xFFFF, &j);
	if(i) { PICClose(port); return 1; }

	// Skip everything before config word(s)
	uint configOffset = device->configWordsAddr - device->configAddr;
	uint userOffset = device->userIDWordsAddr - device->configAddr;
	uint deviceIDOffset = device->deviceIDAddr - device->configAddr;

	// This only works as long as the number of config words is less
	// than CMD_BLOCKSIZE/2 (CMD_BLOCKSIZE bytes)
	if (2*device->configSize < CMD_BLOCKSIZE)
	{
		byte outBuf[4], inBuf[128];
		uint resLen, blockSize = 2*device->configSize;
		outBuf[0] = 0;
		outBuf[1] = 0;
		outBuf[2] = 0;
		outBuf[3] = blockSize;
		i = PICProgram::SendPacket(port, PICProgram::EVA_PIC14RB_CMD,  4,
			outBuf, 128, &resLen, inBuf);
		if (i) { PICClose(port); return 2; }

		for (addr=0;addr<device->configSize;addr++)
		{
			wordVal = (inBuf[2*addr+7] << 8) | inBuf[2*addr+6];

			// Save the contents of the address in the appropriate place (fuse or ID)
			if (addr >= configOffset && (addr - configOffset) < device->configWordsSize)
				hexFile->configWords[addr-configOffset] = wordVal;
			else if (addr >= userOffset && (addr - userOffset) < device->userIDWordsSize)
				hexFile->userIDWords[addr-userOffset] = wordVal;
			else if (addr >= deviceIDOffset && (addr - deviceIDOffset) < device->deviceIDSize)
				hexFile->deviceID[addr-deviceIDOffset] = wordVal;

			if(i) { PICClose(port); return addr + 1; }

			if (statusIF && !statusIF->UpdateProgress(addr, device->configSize))
			{
				PICClose(port);
				return 1;
			}
		}
	}
	else
	{
		for (addr=0;addr<device->configSize;addr++)
		{
			i  = TxCmd(port, EVA_PIC14R_CMD, PIC14_READDATA | 0xFFFF, &wordVal);

			if(i) { PICClose(port); return (addr+1); }

			if (addr >= configOffset && (addr - configOffset) < device->configWordsSize)
			{
				// Save the contents of the address
				hexFile->configWords[addr-configOffset] = wordVal; // ((wordVal >> 1) & 0x3fff);

				assert(addr-configOffset < device->configWordsSize);
			}
			else if (addr >= userOffset && (addr - userOffset) < device->userIDWordsSize)
			{
				hexFile->userIDWords[addr-userOffset] = wordVal; // ((wordVal >> 1) & 0x3fff);
			}
			else if (addr >= deviceIDOffset && (addr - deviceIDOffset) < device->deviceIDSize)
			{
				hexFile->deviceID[addr-deviceIDOffset] = wordVal; // ((wordVal >> 1) & 0x3fff);
			}

			// Increment the address
			i = TxCmd(port, EVA_PIC14_CMD, PIC14_INCADDR | 0xFFFF, &j);

			if(i) { PICClose(port); return addr + 1; }

			if (statusIF && !statusIF->UpdateProgress(addr, device->configSize))
			{
				PICClose(port);
				return 1;
			}
		}
	}

	PICClose(port);

	return 0;
}
int PIC12F6XXProgram::ReadData(const PICDevice *device, CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	uint addr, i;
	if (statusIF) statusIF->InitProgress(_T("Reading EEPROM"));

	// read EEPROM
	PICOpen(port, device->devVppMask);
#if 1
	// Read code memory in CMD_BLOCKSIZE byte chunks.
	byte outBuf[4], inBuf[128];
	uint block, resLen;
	uint blockSize = CMD_BLOCKSIZE;
	if (blockSize > device->EEPROMDataSize)
		blockSize = device->EEPROMDataSize;
	for (addr=0,block=0;addr<device->EEPROMDataSize;block+=blockSize/2)
	{
		outBuf[0] = 0;
		outBuf[1] = 0;
		outBuf[2] = 0;
		outBuf[3] = blockSize;
		i = PICProgram::SendPacket(port, PICProgram::EVA_PIC14RBE_CMD,  4,
			outBuf, 128, &resLen, inBuf, 2*blockSize);
		if (i) { PICClose(port); return addr+1; }

		// Move the block response into the result array
		ushort *valBuf = (ushort *)&inBuf[6]; // Location of first value in response buffer
		for (uint i0=0;i0<blockSize;i0+=2)
		{
			// The programmer sends back the data LSB first so we can just pull
			// the values right out of the buffer.
			hexFile->EEPROMData[addr++] = (*valBuf++) & 0xff;
		}

		if (statusIF && !statusIF->UpdateProgress(addr, device->EEPROMDataSize))
		{
			PICClose(port);
			return 1;
		}
	}
#else
	uint wordVal, j;
	for (addr=0;addr<device->EEPROMDataSize;addr++)
	{
		i  = TxCmd(port, EVA_PIC14R_CMD, PIC14_READE2DATA | 0xFFFF, &wordVal);
		i += TxCmd(port, EVA_PIC14_CMD, PIC14_INCADDR | 0xFFFF, &j);
		// Check for any errors
		if (i) { PICClose(port); return(1); }

		assert(addr < device->EEPROMDataSize);

		// Save the contents of the address
		hexFile->EEPROMData[addr] = (wordVal & 0xFF); // ((wordVal >> 1) & 0xFF);

		if (statusIF && !statusIF->UpdateProgress(addr, device->EEPROMDataSize))
		{
			PICClose(port);
			return 1;
		}
	}
#endif

	PICClose(port);

	return 0;
}

int PIC12F6XXProgram::PartialRead(CIOPort *port, PICHexFile *hexFile, bool codeFlag, bool dataFlag, bool configFlag, CProgrammerProgressIF *statusIF) const
{
	const PICDevice *device = hexFile->Processor;

	int res;
	uint result;
	res = PICProgram::TxCmd(port, PICProgram::EVA_VERSION_CMD, 0, &result);
    if (!SUCCEEDED(ResetProgrammer(port)))
	{
		// Failed to get the programmer into a known good state.
		return 1;
	}

	if (codeFlag && (res = ReadCode(device, port, hexFile, statusIF))) return res;
	if (configFlag && (res = ReadConfig(device, port, hexFile, statusIF))) return res;
	if (dataFlag && (res = ReadData(device, port, hexFile, statusIF))) return res;

	return 0;
}

int PIC12F6XXProgram::Read(CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	const PICDevice *device = hexFile->Processor;

    if (!SUCCEEDED(ResetProgrammer(port)))
	{
		// Failed to get the programmer into a known good state.
		return 1;
	}

	int res;
	if (res = ReadCode(device, port, hexFile, statusIF)) return res;
	if (res = ReadConfig(device, port, hexFile, statusIF)) return res;
	if (res = ReadData(device, port, hexFile, statusIF)) return res;

	return 0;
}

int PIC12F6XXProgram::WriteCode(const PICDevice *device, CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	uint addr, i;
	if (statusIF) statusIF->InitProgress(_T("Writing Program Memory"));

	// Open the programmer
	PICOpen(port, device->devVppMask);

#if 1
	// Write code memory in CMD_BLOCKSIZE byte chunks.
	byte outBuf[128], inBuf[128];
	uint block, resLen;
	uint blockSize = CMD_BLOCKSIZE;
	for (addr=0,block=0;addr<device->programWordsSize;block+=blockSize)
	{
		outBuf[0] = (PIC14_BEGINPROG & 0xff0000) >> 16;
		outBuf[1] = 0;
		outBuf[2] = 0;
		outBuf[3] = blockSize;

		ushort *valBuf = (ushort *)&outBuf[4]; // Location of first value in response buffer

		for (uint i0=0;i0<blockSize;i0+=2)
		{
			*valBuf++ = hexFile->programWords[addr++];
		}
		i = PICProgram::SendPacket(port, PICProgram::EVA_PIC14WB_CMD,  blockSize+4,
			outBuf, 128, &resLen, inBuf, 4*blockSize);
		if (i)
		{
			PICClose(port);
			return addr+1;
		}
		if (inBuf[0] == 0xee)
		{
			// Write failed
			PICClose(port);
			return addr+1;
		}

		if (statusIF && !statusIF->UpdateProgress(addr, device->programWordsSize))
		{
			PICClose(port);
			return 1;
		}
	}
#else
	uint j;
	for(addr=0;addr<device->programWordsSize;addr++)
	{
		// Write the word and check to see the write was valid
		i = WriteWord14(port, hexFile->programWords[addr]);
		if (i) { PICClose(port); return addr+1; }

		// If we got here the write was valid.  Incement the target and continue.
		i  = TxCmd(port, EVA_PIC14_CMD, PIC14_INCADDR | 0xFFFF, &j);

		if (i) { PICClose(port); return addr+1; }

		if (statusIF && !statusIF->UpdateProgress(addr, device->programWordsSize))
		{
			PICClose(port);
			return 1;
		}
	}
#endif

	// Shut down connection to programmer
	PICClose(port);

	return 0;
}

int PIC12F6XXProgram::WriteData(const PICDevice *device, CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	uint addr, i;
	if (statusIF) statusIF->InitProgress(_T("Writing EEPROM Memory"));

	// Open programmer
	PICOpen(port, device->devVppMask);

#if 1
	// Write EEPROM data in CMD_BLOCKSIZE byte chunks.
	byte outBuf[128], inBuf[128];
	uint block, resLen;
	uint blockSize = CMD_BLOCKSIZE;
	if (blockSize > device->EEPROMDataSize)
		blockSize = device->EEPROMDataSize;
	for (addr=0,block=0;addr<device->EEPROMDataSize;block+=blockSize)
	{
		outBuf[0] = (PIC14_BEGINPROG & 0xff0000) >> 16;
		outBuf[1] = 0;
		outBuf[2] = 0;
		outBuf[3] = blockSize;

		byte *valBuf = &outBuf[4]; // Location of first value in response buffer

		for (uint i0=0;i0<blockSize;i0++)
		{
			*valBuf++ = (hexFile->EEPROMData[addr++] & 0xFF);
		}
		i = PICProgram::SendPacket(port, PICProgram::EVA_PIC14WBE_CMD,  blockSize+4,
			outBuf, 128, &resLen, inBuf, 64*blockSize);
		if (i) {
			PICClose(port);
			return addr+1;
		}
		if (inBuf[0] == 0xee)
		{
			// Write failed
			return addr+1;
		}

		if (statusIF && !statusIF->UpdateProgress(addr, device->EEPROMDataSize))
		{
			PICClose(port);
			return 1;
		}
	}
#else
	uint j;
	// Step through EEPROM values, writing and verifying one at a time
	for(addr=0;addr<device->EEPROMDataSize;addr++)
	{
		i = WriteWord14(port, hexFile->EEPROMData[addr], 0x00ff, PIC14_READE2DATA, PIC14_LOADE2DATA);
		if(i) { PICClose(port); return(1); }

		// Write valid, increment target to next EEPROM address
		i = TxCmd(port, EVA_PIC14_CMD, PIC14_INCADDR | 0xFFFF, &j);

		if(i) { PICClose(port); return addr+1; }

		if (statusIF && !statusIF->UpdateProgress(addr, device->EEPROMDataSize))
		{
			PICClose(port);
			return 1;
		}
	}
#endif

	// Shut down connection to programmer
	PICClose(port);

	return 0;
}

int PIC12F6XXProgram::WriteConfig(const PICDevice *device, CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	uint addr, i, j;
	if (statusIF) statusIF->InitProgress(_T("Writing Config Memory"));

	PICOpen(port, device->devVppMask);

	// Determine allowed address ranges for user ID and config word(s)
	uint userWordsBase = device->userIDWordsAddr;
	uint userWordsCeil = device->userIDWordsAddr + device->userIDWordsSize - 1;
	uint configWordsBase = device->configWordsAddr;
	uint configWordsCeil = device->configWordsAddr + device->configWordsSize - 1;

	// Put PIC into configuration space
	i  = TxCmd(port, EVA_PIC14R_CMD, PIC14_LOADCFG | 0xFFFF, &j);
	if(i) { PICClose(port); return 1; }

	// Step through configuration memory, skipping anything that isn't in the
	// valid range for user ID and config/fuse words.
	for (addr=device->configAddr;addr<device->configAddr+device->configSize;addr++)
	{
		if (addr >= userWordsBase && addr <= userWordsCeil)
		{
			int userOffset = addr - userWordsBase;
			// Don't care as much about failure here - keep going even if one or
			// more of the user ID values doesn't get written
			i = WriteWord14(port, hexFile->userIDWords[userOffset]);
			if(i) {  /* PICClose(); return(1); */ }
		}
		else if (addr >= configWordsBase && addr <= configWordsCeil)
		{
			int configOffset = addr - configWordsBase;
			// Config programming must work or we will have a chip with incorrect
			// settings for things like clock, code/data protection, pin assignments, ...
			i = WriteWord14(port, hexFile->configWords[configOffset]);
			if (i) { PICClose(port); return addr+1; }
		}

		// Move to next address in configuration space
		i  = TxCmd(port, EVA_PIC14_CMD, PIC14_INCADDR | 0xFFFF, &j);
		if(i) { PICClose(port); return addr+1; }

		if (statusIF && !statusIF->UpdateProgress(addr, device->configSize))
		{
			PICClose(port);
			return 1;
		}
	}
	PICClose(port);

	return 0;
}

int PIC12F6XXProgram::WriteUser(const PICDevice *device, CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	if (statusIF) statusIF->InitProgress(_T("Writing User IDs"));

	PICOpen(port, device->devVppMask);

	PICClose(port);

	return 0;
}

int PIC12F6XXProgram::Program(CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	int res;

	const PICDevice *device = hexFile->Processor;

    if (!SUCCEEDED(ResetProgrammer(port)))
	{
		// Failed to get the programmer into a known good state.
		return 1;
	}

	// Send the contents of the hex file to the PIC
	if ((res = WriteCode(device, port, hexFile, statusIF)) != 0) return res;
	if ((res = WriteData(device, port, hexFile, statusIF)) != 0) return res;
	if ((res = WriteConfig(device, port, hexFile, statusIF)) != 0) return res;

	// Ensure everything shut down
	PICClose(port);

	return 0;
}

int PIC12F6XXProgram::PartialProgram(CIOPort *port, PICHexFile *hexFile, bool codeFlag, bool dataFlag, bool configFlag, CProgrammerProgressIF *statusIF) const
{
	int res;

	const PICDevice *device = hexFile->Processor;

    if (!SUCCEEDED(ResetProgrammer(port)))
	{
		// Failed to get the programmer into a known good state.
		return 1;
	}

	// Send the contents of the hex file to the PIC
	if (codeFlag && (res = WriteCode(device, port, hexFile, statusIF)) != 0) return res;
	if (dataFlag && (res = WriteData(device, port, hexFile, statusIF)) != 0) return res;
	if (configFlag && (res = WriteConfig(device, port, hexFile, statusIF)) != 0) return res;

	// Ensure everything shut down
	PICClose(port);

	return 0;
}


int PIC12F6XXProgram::Erase(CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	int i;
	uint j;

	const PICDevice *device = hexFile->Processor;

    if (!SUCCEEDED(ResetProgrammer(port)))
	{
		// Failed to get the programmer into a known good state.
		return 1;
	}

	PICOpen(port, device->devVppMask);

	// Enter configuration memory (forces user ID bits to be cleared as well)
	i = TxCmd(port, EVA_PIC14R_CMD, PIC14_LOADCFG | 0xFFFF, &j);
	if(i) { PICClose(port); return 1; }

#if 0
	// Necessary for 16F88 (TBD: other chips as well?).  Note AN910 describes
	// this instruction as necessary for "rare" flash chips.
	i  = TxCmd(port, EVA_PIC14_CMD, PIC14_CHIPERASE | 0xFFFF, &j);
#else
	i  = TxCmd(port, EVA_PIC14_CMD, PIC14_BULKERASEPROG | 0xFFFF, &j);
	::Sleep(10);
	if (i) { PICClose(port); return(1); }
	i += TxCmd(port, EVA_PIC14_CMD, PIC14_BULKERASEE2 | 0xFFFF, &j);
	::Sleep(10);
	if (i) { PICClose(port); return(1); }
#endif

	PICClose(port);

	return 0;
}

int PIC12F6XXProgram::PartialErase(CIOPort *port, PICHexFile *hexFile, bool bootFlag, bool codeFlag, bool dataFlag, bool configFlag, CProgrammerProgressIF *statusIF) const
{
	int i;
	uint j;

	const PICDevice *device = hexFile->Processor;

    if (!SUCCEEDED(ResetProgrammer(port)))
	{
		// Failed to get the programmer into a known good state.
		return 1;
	}

	PICOpen(port, device->devVppMask);
	if (codeFlag)
	{
		i  = TxCmd(port, EVA_PIC14_CMD, PIC14_BULKERASEPROG | 0xFFFF, &j);

		::Sleep(10);

		i  = TxCmd(port, EVA_PIC14_CMD, PIC14_BEGINPROGONLY | 0xFFFF, &j);

		::Sleep(10);

		if (i) { PICClose(port); return(1); }
	}
	if (dataFlag)
	{

	// i = TxCmd(port, EVA_PIC14_CMD, PIC14_LOADE2DATA | 0xFFFF, &j);
	// i = TxCmd(port, EVA_PIC14_CMD, PIC14_LOADDATA | 0xFFFF, &j);

		i  = TxCmd(port, EVA_PIC14_CMD, PIC14_BULKERASEE2 | 0xFFFF, &j);
		if (i) { PICClose(port); return(1); }

		// TBD: is this necessary?
		// i  = TxCmd(port, EVA_PIC14_CMD, PIC14_BEGINPROGONLY | 0xFFFF, &j);
		i  = TxCmd(port, EVA_PIC14_CMD, PIC14_BEGINPROG | 0xFFFF, &j);
	}
	if (configFlag)
	{
		// TBD: erase config bits.  This may require code/data erase if code protection
		// was enabled.
	}
	PICClose(port);

	return 0;
}

int PIC12F6XXProgram::Verify(CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	return 0;
}
