#include "StdAfx.h"
#include <assert.h>
#include "PIC16F88Program.h"

PIC16F88Program PIC16F88Program::m_Driver;

// Get the singleton programmer instance for this class of chips
const PICProgram *PIC16F88Program::GetDriver()
{
	return &m_Driver;
}

int PIC16F88Program::WriteCode(const PICDevice *device, CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	uint addr, i;
	if (statusIF) statusIF->InitProgress(_T("Writing Program Memory"));

	// Open the programmer
	PICOpen(port);

#if 1
	// TBD: Try to adapt this to firmware block write.  The begin/end programming
	// commands for this part are different than many other flash parts.  Might
	// not be worth it for the extra speed if it isn't a frequently used chip.
	// Note: ICProg doesn't even support the 16F88.

	// Write code memory in CMD_BLOCKSIZE byte chunks.
	byte outBuf[128], inBuf[128];
	uint block, resLen;
	uint blockSize = CMD_BLOCKSIZE;
	for (addr=0,block=0;addr<device->programWordsSize;block+=blockSize)
	{
		outBuf[0] = (PIC14_BEGINPROGONLY & 0xff0000) >> 16;
		outBuf[1] = (PIC14_ENDPROGA & 0xff0000) >> 16;
		outBuf[2] = 0;
		outBuf[3] = blockSize;

		ushort *valBuf = (ushort *)&outBuf[4]; // Location of first value in response buffer

		for (uint i0=0;i0<blockSize;i0+=2)
		{
			*valBuf++ = hexFile->programWords[addr++];
		}
		i = PICProgram::SendPacket(port, PICProgram::EVA_PIC14WB_CMD,  blockSize+4, outBuf, 128, &resLen, inBuf, 4*blockSize);
		if (i) { PICClose(port); return addr+1; }
		if (inBuf[0] == 0xee)
		{
			// Write failed
			return addr+1;
		}

		if (statusIF && !statusIF->UpdateProgress(addr, device->programWordsSize))
		{
			PICClose(port);
			return 1;
		}
	}
#else
	// TBD: Get this working first, then see if it can be extended to block
	// write the way other 16F parts can.
	uint j;
	for(addr=0;addr<device->programWordsSize;addr++)
	{
		// Write the word and check to see the write was valid
		i = WriteWord14(port, hexFile->programWords[addr], 0xFFFF,
			PIC14_READDATA, PIC14_LOADDATA,
			PIC14_BEGINPROGONLY, PIC14_ENDPROGA);
		if (i) { PICClose(port); return addr+1; }

		// If we got here the write was valid.  Incement the target and continue.
		i  = TxCmd(port, EVA_PIC14_CMD, PIC14_INCADDR | 0xFFFF, &j);

		if (i) { PICClose(port); return addr+1; }

		if (statusIF && !statusIF->UpdateProgress(addr, device->programWordsSize))
		{
			PICClose(port);
			return 1;
		}
	}
#endif

	// Shut down connection to programmer
	PICClose(port);

	return 0;
}

int PIC16F88Program::WriteData(const PICDevice *device, CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	uint addr, i;
	if (statusIF) statusIF->InitProgress(_T("Writing EEPROM Memory"));

	// Open programmer
	PICOpen(port);

#if 1
	// Write EEPROM data in CMD_BLOCKSIZE byte chunks.
	byte outBuf[128], inBuf[128];
	uint block, resLen;
	uint blockSize = CMD_BLOCKSIZE;
	if (blockSize > device->EEPROMDataSize)
		blockSize = device->EEPROMDataSize;
	for (addr=0,block=0;addr<device->EEPROMDataSize;block+=blockSize)
	{
		outBuf[0] = (PIC14_BEGINPROGONLY & 0xff0000) >> 16;
		outBuf[1] = (PIC14_ENDPROGA & 0xff0000) >> 16;
		outBuf[2] = 0;
		outBuf[3] = blockSize;

		byte *valBuf = &outBuf[4]; // Location of first value in response buffer

		for (uint i0=0;i0<blockSize;i0++)
		{
			*valBuf++ = (hexFile->EEPROMData[addr++] & 0xFF);
		}
		i = PICProgram::SendPacket(port, PICProgram::EVA_PIC14WBE_CMD,  blockSize+4,
			outBuf, 128, &resLen, inBuf, 16*blockSize);
		if (i) { PICClose(port); return addr+1; }
		if (inBuf[0] == 0xee)
		{
			// Write failed
			return addr+1;
		}

		if (statusIF && !statusIF->UpdateProgress(addr, device->EEPROMDataSize))
		{
			PICClose(port);
			return 1;
		}
	}
#else
	uint j;
	// Step through EEPROM values, writing and verifying one at a time
	for(addr=0;addr<device->EEPROMDataSize;addr++)
	{
		i = WriteWord14(port, hexFile->EEPROMData[addr], 0x00FF,
			PIC14_READE2DATA, PIC14_LOADE2DATA,
			PIC14_BEGINPROGONLY, PIC14_ENDPROGA);
		if(i) { PICClose(port); return(1); }

		// Write valid, increment target to next EEPROM address
		i = TxCmd(port, EVA_PIC14_CMD, PIC14_INCADDR | 0xFFFF, &j);

		if(i) { PICClose(port); return addr+1; }

		if (statusIF && !statusIF->UpdateProgress(addr, device->EEPROMDataSize))
		{
			PICClose(port);
			return 1;
		}
	}
#endif
	// Shut down connection to programmer
	PICClose(port);

	return 0;
}

int PIC16F88Program::WriteConfig(const PICDevice *device, CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	uint addr, i, j;
	if (statusIF) statusIF->InitProgress(_T("Writing Config Memory"));

	PICOpen(port);

	// Determine allowed address ranges for user ID and config word(s)
	uint userWordsBase = device->userIDWordsAddr;
	uint userWordsCeil = device->userIDWordsAddr + device->userIDWordsSize - 1;
	uint configWordsBase = device->configWordsAddr;
	uint configWordsCeil = device->configWordsAddr + device->configWordsSize - 1;

	// Put PIC into configuration space
	i  = TxCmd(port, EVA_PIC14R_CMD, PIC14_LOADCFG | 0xFFFF, &j);
	if(i) { PICClose(port); return 1; }

	// Step through configuration memory, skipping anything that isn't in the
	// valid range for user ID and config/fuse words.
	for (addr=device->configAddr;addr<device->configAddr+device->configSize;addr++)
	{
		if (addr >= userWordsBase && addr <= userWordsCeil)
		{
			int userOffset = addr - userWordsBase;
			// Don't care as much about failure here - keep going even if one or
			// more of the user ID values doesn't get written
			i = WriteWord14(port, hexFile->userIDWords[userOffset], 0xFFFF,
				PIC14_READDATA, PIC14_LOADDATA,
				PIC14_BEGINPROGONLY, PIC14_ENDPROGA);
			if(i) {  /* PICClose(); return(1); */ }
		}
		else if (addr >= configWordsBase && addr <= configWordsCeil)
		{
			int configOffset = addr - configWordsBase;
			// Config programming must work or we will have a chip with incorrect
			// settings for things like clock, code/data protection, pin assignments, ...
			i = WriteWord14(port, hexFile->configWords[configOffset], 0xFFFF,
				PIC14_READDATA, PIC14_LOADDATA,
				PIC14_BEGINPROGONLY, PIC14_ENDPROGA);
			if (i) { PICClose(port); return addr+1; }
		}

		// Move to next address in configuration space
		i  = TxCmd(port, EVA_PIC14_CMD, PIC14_INCADDR | 0xFFFF, &j);
		if(i) { PICClose(port); return addr+1; }

		if (statusIF && !statusIF->UpdateProgress(addr, device->configSize))
		{
			PICClose(port);
			return 1;
		}
	}
	PICClose(port);

	return 0;
}

int PIC16F88Program::Program(CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	int res;

	const PICDevice *device = hexFile->Processor;

    if (!SUCCEEDED(ResetProgrammer(port)))
	{
		// Failed to get the programmer into a known good state.
		return 1;
	}

	// Send the contents of the hex file to the PIC
	if ((res = WriteCode(device, port, hexFile, statusIF)) != 0) return res;
	if ((res = WriteData(device, port, hexFile, statusIF)) != 0) return res;
	if ((res = WriteConfig(device, port, hexFile, statusIF)) != 0) return res;

	// Ensure everything shut down
	PICClose(port);

	return 0;
}

int PIC16F88Program::PartialProgram(CIOPort *port, PICHexFile *hexFile, bool codeFlag, bool dataFlag, bool configFlag, CProgrammerProgressIF *statusIF) const
{
	int res;

	const PICDevice *device = hexFile->Processor;

    if (!SUCCEEDED(ResetProgrammer(port)))
	{
		// Failed to get the programmer into a known good state.
		return 1;
	}

	// Send the contents of the hex file to the PIC
	if (codeFlag && (res = WriteCode(device, port, hexFile, statusIF)) != 0) return res;
	if (dataFlag && (res = WriteData(device, port, hexFile, statusIF)) != 0) return res;
	if (configFlag && (res = WriteConfig(device, port, hexFile, statusIF)) != 0) return res;

	// Ensure everything shut down
	PICClose(port);

	return 0;
}


int PIC16F88Program::Erase(CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF) const
{
	int i;
	uint j;

    if (!SUCCEEDED(ResetProgrammer(port)))
	{
		// Failed to get the programmer into a known good state.
		return 1;
	}

	PICOpen(port);

	// Necessary for 16F88 (TBD: other chips as well?).  Note AN910 describes
	// this instruction as necessary for "rare" flash chips.
	i  = TxCmd(port, EVA_PIC14_CMD, PIC14_CHIPERASE | 0xFFFF, &j);

	PICClose(port);

	return 0;
}

int PIC16F88Program::PartialErase(CIOPort *port, PICHexFile *hexFile, bool bootFlag, bool codeFlag, bool dataFlag, bool configFlag, CProgrammerProgressIF *statusIF) const
{
	int i;
	uint j;

    if (!SUCCEEDED(ResetProgrammer(port)))
	{
		// Failed to get the programmer into a known good state.
		return 1;
	}

	PICOpen(port);
	if (codeFlag)
	{
		i  = TxCmd(port, EVA_PIC14_CMD, PIC14_BULKERASEPROG | 0xFFFF, &j);

		if (i) { PICClose(port); return(1); }

		i  = TxCmd(port, EVA_PIC14_CMD, PIC14_BEGINPROG | 0xFFFF, &j);

		if (i) { PICClose(port); return(1); }

		::Sleep(3);
	}
	if (dataFlag)
	{
		i  = TxCmd(port, EVA_PIC14_CMD, PIC14_BULKERASEE2 | 0xFFFF, &j);
		if (i) { PICClose(port); return(1); }

		i  = TxCmd(port, EVA_PIC14_CMD, PIC14_BEGINPROG | 0xFFFF, &j);

		if (i) { PICClose(port); return(1); }

		::Sleep(3);
	}
	if (configFlag)
	{
		// TBD: erase config bits.  This may require code/data erase if code protection
		// was enabled.
	}
	PICClose(port);

	return 0;
}
