#pragma once

#include <vector>
#include <list>
#include <map>
#include "PIC/PICDevice.h"
#include "String/TextList.h"

#if !defined(ushort)
#define ushort unsigned short
#endif
#if !defined(uint)
#define uint unsigned int
#endif

// Types used by PICHexFile for disassembly of code

	// Each entry defines a particular pattern of arguments for an instruction.
	// These are taken from the datasheet for 16 bit PIC processors.
	enum opcodeArgType {
		noArgOp,      // 16 - [opcode<15:0>]
					  // 14 - [opcode<13:0>]
		byteFileOp,   // 16 - [opcode<15:10>|d<9>|a<8>|f<7:0>]
					  // 14 - [opcode<13:8>|d<7>|f<6:0>]
		byteFileOpA,  // 16 - [opcode<15:9>|a<8>|f<7:0>]
					  // 14 - [opcode<13:7>|f<6:0>]
		byteToByte,   // 16 - [opcode<15:12>|f<11:0>]+[0xF<15:12>|f<11:0>]
		bitFileOp,    // 16 - [opcode<15:12>|b<11:9>|a<8>|f<7:0>]
					  // 14 - [opcode<13:10>|b<9:7>|f<6:0>]
		literalOp,    // 16 - [opcode<15:8>|k<7:0>]
					  // 14 - [opcode<13:8>|k<7:0>]
		literalOp4,   // 16 - [opcode<15:4>|k<3:0>]
		literalOp12,  // 16 - [opcode<15:6>|f<5:4>|k<3:0>]+[0xf0<15:8>|k<7:0>]
		controlOp20,  // 16 - [opcode<15:8>|n<7:0>]+[0xF|n<11:0>]
		controlOp20S, // 16 - [opcode<15:9>|S<8>|n<7:0>]+[unset?<15:12>|n<11:0>]
		controlOp11,  // 16 - [opcode<15:11>|n<10:0>]
					  // 14 - [opcode<13:11>|n<10:0>]
		controlOp8,   // 16 - [opcode<15:8>|n<7:0>]
		controlOpS,   // 16 - [opcode<15:1>|s<0>]
	};

	typedef struct s_instructionSetEntry {
	  TCHAR *name;
	  enum opcodeArgType opCodeType;
	  ushort opCodeMask;
	  ushort opCode;
	} instructionSetEntry;

	// Keep track of all addresses that are targets of call/goto
	// statements.  The value stored along with the address is
	// a simple index for generating unique target labels.
	typedef std::map<uint, uint> callTargetArray;
	typedef std::pair<uint, uint> callTargetEntry;

/// <summary>
/// PICHexFile holds all of the information to write to a PIC microprocessor (or
/// conversly holds all the information we read from a PIC).  TBD: change public
/// members to properties and/or find a way to grant just PICHexFileIO access to
/// the members.
/// </summary>
class PICHexFile
{
public:
	// Read a hex file and produce a disassembly listing
	CTextList *Disassemble(bool showAddr = false);

	// Construct and initialize information for this processor.
	PICHexFile(const PICDevice *proc);
	~PICHexFile(void);

	// File we got the hex info from
	LPTSTR fileName;

	// Processor
	const PICDevice *Processor;

	// Configuration data
	ushort *configWords;

	// User ID data
	ushort *userIDWords;

	// Device ID words
	ushort *deviceID;

	// Program data
	ushort *programWords;	// Entire contents of PIC program (instruction) memory
	uint programWordsBase;	// Lowest used address of a word in program memory
	uint programWordsCeil;	// Highest used address of a word in program memory
	std::vector<bool> programWordsUsed;	// Flags for which addresses actually have contents

	// EEPROM data
	ushort *EEPROMData;
	uint EEPROMDataBase;		// Lowest used address of a word in data memory
	uint EEPROMDataCeil;		// Highest used address of a word in data memory

	int LoadFromFile(LPCTSTR hexFileName);

protected:
	// Fill the buffer with the text equivalent of the opcode "memWord".  The
	// return value is the number of words associated with this instruction
	// (only used for 16 bit devices).
	int emitOpcode(uint addr, ushort memWord, TCHAR *buf, callTargetArray &callTargets);
	// Add a string to the end of a list of strings.  The string "str" is
	// copied before being put in the list.
	void appendStr(TCHAR *str, CTextList *lineList);
	// Find all addresses that are targets of call/goto in a 14 bit PIC
	void findCallTargets(uint startAddr, uint endAddr, callTargetArray &callTargets);

	// Instruction set tables for 14 and 16 bit PIC processors
	static instructionSetEntry PIC14Instructions[];
	static instructionSetEntry PIC16Instructions[];


};
