#include "StdAfx.h"
#include <Mmsystem.h>
#include "PICProgram.h"

const int PICProgram::RETRY_LIMIT = 25;
#if defined(_WIN32_WCE)
const byte PICProgram::CMD_BLOCKSIZE = 32;
#else
const byte PICProgram::CMD_BLOCKSIZE = 64;
#endif

const byte PICProgram::MAX_PACKET_LENGTH = 127;
const byte PICProgram::ETX = 0x04;
const byte PICProgram::DLE = 0x05;
const byte PICProgram::STX = 0x55;

byte PICProgram::outBuf[128];
byte PICProgram::inBuf[128];

const int PICProgram::IO_OK = 0;				// No problems with IO
const int PICProgram::ERROR_IO_TIMEOUT = -1;	// Failed to Rx/Tx within timeout period
const int PICProgram::ERROR_IO_CHECKSUM = -2;	// Checksum failure in data packet
const int PICProgram::ERROR_IO_DATALEN = -3;	// Length of packet was wrong
const int PICProgram::ERROR_IO_BADHEXFILE = -4;	// Null or invalid hex file parameter
const int PICProgram::ERROR_IO_BUFFER = -5;		// Data buffer too small for response (overrun)
const int PICProgram::ERROR_IO_UNFINISHED = -6;	// Got an <STX> in the middle of response
const int PICProgram::ERROR_IO_NOSTX = -7;		// Didn't get a <STX> at beginning of response
const int PICProgram::ERROR_IO_BADADDRESS = -8;	// Addresses out of range for the PIC
const int PICProgram::ERROR_IO_UNALIGNED = -9;	// Attempt to write program memory with misaligned data (not on 4 word boundary)

const byte PICProgram::EVA_VERSION_CMD      = 0x01; // Version of firmware in programmer
const byte PICProgram::EVA_SET_LINES_CMD    = 0x02; // Set all signal lines/LEDs on/off
const byte PICProgram::EVA_VPP_CMD          = 0x03; // Turn Vpp on/off
// (unused - null command)
const byte PICProgram::EVA_PRGSEL_CMD       = 0x05; // Turn PRG_SEL (dsPIC select) on/off
const byte PICProgram::EVA_VDD_CMD          = 0x06; // Turn Vdd on/off
const byte PICProgram::EVA_CLKDAT_CMD       = 0x07; // Set the ICSP CLK and DAT lines
const byte PICProgram::EVA_READ_LINES_CMD   = 0x08; // Read current value of signal/LED lines
const byte PICProgram::EVA_READ_REGS_CMD    = 0x09; // Read value of programmer registers
const byte PICProgram::EVA_ECHO_CMD         = 0x0a; // Send back the input
const byte PICProgram::EVA_BEGINPROG_CMD    = 0x0b; // Enter prog mode: lower DAT/CLK, raise Vdd and Vpp
const byte PICProgram::EVA_ENDPROG_CMD      = 0x0c; // Exit prog mode: lower Vdd/Vpp, tristate CLK/DAT
const byte PICProgram::EVA_PIC14_CMD        = 0x0d; // Send command to 14 bit PIC
const byte PICProgram::EVA_PIC14R_CMD       = 0x0e; // Send cmd to 14 bit PIC, read response
// (unused - null command)
const byte PICProgram::EVA_READ_ANALOG_CMD  = 0x10; // Read the current value of the A/D converter
const byte PICProgram::EVA_SET_ANALOG_CMD   = 0x11; // Set desired voltage bounds
const byte PICProgram::EVA_PIC16_ADDR_CMD   = 0x12; // Set the current address of a 16 bit PIC
const byte PICProgram::EVA_PIC16_CMD        = 0x13; // Send command to 16 bit PIC
const byte PICProgram::EVA_PIC16R_CMD       = 0x14; // Send cmd to 16 bit PIC, read response
const byte PICProgram::EVA_PIC16W_CMD       = 0x15; // Send write cmd to 16 bit PIC, hold CLK for time P9 (1-5 msec)
const byte PICProgram::EVA_READ_MASKS_CMD   = 0x16; // Read current value of TRIS registers
const byte PICProgram::EVA_READ_PORTS_CMD   = 0x17; // Read current value of PORT registers
const byte PICProgram::EVA_READ_PWM_CMD     = 0x18; // Read current PWM duty cycle
const byte PICProgram::EVA_PIC14RB_CMD      = 0x19; // Read block of words from 14 bit PIC
const byte PICProgram::EVA_PIC14WB_CMD      = 0x1a; // Write block of words to 14 bit PIC
const byte PICProgram::EVA_PIC14RBE_CMD     = 0x1b; // Read block of data from 14 bit PIC
const byte PICProgram::EVA_PIC14WBE_CMD     = 0x1c; // Write block of data to 14 bit PIC
const byte PICProgram::EVA_PIC16RB_CMD      = 0x1d; // Read block of words from 16 bit PIC
const byte PICProgram::EVA_PIC16WB_CMD      = 0x1e; // Write block of words to 16 bit PIC
const byte PICProgram::EVA_PIC16RBE_CMD     = 0x1f; // Read a block of words from a 16 bit PIC EEPROM
const byte PICProgram::EVA_PIC16WBE_CMD     = 0x20; // Write a  block of words to 16 bit PIC EEPROM
const byte PICProgram::EVA_PIC16WBUF_CMD    = 0x21; // Write one buffer of bytes to a 16 bit PIC
const byte PICProgram::EVA_PIC16WEBYTE_CMD  = 0x22; // Write one byte to a 16 bit PIC EEPROM
const byte PICProgram::EVA_PIC16EBUF_CMD    = 0x23; // Erase one window of bytes on a 16 bit PIC
const byte PICProgram::EVA_PIC24_CMD        = 0x24; // Send command to 24 bit PIC
const byte PICProgram::EVA_PIC24R_CMD       = 0x25; // Read 16 bit register from 24 bit PIC
const byte PICProgram::EVA_PIC24RBUF_CMD    = 0x26; // Read one buffer of instructino words from 24 bit PIC
const byte PICProgram::EVA_PIC24WBUF_CMD    = 0x27; // Write one buffer of instruction words to 24 bit PIC
const byte PICProgram::EVA_PIC24WBUF_SETUP  = 0x28; // Setup to write one buffer to 24 bit PIC
const byte PICProgram::EVA_PIC24WBUF_FINISH = 0x29; // Complete write of one buffer to 24 bit PIC
// The following commands are specific to the 18F2455 version of the firmware - they are
// used to read/write portions of the firmware in the running programmer.  The values of these
// commands must be above the standard commands above.
const byte PICProgram::EVA_IREAD_CMD        = 0x41; // Read programmer memory
const byte PICProgram::EVA_IWRITE_CMD       = 0x42; // Write programmer memory
const byte PICProgram::EVA_IERASE_CMD       = 0x43; // Erase programmer memory
const byte PICProgram::EVA_IRUN_CMD         = 0x44; // Run code in internal programmer memory
const byte PICProgram::EVA_IDEBUG_PRINT_CMD = 0x45; // Turn debug printing on/off

// ICSP commands - bit patterns taken from 16F and 18F programming specifications.
// The top byte of each value is the real bit pattern - they are left shifted 16
// bits to make masking easier in the programming code.
const uint PICProgram::PIC14_LOADCFG       = 0x000000;
const uint PICProgram::PIC14_LOADDATA      = 0x020000;
const uint PICProgram::PIC14_READDATA      = 0x040000;
const uint PICProgram::PIC14_INCADDR       = 0x060000;
const uint PICProgram::PIC14_BEGINPROG     = 0x080000;
const uint PICProgram::PIC14_ENDPROG       = 0x0E0000;
const uint PICProgram::PIC14_ENDPROGA      = 0x170000;
const uint PICProgram::PIC14_BEGINPROGONLY = 0x180000;
const uint PICProgram::PIC14_LOADE2DATA    = 0x030000;
const uint PICProgram::PIC14_READE2DATA    = 0x050000;
const uint PICProgram::PIC14_BULKERASEPROG = 0x090000;
const uint PICProgram::PIC14_BULKERASEE2   = 0x0B0000;
const uint PICProgram::PIC14_CHIPERASE     = 0x1F0000;

const uint PICProgram::PIC16_CORE           = 0x000000;
const uint PICProgram::PIC16_SHIFT_TBLAT    = 0x020000;
const uint PICProgram::PIC16_TBLRD          = 0x080000;
const uint PICProgram::PIC16_TBLRD_POSTINC  = 0x090000;
const uint PICProgram::PIC16_TBLRD_POSTDEC  = 0x0A0000;
const uint PICProgram::PIC16_TBLRD_PREINC   = 0x0B0000;
const uint PICProgram::PIC16_TBLWRT         = 0x0C0000;
const uint PICProgram::PIC16_PBLWRT_POSTINC = 0x0D0000;
const uint PICProgram::PIC16_TBLWRT_POSTDEC = 0x0E0000;
const uint PICProgram::PIC16_STARTPROG      = 0x0F0000;

int errIndex;

int PICProgram::ReadResponse(CIOPort *port, byte cmd, uint *resultLen, byte *fromPICData)
{
	// Read the response <STX><STX><CMD><DLEN><...Data...><ETX>
	errIndex = -1;
	int dataCount = 0;

	if (port->m_Serial.IsOpen())
	{
		int inByte, inByte0, inByte1;
		bool escFlag = false;

		if ((inByte0 = RxByte(port)) < 0)
			return ERROR_IO_TIMEOUT;
		if (inByte0 != STX)
			return ERROR_IO_NOSTX;
		errIndex--;
		if ((inByte1 = RxByte(port)) < 0)
			return ERROR_IO_TIMEOUT;
		if (inByte1 != STX)
			return ERROR_IO_NOSTX;

		do {
			errIndex--;
			if ((inByte = RxByte(port)) < 0)
				return ERROR_IO_TIMEOUT;
			if (escFlag)
			{
				// Last byte was a DLE - we want to store this byte regardless
				// of it's value.
				fromPICData[dataCount++] = inByte;
				escFlag = false;
			}
			else if (inByte == DLE)
			{
				escFlag = true;
			}
			else if (inByte == STX)
			{
				// Bad - we got an STX in the middle of a response.
				return ERROR_IO_UNFINISHED;
			}
			else if (inByte == ETX)
			{
				// Done reading the response
				break;
			}
			else
			{
				fromPICData[dataCount++] = inByte;
			}
		} while (dataCount < MAX_PACKET_LENGTH);

		if (dataCount >= MAX_PACKET_LENGTH)
			return ERROR_IO_DATALEN;
	}
	else if (port->m_USB.IsOpen())
	{
		// Read the response from the USB device
		byte reportID = 0; // (cmd == EVA_INTERNAL_CMD) ? 2 : 0;
		dataCount = port->m_USB.Read(reportID, fromPICData, MAX_PACKET_LENGTH);
	}

	// TBD: The second to last byte that came in was the
	// checksum.  Add a check to make sure data length and
	// checksum came out right.

	*resultLen = dataCount;

	return IO_OK;
}

int PICProgram::SendPacket(CIOPort *port, byte cmd, uint dataLen, byte *toPICData,
	uint maxResultLen, uint *resultLen, byte *fromPICData, uint msecTimeout)
{
	*resultLen = -1;
	byte checksum;
	byte bDataLen = (byte)(dataLen & 0xff);
	if (bDataLen != dataLen || bDataLen > MAX_PACKET_LENGTH)
		return -1;

	errIndex = 0;

	// Send two STX bytes
	errIndex = 1;

	int bind = 0;
	if (port->m_Serial.IsOpen())
	{
		// Build the output packet and send in a single WriteFile() call.
		// TBD: Doesn't work using internal 8MHz clock in PIC (perhaps can't
		// handle speed that bits come in).  Try to fix in firmware.
		bind = 2;
		outBuf[0] = STX; outBuf[1] = STX;
		if (cmd == STX || cmd == DLE || cmd == ETX)
			outBuf[bind++] = DLE;
		outBuf[bind++] = cmd;
		if (dataLen == STX || dataLen == DLE || dataLen == ETX)
			outBuf[bind++] = DLE;
		outBuf[bind++] = dataLen;

		// Send the packet (escaping special values as we go)
		byte outByte;
		checksum = cmd + dataLen;
		for (uint i=0;i<dataLen;i++)
		{
			outByte = *toPICData++;
			errIndex++;
			if (outByte == STX || outByte == DLE || outByte == ETX)
				outBuf[bind++] = DLE;
			outBuf[bind++] = outByte;
			checksum += outByte;
		}

		// Send the checksum
		byte checksumOut = -checksum;
		errIndex++;
		if (checksumOut == STX || checksumOut == DLE || checksumOut == ETX)
			outBuf[bind++] = DLE;
		outBuf[bind++] = checksumOut;
		outBuf[bind++] = ETX;
		}
	else if (port->m_USB.IsOpen())
	{
		bind = dataLen + 2;
		outBuf[0] = cmd;
		outBuf[1] = dataLen;
		memcpy(&outBuf[2], toPICData, dataLen);
	}
	else
	{
		return -1;
	}

	// Send the command
	errIndex = 2;
	if (TxRxBuffer(port, bind, outBuf) <= 0)
		return ERROR_IO_TIMEOUT;

#if 0
	if (port->m_USB.IsOpen())
		::Sleep(2);
#endif

	uint endTime;
	uint startTime = ::timeGetTime();
	int res = IO_OK;
	do
	{
		res = ReadResponse(port, cmd, resultLen, fromPICData);
		endTime = timeGetTime();
		if (res != IO_OK)
		{
			if (msecTimeout > 0)
				::Sleep(1);
		}
	} while (((res != IO_OK) || (*resultLen == 0)) && ((endTime - startTime) < msecTimeout));
	
	if ((res == IO_OK) && (*resultLen > 0))
		return 0;
	else
		return -1;
}

int PICProgram::ResetProgrammer(CIOPort *port)
{
	uint j;
	port->Purge();
	TxByte(port, ETX);
	port->Flush();
	int res = TxCmd(port, EVA_ENDPROG_CMD, 0, &j);    // Pull Vpp low, tristate CLK & DAT
	if (res != 0)
	{
		// Try purging the comm port and sending again.
		// port->Flush();
		port->Purge();
		res = TxCmd(port, EVA_ENDPROG_CMD, 0, &j);    // Pull Vpp low, tristate CLK & DAT
	}

	return res;
}

int PICProgram::SetTablePointer(CIOPort *port, uint addr)
{
	uint i, resLen = 0;
	byte TBLPTRU = (addr & 0xFF0000) >> 16;
	byte TBLPTRH = (addr & 0x00FF00) >> 8;
	byte TBLPTRL = (addr & 0x0000FF);

#if 1
	byte cmdBuf[4], resBuf[128];
	cmdBuf[0] = TBLPTRL;
	cmdBuf[1] = TBLPTRH;
	cmdBuf[2] = TBLPTRU;
	cmdBuf[3] = 0x00;
	i = SendPacket(port, EVA_PIC16_ADDR_CMD, 3, cmdBuf, sizeof(resBuf), &resLen, resBuf, 4);
#else
	i  = TxCmd(port, EVA_PIC16_CMD, PIC16_CORE | (0x0E00 | TBLPTRU), &resLen);
	i += TxCmd(port, EVA_PIC16_CMD, PIC16_CORE | 0x6EF8, &resLen);
	i += TxCmd(port, EVA_PIC16_CMD, PIC16_CORE | (0x0E00 | TBLPTRH), &resLen);
	i += TxCmd(port, EVA_PIC16_CMD, PIC16_CORE | 0x6EF7, &resLen);
	i += TxCmd(port, EVA_PIC16_CMD, PIC16_CORE | (0x0E00 | TBLPTRL), &resLen);
	i += TxCmd(port, EVA_PIC16_CMD, PIC16_CORE | 0x6EF6, &resLen);
	i += TxCmd(port, EVA_PIC16_CMD, PIC16_CORE, &resLen);
#endif

	return i;
}

int PICProgram::PICOpen(CIOPort *port, byte VppMask, bool PIC14BitFlag)
{
	uint i, j;

	// Pull Vpp low, tristate CLK & DAT
	i = TxCmd(port, EVA_ENDPROG_CMD, 0, &j);

	// Pull Vpp high, CLK & DAT set low
	i = TxCmd(port, EVA_BEGINPROG_CMD, (VppMask << 16), &j);

	// If using inductor based hardware?
	// ::Sleep(200);
#if 0
	// ARE: Why do extra reads following PicOpen help things?
	for (int tItr=0;tItr<4;tItr++)
	{
		if (PIC14BitFlag)
			i  = TxCmd(port, EVA_PIC14R_CMD, PIC14_READDATA | 0xFFFF, &j);
		else
		{
			// TBD; verify this PIC16 read command works
			// i  = TxCmd(port, EVA_PIC16R_CMD, PIC16_TBLRD, &j);
		}
	}
	// port->Flush();
#endif

	// Successful return
	return 0;
}

int PICProgram::PICClose(CIOPort *port)
{
	int i;
	uint j;

	port->Flush();

	i = TxCmd(port, EVA_ENDPROG_CMD, 0, &j);

	// Successful
	return 0;
}

void
printPacketByte(byte b)
{
	if (b == PICProgram::STX || b == PICProgram::DLE || b == PICProgram::ETX)
		printf("%02x ", 5);
	printf("%02x ", b);
}

void
printCmdPacket(int err, int index, byte EvaCmd, byte b0, byte b1, byte b2)
{
	byte checksum = EvaCmd + 3 + b0 + b1 + b2;
	byte checksumOut = -checksum;
	printf("\nP(%d:%d): 55 55 ", err, index);
	printPacketByte(EvaCmd);
	printPacketByte(3);
	printPacketByte(b0);
	printPacketByte(b1);
	printPacketByte(b2);
	printPacketByte(checksumOut);
	printf("04\n");
}

void
printPacket(uint len, byte *inBuf)
{
	bool DLEFlag = false;
	printf("\nP:");
	for (uint i=0;i<len;i++)
	{
		printf("%02x ", inBuf[i]);
		if (inBuf[i] == PICProgram::DLE)
			DLEFlag = true;
		else if (!DLEFlag && inBuf[i] == PICProgram::ETX)
			break;
		else
			DLEFlag = false;
	}
	printf("\n");
}

int PICProgram::TxCmd(CIOPort *port, byte EvaCmd, uint data, uint *in)
{
	byte cmdBuf[4];
	cmdBuf[0] = (byte)((data & 0xFF0000) >> 16);   // ICSP command   (or ADDRL)
	cmdBuf[1] = (byte)(data & 0xFF);               // Data low byte  (or ADDRH)
	cmdBuf[2] = (byte)((data & 0xFF00) >> 8);      // Data high byte (or ADDRU)
	cmdBuf[3] = 0;
	int i = 0;
	uint resLen = 0;
#if 1
	int res = SendPacket(port, EvaCmd, 4, cmdBuf, MAX_PACKET_LENGTH, &resLen, inBuf, 100);
#else
	int res;
	for (i=0;i<4;i++)
	{
		// port->Purge();
		if ((res = SendPacket(port, EvaCmd, 3, cmdBuf, MAX_PACKET_LENGTH, &resLen, inBuf)) == 0)
			break;
		// Show the packet that failed
		printCmdPacket(res, errIndex, EvaCmd, cmdBuf[0], cmdBuf[1], cmdBuf[2]);
	}
#endif
	// if (res == 0) printPacket(resLen, inBuf);

	if (res)
	{
		port->Purge();
		*in = 0;
	}
	else
		*in = ((inBuf[4] << 8) | inBuf[3]);

	return res;
}

int PICProgram::SetICSPLevels(CIOPort *port, bool CLKFlag, bool DATFlag)
{
	uint resLen = 0;
	outBuf[0] = (CLKFlag ? 0x01 : 0) | ((DATFlag ? 0x02 : 0x00) << 1);
	return SendPacket(port, EVA_CLKDAT_CMD, 1, outBuf, MAX_PACKET_LENGTH, &resLen, inBuf);
}

int PICProgram::WriteWord14(CIOPort *port, uint val, ushort mask, uint readCmd, uint loadCmd,
							uint progCmd, uint endProgCmd, int limit)
{
	int i, retry;
	uint j, k, r;

	// prepare val
	// k = (val & 0x3fff) << 1;
	k = val;

	// try writing
	retry=0;
	while(retry<limit)
	{
		// read current data
		i = TxCmd(port, EVA_PIC14R_CMD, readCmd | 0xFFFF, &r);
		if (i)
			return(i);

		// r=(r>>1) & 0x3fff; r<<=1;

		// If the memory already has this word, then there is no need
		// to proceed with the write.
		if ((r & mask) == (k & mask))
			break;

		// Load the data to the PIC prior to the write cycle
		i = TxCmd(port, EVA_PIC14R_CMD, loadCmd | k, &r);
		if(i)
			return(i);

		// Begin the programming cycle
		i = TxCmd(port, EVA_PIC14_CMD, progCmd | 0xFFFF, &j);
		if (i)
			return(i);
		// actualWrites++;
		// Wait TPROG1 for program cycle to complete
		Sleep(3);

		// Some parts need and end programming command.  If this is
		// one of them, then issue it
		if (endProgCmd != 0)
		{
			i = TxCmd(port, EVA_PIC14_CMD, endProgCmd | 0xFFFF, &j);
			if (i)
				return(i);
		}

		// inc retry counter
		retry++;
	}

	// failed?
	if(retry==limit)
	{
		printf("\n failure: %04lx %04lx\n",k>>1,r>>1);
		return 1;
	}

	// Successfully wrote the value
	return 0;
}

int PICProgram::WriteWord16(CIOPort *port, ushort *vals, int limit)
{
	int i;
	uint r;

	// Load the write buffer at the current address
	i  = TxCmd(port, EVA_PIC16_CMD, PIC16_PBLWRT_POSTINC | vals[0], &r);
	i += TxCmd(port, EVA_PIC16_CMD, PIC16_PBLWRT_POSTINC | vals[1], &r);
	i += TxCmd(port, EVA_PIC16_CMD, PIC16_PBLWRT_POSTINC | vals[2], &r);
	i += TxCmd(port, EVA_PIC16_CMD, PIC16_STARTPROG | vals[3], &r);

	// At this point we need to emit a NOP command and hold the CLK line high
	// for the duration of the write.

	i += TxCmd(port, EVA_PIC16W_CMD, PIC16_CORE | 0x2000, &r);

	// Increment the address pointer (presumes that WriteWord16 only incremented 3 times
	i += TxCmd(port, EVA_PIC16_CMD, PIC16_PBLWRT_POSTINC, &r);

	// Successfully wrote the value
	return i;
}

int PICProgram::RxByte(CIOPort *port)
{
	int i;
	DWORD len;
	unsigned char inChar;
	i = port->m_Serial.Read(&inChar, 1, &len);
// printf("R%02x ", (i != -1 && len > 0) ? inChar : (unsigned char)i);
	if (i != -1 && len > 0)
		return inChar;
	else
	{
#if 1
		// ::Sleep(1);
		// port->Flush();
		unsigned char inChar2;
		i = port->m_Serial.Read(&inChar2, 1, &len);
		if (i != -1 && len > 0)
			return inChar2;
		else
			return -1;
#else
		return -1;
#endif
	}
}

int PICProgram::TxByte(CIOPort *port, int ch)
{
	static int cntr = 0;
	char outCh = (char)ch;
	DWORD wFlag;
	BOOL res = port->m_Serial.Write(&outCh, 1, &wFlag);

	if (res == -1 || wFlag != 1)
	{
		// Try once more...
		res = port->m_Serial.Write(&outCh, 1, &wFlag);
	}

	return res;
}

int PICProgram::TxRxBuffer(CIOPort *port, int len, byte *buffer)
{
	DWORD wFlag;
	BOOL res0;

	// Send the buffer
#if 1
	if (port->m_Serial.IsOpen())
	{
#if 1
		// Entire packet at once
		res0 = port->m_Serial.Write(buffer, len, &wFlag);
		if (res0 <= 0 || wFlag != len)
			return res0;
#else
		// One byte at a time
		for (int i=0;i<len;i++)
		{
			res0 = port->m_Serial.Write(&buffer[i], 1, &wFlag);
			// Sleep(1);
			if (res0 <= 0 || wFlag <= 0)
				return res0;
		}
#endif
	}
	else if (port->m_USB.IsOpen())
	{
		byte reportID = 0; // (buffer[0] == EVA_INTERNAL_CMD) ? 2 : 0;
		if (port->m_USB.Write(reportID, buffer, len))
			return 1;
		else
			return -1;
	}
#else
	for (int i=0;i<len;i++)
	{
		res0 = port->Write(&buffer[i], 1, &wFlag);
		// Sleep(1);
		if (res0 <= 0 || wFlag <= 0)
			return res0;
	}
#endif
	// printf("tt: "); for (int q0=0;q0<len;q0++) printf("%02x ", buffer[q0]);
#if 0
	// Read the response
	// ::Sleep(1);
#if 1
	bool flag = true;
	int ch1;
	// printf("rr: ");
	for (int q1=0;q1<len;q1++)
	{
		ch1 = RxByte(port);
		if (ch1 == -1)
			flag = false;
		// printf("%02x ", (unsigned char)ch1);
	}
	// printf("\n");
	return (flag ? 1 : -1);
#else
	DWORD inLen;
	unsigned char tbuf[1024];
	int i;
	i = port->Read(tbuf, 2*len, &inLen);
	if (i != -1 && inLen > 0)
	{
		printf("rr(%d): ", inLen); for (int q1=0;q1<inLen;q1++) printf("%02x ", tbuf[q1]);
	}
	else
	{
		printf("rr(%d): failed ", inLen);
	}
	if (len != inLen)
	{
#if 1
		printf("n: ");
		while (RxByte(port) != -1)
			;
#else
		printf("n: %02x ", (unsigned char)RxByte(port));
#endif
	}
	printf("\n");

	if (i != -1 && len > 0)
		return 1;
	else
		return -1;
#endif
#else
	return 1;
#endif
}

// Simple delay routine
void PICProgram::Delay(int p)
{
	int i, j;
	j = 0;
	for (i=0;i<p*10;i++)
		j++;
}

// Read 8 words from address 0x2000 (the location of configuration words
// in 14 bit PIC devices).
bool PICProgram::readPIC14Config(CIOPort *port, ushort *buf)
{
	uint i, j;

	// Enter configuration memory
	i = TxCmd(port, EVA_PIC14R_CMD, PIC14_LOADCFG | 0xFFFF, &j);
	if(i) { PICClose(port); return false; }

	byte outBuf[4], inBuf[128];
	uint resLen, blockSize = 16;
	outBuf[0] = 0;
	outBuf[1] = 0;
	outBuf[2] = 0;
	outBuf[3] = blockSize;
	i = SendPacket(port, PICProgram::EVA_PIC14RB_CMD,  4,
		outBuf, sizeof(inBuf), &resLen, inBuf);
	if (i) { PICClose(port); return false; }

	for (int addr=0;addr<8;addr++)
	{
		ushort wordVal = (inBuf[2*addr+7] << 8) | inBuf[2*addr+6];
		buf[addr] = wordVal;
	}

	return true;
}

// Read the device ID from a 16 bit PIC
bool PICProgram::readPIC16DevID(CIOPort *port, ushort *devID)
{
	uint b0, b1;

	SetTablePointer(port, 0x3FFFFE);
	TxCmd(port, EVA_PIC16R_CMD, PIC16_TBLRD_POSTINC, &b0);
	TxCmd(port, EVA_PIC16R_CMD, PIC16_TBLRD_POSTINC, &b1);

	b0 &= 0xff;
	b1 &= 0xff;
	if ((b0 != 0xff && b0 != 0x80 && b0 != 0x00)
	 || (b1 != 0xff && b1 != 0x80 && b1 != 0x00))
	{
		*devID = (b1 << 8) | b0;
		return true;
	}
	else
	{
		*devID = 0x0000;
		return false;
	}
}

bool PICProgram::readPIC24DevID(CIOPort *port, ushort *devID)
{
	int i = 0;
	uint j, res;

	// Read the device ID registers at address 0xFF0000:0xff0002
	// Exit reset vector
	i  = TxCmd(port, EVA_PIC24_CMD, 0x000000, &j);  // nop
	i += TxCmd(port, EVA_PIC24_CMD, 0x000000, &j);  // nop
	i += TxCmd(port, EVA_PIC24_CMD, 0x040100, &j);  // goto 0x100
	i += TxCmd(port, EVA_PIC24_CMD, 0x000000, &j);  // nop

	// Initialize TBLPAG, the read pointer (W6), and the write
	// pointer (W7) for TBLRD instruction
	i = TxCmd(port, EVA_PIC24_CMD, 0x200ff0, &j);  // mov 0xff, W0
	i = TxCmd(port, EVA_PIC24_CMD, 0x880190, &j);  // mov W0, TBLPAG
	i = TxCmd(port, EVA_PIC24_CMD, 0xeb0300, &j);  // clr W6
	i = TxCmd(port, EVA_PIC24_CMD, 0xeb0380, &j);  // clr W7


	// Read the two words of the device ID
	for (int addr=0;addr<2;addr++)
	{
		// Read the configuration register and write it to the VISI register
		// (located at 0x784)
		i += TxCmd(port, EVA_PIC24_CMD, 0xba0bb6, &j);  // TBLRDL [W6++], [W7]
		i += TxCmd(port, EVA_PIC24_CMD, 0x000000, &j);  // nop
		i += TxCmd(port, EVA_PIC24_CMD, 0x000000, &j);  // nop
		i += TxCmd(port, EVA_PIC24_CMD, 0x883c20, &j);  // mov W0, VISI
		i += TxCmd(port, EVA_PIC24_CMD, 0xeb0380, &j);  // clr W7

		// Output the VISI register using the REGOUT command
		i += TxCmd(port, EVA_PIC24R_CMD, 0x000000, &res); // Read register
		devID[addr] = res;
		i += TxCmd(port, EVA_PIC24_CMD, 0x000000, &j);  // nop

		// Reset device internal PC
		i += TxCmd(port, EVA_PIC24_CMD, 0x040100, &j);  // goto 0x100
		i += TxCmd(port, EVA_PIC24_CMD, 0x000000, &j);  // nop
	}

	if ((i != 0) || (devID[0] == 0) || (devID[0] == 0xffff))
		return false;
	else
		return true;

}

bool PICProgram::ReadDeviceID(CIOPort *port, uint *devID, uint *wordSize)
{
	bool res;
	ushort configBuf[32];

	// First try treating it as a 40 pin chip

	// See if it is a 14 bit device (PIC12F/PIC16F)
	PICOpen(port, 0x02);
	res = readPIC14Config(port, &configBuf[0]);
	PICClose(port);
	if (res)
	{
		ushort tdevID = configBuf[6];
		if (tdevID != 0xFFFF && tdevID != 0x0000)
		{
			*devID = tdevID;
			*wordSize = 14;
			return true;
		}
	}

	// See if it is a 24 bit device (dsPIC)
	PICOpen(port, 0x06);
	res = readPIC24DevID(port, configBuf);
	PICClose(port);
	if (res)
	{
		*devID = (configBuf[0] | (configBuf[1] << 16));
		*wordSize = 24;
		return true;
	}

	// See if it is a 16 bit device (PIC18F)
	PICOpen(port, 0x02);
	res = readPIC16DevID(port, configBuf);
	PICClose(port);
	if (res)
	{
		*devID = configBuf[0];
		*wordSize = 16;
		return true;
	}


	// Now look at 8-28 pin devices (14 and 16 bit only - dsPICs should
	// all respond to the single query above).

	// See if it is a 14 bit device (PIC12F/PIC16F)
	PICOpen(port, 0x03);
	res = readPIC14Config(port, configBuf);
	PICClose(port);
	if (res)
	{
		ushort tdevID = configBuf[6];
		if (tdevID != 0xFFFF && tdevID != 0x0000)
		{
			*devID = tdevID;
			*wordSize = 14;
			return true;
		}
	}

	// See if it is a 24 bit device (dsPIC)
	PICOpen(port, 0x07);
	res = readPIC24DevID(port, configBuf);
	PICClose(port);
	if (res)
	{
		*devID = (configBuf[0] | (configBuf[1] << 16));
		*wordSize = 24;
		return true;
	}

	// See if it is a 16 bit device (PIC12F/PIC16F)
	PICOpen(port, 0x03);
	res = readPIC16DevID(port, configBuf);
	PICClose(port);
	if (res)
	{
		*devID = configBuf[0];
		*wordSize = 16;
		return true;
	}

	return false;
}
