#pragma once

#include "Serial/IOPort.h"
#include "PIC/PICHexFile.h"
#include "PIC/ProgrammerProgressIF.h"

#if !defined(uint)
#define uint unsigned int
#endif

#if !defined(byte)
#define byte unsigned char
#endif

// Base class for programming PIC chips using Eva programmer.  Only low level
// routines are supplied by this class.  Chip (or family of chip) specific
// routines must be implemented by a subclass of PICProgram.
class PICProgram
{
public:
	// Interface functions.  These must be implemented by a class that inherits
	// from PICProgram.  These functions are unique to families of processors
	// so there is no default implementation.

	// Read the contents of the chip into hexFile.
	virtual int Read(CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF = NULL) const = 0;
	// Read the contents of the chip into hexFile.
	virtual int PartialRead(CIOPort *port, PICHexFile *hexFile, bool codeFlag = false, bool dataFlag = false, bool configFlag = false, CProgrammerProgressIF *statusIF = NULL) const = 0;
	// Write the contents of hexFile to the chip.
	virtual int Program(CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF = NULL) const = 0;
	// Read the contents of the chip into hexFile.
	virtual int PartialProgram(CIOPort *port, PICHexFile *hexFile, bool codeFlag = false, bool dataFlag = false, bool configFlag = false, CProgrammerProgressIF *statusIF = NULL) const = 0;
	// Bulk erase the chip (TBD: add interface for erase of specific parts
	// of chip memory.
	virtual int Erase(CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF = NULL) const = 0;
	// Erase a selected part of memory.  (Note: bootFlag only makes sense for 16 bit PIC devices.)
	virtual int PartialErase(CIOPort *port, PICHexFile *hexFile, bool bootFlag = false, bool codeFlag = false, bool dataFlag = false, bool configFlag = false, CProgrammerProgressIF *statusIF = NULL) const = 0;
	// Verify that the chip exactly matches the contents of hexFile.  (TBD: account
	// for unimplemented bits in config words.)
	virtual int Verify(CIOPort *port, PICHexFile *hexFile, CProgrammerProgressIF *statusIF = NULL) const = 0;

	// Max number of times to try writing before abandoning
	static const int RETRY_LIMIT;

    // ICSP commands for 14 bit PIC chips (12F, 16F).  The
	// 14 bit parts use a 6 bit command.  The command bits
	// are right aligned into bits <23:16> of the 32 bit packet
	// sent to the programmer.  These definitions have the bits
	// aligned to allow simple bitwise "or" into the final
	// packet.

    static const uint PIC14_LOADCFG; //       = 0x000000;
    static const uint PIC14_LOADDATA; //      = 0x020000;
    static const uint PIC14_READDATA; //      = 0x040000;
    static const uint PIC14_INCADDR; //       = 0x060000;
    static const uint PIC14_BEGINPROG; //     = 0x080000;
    static const uint PIC14_ENDPROG; //       = 0x0E0000;
    static const uint PIC14_ENDPROGA; //      = 0x170000;
    static const uint PIC14_BEGINPROGONLY; // = 0x180000;
    static const uint PIC14_LOADE2DATA; //    = 0x030000;
    static const uint PIC14_READE2DATA; //    = 0x050000;
    static const uint PIC14_BULKERASEPROG; // = 0x090000;
    static const uint PIC14_BULKERASEE2; //   = 0x0B0000;
    static const uint PIC14_CHIPERASE;   //   = 0x1F0000;

    // ICSP commands for 16 bit PIC chips (18F).  The 16 bit
	// parts use a 4 bit command.  The command bits are right
	// aligned into bits <23:16> of the 32 bit packet sent
	// to the programmer.

	// Core Instruction (Shift in 16-bit instruction)
	static const uint PIC16_CORE; //           = 0x000000;
	// Shift out TABLAT register
	static const uint PIC16_SHIFT_TBLAT; //    = 0x020000;
	// Table Read
	static const uint PIC16_TBLRD; //          = 0x080000;
	// Table Read, post-increment
	static const uint PIC16_TBLRD_POSTINC; //  = 0x090000;
	// Table Read, post-decrement
	static const uint PIC16_TBLRD_POSTDEC; //  = 0x0A0000;
	// Table Read, pre-increment
	static const uint PIC16_TBLRD_PREINC; //   = 0x0B0000;
	// Table Write
	static const uint PIC16_TBLWRT; //         = 0x0C0000;
	// Table Write, post-increment by 2
	static const uint PIC16_PBLWRT_POSTINC; // = 0x0D0000;
	// Table Write, post-decrement by 2
	static const uint PIC16_TBLWRT_POSTDEC; // = 0x0E0000;
	// Table Write, start programming
	static const uint PIC16_STARTPROG; //      = 0x0F0000;

	static const byte ETX;
	static const byte DLE;
	static const byte STX;

	static const byte MAX_PACKET_LENGTH;  // Largest data packet we can send

	static const byte CMD_BLOCKSIZE;       // Largest block for read/write of code and data

	// Error return values
	static const int IO_OK;// = 0;					// No problems with IO
	static const int ERROR_IO_TIMEOUT;// = -1;		// Failed to Rx/Tx within timeout period
	static const int ERROR_IO_CHECKSUM;// = -2;		// Checksum failure in data packet
	static const int ERROR_IO_DATALEN;// = -3;		// Length of packet was wrong
	static const int ERROR_IO_BADHEXFILE;// = -4;	// Null or invalid hex file parameter
	static const int ERROR_IO_BUFFER;// = -5;		// Data buffer too small for response (overrun)
	static const int ERROR_IO_UNFINISHED;// = -6;	// Got an <STX> in the middle of response
	static const int ERROR_IO_NOSTX;// = -7;		// Didn't get a <STX> at beginning of response
	static const int ERROR_IO_BADADDRESS;// = -8;	// Addresses out of range for the PIC
	static const int ERROR_IO_UNALIGNED;// = -9;	// Attempt to write program memory with misaligned data (not on 4 word boundary)

	// Commands for version 2 programmer.
    static const byte EVA_VERSION_CMD;     // = 0x01; // Version of firmware in programmer
    static const byte EVA_SET_LINES_CMD;   // = 0x02; // Set all signal lines/LEDs on/off
    static const byte EVA_VPP_CMD;         // = 0x03; // Turn Vpp on/off
// (unused - null command)
    static const byte EVA_PRGSEL_CMD;      // = 0x05; // Turn PRG_SEL (dsPIC select) on/off
    static const byte EVA_VDD_CMD;         // = 0x06; // Turn Vdd on/off
    static const byte EVA_CLKDAT_CMD;      // = 0x07; // Set the ICSP CLK and DAT lines
    static const byte EVA_READ_LINES_CMD;  // = 0x08; // Read current value of signal/LED lines
    static const byte EVA_READ_REGS_CMD;   // = 0x09; // Read value of programmer registers
    static const byte EVA_ECHO_CMD;        // = 0x0a; // Send back the input
    static const byte EVA_BEGINPROG_CMD;   // = 0x0b; // Enter prog mode: lower DAT/CLK, raise Vdd and Vpp
    static const byte EVA_ENDPROG_CMD;     // = 0x0c; // Exit prog mode: lower Vdd/Vpp, tristate CLK/DAT
    static const byte EVA_PIC14_CMD;       // = 0x0d; // Send command to 14 bit PIC
    static const byte EVA_PIC14R_CMD;      // = 0x0e; // Send cmd to 14 bit PIC, read response
// (unused - null command)
    static const byte EVA_READ_ANALOG_CMD; // = 0x10; // Read the current value of the A/D converter
    static const byte EVA_SET_ANALOG_CMD;  // = 0x11; // Set desired voltage bound
    static const byte EVA_PIC16_ADDR_CMD;  // = 0x12; // Set the current address of a 16 bit PIC
    static const byte EVA_PIC16_CMD;       // = 0x13; // Send command to 16 bit PIC
    static const byte EVA_PIC16R_CMD;      // = 0x14; // Send cmd to 16 bit PIC, read response
    static const byte EVA_PIC16W_CMD;      // = 0x15; // Send write cmd to 16 bit PIC, hold CLK for time P9 (1-5 msec)
    static const byte EVA_READ_MASKS_CMD;  // = 0x16; // Read current value of TRIS registers
    static const byte EVA_READ_PORTS_CMD;  // = 0x17; // Read current value of PORT registers
    static const byte EVA_READ_PWM_CMD;    // = 0x18; // Read current PWM duty cycle
    static const byte EVA_PIC14RB_CMD;     // = 0x19; // Read block of code from 14 bit PIC
    static const byte EVA_PIC14WB_CMD;     // = 0x1a; // Write block of code to 14 bit PIC
    static const byte EVA_PIC14RBE_CMD;    // = 0x1b; // Read block of data from 14 bit PIC EEPROM
    static const byte EVA_PIC14WBE_CMD;    // = 0x1c; // Write block of data to 14 bit PIC EEPROM
    static const byte EVA_PIC16RB_CMD;     // = 0x1d; // Write block of words to 16 bit PIC
    static const byte EVA_PIC16WB_CMD;     // = 0x1e; // Read block of words from 16 bit PIC
    static const byte EVA_PIC16RBE_CMD;    // = 0x1f; // Write block of words to 16 bit PIC EEPROM
    static const byte EVA_PIC16WBE_CMD;    // = 0x20; // Read block of words from 16 bit PIC EEPROM
    static const byte EVA_PIC16WBUF_CMD;   // = 0x21; // Write one buffer of bytes to a 16 bit PIC
	static const byte EVA_PIC16WEBYTE_CMD; // = 0x22; // Write one byte to a 16 bit PIC EEPROM
    static const byte EVA_PIC16EBUF_CMD;   // = 0x23; // Erase one window of bytes on a 16 bit PIC
    static const byte EVA_PIC24_CMD;       // = 0x24; // Send command to 24 bit PIC
    static const byte EVA_PIC24R_CMD;      // = 0x25; // Read 16 bit register from 24 bit PIC
    static const byte EVA_PIC24RBUF_CMD;   // = 0x26; // Read one buffer of instructino words from 24 bit PIC
    static const byte EVA_PIC24WBUF_CMD;   // = 0x27; // Write one buffer of instruction words to 24 bit PIC
    static const byte EVA_PIC24WBUF_SETUP; // = 0x28; // Setup to write one buffer to 24 bit PIC
    static const byte EVA_PIC24WBUF_FINISH;// = 0x29; // Complete write of one buffer to 24 bit PIC
// The following commands are specific to the 18F2455 version of the firmware - they are
// used to read/write portions of the firmware in the running programmer.  The values of these
// commands must be above the standard commands above.
    static const byte EVA_IREAD_CMD;       // = 0x41; // Read programmer memory
    static const byte EVA_IWRITE_CMD;      // = 0x42; // Write programmer memory
    static const byte EVA_IERASE_CMD;      // = 0x43; // Erase programmer memory
    static const byte EVA_IRUN_CMD;        // = 0x44; // Run code in internal programmer memory
    static const byte EVA_IDEBUG_PRINT_CMD;// = 0x45; // Run code in internal programmer memory

public:
	// Generic programmer routines
	// Tell the programmer to return to a known good state
	int static ResetProgrammer(CIOPort *port);
	// Set voltage, data, and clock lines into the correct state to
	// begin read/write operations.  (Normally: Vdd=5V, Vpp=12-13V,
	// clock low, data low.)  Set PIC14BitFlag to false if opening
	// a 16 bit PIC.
	int static PICOpen(CIOPort *port, byte VppMask = 0x03, bool PIC14BitFlag = true);
	// Set voltage, data, and clock lines to offline state. (Normally:
	// Vdd low, Vpp low, clock/data don't care or tristate.)
	int static PICClose(CIOPort *port);

	// Send a command packet to the programmer and return the response packet.
	// On success this function will return the number of bytes in the response
	// packet "fromPICData".  Negative values indicate errors.
	int static SendPacket(CIOPort *port, byte cmd, uint dataLen, byte *toPICData,
	uint maxResultLen, uint *resultLen, byte *fromPICData, uint msecTimeout = 0);
	int static ReadResponse(CIOPort *port, byte cmd, uint *resultLen, byte *fromPICData);
	// Transmit a command/data packet to the PIC and read the result
	int static TxCmd(CIOPort *port, byte EvaCmd, uint data, uint *in);
	// Try to read the device ID from the chip.  Returns true on success.
	bool static ReadDeviceID(CIOPort *port, uint *devID, uint *wordSize);

	// Write a single word to a 14 bit PIC processor.  All of the supported
	// parts can write one word at a time, so this can be used in a loop
	// to program code, data, and configuration memory.
	int static WriteWord14(CIOPort *port, uint val, ushort mask = 0xffff,
		uint readCmd = PIC14_READDATA, uint loadCmd = PIC14_LOADDATA,
		uint progCmd = PIC14_BEGINPROG, uint endProgCmd = 0,
		int limit = RETRY_LIMIT);
	// Write 4 words (8 bytes) to a 16 bit PIC processor starting at the current address
	int static WriteWord16(CIOPort *port, ushort *vals, int limit);
	// 16 bit PIC specific routine to set the table pointer to a specific address
	int static SetTablePointer(CIOPort *port, uint addr);
	// Simple delay loop
	void static Delay(int p);
	// Send a character to the serial port
	int static TxByte(CIOPort *port, int c);
	// Receive a character from the serial port.  The return value will
	// be -1 if there was an error (e.g., timeout, framing error).
	int static RxByte(CIOPort *port);
	int static TxRxBuffer(CIOPort *port, int len, byte *buffer);
	// Set CLK and DAT lines explicitly
	int static SetICSPLevels(CIOPort *port, bool CLKFlag, bool DATFlag);

	// Read the config words (8 are returned) from a 14 bit PIC.
	// The device ID is located in them.
	bool static readPIC14Config(CIOPort *port, ushort *buf);
	// Read the device ID (one word is returned) from a 16 bit PIC
	bool static readPIC16DevID(CIOPort *port, ushort *devID);
	// Read the device ID (two words are returned) from a 24 bit PIC (dsPIC)
	bool static readPIC24DevID(CIOPort *port, ushort *devID);

private:
	static byte outBuf[128], inBuf[128];
};
