//	Serial0.cpp - Simple RS232 serial I/O class
#include "stdafx.h"

#if !defined(STRICT)
#define STRICT
#endif

#include <tchar.h>
#include <windows.h>
#include <winreg.h>

#include "Serial0.h"

//////////////////////////////////////////////////////////////////////
// Code

CSerial::CSerial ()
{
	m_hFile = NULL;
	charBuffer.clear();
}

CSerial::~CSerial ()
{
	// If the device is already closed,
	// then we don't need to do anything.
	if (m_hFile)
	{
		Close();
	}
}

bool CSerial::IsOpen()
{
	return (m_hFile != NULL);
}

/// <summary>
/// Returns the alternative name of a com port i.e. \\.\COM1 for COM1:
/// Some systems require this form for double or more digit com port numbers.
/// </summary>
/// <param name="s">Name in form COM1 or COM1:</param>
/// <returns>Name in form \\.\COM1</returns>
static LPTSTR AltName(LPCTSTR str)
{
	size_t strLen = _tcslen(str);
	// Make sure we aren't dealing with an empty string
	if (strLen < 1)
		return NULL;

	// If there is a ':' at the end, then get rid of it.
	if (str[strLen-1] == _T(':'))
		strLen--;

	// No alternative for a string that already starts with a '\'.
	if (str[0] == _T('\\'))
		return NULL;

	// Increase the length of the string by the number of characters we
	// are adding.
	size_t newLen = strLen + 4;

	LPTSTR newStr = new TCHAR[newLen+1]; // length of old string plus terminating 0.

	// Copy the string to the new buffer
	_tcsncpy(newStr, _T("\\\\.\\"), 4);
	_tcsncpy(newStr+4, str, strLen);
	newStr[strLen+4] = _T('\0');

	return newStr;
}

HRESULT CSerial::Open (LPCTSTR pDeviceName, int baudRate, int parity, int stopBits, bool handshake)
{
	// Check if the port isn't already opened
	if (m_hFile)
	{
		return -1;
	}

	// Get a handle to the com port.  Note: not using overlapped I/O.
	m_hFile = ::CreateFile(pDeviceName, GENERIC_READ|GENERIC_WRITE, 0, 0,
		OPEN_EXISTING, 0 /* FILE_FLAG_NO_BUFFERING */ , 0);
	if (m_hFile == INVALID_HANDLE_VALUE)
	{
		HRESULT err = ::GetLastError();
		if (err == ERROR_ACCESS_DENIED)
		{
			m_hFile = NULL;
			return ::GetLastError();
		}

		// May need alternative name (e.g., for COM ports with double digit values).
		LPTSTR newName = AltName(pDeviceName);
		if (newName != NULL)
		{
			m_hFile = ::CreateFile(newName, GENERIC_READ|GENERIC_WRITE, 0, 0,
				OPEN_EXISTING, 0 /* FILE_FLAG_NO_BUFFERING */ , 0);
			delete[] newName;
			if (m_hFile == INVALID_HANDLE_VALUE)
			{
				m_hFile = NULL;
				return ::GetLastError();
			}
		}
		else
		{
			m_hFile = NULL;
			return ::GetLastError();
		}
	}

	::SetupComm(m_hFile, 128, 128);

	// Setup characteristics of the com port (e.g., speed, parity, handshaking, ...)
	DCB dcb;
	::GetCommState(m_hFile,&dcb);

	// Standard characteristics
	dcb.BaudRate = baudRate;
	dcb.ByteSize = 8;
	dcb.Parity   = parity;
	dcb.StopBits = stopBits;
	dcb.fParity  = (parity != NOPARITY);

	// Handshaking
	if (!handshake)
	{
		// No flow control
		dcb.fOutxCtsFlow = false;
		dcb.fOutxDsrFlow = false;
		dcb.fDtrControl = DTR_CONTROL_DISABLE;
		dcb.fOutX = false;
		dcb.fInX = false;
		dcb.fRtsControl = RTS_CONTROL_DISABLE;
	}
	else
	{
		// RTS/CTS only
		dcb.fOutxCtsFlow = true;
		dcb.fOutxDsrFlow = false;
		dcb.fDtrControl = DTR_CONTROL_DISABLE;
		// dcb.fDtrControl = DTR_CONTROL_ENABLE;
		dcb.fOutX = false;
		dcb.fInX = false;
		// dcb.fRtsControl = RTS_CONTROL_TOGGLE;
		// dcb.fRtsControl = RTS_CONTROL_DISABLE;
		dcb.fRtsControl = RTS_CONTROL_ENABLE;
	}

	// Tell Windows to make the changes
	::SetCommState(m_hFile,&dcb);

#if 1
	COMMTIMEOUTS commTimeouts;
	::GetCommTimeouts (m_hFile, &commTimeouts);

#if 1
	// Both of these seem to work.  The first has a single timeout,
	// the second uses a multiplier based on the number of bytes
	// requested in a read.
#if 0
	commTimeouts.ReadIntervalTimeout = MAXDWORD;
	commTimeouts.ReadTotalTimeoutConstant = 100;
	commTimeouts.ReadTotalTimeoutMultiplier = MAXDWORD;
	commTimeouts.WriteTotalTimeoutConstant = 4;
	commTimeouts.WriteTotalTimeoutMultiplier = 0;
#else
	commTimeouts.ReadIntervalTimeout = 1;
	commTimeouts.ReadTotalTimeoutConstant = 0;
	commTimeouts.ReadTotalTimeoutMultiplier = 1;
	commTimeouts.WriteTotalTimeoutConstant = 4;
	commTimeouts.WriteTotalTimeoutMultiplier = 1;
#endif
#else
	// Using MAXDWORD can result in stalled code - doesn't
	// seem to be a good idea.
	commTimeouts.ReadIntervalTimeout = MAXDWORD;
	commTimeouts.ReadTotalTimeoutConstant = MAXDWORD;
	commTimeouts.ReadTotalTimeoutMultiplier = MAXDWORD;
	commTimeouts.WriteTotalTimeoutConstant = MAXDWORD;
	commTimeouts.WriteTotalTimeoutMultiplier = MAXDWORD;
#endif
	::SetCommTimeouts(m_hFile, &commTimeouts);
#endif

	return 0;
}

HRESULT CSerial::Close (void)
{
	if (m_hFile == NULL)
		// Trying to close a file that isn't open
		return -1;

	::CloseHandle(m_hFile);
	m_hFile = NULL;

	return 0;
}

BOOL CSerial::Write(LPCVOID lpBuffer, DWORD bufLen, DWORD* numWritten)
{
	if (m_hFile == NULL)
		return -1;

	int res = ::WriteFile(m_hFile, lpBuffer, bufLen, numWritten, NULL);

	if (res == 0)
	{
		int err = ::GetLastError();
		LPVOID lpMsgBuf;
		FormatMessage( 
			FORMAT_MESSAGE_ALLOCATE_BUFFER | 
			FORMAT_MESSAGE_FROM_SYSTEM | 
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			GetLastError(),
			0, // Default language
			(LPTSTR) &lpMsgBuf,
			0,
			NULL 
		);
		// Process any inserts in lpMsgBuf.
		// ...
		// Display the string.
		MessageBox( NULL, (LPCTSTR)lpMsgBuf, _T("Error"), MB_OK | MB_ICONINFORMATION );
		// Free the buffer.
		LocalFree( lpMsgBuf );
	}
	else if (*numWritten < bufLen)
	{
		// Try writing the rest????
		res = ::WriteFile(m_hFile, (LPCVOID)((char *)lpBuffer + *numWritten), bufLen - *numWritten, numWritten, NULL);
	}

	return res;
}

BOOL CSerial::Write(LPCSTR pString, DWORD* pdwWritten)
{
	// Determine the length of the string
	return Write(pString, (DWORD)strlen(pString), pdwWritten);
}

BOOL CSerial::Read(LPVOID lpBuffer, DWORD bufLen, DWORD* numRead)
{
	// Check if the device is open
	if (m_hFile == NULL)
		return FALSE;
#if 0 // !defined(_WIN32_WCE)
	return ::ReadFile(m_hFile, lpBuffer, bufLen, numRead, NULL);
#else
#if 0
	// Read the data
	int tmpLen = 128;
	DWORD tmpRead;
	BOOL res = ::ReadFile(m_hFile, (LPVOID)tmpBuf, tmpLen, &tmpRead, NULL);

	if (res > 0 && tmpRead > 0)
	{
		// More characters to stuff into the buffer
		unsigned char *buf = (unsigned char *)lpBuffer;
		*buf = tmpBuf[0];

		for (DWORD i=0;i<tmpRead;i++)
			charBuffer.push_back(tmpBuf[i]);
	}

	if (!charBuffer.empty())
	{
		unsigned char *buf = (unsigned char *)lpBuffer;
		*buf = (unsigned char)charBuffer.front();
		charBuffer.pop_front();
		*numRead = 1;
		return 0;
	}
	else
	{
		return -1;
	}

#else
	// This code will return a character from the queue if one is available,
	// and will try a read from the COM port if there aren't any.  If chars
	// come in from the COM port, they will be pushed onto the end of the queue.
	if (!charBuffer.empty())
	{
		unsigned char *buf = (unsigned char *)lpBuffer;
		*buf = (unsigned char)charBuffer.front();
		charBuffer.pop_front();
		*numRead = 1;
		return 0;
	}

	// Read the data
	int tmpLen = 128;
	DWORD tmpRead;
	BOOL res = ::ReadFile(m_hFile, (LPVOID)tmpBuf, tmpLen, &tmpRead, NULL);

	if (res > 0)
	{
		if (tmpRead == 0)
		{
			return -1;
		}

		unsigned char *buf = (unsigned char *)lpBuffer;
		*buf = tmpBuf[0];

		if (tmpRead > 1)
		{
			for (DWORD i=1;i<tmpRead;i++)
				charBuffer.push_back(tmpBuf[i]);
		}

		*numRead = 1;

		return 0;
	}
	else
	{
		int err = ::GetLastError();
		LPVOID lpMsgBuf;
		FormatMessage( 
			FORMAT_MESSAGE_ALLOCATE_BUFFER | 
			FORMAT_MESSAGE_FROM_SYSTEM | 
			FORMAT_MESSAGE_IGNORE_INSERTS,
			NULL,
			GetLastError(),
			0, // Default language
			(LPTSTR) &lpMsgBuf,
			0,
			NULL 
		);
		// Process any inserts in lpMsgBuf.
		// ...
		// Display the string.
		MessageBox( NULL, (LPCTSTR)lpMsgBuf, _T("Error"), MB_OK | MB_ICONINFORMATION );
		// Free the buffer.
		LocalFree( lpMsgBuf );

		return -1;
	}
#endif
#endif
}

BOOL CSerial::Flush()
{
	if (!charBuffer.empty())
	{
		charBuffer.clear();
	}
	memset(tmpBuf, 0, sizeof(tmpBuf)/sizeof(tmpBuf[0]));
	return FlushFileBuffers(m_hFile);
}

BOOL CSerial::Purge()
{
	if (!charBuffer.empty())
		charBuffer.clear();
	memset(tmpBuf, 0, sizeof(tmpBuf)/sizeof(tmpBuf[0]));
	return ::PurgeComm(m_hFile, PURGE_TXCLEAR | PURGE_RXCLEAR);
}


bool CSerial::GetCTS (void)
{
	DWORD modemStat = 0;
	::GetCommModemStatus(m_hFile, &modemStat);

	return (modemStat & MS_CTS_ON) != 0;
}

bool CSerial::GetDSR (void)
{
	DWORD modemStat = 0;
	::GetCommModemStatus(m_hFile, &modemStat);

	return (modemStat & MS_DSR_ON) != 0;
}

bool CSerial::GetRing (void)
{
	DWORD modemStat = 0;
	::GetCommModemStatus(m_hFile, &modemStat);

	return (modemStat & MS_RING_ON) != 0;
}

bool CSerial::GetRLSD (void)
{
	DWORD modemStat = 0;
	::GetCommModemStatus(m_hFile, &modemStat);

	return (modemStat & MS_RLSD_ON) != 0;
}

void CSerial::SetBRK(bool flag)
{
	::EscapeCommFunction(m_hFile, (flag ? SETBREAK : CLRBREAK));
}

void CSerial::SetRTS(bool flag)
{
	::EscapeCommFunction(m_hFile, (flag ? SETRTS : CLRRTS));
}

void CSerial::SetDTR(bool flag)
{
	::EscapeCommFunction(m_hFile, (flag ? SETDTR : CLRDTR));
}

// Code for enumerating COM port on Windows machines

#if 1
typedef struct
{
  LPCTSTR lpPortName;     /* "COM1", etc. */
  LPCTSTR lpFriendlyName; /* Suitable to describe the port, as for  */
                          /* instance "Infrared serial port (COM4)" */
}LISTPORTS_PORTINFO;

typedef BOOL (CALLBACK* LISTPORTS_CALLBACK)(LPVOID              lpCallbackValue,
                                            LISTPORTS_PORTINFO* lpPortInfo);
/* User provided callback funtion that receives the information on each
 * serial port available.
 * The strings provided on the LISTPORTS_INFO are not to be referenced after
 * the callback returns; instead make copies of them for later use.
 * If the callback returns FALSE, port enumeration is aborted.
 */

BOOL ListPorts(LISTPORTS_CALLBACK lpCallback,LPVOID lpCallbackValue);
/* Lists serial ports available on the system, passing the information on
 * each port on succesive calls to lpCallback.
 * lpCallbackValue, treated opaquely by ListPorts(), is intended to carry
 * information internal to the callback routine.
 * Returns TRUE if succesful, otherwise error code can be retrieved via
 * GetLastError().
 */

static BOOL WinCEListPorts(LISTPORTS_CALLBACK lpCbk, LPVOID lpCbkValue);
static BOOL Win9xListPorts(LISTPORTS_CALLBACK lpCbk,LPVOID lpCbkValue);
static BOOL WinNT40ListPorts(LISTPORTS_CALLBACK lpCbk,LPVOID lpCbkValue);
static BOOL Win2000ListPorts(LISTPORTS_CALLBACK lpCbk,LPVOID lpCbkValue);
static BOOL ScanEnumTree(LPCTSTR lpEnumPath,LISTPORTS_CALLBACK lpCbk,LPVOID lpCbkValue);
static LONG OpenSubKeyByIndex(HKEY hKey,DWORD dwIndex,REGSAM samDesired,PHKEY phkResult); 
static LONG QueryStringValue(HKEY hKey,LPCTSTR lpValueName, LPTSTR* lppStringValue);

BOOL ListPorts(LISTPORTS_CALLBACK lpCbk,LPVOID lpCbkValue)
{
  OSVERSIONINFO osvinfo;

  /* check parameters */

  if(!lpCbk){
    SetLastError(ERROR_INVALID_PARAMETER);
    return FALSE;
  }

  /* determine which platform we're running on and forward
   * to the appropriate routine
   */

  ZeroMemory(&osvinfo,sizeof(osvinfo));
  osvinfo.dwOSVersionInfoSize=sizeof(osvinfo);

  GetVersionEx(&osvinfo);

  switch(osvinfo.dwPlatformId){
    case VER_PLATFORM_WIN32_WINDOWS:
      return Win9xListPorts(lpCbk,lpCbkValue);
      break;
    case VER_PLATFORM_WIN32_NT:
      if(osvinfo.dwMajorVersion<4){
        SetLastError(ERROR_OLD_WIN_VERSION);
        return FALSE;
      }
#if 0
	  else
	  {
        return WinNT40ListPorts(lpCbk,lpCbkValue);
	  }
#else
      else if(osvinfo.dwMajorVersion==4){
        return WinNT40ListPorts(lpCbk,lpCbkValue);
      }
      else{
        return Win2000ListPorts(lpCbk,lpCbkValue); /* hopefully it'll also work for XP */
      }
#endif
      break;
#if defined(_WIN32_WCE)
	case VER_PLATFORM_WIN32_CE:
      return WinCEListPorts(lpCbk, lpCbkValue);
		break;
#endif
    default:
      SetLastError(ERROR_OLD_WIN_VERSION);
      return FALSE;
      break;
  }
}

// 2) On PocketPC, iterate through all of the folders in:
//       'HKLM\\Drivers\\Active'
//      - Grab the value of "key" in each folder (e.g., 'HKLM\\Drivers\\Active\\05\\key")
//      - Use that value to look in the registry (e.g., key = 'Drivers\BuiltIn\Serial5',
//        so look in 'HKLM\\Drivers\\BuiltIn\\Serial5).
//      - If the value of "Prefix" in that location is "COM", then add it to the list,
//        with the value in "Index" appended.  This should match the value of "Name" in
//        the active drivers key.
//      - Can use the "FriendlyName" value in the BuiltIn key for something easier to read.
static BOOL WinCEListPorts(LISTPORTS_CALLBACK lpCbk, LPVOID lpCbkValue)
{
	DWORD  dwError = 0;
	HKEY   hkEnum = NULL;
	HKEY   hKey = NULL;
	HKEY   hActiveKey = NULL;
	DWORD  dwIndex;
	LPTSTR lpPortName = NULL;
	LPTSTR lpFriendlyName = NULL;
	LPTSTR lpKeyName = NULL;

    dwError = RegOpenKeyEx(HKEY_LOCAL_MACHINE, TEXT("Drivers\\Active"),
						   0, KEY_ENUMERATE_SUB_KEYS, &hkEnum);
	if (dwError != ERROR_SUCCESS)
	{
		return FALSE;
	}

	for (dwIndex=0;;dwIndex++)
	{
		dwError = OpenSubKeyByIndex(hkEnum, dwIndex, KEY_ENUMERATE_SUB_KEYS, &hActiveKey);
		if (dwError == ERROR_NO_MORE_ITEMS)
		{
		    dwError=0;
			break;
		}
		else if (dwError != ERROR_SUCCESS)
		{
			break;
		}

		// Got a usable subkey.
		dwError = QueryStringValue(hActiveKey, TEXT("Key"), &lpKeyName);
		if (dwError != ERROR_SUCCESS)
		{
			RegCloseKey(hActiveKey);
			continue;
		}

		// Get it's basic name.
		dwError = QueryStringValue(hActiveKey, TEXT("Name"), &lpPortName);
		if (dwError != ERROR_SUCCESS)
		{
			// Not one we want - keep looking
			RegCloseKey(hActiveKey);
			continue;
		}

		// Don't need this key any more, only the contents we just got.
		RegCloseKey(hActiveKey);

		// Use the contents of the key as a key in the registry
		dwError = RegOpenKeyEx(HKEY_LOCAL_MACHINE, lpKeyName, 0, NULL, &hKey);
		// Free memory associated with the key
		free(lpKeyName);
		if (dwError != ERROR_SUCCESS)
		{
			break;
		}

		dwError = QueryStringValue(hKey, _T("Prefix"), &lpKeyName);
		if (dwError != ERROR_SUCCESS)
		{
			RegCloseKey(hKey);
			continue;
		}

		bool haveFriendlyName = true;
		dwError = QueryStringValue(hKey, _T("FriendlyName"), &lpFriendlyName);
		if (dwError != ERROR_SUCCESS)
			haveFriendlyName = false;
		RegCloseKey(hKey);

		// See if this is a COM port
        if(_tcsncicmp(lpKeyName, _T("COM"), 3) !=0)
			continue;

		// We have a port (or at least something that fits the description.
		LISTPORTS_PORTINFO portinfo;
		portinfo.lpPortName = lpPortName;
		portinfo.lpFriendlyName = (haveFriendlyName ? lpFriendlyName : lpPortName);
		if (!lpCbk(lpCbkValue,&portinfo))
		{
			// listing aborted by callback
			break;
		}
	}

	if (hkEnum != NULL)
		RegCloseKey(hkEnum);

	free(lpKeyName);
	free(lpPortName);
	free(lpFriendlyName);
	if (dwError!=0)
	{
		SetLastError(dwError);
		return FALSE;
	}
	else
		return TRUE;


}

BOOL Win9xListPorts(LISTPORTS_CALLBACK lpCbk,LPVOID lpCbkValue)
/* Information on communications ports is stored in various places in the registry under
 * the HKEY_LOCAL_MACHINE\ENUM key. The structure of the HKLM\ENUM key, where information
 * on all hw devices present on the system is stored, is as follows
 *
 * HKLM\ENUM
 *   |-LEVEL1
 *     |-LEVEL2
 *       |-LEVEL3
 *          CLASS
 *
 * LEVEL1 keys indicate the type of technology (vg., "PCI", "BIOS", "SCSI"). LEVEL2
 * keys refer to the type of device. So, for instance, "*PNP0501" is for 16550A UARTs.
 * LEVEL3 keys enumerate actual devices installed on the system. The value "CLASS",
 * present in all LEVEL3 keys, informs about the type of driver. For our purposes, we must
 * pay attention to the Class "Ports", that includes serial as well as parallel ports.
 * In these cases, another two values, "PORTNAME" and "FRIENDLYNAME" are expected.
 * An example follows. Typically, COM1 is recognized as a PnP device by the system BIOS
 * and as such placed on the registry like this:
 *
 * HKLM\ENUM
 *   |-BIOS
 *     |-*PNP0501
 *       |-0D (or any other value, this is not important for us)
 *          CLASS=        "Ports"
 *          PORTNAME=     "COM1"
 *          FRIENDLYNAME= "Communications Port (COM1)"
 *
 * Win9xListPorts() scans all LEVEL3 keys under HKLM\ENUM searching for the Class
 * for ports, and when this is found determines, by means of "PORTNAME", whether the
 * device is a serial port or not.
 */
{
  return ScanEnumTree(TEXT("ENUM"),lpCbk,lpCbkValue);
}

static BOOL WinNT40ListPorts(LISTPORTS_CALLBACK lpCbk,LPVOID lpCbkValue)
/* Unlike Win9x, information on serial ports registry storing in NT 4.0 is
 * scarce. HKEY_LOCAL_MACHINE\HARDWARE\DEVICEMAP\SERIALCOMM is reliably
 * documented to hold the names of all installed serial ports, but I haven't
 * confirmed this for infrared ports and other nonstandard drivers.
 * Descriptions stored under SERIALCOMM lack the "FRIENDLYNAME" found
 * in Win9x, there's only the bare names of the ports.
 */
{
  DWORD  dwError=0;
  HKEY   hKey=NULL;
  DWORD  dwIndex;
  LPTSTR lpValueName=NULL;
  LPTSTR lpPortName=NULL;

  if(dwError=RegOpenKeyEx(HKEY_LOCAL_MACHINE,TEXT("HARDWARE\\DEVICEMAP\\SERIALCOMM"),
                          0,KEY_READ,&hKey)){
    /* it is really strange that this key does not exist, but could happen in theory */
    if(dwError==ERROR_FILE_NOT_FOUND)dwError=0;
    goto end;
  }

  for(dwIndex=0;;++dwIndex){
    DWORD              cbValueName=32*sizeof(TCHAR);
    DWORD              cbPortName=32*sizeof(TCHAR);
    LISTPORTS_PORTINFO portinfo;

    /* loop asking for the value data til we allocated enough memory */

    for(;;){
      free(lpValueName);
      if(!(lpValueName=(LPTSTR)malloc(cbValueName))){
        dwError=ERROR_NOT_ENOUGH_MEMORY;
        goto end;
      }
      free(lpPortName);
      if(!(lpPortName=(LPTSTR)malloc(cbPortName))){
        dwError=ERROR_NOT_ENOUGH_MEMORY;
        goto end;
      }
      if(!(dwError=RegEnumValue(hKey,dwIndex,lpValueName,&cbValueName,
                                NULL,NULL,(LPBYTE)lpPortName,&cbPortName))){
        break; /* we did it */
      }
      else if(dwError==ERROR_MORE_DATA){ /* not enough space */
        dwError=0;
        /* no indication of space required, we try doubling */
        cbValueName=(cbValueName+sizeof(TCHAR))*2;
      }
      else goto end;
    }

    portinfo.lpPortName=lpPortName;
    portinfo.lpFriendlyName=lpPortName; /* no friendly name in NT 4.0 */
    if(!lpCbk(lpCbkValue,&portinfo)){
      goto end; /* listing aborted by callback */
    }
  } 

end:
  free(lpValueName);
  free(lpPortName);
  if(hKey!=NULL)RegCloseKey(hKey);
  if(dwError!=0){
    SetLastError(dwError);
    return FALSE;
  }
  else return TRUE;
}

BOOL Win2000ListPorts(LISTPORTS_CALLBACK lpCbk,LPVOID lpCbkValue)
/* Information on serial ports is stored in the Win2000 registry in a manner very
 * similar to that of Win9x, with three differences:
 *    The ENUM tree is not located under HKLM, but under HKLM\SYSTEM\CURRENTCONTROLSET.
 *    The parameter "PORTNAME" is not a LEVEL3 value like "CLASS" and "FRIENDLYNAME";
 *     instead, it is located under an additional subkey named "DEVICE PARAMETERS"
 *    "CLASS" seems to be deprecated in favor of "CLASSGUID" wich contains a GUID
 *     identifying each type of device. I've seen Win200 installations with and
 *     without "CLASS" values. Moreover, I've seen Win9x installations containing
 *     "CLASSGUID" values, so, to be as robust as possible, ScanEnumTree() accept either
 *     type of value.
 *
 * An example follows:
 *
 * HKLM\SYSTEM\CURRENTCONTROLSET
 *   |-BIOS
 *     |-*PNP0501
 *       |-0D (or any other value, this is not important for us)
 *          CLASSGUID=    "{4D36E978-E325-11CE-BFC1-08002BE10318}"
 *          FRIENDLYNAME= "Communications Port (COM1)"
 *         |-DEVICE PARAMETERS
 *            PORTNAME=   "COM1"
 */
{
  return ScanEnumTree(TEXT("SYSTEM\\CURRENTCONTROLSET\\ENUM"),lpCbk,lpCbkValue);
}

BOOL ScanEnumTree(LPCTSTR lpEnumPath,LISTPORTS_CALLBACK lpCbk,LPVOID lpCbkValue)
{
  static const TCHAR lpstrPortsClass[]=    TEXT("PORTS");
  static const TCHAR lpstrPortsClassGUID[]=TEXT("{4D36E978-E325-11CE-BFC1-08002BE10318}");

  DWORD  dwError=0;
  HKEY   hkEnum=NULL;
  DWORD  dwIndex1;
  HKEY   hkLevel1=NULL;
  DWORD  dwIndex2;
  HKEY   hkLevel2=NULL;
  DWORD  dwIndex3;
  HKEY   hkLevel3=NULL;
  HKEY   hkDeviceParameters=NULL;
  TCHAR  lpClass[sizeof(lpstrPortsClass)/sizeof(lpstrPortsClass[0])];
  DWORD  cbClass;
  TCHAR  lpClassGUID[sizeof(lpstrPortsClassGUID)/sizeof(lpstrPortsClassGUID[0])];
  DWORD  cbClassGUID;
  LPTSTR lpPortName=NULL;
  LPTSTR lpFriendlyName=NULL;

  if(dwError=RegOpenKeyEx(HKEY_LOCAL_MACHINE,lpEnumPath,0,KEY_ENUMERATE_SUB_KEYS,&hkEnum)){
    goto end;
  }

  for(dwIndex1=0;;++dwIndex1){
    if(hkLevel1!=NULL){
      RegCloseKey(hkLevel1);
      hkLevel1=NULL;
    }
    if(dwError=OpenSubKeyByIndex(hkEnum,dwIndex1,KEY_ENUMERATE_SUB_KEYS,&hkLevel1)){
      if(dwError==ERROR_NO_MORE_ITEMS){
        dwError=0;
        break;
      }
      else goto end;
    }

    for(dwIndex2=0;;++dwIndex2){
      if(hkLevel2!=NULL){
        RegCloseKey(hkLevel2);
        hkLevel2=NULL;
      }
      if(dwError=OpenSubKeyByIndex(hkLevel1,dwIndex2,KEY_ENUMERATE_SUB_KEYS,&hkLevel2)){
        if(dwError==ERROR_NO_MORE_ITEMS){
          dwError=0;
          break;
        }
        else goto end;
      }

      for(dwIndex3=0;;++dwIndex3){
        BOOL               bFriendlyNameNotFound=FALSE;
        LISTPORTS_PORTINFO portinfo;

        if(hkLevel3!=NULL){
          RegCloseKey(hkLevel3);
          hkLevel3=NULL;
        }
        if(dwError=OpenSubKeyByIndex(hkLevel2,dwIndex3,KEY_READ,&hkLevel3)){
          if(dwError==ERROR_NO_MORE_ITEMS){
            dwError=0;
            break;
          }
          else goto end;
        }

        /* Look if the driver class is the one we're looking for.
         * We accept either "CLASS" or "CLASSGUID" as identifiers.
         * No need to dynamically arrange for space to retrieve the values,
         * they must have the same length as the strings they're compared to
         * if the comparison is to be succesful.
         */

        cbClass=sizeof(lpClass);
        if(RegQueryValueEx(hkLevel3,TEXT("CLASS"),NULL,NULL,
                           (LPBYTE)lpClass,&cbClass)==ERROR_SUCCESS&&
           _tcsicmp(lpClass,lpstrPortsClass)==0){
          /* ok */
        }
        else{
          cbClassGUID=sizeof(lpClassGUID);
          if(RegQueryValueEx(hkLevel3,TEXT("CLASSGUID"),NULL,NULL,
                             (LPBYTE)lpClassGUID,&cbClassGUID)==ERROR_SUCCESS&&
             _tcsicmp(lpClassGUID,lpstrPortsClassGUID)==0){
            /* ok */
          }
          else continue;
        }

        /* get "PORTNAME" */

        dwError=QueryStringValue(hkLevel3,TEXT("PORTNAME"),&lpPortName);
        if(dwError==ERROR_FILE_NOT_FOUND){
          /* In Win200, "PORTNAME" is located under the subkey "DEVICE PARAMETERS".
           * Try and look there.
           */

          if(hkDeviceParameters!=NULL){
            RegCloseKey(hkDeviceParameters);
            hkDeviceParameters=NULL;
          }
          if(RegOpenKeyEx(hkLevel3,TEXT("DEVICE PARAMETERS"),0,KEY_READ,
                          &hkDeviceParameters)==ERROR_SUCCESS){
             dwError=QueryStringValue(hkDeviceParameters,TEXT("PORTNAME"),&lpPortName);
          }
        }
        if(dwError){
          if(dwError==ERROR_FILE_NOT_FOUND){ 
            /* boy that was strange, we better skip this device */
            dwError=0;
            continue;
          }
          else goto end;
        }

        /* check if it is a serial port (instead of, say, a parallel port) */

        if(_tcsncicmp(lpPortName,TEXT("COM"),3)!=0)continue;

        /* now go for "FRIENDLYNAME" */

        if(dwError=QueryStringValue(hkLevel3,TEXT("FRIENDLYNAME"),&lpFriendlyName)){
          if(dwError==ERROR_FILE_NOT_FOUND){
            bFriendlyNameNotFound=TRUE;
            dwError=0;
          }
          else goto end;
        }
        
        /* Assemble the information and pass it on to the callback.
         * In the unlikely case there's no friendly name available,
         * use port name instead.
         */
        portinfo.lpPortName=lpPortName;
        portinfo.lpFriendlyName=bFriendlyNameNotFound?lpPortName:lpFriendlyName;
        if(!lpCbk(lpCbkValue,&portinfo)){
          goto end; /* listing aborted by callback */
        }
      }
    }
  }

end:
  free(lpFriendlyName);
  free(lpPortName);
  if(hkDeviceParameters!=NULL)RegCloseKey(hkDeviceParameters);
  if(hkLevel3!=NULL)          RegCloseKey(hkLevel3);
  if(hkLevel2!=NULL)          RegCloseKey(hkLevel2);
  if(hkLevel1!=NULL)          RegCloseKey(hkLevel1);
  if(hkEnum!=NULL)            RegCloseKey(hkEnum);
  if(dwError!=0){
    SetLastError(dwError);
    return FALSE;
  }
  else return TRUE;
}

LONG OpenSubKeyByIndex(HKEY hKey,DWORD dwIndex,REGSAM samDesired,PHKEY phkResult)
{
  DWORD              dwError=0;
  LPTSTR             lpSubkeyName=NULL;
  DWORD              cbSubkeyName=128*sizeof(TCHAR); /* an initial guess */
  FILETIME           filetime;

  /* loop asking for the subkey name til we allocated enough memory */

  for(;;){
    free(lpSubkeyName);
    if(!(lpSubkeyName=(LPTSTR)malloc(cbSubkeyName))){
       dwError=ERROR_NOT_ENOUGH_MEMORY;
       goto end;
    }
    if(!(dwError=RegEnumKeyEx(hKey,dwIndex,lpSubkeyName,&cbSubkeyName,
                              0,NULL,NULL,&filetime))){
      break; /* we did it */
    }
    else if(dwError==ERROR_MORE_DATA){ /* not enough space */
      dwError=0;
      /* no indication of space required, we try doubling */
      cbSubkeyName=(cbSubkeyName+sizeof(TCHAR))*2;
    }
    else goto end;
  }

  dwError=RegOpenKeyEx(hKey,lpSubkeyName,0,samDesired,phkResult);

end:
  free(lpSubkeyName);
  return dwError;
}

LONG QueryStringValue(HKEY hKey,LPCTSTR lpValueName,LPTSTR* lppStringValue)
{
  DWORD cbStringValue=128*sizeof(TCHAR); /* an initial guess */

  for(;;){
    DWORD dwError;

    free(*lppStringValue);
    if(!(*lppStringValue=(LPTSTR)malloc(cbStringValue))){
      return ERROR_NOT_ENOUGH_MEMORY;
    }
    if(!(dwError=RegQueryValueEx(hKey,lpValueName,NULL,NULL,
                                 (LPBYTE)*lppStringValue,&cbStringValue))){
      return ERROR_SUCCESS;
    }
    else if(dwError==ERROR_MORE_DATA){
      /* not enough space, keep looping */
    }
    else return dwError;
  }
}


#else
#include <setupapi.h>

#define MAX_NAME_PORTS 7
#define RegDisposition_OpenExisting (0x00000001)   // open key only if exists
#define CM_REGISTRY_HARDWARE        (0x00000000)

typedef DWORD WINAPI
  (* CM_Open_DevNode_Key)(DWORD, DWORD, DWORD, DWORD, ::PHKEY, DWORD);

static HANDLE BeginEnumeratePorts(VOID)
{
	BOOL guidTest=FALSE;
	DWORD RequiredSize=0;
	int j;
	HDEVINFO DeviceInfoSet;
	char *buf;

	// Ask Windows how many GUIDs are associated with the class name "Ports".
	guidTest = SetupDiClassGuidsFromNameA("Ports", 0, 0, &RequiredSize);
	if (RequiredSize < 1)
		return INVALID_HANDLE_VALUE;

	// Allocate a buffer to hold the GUIDs
	buf = new char[RequiredSize * sizeof(GUID)];

	// Ask Windows for the array of GUIDs associated with the class name "Ports"
	guidTest = SetupDiClassGuidsFromNameA("Ports", (GUID *)buf,
		RequiredSize * sizeof(GUID), &RequiredSize);

	// See if the request worked
	if (guidTest == FALSE)
	{
		delete[] buf;
		return INVALID_HANDLE_VALUE;
	}

	// Get a device information set that contains all devices of the "Ports" class
	DeviceInfoSet = SetupDiGetClassDevs((GUID *)buf, NULL, NULL, DIGCF_PRESENT);

	// Free list of GUIDS
	delete[] buf;

	// Ensure we have a valid result
	if (DeviceInfoSet == INVALID_HANDLE_VALUE)
	{
		return INVALID_HANDLE_VALUE;
	}

	// Return the handle to the device information set
	return DeviceInfoSet;
}

static BOOL EnumeratePortsNext(HANDLE DeviceInfoSet, LPTSTR lpBuffer)
{
	static CM_Open_DevNode_Key OpenDevNodeKey=NULL;
	static HINSTANCE CfgMan;

	int res1;
	char DevName[MAX_NAME_PORTS]={0};
	static int numDev=0;
	int numport;

	SP_DEVINFO_DATA DeviceInfoData={0};
	DeviceInfoData.cbSize=sizeof(SP_DEVINFO_DATA);


	if(!DeviceInfoSet || !lpBuffer)return -1;
	if(!OpenDevNodeKey)
	{
		CfgMan = LoadLibrary("cfgmgr32");
		if(!CfgMan)return FALSE;
		OpenDevNodeKey = (CM_Open_DevNode_Key)GetProcAddress(CfgMan, "CM_Open_DevNode_Key");
		if(!OpenDevNodeKey)
		{
			FreeLibrary(CfgMan);
			return FALSE;
		}
	}

	while(TRUE)
	{

		HKEY KeyDevice;
		DWORD len;
		res1=SetupDiEnumDeviceInfo(
			DeviceInfoSet,numDev,&DeviceInfoData);

		if(!res1)
		{
			SetupDiDestroyDeviceInfoList(DeviceInfoSet);
			FreeLibrary(CfgMan);
			OpenDevNodeKey=NULL;
			return FALSE;
		}


		res1 = OpenDevNodeKey(DeviceInfoData.DevInst, KEY_QUERY_VALUE,0,
			RegDisposition_OpenExisting, &KeyDevice, CM_REGISTRY_HARDWARE);
		if(res1 != ERROR_SUCCESS)return NULL;
		len=MAX_NAME_PORTS;

		res1=RegQueryValueEx(
			KeyDevice,    // handle of key to query
			"portname",    // address of name of value to query
			NULL,    // reserved
			NULL,    // address of buffer for value type
			DevName,    // address of data buffer
			&len     // address of data buffer size
			);

		RegCloseKey(KeyDevice);
		if(res1 != ERROR_SUCCESS)return NULL;
		numDev++;
		if(memicmp(DevName, "com", 3))continue;
		numport=atoi(DevName+3);
		if(numport > 0 && numport <= 256)
		{
			strcpy(lpBuffer,DevName);
			return TRUE;
		}

		FreeLibrary(CfgMan);
		OpenDevNodeKey=NULL;
		return FALSE;
	}
}

BOOL EndEnumeratePorts(HANDLE DeviceInfoSet)
{
if(SetupDiDestroyDeviceInfoList(DeviceInfoSet))return TRUE;
        else return FALSE;
}
#endif

static BOOL CALLBACK listPortsCallbackFn(LPVOID lpCallbackValue, LISTPORTS_PORTINFO* lpPortInfo)
{
	std::list<LPTSTR> *COMPorts = (std::list<LPTSTR> *)lpCallbackValue;

	if (COMPorts == NULL)
	{
		COMPorts = new std::list<LPTSTR>;
	}

	size_t nameLen = _tcslen(lpPortInfo->lpPortName) + 1;
	TCHAR *portName = new TCHAR[nameLen];
	_tcsncpy(portName, lpPortInfo->lpPortName, nameLen);

	COMPorts->push_back(portName);

	return TRUE;
}

std::list<LPTSTR> *CSerial::GetCOMPortList()
{
	std::list<LPTSTR> *COMPorts = new std::list<LPTSTR>;
	if (COMPorts == NULL)
		return NULL;

	BOOL result = ListPorts(listPortsCallbackFn, COMPorts);

	return COMPorts;
}
