;
; Code for PIC programmer that generates it's own 13V programming
; voltage using the PWM to drive a boost circuit.  Communications
; with the host PC is over RS232 with a bit-bang implementation
; (the PIC 16F684 doesn't have a hardware UART).
;
   ; processor __16F684
   #include "p16f684.inc"

;
; Configuration definitions - adjust according to processor
; and hardware features.
;
#define _USE_XTAL_18432 0 ; 0 = Internal 8MHz clock, 1 = 18.432 MHz crystal
#define _INDUCTOR_BASED 1 ; 0 = MAX232, 1 = Inductor for voltage boost
; RS-232 Parameters
#define _USE_RTSCTS 1     ; 1 = Use hardware flow control, 0 = don't use it
#define _USE_UART 0       ; 0 = Don't use/don't have UART
#define ECHO_ALL 0        ; Echo any characters we get
#define BAUD_RATE_1152 1  ; 0 = 38.4K Baud, 1 = 115.2K Baud

; Watch dog off, Power Up Timer on, Code Protect off,
; MCLR pin disabled - used as general IO
#if _USE_XTAL_18432
    ; External high-speed crystal
   __config 0x30c2
#else
   ; Internal 4MHz Osc with pin used at general IO.  Clock is
   ; increased to 8MHz during setup.
   __config 0x30c4
#endif

; Current version: 2.9
VERSION0 equ 0x09 ; Minor version #
VERSION1 equ 0x02 ; Major version #

;
; Pin assignments
;
;    RA0  -  Rx RS232 data
;    RA1  -  Tx RS232 data
;    RA2  -  ICSP data
;    RA3  -  (optional - RS232 RTS)
;    RA4  -  [unused - leave open for crystal]
;    RA5  -  [unused - leave open for crystal]
;    RC0  -  Analog feedback from voltage boost converter
;    RC1  -  13V chip programming voltage control
;    RC2  -  ICSP clock line
;    RC3  -  5V supply voltage control
;    RC4  -  (optional - RS232 CTS)
;    RC5  -  PWM output
;
#define RX_MASK TRISA, 0
#define RX PORTA, 0
#define TX_MASK TRISA, 1
#define TX PORTA, 1
#define ICSP_DAT_MASK TRISA, 2
#define ICSP_DAT PORTA, 2
#define AD_FEEDBACK_MASK TRISC, 0
#define AD_FEEDBACK_ENABLE ANSEL, 4
#define AD_FEEDBACK PORTC, 0
#define VPP1_MASK TRISC, 1
#define VPP1 PORTC, 1
#if 0
; TBD: separate control of Vpp for 40 pin versus smaller parts
#define VPP2_MASK TRISX, X
#define VPP2 PORTX, X
#endif
#define ICSP_CLK_MASK TRISC, 2
#define ICSP_CLK PORTC, 2
#define VDD_MASK TRISC, 3
#define VDD PORTC, 3
#if _USE_RTSCTS
#define CTS_MASK TRISC, 4
#define CTS PORTC, 4
#define RTS_MASK TRISA, 3
#define RTS PORTA, 3
#endif
#if _INDUCTOR_BASED
#define PWM_MASK TRISC, 5
#define PWM PORTC, 5 ; RC5 is PWM output line
#endif

;
; Note: an 18.432 MHz crystal would be ideal for keeping serial
; I/O working smoothly.  That value would give an even number of
; instruction cycles for both 38.4K and 115.2K Baud rates.
; (Digi-Key part: X061-ND).
;    

; Bank 0 variables
x          equ 0x20
y          equ 0x22

; Registers for I/O and programming variables (bank 0)
TEMP0       equ 0x24 ; Low byte of temp variable
TEMP1       equ 0x25 ; High byte of temp variable
PCADDR0     equ 0x26 ; Low byte of current ICSP address
PCADDR1     equ 0x27 ; High byte of current ICSP address
PCADDR2     equ 0x28 ; Upper byte of current ICSP address
ICTR        equ 0x29 ; Temp counter used for ICSP and Serial I/O
ERRORS      equ 0x2a ; Error result from RS232 routines
TEMP        equ 0x2b ; Temp value
DLYTEMP     equ 0x2c ; Temp counter used in delay routine
WRTEMP      equ 0x2d ; Temp bit counter used for RS232 IO
SERIAL_OUT  equ 0x2e ; Last value read from RS232
SERIAL_IN   equ 0x2f ; Value to send over RS232 (not valid after send0
COUNTER     equ 0x30 ; Temp counter used for packet IO
CHECKSUM    equ 0x31 ; Checksum for command packet
CMDTEMP     equ 0x32 ; Temp ICSP command value (for block programming)
PROGTEMP    equ 0x33 ; Temp ICSP begin programming command value
ENDPROGTEMP equ 0x34 ; Temp ICSP end programming command value
RETRYCTR    equ 0x35 ; # of times write attempted at one location
PICDATA0    equ 0x36 ; Low byte returned from target PIC
PICDATA1    equ 0x37 ; High byte returned from target PIC
; TBD: equ 0x38
; TBD: equ 0x39
; Storage for command packet:
;    <cmd><data len><addrL><addrH><addrU><cmdOPT><...data...>
;    <cmd><data len><data0><data1><data2>
;    <cmd><data len><ICSP cmd><data high byte><data low byte>
;
PACKET      equ 0x3a ; Start of packet
COMMAND     equ 0x3a ; Command value is first entry in the packet
DATALEN     equ 0x3b ; Length of data to follow
DATABUF0    equ 0x3c ; Byte 0 of generic data buffer
ADDRL       equ 0x3c ; Low byte of target address
ICSP_CMD    equ 0x3c ; command value to send to PIC
DATABUF1    equ 0x3d ; Byte 1 of generic data buffer
ADDRH       equ 0x3d ; High byte of target address
DATA0       equ 0x3d ; Low byte of data to send to PIC
DATABUF2    equ 0x3e ; Byte 2 of generic data buffer
ADDRU       equ 0x3e ; Upper byte of target address
DATA1       equ 0x3e ; High byte of data to send to PIC
CMDOPT      equ 0x3f ; Optional value (e.g., block length)
DATA_BUF    equ 0x40 ; Up to 64 bytes of data to send/recieve
;
; Address 0x40..0x7f are used for data payload.
;

; Bank 1 variables
analogVal  equ 0xa0 ; Last value read from A/D converter
delayVal   equ 0xa1 ; How long to wait for A/D conversion
#if _INDUCTOR_BASED
PWMCtr     equ 0xa2 ; Number of passes to wait before checking PWM
PWMVal     equ 0xa3 ; PWM duty cycle value
pwm_tmp    equ 0xa4 ; Temp value used when updating PWM duty cycle
#endif

MAX_RETRIES equ 0x04 ; Max times write tried before failure

MAX_BUFLEN  equ 0x40 ; Limit buffer count to 64

; ICSP commands for 14 bit PIC devices
PIC14_LOADCFG       equ 0x00
PIC14_LOADDATA      equ 0x02
PIC14_READDATA      equ 0x04
PIC14_INCADDR       equ 0x06
PIC14_BEGINPROG     equ 0x08
PIC14_ENDPROG       equ 0x0E
PIC14_ENDPROGA      equ 0x17
PIC14_BEGINPROGONLY equ 0x18
PIC14_LOADE2DATA    equ 0x03
PIC14_READE2DATA    equ 0x05
PIC14_BULKERASEPROG equ 0x09
PIC14_BULKERASEE2   equ 0x0B
PIC14_CHIPERASE     equ 0x1F

; ICSP commands for 16 bit PIC devices
PIC16_CORE           equ 0x00 ; 16 bit PIC core Instruction
PIC16_SHIFT_TBLAT    equ 0x02 ; Shift out TABLAT register
PIC16_TBLRD          equ 0x08 ; Table Read
PIC16_TBLRD_POSTINC  equ 0x09 ; Table Read, post-increment
PIC16_TBLRD_POSTDEC  equ 0x0A ; Table Read, post-decrement
PIC16_TBLRD_PREINC   equ 0x0B ; Table Read, pre-increment
PIC16_TBLWRT         equ 0x0C ; Table Write
PIC16_PBLWRT_POSTINC equ 0x0D ; Table Write, post-increment by 2
PIC16_TBLWRT_POSTDEC equ 0x0E ; Table Write, post-decrement by 2
PIC16_STARTPROG      equ 0x0F ; Table Write, start programming

#if _INDUCTOR_BASED
#if _USE_XTAL_18432
; Constants for PWM management (for 18.432 MHz clock)
PWMPeriod equ 0x3f ; Setting for PR2 to get 62.5 KHz PWM
maxPWM    equ 0x40 ; Max duty cycle allowed
minAnalog equ 0x92 ; 15V - Minimum voltage before we increase duty cycle
fixVal    equ 0x94 ; ~15.24V (reading of about 2.9V from A/D converter)
maxAnalog equ 0x9b ; 16V - Max voltage before we decrease duty cycle
startPWM  equ 0x40 ; about 3% duty cycle to start with
#else
; Constants for PWM management (assuming 8MHz clock)
PWMPeriod equ 0x1f ; Setting for PR2 to get 62.5 KHz PWM
maxPWM    equ 0x20 ; 0x4c ; 0.6 * 127 = 76 (decimal) = 0x4c (hex)
minAnalog equ 0x92 ; 15V
fixVal    equ 0x94 ; ~15.24V (reading of about 2.9V from A/D converter)
maxAnalog equ 0x9b ; 16V
startPWM  equ 0x18 ; about 3% duty cycle to start with
#endif
#endif

; sampleTime is the number of instruction cycles
; it will take to ensure an A/D channel has been
; acquired.  A delay of that long will be
; performed before starting conversion on a
; A/D channel.
#if _USE_XTAL_18432
sampleTime equ 0x40 ; Datasheet gives 11.5us for Tcnv
#else
sampleTime equ 12 ; Datasheet gives 11.5us for Tcnv
#endif

;
; Incoming data format:
;
;    <STX><STX><DATA><CHKSUM><ETX>
;              /    \
;     ________/      \____________________________
;    /                                         \
;    <COMMAND><DLEN><ADDRL><ADDRH><ADDRU><DATA>...
;
;
; Definitions:
;
;     STX     -  Start of packet indicator
;     ETX     -  End of packet indicator
;     DATA    -  General data up to 64 bytes
;     CHKSUM  -  The 8-bit two's compliment sum of DLEN & DATA
;     COMMAND -  Base command
;     DLEN    -  Length of data buffer
;     ADDR    -  Address up to 24 bits
;     DATA    -  Data buffer (if any)
STX      equ 0x55
ETX      equ 0x04
DLE      equ 0x05
;
;
; Commands:
;
VERSION_CMD      equ 0x01 ; 
RESET_CMD        equ 0x02 ; Reset chip and start over
VPP_CMD          equ 0x03 ; Turn Vpp on/off
; (unused - ETX)
; (unused - DLE)
VDD_CMD          equ 0x06 ; Turn Vdd on/off
CLKDAT_CMD       equ 0x07 ; Set the ICSP CLK and DAT lines
READ_CLKDAT_CMD  equ 0x08 ; Read current value of CLK/DAT lines
; (unused)
ECHO_CMD         equ 0x0a ; Send back the input
BEGINPROG_CMD    equ 0x0b ; Enter prog mode: lower DAT/CLK, raise Vdd and Vpp
ENDPROG_CMD      equ 0x0c ; Exit prog mode: lower Vdd/Vpp, tristate CLK/DAT
PIC14_CMD        equ 0x0d ; Send command to 14 bit PIC
PIC14R_CMD       equ 0x0e ; Send cmd to 14 bit PIC, read response
; TBD equ 0x0f
READ_ANALOG_CMD  equ 0x10 ; Read the current value of the A/D converter
SET_ANALOG_CMD   equ 0x11 ; Set desired voltage bounds
PIC16ADDR_CMD    equ 0x12 ; Set address of a 16 bit PIC
PIC16_CMD        equ 0x13 ; Send command to 16 bit PIC
PIC16R_CMD       equ 0x14 ; Send cmd to 16 bit PIC, read response
PIC16W_CMD       equ 0x15 ; Send write cmd to 16 bit PIC, hold CLK for time P9 (1-5 msec)
READ_MASKS_CMD   equ 0x16 ; Read current value of TRIS registers
READ_PORTS_CMD   equ 0x17 ; Read current value of PORT registers
READ_PWM_CMD     equ 0x18 ; Read current PWM duty cycle
PIC14RBP_CMD     equ 0x19 ; Read a block from a 14 bit PIC
PIC14WBP_CMD     equ 0x1a ; Write a block to a 14 bit PIC
PIC14RBE_CMD     equ 0x1b ; Read a block from a 14 bit PIC EEPROM
PIC14WBE_CMD     equ 0x1c ; Write a block to a 14 bit PIC EEPROM
PIC16RB_CMD      equ 0x1d ; Read a block from a 16 bit PIC
PIC16WB_CMD      equ 0x1e ; Write a block to a 16 bit PIC
PIC16RBE_CMD     equ 0x1f ; Read a block from a 16 bit PIC EEPROM
PIC16WBE_CMD     equ 0x20 ; Write a block to a 16 bit PIC EEPROM
LAST_CMD         equ 0x20 ; [Adjust if commands added/deleted]

ERROR_OK       equ 0   ; Error codes
ERROR_BADCMD   equ 1
ERROR_TIMEOUT  equ 3

; Macro to send a core command to a 16 bit PIC target.  The
; argument is a 16 bit literal.
; Affects: ICSP_CMD, DATA0, DATA1, ICTR
coreCmd1 macro a
    clrf ICSP_CMD
    movlw (((a) & 0xff00) >> 8)
    movwf DATA1
    movlw ((a) & 0xff)
    movwf DATA0
    call PIC16CmdBase
    endm

; Macro to send a core command to a 16 bit PIC target.  The
; first argument is an 8 bit literal value to be used as the
; high byte of the command.  The second argument is the address
; of an 8 bit value to be used as the low byte of the command
; Affects: ICSP_CMD, DATA0, DATA1, ICTR
coreCmd2 macro h, l
    clrf ICSP_CMD
    movlw (h)
    movwf DATA1
    movfw (l)
    movwf DATA0
    call PIC16CmdBase
    endm

; Macro to set PWM period.  Note that this macro assumes that pwm_tmp is
; in bank 1 and that we are already in bank 1 when the macro is invoked.
setPeriod macro a ; a is 1 SFR
    swapf a, w     ; This macro will change w
    andlw 0x30     ; Set PWM duty cycle LSBs
    iorlw 0x0c     ; ECCP mode - use P1A active-high, single output
    bcf STATUS, RP0 ; bank 0
    movwf CCP1CON
    bsf STATUS, RP0 ; bank 1
    bcf STATUS, C  ; Ensure no carry goes into the result
    rrf a, w       ; Shift out LSBs of period
    movwf pwm_tmp
    bcf STATUS, C
    rrf pwm_tmp,w
    bcf STATUS, RP0
    movwf CCPR1L   ; Top 6 bits of period now in CCPR1L
    bsf STATUS, RP0
    endm

; Macro to adjust PWM.  Must be in bank 1 when invoked
adjustPWM macro
    call readAN4  ; Read analog value from AN4
    call PWMAdjust
    movlw 0x20
    movwf PWMCtr
    endm

; Macro to subtract 1 from a 16 bit value
DEC16 macro DST
    decf   (DST), F       ; 0   Decrement low byte
    incfsz (DST), W       ; 1   Check for underflow
    incf   (DST)+1, F     ; 2   Update
    decf   (DST)+1, F     ; 3   Fixup
    movfw  (DST)          ; 4  
    iorwf  (DST)+1, W     ; 5   Set Z bit
    endm

;
; Start of code
;

; Reset vector
    org 0x00

    clrf INTCON   ; Reset vector
    goto startup

; Interrupt vector
    org 0x04
    goto intServ

    org 0x05

; Main code
startup
    ; Disable interrupts until everything
    ; has been configured.
    clrf INTCON
    call setupClock
    call setupPins
#if _INDUCTOR_BASED
    call setupPWM
#endif
    call setupVars
    ; call setupInterrupts
    bcf STATUS, RP0 ; Ensure we are in RAM bank 0

programLoop
#if _INDUCTOR_BASED
    ; Every few passes through the loop we measure
    ; the output voltage and adjust PWM duty cycle
    ; as necessary.
    bsf STATUS, RP0
    decfsz PWMCtr, F
    goto noPWMCheck
    adjustPWM      ; Adjust PWM/Voltage levels

noPWMCheck
    bcf STATUS, RP0
#endif

;
; Read a packet from the PC and process it.
;
readCommand
    ; The instruction counts shown below are to track how much
    ; time it takes to get back to the RS232 receive routine.
    ; There are only about 20 instruction cycles available (at
    ; 18MHz) before the next character could arrive.  Since
    ; Windows only seems to look at CTS at the start of a
    ; WriteFile() commmand, we could be seeing an entire packet
    ; arrive without any pauses between bytes.
    ;
    ; Look for the first <STX>
    call RdRS232       ; Start counting time after getting a byte
    btfsc ERRORS, 0    ; 0
    goto programLoop   ; Timeout on first STX - do other processing
    movfw SERIAL_IN    ; 2
    xorlw STX          ; 3 first <STX>
    btfss STATUS, Z    ; 4
    goto programLoop

;
; Read and parse the data.
;
StartOfLine    
    call RdRS232       ; 6..7  (back in plenty of time)
    btfsc ERRORS, 0    ; 0
    ; TBD: Should we break out of this loop after a while?
    goto StartOfLine

checkSTX2
    movfw SERIAL_IN    ; 2
    xorlw STX          ; 3   second <STX>
    btfss STATUS, Z    ; 4
    goto programLoop   ;      Unexpected value - start all over

STXOk
    ; Setup indirect addressing so that the packet contents go
    ; the right place
    movlw PACKET       ; 6 Point to the buffer
    movwf FSR          ; 7

    clrf CHECKSUM      ; 8 Reset checksum

GetNextDat
    call RdRS232       ; 9..10    16..17   Get the data
    btfsc ERRORS, 0    ; 0
    goto GetNextDat
    movfw SERIAL_IN    ; 2

    xorlw ETX          ; 3  Check for a ETX
    btfsc STATUS, Z    ; 4
    goto CheckSum      ; Done with packet.  Verify checksum

NotETX
    movfw SERIAL_IN    ; 6
    xorlw DLE          ; 7  Check for a DLE
    btfsc STATUS, Z    ; 8
    goto DLEByte       ; 9..10  Got DLE - use next byte

AddPacketByte
    movfw SERIAL_IN    ; 10  (or 4 if DLEByte was executed)
    movwf INDF         ; 11  Store the data
    addwf CHECKSUM, F  ; 12  Get sum

    ; Increment pointer to next position in input buffer
    incf FSR, F        ; 13

    ; Get next byte in the buffer
    goto GetNextDat    ; 14..15

; Read an escaped byte and put into buffer
DLEByte
    call RdRS232       ; 11..12  Yes, Get the next byte (tight timing)
    btfsc ERRORS, 0    ; 0
    goto DLEByte       ; Wait till we get it
    goto AddPacketByte ; 2..3 (quicker than standard read/check)

CheckSum    
    movf CHECKSUM, F   ; Checksum test
    btfss STATUS, Z
    goto programLoop   ; Checksum failed.  Wait for next packet

processCommand
    ; Make sure command is in the correct range
    movfw COMMAND
    sublw LAST_CMD
    btfss STATUS, C
    goto badCommand

    ; Computed goto, based on value of command
    movfw COMMAND
    addwf PCL, F
    goto programLoop    ; 0x00 - NOP
    goto readVersionCmd ; 0x01 - Return the version of this firmware
    goto startup        ; 0x02 - Reset chip and start over
    goto vppCmd         ; 0x03 - Turn Vpp on/off
    goto programLoop    ; 0x04 - (unused - ETX)
    goto programLoop    ; 0x05 - (unused - DLE)
    goto vddCmd         ; 0x06 - Turn Vdd on/off
    goto setCLKDATCmd   ; 0x07 - Set the ICSP CLK and DAT lines
    goto readCLKDATCmd  ; 0x08 - Read current value of CLK/DAT lines
    goto programLoop    ; 0x09 - (unused)
    goto echoCmd        ; 0x0a - Send back the input
    goto beginProgCmd   ; 0x0b - Enter prog mode: lower DAT/CLK, raise Vdd and Vpp
    goto endProgCmd     ; 0x0c - Exit prog mode: lower Vdd/Vpp, tristate CLK/DAT
    goto PIC14Cmd       ; 0x0d - Send command to 14 bit PIC
    goto PIC14RCmd      ; 0x0e - Send cmd to 14 bit PIC, read response
    goto programLoop    ; 0x0f - (unused)
    goto readAnalogCmd  ; 0x10 - Read the current value of the A/D converter
    goto setAnalogCmd   ; 0x11 - Set desired voltage bounds
    goto PIC16SetAddrCmd; 0x12 - Set address of a 16 bit PIC
    goto PIC16Cmd       ; 0x13 - Send command to 16 bit PIC
    goto PIC16ReadCmd   ; 0x14 - Send cmd to 16 bit PIC, read response
    goto PIC16WriteCmd  ; 0x15 - Send write cmd to 16 bit PIC, hold CLK for time P9 (1-5 msec)
    goto readMasksCmd   ; 0x16 - Read current value of TRIS registers
    goto readPortsCmd   ; 0x17 - Read current value of PORT registers
    goto readPWMCmd     ; 0x18 - Read current PWM duty cycle
    goto PIC14RBPgm     ; 0x19 - Read a block of code from 14 bit PIC
    goto PIC14WBPgm     ; 0x1a - Write a block of code to 14 bit PIC
    goto PIC14RBDat     ; 0x1b - Read a block of data from 14 bit PIC EEPROM
    goto PIC14WBDat     ; 0x1c - Write a block of data to 14 bit PIC EEPROM
    goto PIC16RBlock    ; 0x1d - Read a block of bytes from 16 bit PIC
    goto PIC16WBlock    ; 0x1e - Write a block of bytes to 16 bit PIC
    goto PIC16REBlock   ; 0x1f - Read a block of data from a 16 bit PIC EEPROM
    goto PIC16WEBlock   ; 0x20 - Write a block of data to a 16 bit PIC EEPROM

;
; Set error value to indicate we got a command we can't interpret
;
badCommand
    movlw ERROR_BADCMD
    movwf ERRORS
    goto programLoop
;
; Commands
;

; Send back the current firmware version
; In : <STX><STX>[<0x01><0x01>]<0xfe><CHKSUM><ETX>
; Out: <STX><STX>[<0x01><DLEN><N/A><VERL><VERH>]<CHKSUM><ETX>
readVersionCmd
    movlw 0x03
    movwf DATALEN
    movlw VERSION0
    movwf DATA0
    movlw VERSION1
    movwf DATA1

    movlw 0x05
    goto WritePacket


;
; Send back exactly what we got
;
; In : <STX><STX>[<cmd><DATALEN><DATA...>]<CHKSUM><ETX>
; Out: <STX><STX>[<cmd><DATALEN><DATA...>]<CHKSUM><ETX>
echoCmd
    movfw DATALEN
    addlw 0x02
    goto WritePacket

;
; Turn Vpp on/off based on bit 1 of the data word
;
; In : <STX><STX>[<cmd><0x01><flags>]<CHKSUM><ETX>
; Out: <STX><STX>[<cmd>]<CHKSUM><ETX>
vppCmd
    btfsc DATABUF0, 1
    goto Vpp1High
    bsf VPP1      ; Pull Vpp1 low (action is opposite of logic level)
    goto Vpp2Cmd
Vpp1High
    bcf VPP1      ; Pull Vpp1 high

    ; TBD: Turn Vpp on/off separately on the two
    ; pins of the programming socket that need it.
    ; Rationale: 40 pin chips are unhappy with the
    ; 13V comeing in on one of the data lines - they
    ; lock up instead of allowing programming.  Smaller
    ; chips don't have this problem - only one Vpp
    ; line ever touches them.
Vpp2Cmd
#if 0
    btfsc DATABUF0, 2
    goto Vpp2High
    bsf VPP2
    goto VppDone
Vpp2High
    bcf VPP2
#endif

VppDone
    movlw 0x01
    movwf DATALEN
    movlw 0x04
    goto WritePacket

;
; Turn Vdd on/off based on bit 0 of the data word
;
; In : <STX><STX>[<cmd><0x01><flags>]<CHKSUM><ETX>
; Out: <STX><STX>[<cmd>]<CHKSUM><ETX>
vddCmd
    movlw 0x01
    movwf DATALEN
    btfsc DATABUF0, 1
    goto VddHigh
    bsf VDD
    movlw 0x04
    goto WritePacket
VddHigh
    bcf VDD
    movlw 0x04
    goto WritePacket

;
; Return values of Vpp and Vdd control lines - the
; voltage levels are inverted from logic levels, so
; report them as on if the actual pin is off and vice
; versa.
;
; In : <STX><STX>[<cmd><0x00><ETX>
; Out: <STX><STX>[<cmd><DLEN><N/A><flags><N/A>]<CHKSUM><ETX>
readCLKDATCmd
    movlw 0x01
    movwf DATALEN
    clrf DATA0
    btfss VDD
    bsf DATA0, 0
    btfss VPP1
    bsf DATA0, 1
#if 0
    btfss VPP2
    bsf DATA0, 2
#endif
    btfsc ICSP_CLK
    bsf DATA0, 3
    btfss ICSP_DAT
    bsf DATA0, 4
    clrf DATA1

    movlw 0x05
    goto WritePacket

;
; Turn CLK on/off based on bit 0 of data word; turn DAT on/off
; based on bit 1 of data word
;
; In : <STX><STX>[<cmd><0x01><flags><CHKSUM><ETX>
; Out:  <STX><STX>[<cmd>]<CHKSUM><ETX>
setCLKDATCmd
    btfsc DATABUF0, 0
    goto CLKHigh
    bcf ICSP_CLK
    goto checkDAT
CLKHigh
    bsf ICSP_CLK
checkDAT
    btfsc DATABUF0, 1
    goto DATHigh
    bcf ICSP_DAT
    goto sendAck
DATHigh
    bsf ICSP_DAT
    goto sendAck

;
; Read the current value of the A/D converter
;
; In : <STX><STX>[<cmd><0x00><CHKSUM><ETX>
; Out: <STX><STX>[<cmd><DLEN><N/A><AD Val><N/A>]<CHKSUM><ETX>
readAnalogCmd
    call readAN4
    bsf STATUS, RP0
    movfw analogVal
    bcf STATUS, RP0
    movwf DATA0
    movlw 0x03
    movwf DATALEN

    movlw 0x05
    goto WritePacket

;
; Set the minimum and maximum voltage thresholds used for PWM
; adjustment.
;
; In : <STX><STX>[<cmd><0x02><Min V><Max V><CHKSUM><ETX>
; Out: <STX><STX>[<cmd>]<CHKSUM><ETX>
setAnalogCmd
    movlw 0x00
    goto sendAck
;
; Read the current value of the PWM duty cycle
;
; In : <STX><STX>[<cmd><0x00><CHKSUM><ETX>
; Out: <STX><STX>[<cmd><DLEN><N/A><PWM Val><N/A>]<CHKSUM><ETX>
readPWMCmd
#if _INDUCTOR_BASED
    bsf STATUS, RP0
    movfw PWMVal
    bcf STATUS, RP0
    movwf DATA0
    movlw 0x03
    movwf DATALEN
    clrf DATA1

    movlw 0x05
    goto WritePacket
#endif
    movlw 0x00
    goto sendAck

;
; Read the current value of port A and C direction masks
;
; In : <STX><STX>[<cmd><0x00><CHKSUM><ETX>
; Out: <STX><STX>[<cmd><0x02><N/A><mask0><mask1>]<CHKSUM><ETX>
readMasksCmd
    bsf STATUS, RP0
    movfw TRISA
    bcf STATUS, RP0
    andlw 0x1f
    movwf DATA0
    ; Move bits left in preparation for rotating in bits
    rlf DATA0, F
    rlf DATA0, F
    rlf DATA0, F
    bsf STATUS, RP0
    movfw TRISC
    bcf STATUS, RP0
    movwf DATA1
    ; rotate the bottom 3 bits of TRISC into RESULTL
    rrf DATA1, F
    rrf DATA0, F
    rrf DATA1, F
    rrf DATA0, F
    rrf DATA1, F
    rrf DATA0, F
    movfw DATA1
    andlw 0x03
    movwf DATA1

    movlw 0x03
    movwf DATALEN
    movlw 0x05
    goto WritePacket

;
; Read the current value of port A and C
;
; In : <STX><STX>[<cmd><0x00><CHKSUM><ETX>
; Out: <STX><STX>[<cmd><0x03><N/A><val0><val1>]<CHKSUM><ETX>
readPortsCmd
    movlw 0x03
    movwf DATALEN
    movfw PORTA
    movwf DATA0
    movfw PORTC
    movwf DATA1
    movlw 0x05
    goto WritePacket

;
; Enter programming mode
;
; Vdd and Vpp must be off when entering this routine
;
; In : <STX><STX>[<cmd><0x00><CHKSUM><ETX>
; Out: <STX><STX>[<cmd>]<CHKSUM><ETX>
beginProgCmd
    ; Common code - pull CLK and DAT lines low
    bsf STATUS,RP0    ; bank 1
    bcf ICSP_CLK_MASK
    bcf ICSP_DAT_MASK
    bcf STATUS,RP0    ; bank 0
    bcf ICSP_CLK      ; lower bits
    bcf ICSP_DAT
    movlw 0x02
    call delay        ; 22 instructions - 11 usec @ 8MHz, ~4.8 usec @ 18.432MHz
    ; Generic entry: raise Vdd, wait 5 usec (14 bit)/100 nsec (16 bit);
    ; raise Vpp; wait 5 usec (14 bit) 2 usec (16 bit)
    ; TBD: add code for Vdd change
    bcf VPP1           ; Pull Vpp high
    movlw 0x06
    call delay        ; 22 instructions - 11 usec @ 8MHz, ~4.8 usec @ 18.432MHz
    ; After entering programming mode, assume PC is at address 0
    clrf PCADDR0
    clrf PCADDR1
    clrf PCADDR2
    ; Now safe to toggle CLK/DAT lines
    goto sendAck      ; done

;
; Tristate CLK and DATA, turn Vdd and Vpp off
;
; In : <STX><STX>[<cmd><0x00><CHKSUM><ETX>
; Out: <STX><STX>[<cmd>]<CHKSUM><ETX>
endProgCmd
    bsf STATUS,RP0    ; bank 1
    bsf ICSP_CLK_MASK ; tristate ICSP_CLK line
    bsf ICSP_DAT_MASK ; tristate ICSP_DAT line
    bcf STATUS,RP0    ; bank0
    bcf ICSP_CLK      ; ensure bits are low
    bcf ICSP_DAT
    bsf VPP1           ; Pull Vpp low
    ;bcf VDD           ; TBD: pull Vdd low
    goto sendAck      ; done


;
;
; 14 bit PIC code
;
;

;
; Transmit command to the PIC - 6 bits.  Command must be in
; ICSP_CMD.  Affects: ICTR, ICSP_CMD.
;
PIC14TxCmd
    ; Ensure ICSP_DAT line is an output and starts low.
    bsf STATUS, RP0
    bcf ICSP_DAT_MASK
    bcf STATUS, RP0
    bcf ICSP_DAT

    movlw 0x06
    movwf ICTR

P14CmdLoop1
    bsf ICSP_CLK      ; raise clock
    bcf ICSP_DAT      ; default data low
    rrf ICSP_CMD, F   ; get next bit in carry
    btfsc STATUS,C    ; is the bit set?
    bsf ICSP_DAT      ; yes - raise ICSP_DAT
    bcf ICSP_CLK      ; lower clock, PIC latches data
    decfsz ICTR, F    ; dec ctr
    goto P14CmdLoop1  ; loop if not zero
    bcf ICSP_DAT      ; Ensure ICSP_DAT low
    return

;
; Transmit a data payload to the target 14 bit PIC and
; read any response data.  Response will be in
; PICDATA1:PICDATA0 (lsb in PICDATA0).  PIC14TxCmd must
; have been called immediately before this routine.
; Affects: ICTR, DATA0, DATA1, PICDATA0, PICDATA1.
PIC14TxRxData
    movlw 0x10
    movwf ICTR

    ; Prepare 14 bit value to send to target.  We want it
    ; to sit in the middle of the 16 bits we send (0 for
    ; bit 0 and bit 15).
    bcf STATUS, C
    rlf DATA0, F
    rlf DATA1, F

; Transmit/Receive data to/from the PIC - 16 bits
P14DatLoop1
    bsf ICSP_CLK      ; raise clock
    bcf ICSP_DAT      ; default data low
    rrf DATA1, F      ; rotate data and read low bit
    rrf DATA0, F
    btfsc STATUS,C
    bsf ICSP_DAT      ; pull data high
    nop               ; ensure TSET1 is satisfied
incCtr2
    bcf ICSP_CLK      ; lower clock, PIC latches data
    ; Read any output from the PIC
    rrf PICDATA1, F
    rrf PICDATA0, F
    bcf PICDATA1, 7   ; Default is clear bit

    bsf STATUS, RP0   ; bank 1
    bsf ICSP_DAT_MASK ; Ensure ICSP_DAT is an input
    bcf STATUS, RP0   ; bank 0
    btfsc ICSP_DAT    ; check the value of output from PIC
    bsf PICDATA1, 7   ; set if PIC is sending a 1

    bsf STATUS, RP0   ; bank 1
    bcf ICSP_DAT_MASK ; Set ICSP_DAT as output to send data
    bcf STATUS, RP0   ; bank 0

    decfsz ICTR, F    ; dec ctr
    goto P14DatLoop1  ; loop if not zero
    bcf ICSP_DAT      ; pull ICSP_DAT low

    ; Bits 0 and 15 aren't valid.  Rotate the result and mask off
    ; the two high bits of PICDATA1.  Final result sent back will
    ; the actual 14 bit value in code memory.
    rrf PICDATA1, F
    rrf PICDATA0, F
    movlw 0x3f
    andwf PICDATA1, F

    return;

;
; Clock and data should both be OUTPUT/LOW at this stage...
; we're going to clock 6 bits of command data from ICSP_CMD, then
; clock 16 bits of data from DATA1/LO - in the case of read commands,
; data will be tristate anyway (handy pullup).
;
; In : <STX><STX>[<cmd><0x03><ICSP Cmd><Data0><Data1>]<CHKSUM><ETX>
; Out: <STX><STX>[<cmd>]<CHKSUM><ETX>
; Out: <STX><STX>[<cmd><0x03><ICSP Cmd><Data0><Data1>]<CHKSUM><ETX>
PIC14Cmd
    call PIC14TxCmd   ; Send the ICSP command
    goto sendAck      ; Send ack response

PIC14RCmd
    call PIC14TxCmd
    call PIC14TxRxData

    ; Send back the command #, the ICSP command, and any data
    ; we got from the PIC (which will be in DATA0 and DATA1)
    movlw 0x03
    movwf DATALEN
    movfw PICDATA0
    movwf DATA0
    movfw PICDATA1
    movwf DATA1
    movlw 0x05
    goto WritePacket

;
; Read a block of data from a 14 bit PIC
;
; In:    <STX><STX>[<0x19><DLEN><ADDRL><ADDRH><ADDRU><DATA>...]<CHKSUM><ETX>
; Out:   <STX><STX>[<0x19><DLEN><ADDRL><ADDRH><ADDRU>]<CHKSUM><ETX>
;
; DLEN on input tells us how many words of information are being
; requested.  The maximum is 32 (64 bytes) to prevent overflow.
; TBD: Use the input address to find the data...
;
; Affects: TEMP
;
PIC14RBlock
    ; TBD: Save contents of ADDRU:ADDRH:ADDRL so they can be
    ; restored after we use the overlaid locations for
    ; ICSP_CMD:DATA0:DATA1

    ; TBD: Pull out requested address and compare to the
    ; one saved in PCADDR2:PCADDR1:PCADDR0.  If they don't
    ; match, then either increment until we get there, or
    ; reset the target and increment.  (Note, on 16 bit parts
    ; we can directly change the target address.)

    ; Prepare response packet
    movlw DATA_BUF   ; Point to the output buffer
    movwf FSR        ;

    movfw CMDOPT     ; Get # of bytes requested
    sublw 0x40       ; Is # requested more than 64?
    btfsc STATUS, C
    goto PIC14RB1    ; Nope, continue on

    ; Too many bytes requested.  Limit to 64
    movlw 0x40
    movwf CMDOPT     ; Send back the reduced #

PIC14RB1
    movfw CMDOPT     ; Move requested # of bytes into TEMP
    movwf TEMP       ; to use for loop counter
    bcf STATUS, C    ; Don't rotate in any leftovers
    rrf TEMP, F      ; Divide by 2 for # of words

PIC14RBLoop
    ; Read data from the current memory location
    movfw CMDTEMP
    movwf ICSP_CMD
    call PIC14TxCmd
    call PIC14TxRxData

    ; Move response to output buffer (LSB first)
    movfw PICDATA0
    movwf INDF
    incf FSR, F
    movfw PICDATA1
    movwf INDF
    incf FSR, F

    ; Increment the memory address
    movlw PIC14_INCADDR ; Increment address command for 14 bit PICs
    movwf ICSP_CMD
    call PIC14TxCmd

    ; See if there are any more words to read
    decfsz TEMP, F
    goto PIC14RBLoop

    movfw CMDOPT    ; Output size is CMDOPT + cmd/dlen/address/opt
    addlw 0x6
    goto WritePacket ; Send back the data
    return;

PIC14RBPgm
    movlw PIC14_READDATA   ; Read code command for 14 bit PICs
    movwf CMDTEMP
    goto PIC14RBlock

PIC14RBDat
    movlw PIC14_READE2DATA ; Read data command for 14 bit PICs
    movwf CMDTEMP
    goto PIC14RBlock

;
; Write a block of code to a 14 bit PIC
;
; In:    <STX><STX>[<0x01><DLEN><ADDRL><ADDRH><ADDRU><CMDOPT><DATA>]<CHKSUM><ETX>
; In:    <STX><STX>[<0x01><DLEN><PROGCMD><ENDPROGCMD><N/A><CMDOPT><DATA>]<CHKSUM><ETX>
; OUT:    <STX><STX>[<0x01><DLEN><N/A><N/A><N/A><DATA>]<CHKSUM><ETX>
;

PIC14WBPgm
    movfw DATABUF0
    movwf PROGTEMP
    movfw DATABUF1
    movwf ENDPROGTEMP

    movlw DATA_BUF   ; Point to the output buffer
    movwf FSR        ;

    movfw CMDOPT     ; Get # of bytes requested
    sublw 0x40       ; Is # requested more than 64?
    btfsc STATUS, C
    goto PIC14WB1    ; Nope, continue on

    ; Too many bytes requested.  Limit to 64
    movlw 0x40
    movwf CMDOPT     ; Send back the reduced #

PIC14WB1
    movfw CMDOPT     ; Move requested # of bytes into TEMP
    movwf TEMP       ; to use for loop counter
    bcf STATUS, C    ; Don't rotate in any leftovers
    rrf TEMP, F      ; Divide by 2 for # of words

PIC14WBLoop
    ; Up to MAX_RETRIES before abandoning write
    movlw MAX_RETRIES
    movwf RETRYCTR
PIC14WBRetryLoop

    ; Read data from the current memory location
    movlw PIC14_READDATA
    movwf ICSP_CMD
    call PIC14TxCmd
    call PIC14TxRxData

    ; Compare to desired value. If same then
    ; go to next one in block.
    ; Move response to output buffer
    movfw PICDATA0    ; Compare low bytes
    xorwf INDF, W
    btfss STATUS, Z
    goto PIC14WBRetry1
    incf FSR, F
    movfw PICDATA1    ; Compare high bytes
    xorwf INDF, W
    btfss STATUS, Z
    goto PIC14WBRetry2
    incf FSR, F

    goto PIC14WBOk    ; Write confirmed ok

PIC14WBRetry2
    decf FSR, F      ; Back up pointer
PIC14WBRetry1
    ; If not a match then see if we have tried
    ; enough times
    decfsz RETRYCTR, F
    goto PIC14WBWrite
PIC14WBFail
    ; Write failed, send back neg ack
    movlw 0xee
    movwf COMMAND
    ; clrf COMMAND
    goto sendAck

PIC14WBWrite
    ; Move current data into DATA0:DATA1 without moving INDF
    ; past the current data value.
    movfw INDF
    movwf DATA0
    incf FSR, F
    movfw INDF
    movwf DATA1
    decf FSR, F

    ; Load the data
    movlw PIC14_LOADDATA
    movwf ICSP_CMD
    call PIC14TxCmd
    call PIC14TxRxData

    ; Initiate programming
    ; movlw PIC14_BEGINPROG
    movfw PROGTEMP
    movwf ICSP_CMD
    call PIC14TxCmd

    ; Wait for programming to complete (max 2.5 msec according to
    ; datasheets).  Passes needed is (clock / 4) * (# msec / 9)
    ; (9 instructions per pass through the loop, 4 clocks per
    ; instruction.
    ;   At 18.432 MHz: (18.432 / 4) * (2500 / 9) = 1280 (0x500)
    ;   At 8 MHz: (8 / 4) * (2500 / 9) = 555.5 (0x22c)
    ; TBD: Make the duration of this delay based on an input.  It
    ; can change quite a bit
#if _USE_XTAL_18432
    clrf x
    movlw 0x05
    movwf x+1
#else
    movlw 0x2c
    movwf x
    movlw 0x02
    movwf x+1
#endif
PIC14ProgLoop
    DEC16 x
    btfss STATUS, Z
    goto PIC14ProgLoop

    ; After the programming cycle is done, some parts need to
    ; have an end program command (e.g., 16F88).  If the value
    ; of ENDPROGTEMP is nonzero then send it to the PIC.
    movfw ENDPROGTEMP
    btfsc STATUS, Z        ; See if ENDPROGTEMP is zero
    goto PIC14EndProgDone  ; It is - don't need end prog command

    ; ENDPROGTEMP is set, so we need to send it to the chip
    movwf ICSP_CMD
    call PIC14TxCmd

PIC14EndProgDone
    ; Programming delay done.  See if the word was successfully written
    goto PIC14WBRetryLoop

PIC14WBOk
    ; Increment the memory address
    movlw PIC14_INCADDR    ; Increment address command for 14 bit PICs
    movwf ICSP_CMD
    call PIC14TxCmd

    ; See if there are any more words to read
    decfsz TEMP, F
    goto PIC14WBLoop

#if 1
    goto sendAck
#else
    movfw CMDOPT           ; Output size is CMDOPT + cmd/dlen/address/opt
    addlw 0x6
    goto WritePacket       ; Send back the data
#endif
    return;

;
; Write a block of data to a 14 bit PIC
;
; In:    <STX><STX>[<0x01><DLEN><ADDRL><ADDRH><ADDRU><CMDOPT><DATA>]<CHKSUM><ETX>
; OUT:    <STX><STX>[<0x01><DLEN><ADDRL><ADDRH><ADDRU><DATA>]<CHKSUM><ETX>
;
PIC14WBDat
    movfw DATABUF0
    movwf PROGTEMP
    movfw DATABUF1
    movwf ENDPROGTEMP

    movlw DATA_BUF   ; Point to the output buffer
    movwf FSR        ;

    movfw CMDOPT     ; Get # of bytes requested
    sublw 0x40       ; Is # requested more than 64?
    btfsc STATUS, C
    goto PIC14WBE1   ; Nope, continue on

    ; Too many bytes requested.  Limit to 64
    movlw 0x40
    movwf CMDOPT     ; Send back the reduced #

PIC14WBE1
    movfw CMDOPT     ; Move requested # of bytes into TEMP
    movwf TEMP       ; to use for loop counter

PIC14WBELoop
    ; Up to MAX_RETRIES before abandoning write
    movlw MAX_RETRIES
    movwf RETRYCTR
PIC14WBERetryLoop

    ; Read data from the current memory location
    movlw PIC14_READE2DATA
    movwf ICSP_CMD
    call PIC14TxCmd
    call PIC14TxRxData

    ; Compare low byte to desired value. If same then
    ; move response to output buffer
    movfw PICDATA0    ; Compare low bytes
    xorwf INDF, W
    btfsc STATUS, Z
    goto PIC14WBEOk   ; Write confirmed ok

PIC14WBERetry1
    ; If not a match then see if we have tried
    ; enough times
    decfsz RETRYCTR, F
    goto PIC14WBEWrite
PIC14WBEFail
    ; Write failed, send back neg ack
    movlw 0xee
    movwf COMMAND
#if 0
    goto sendAck
#else
    movlw 0x01
    movwf DATALEN
    movfw TEMP
    movwf DATABUF0
    movlw 0x03
    goto WritePacket
#endif

PIC14WBEWrite
    ; Move current data into DATA0 without moving INDF
    ; past the current data value.  Clear high byte (not
    ; used when writing EEPROM data.
    movfw INDF
    movwf DATA0
    clrf DATA1

    ; Load the data
    movlw PIC14_LOADE2DATA
    movwf ICSP_CMD
    call PIC14TxCmd
    call PIC14TxRxData

    ; Initiate programming
    movfw PROGTEMP
    movwf ICSP_CMD
    call PIC14TxCmd

    ; Wait for programming to complete (5 to 6 msec according to
    ; datasheets).  Passes needed is (clock / 4) * (# msec / 9)
    ; (9 instructions per pass through the loop, 4 clocks per
    ; instruction.
    ;   At 18.432 MHz: (18.432 / 4) * (6000 / 9) = 3072 (0xC00)
    ;   At 8 MHz: (8 / 4) * (6000 / 9) = 1333.3 (0x536)
#if _USE_XTAL_18432
    clrf x
    movlw 0xc0
    movwf x+1
#else
    movlw 0x36
    movwf x
    movlw 0x05
    movwf x+1
#endif
PIC14ProgELoop
    DEC16 x
    btfss STATUS, Z
    goto PIC14ProgELoop

    ; After the programming cycle is done, some parts need to
    ; have an end program command (e.g., 16F88).  If the value
    ; of ENDPROGTEMP is nonzero then send it to the PIC.
    movfw ENDPROGTEMP
    btfsc STATUS, Z        ; See if ENDPROGTEMP is zero
    goto PIC14EndProgEDone ; It is - don't need end prog command

    ; ENDPROGTEMP is set, so we need to send it to the chip
    movwf ICSP_CMD
    call PIC14TxCmd

PIC14EndProgEDone
    ; Programming delay done.  See if the word was successfully written
    goto PIC14WBERetryLoop

PIC14WBEOk
    ; Increment the memory address
    incf FSR, F            ; Move to next byte in the buffer
    movlw PIC14_INCADDR    ; Increment address command for 14 bit PICs
    movwf ICSP_CMD
    call PIC14TxCmd

    ; See if there are any more bytes to read
    decfsz TEMP, F
    goto PIC14WBELoop

    movfw CMDOPT           ; Output size is CMDOPT + cmd/dlen/address/opt
    addlw 0x6
    goto WritePacket       ; Send back the data
    return;

;
;
; 16 bit PIC code
;
;

;
; Set the target chips address to ADDRU:ADDRH:ADDRL
PIC16SetTablePtr
    ; Save the address
    movfw ADDRU
    movwf PCADDR2
    movfw ADDRH
    movwf PCADDR1
    movfw ADDRL
    movwf PCADDR0

    coreCmd2 0x0e, PCADDR2
    coreCmd1 0x6ef8
    coreCmd2 0x0e, PCADDR1
    coreCmd1 0x6ef7
    coreCmd2 0x0e, PCADDR0
    coreCmd1 0x6ef6

    ; A final NOP to finish things off
    coreCmd1 0x0000

    return

; Set the target address
; In:    <cmd><DLEN><ADDRL><ADDRH><ADDRU>
PIC16SetAddrCmd
    call PIC16SetTablePtr

    movlw 0x04
    movwf DATALEN
    movfw PCADDR2
    movwf ADDRU
    movfw PCADDR1
    movwf ADDRH
    movfw PCADDR0
    movwf ADDRL
    movlw 0x06
    goto WritePacket

; Setup CLK/DAT lines and send 4 bit command to a 16 bit
; PIC processor.  ICSP_CLK and ICSP_DAT should both be
; low at this point.  Serial clock period P2 ranges from
; 100 nsec to 1 usec, depending on Vdd.  Even at 18MHz,
; the loop below will not be shorter than the minimum
; (loop is 10 instruction cycles long, ~2 usec at 18MHz,
; ~4.6 usec at 8MHz).
PIC16CmdSetup
    ; Ensure ICSP_DAT line is an output and starts low.
    bsf STATUS, RP0
    bcf ICSP_DAT_MASK
    bcf STATUS, RP0
    bcf ICSP_DAT

    movlw 0x04            ; 4 bits in the command
    movwf ICTR
pic16CmdLoop
    bsf ICSP_CLK          ; Pull clock high
    bcf ICSP_DAT          ; Start data low, raise if cmd bit set
    rrf ICSP_CMD, F       ; Read bit in command
    btfsc STATUS,C        ; If set then raise ICSP_DAT
    bsf ICSP_DAT          ; Pull ICSP_DAT high
    bcf ICSP_CLK          ; PIC latches in data when clock goes low
    nop
    decfsz ICTR, F        ; Fall out of bottom if all 4 bits sent
    goto pic16CmdLoop     ; Keep sending the command

    ; Need to ensure P5 time (20 nsec) is met here.  Shouldn't
    ; be an issue if this code is running at less than 200 MHz
    bcf ICSP_DAT          ; Ensure DAT is low before continuing
    rrf ICSP_CMD, F       ; top and bottom of byte should be swapped
    return

; Transmit command to the PIC - 4 bit command for 16 bit devices
; followed by 16 bits of data output.  This routine is for commands
; that do not require a data response from the PIC.  Must be in
; bank 0, ICSP_CLK aand ICSP_DAT must be low.
; Affects: ICTR, DATA0, DATA1.
PIC16CmdBase
    ; Send the 4 bit command to the target PIC
    call PIC16CmdSetup

    ; Send 16 bit data payload to the target
    movlw 0x10
    movwf ICTR
pic16DATLoop1
    ; Serial clock period must be > P2 (100 nsec to 1 usec)
    bsf ICSP_CLK      ; raise clock
    bcf ICSP_DAT      ; default data low
    rrf DATA1, F      ; rotate data and read low bit
    rrf DATA0, F
    btfsc STATUS,C
    bsf ICSP_DAT      ; pull data high
    bcf ICSP_CLK      ; lower clock, PIC latches data
    nop
    decfsz ICTR, F    ; dec ctr
    goto pic16DATLoop1; loop if not zero
    bcf ICSP_DAT      ; pull DAT low

    return


; Send a command and 16 bit payload to the target.
; In:  <cmd><data len><ICSP cmd><data high byte><data low byte>
; Out: <STX><STX>[<cmd><0x03><ICSP cmd><data0><data1>]<CHKSUM><ETX>
PIC16Cmd
    ; Save off the input data
    movfw ICSP_CMD
    movwf CMDTEMP
    movfw DATA0
    movwf PICDATA0
    movfw DATA1
    movwf PICDATA1

    call PIC16CmdBase

    ; Send back input as an ack
    movlw 0x03
    movwf DATALEN
    movfw CMDTEMP
    movwf ICSP_CMD
    movfw PICDATA0
    movwf DATA0
    movfw PICDATA1
    movwf DATA1

    movlw 0x05        ; 5 - cmd, data len, ICSP cmd, Data0, Data1
    goto WritePacket

; Transmit a command to the PIC, then hold the CLK line high for
; the maximum programming time (P9).  This routine is necessary
; when initiating a memory write on the target.  
PIC16WriteBase
    movlw 0x04            ; 4 bits in the command
    movwf ICTR
pic16WriteLoop
    bsf ICSP_CLK          ; Pull clock high
    bcf ICSP_DAT          ; Start data low, raise if cmd bit set
    rrf ICSP_CMD, F       ; Read bit in command
    btfsc STATUS,C        ; If set then raise ICSP_DAT
    bsf ICSP_DAT          ; Pull ICSP_DAT high
    decfsz ICTR, F        ; Done writing if all 4 bits sent
    goto continueWrite    ; Not done - pull clock low and continue
    goto cmdWriteDone
continueWrite
    bcf ICSP_CLK
    goto pic16WriteLoop
cmdWriteDone
; Need to delay for time P9 (1 msec) after clocking out the instruction.
;
; Delay loop for 9*N instructions.  The duration of the delay
; will be determined by the clock speed.  2 instruction cycles
; per microsecond for an 8MHz clock, about 5 for an 18MHz clock.
#if _USE_XTAL_18432
    ; 4608 (0x1200) instructions per msec at 18.432 MHz.
    ; 9 instructions per loop, giving 512 (0x200) loops.
    clrf x
    movlw 0x02
    movwf x+1
#else
    ; 2000 (0x07d0) instructions per msec at 8MHz.  (223 loops)
    movlw 0xdf
    movwf x
    clrf x+1
#endif

cmdWriteDelay
    DEC16 x            ; 0..5
    btfss STATUS, Z    ; 6
    goto cmdWriteDelay ; 7..8

    bcf ICSP_CLK        ; now pull clock low

   ; delay for time P10 (5 usec) for high voltage discharge.
#if _USE_XTAL_18432
    ; At 18MHz, there are about 5 instructions per usec, so
    ; we need to wait about 25 instruction cycles for 5 usec.
    movlw 0x07        ; 0
    call delay        ; 1..25
#else
    ; At 8MHz, there are 2 instructions per usec, so
    ; we need to wait 10 instruction cycles for 5 usec.
    movlw 0x02        ; 0
    call delay        ; 1..10
#endif

    ; Clear data registers and clock out 16 bits.  No data
    ; going either way - all commands have a payload, even
    ; if they are ignored.
    clrf DATA1
    clrf DATA0
    movlw 0x10
    movwf ICTR
    call pic16DATLoop1

    ; All done with programming write
    return

; In:  <cmd><data len><ICSP cmd><data0><data1>
; Out: <cmd><0x03><ICSP cmd><data0><data1>
PIC16WriteCmd
    call PIC16WriteBase
    goto sendAck

; Transmit command to the PIC - 4 bit command for 16 bit devices
; followed by 16 bits of data I/O.  We are looking for 8 bits of
; data in the second set of 8 bits
; Affects; ICTR, ICSP_CMD, PICDATA0
PIC16ReadBase
    call PIC16CmdSetup

    ; Toggle the clock 8 times and ignore anything that comes back
    movlw 0x08
    movwf ICTR
    bcf ICSP_DAT      ; Hold ICSP_DAT low for first 8 bits
 
    ; Serial clock period must be > P2 (100 nsec to 1 usec)  With the
    ; 18MHz crystal, each instruction takes 217 nsec.  The following
    ; loop takes 7 instructions per pass for a minimum time of about
    ; 1.3 usec for the serial clock period.
pic16RdLoop1
    bsf ICSP_CLK      ; 0      raise clock
    nop               ; 1
    nop               ; 2
    bcf ICSP_CLK      ; 3      lower clock, PIC latches data
    decfsz ICTR, F    ; 4      dec ctr
    goto pic16RdLoop1 ; 5..6   loop if not zero

    ; Ensure P6 (20 nsec) between first 8 bits and second

    ; Toggle 8 more times and read the result byte
    bsf STATUS, RP0   ; Bank 1
    bsf ICSP_DAT_MASK ; Set DAT line as input
    bcf STATUS, RP0   ; Bank 0
    movlw 0x08
    movwf ICTR
    clrf PICDATA0
pic16RdLoop2
    ; Serial clock period must be > P2 (100 nsec to 1 usec)
    bsf ICSP_CLK      ; Raise ICSP_CLK
    rrf PICDATA0, F   ; rotate previous contents before setting bit
    bcf PICDATA0, 7   ; Start output bit as 0
    nop
    nop
    nop
    nop
    bcf ICSP_CLK      ; Lower ICSP_CLK, PIC shifts data out
    nop
    btfsc ICSP_DAT    ; Check value of output from PIC
    bsf PICDATA0, 7   ; Set bit if PIC is sending a 1
    decfsz ICTR, F    ; 
    goto pic16RdLoop2 ; loop if not zero

    bsf STATUS, RP0   ; Bank 1
    bcf ICSP_DAT_MASK ; Set DAT back to output
    bcf STATUS, RP0   ; Bank 0
    bcf ICSP_DAT      ; pull DAT low

    return


; 16 bit PIC table read command has to already be in ICSP_CMD
; In:  <cmd><data len><ICSP cmd><N/A><N/A>
; Out: <cmd><DLEN><N/A><data val><N/A>
PIC16ReadCmd
    movfw ICSP_CMD
    movwf CMDTEMP

    ; Read a byte from the target
    call PIC16ReadBase

    ; Move value read from PIC to output buffer
    movfw PICDATA0
    movwf DATA0
    movfw PICDATA1
    movwf DATA1
    movlw 0x03
    movwf DATALEN

    movlw 0x05
    goto WritePacket

; Routine to prepare a packet for block read/write
packetInit
    ; Prepare response packet
    movlw DATA_BUF   ; Point to the output buffer
    movwf FSR        ;

    ; Save the address
    movfw ADDRL
    movwf PCADDR0
    movfw ADDRH
    movwf PCADDR1
    movfw ADDRU
    movwf PCADDR2

    movfw CMDOPT     ; Get # of bytes requested
    sublw 0x40       ; Is # requested more than 64?
    btfsc STATUS, C
    return           ; Size is ok - continue on

    ; Too many bytes requested.  Limit to 64
    movlw 0x40
    movwf CMDOPT     ; Send back the reduced #
    return

;
; Read a block of data from a 16 bit PIC
;
; In:    <cmd><DLEN><ADDRL><ADDRH><ADDRU><CMDOPT>
; Out:   <cmd><DLEN><ADDRL><ADDRH><ADDRU><cmdopt><data>
;
; DLEN on input tells us how many words of information are being
; requested.  The maximum is 32 (64 bytes) to prevent overflow.
; TBD: Use the input address to find the data...
;
; Affects: TEMP
;
PIC16RBlock
    call packetInit
    call PIC16SetTablePtr ; Set the table pointer to ADDRU:ADDRH:ADDRL
PIC16RB1
    movfw CMDOPT     ; Move requested # of bytes into TEMP
    movwf TEMP       ; to use for loop counter

PIC16RBLoop
    ; Read data from the current memory location
    movlw PIC16_TBLRD_POSTINC
    movwf ICSP_CMD
    call PIC16ReadBase

    ; Move response to output buffer (LSB first)
    movfw PICDATA0
    movwf INDF
    incf FSR, F

    ; See if there are any more words to read
    decfsz TEMP, F
    goto PIC16RBLoop

    movfw CMDOPT    ; Output size is CMDOPT + cmd/dlen/address/opt
    addlw 0x6
    goto WritePacket ; Send back the data

;
; Write a block of 64 bytes of code to a 16 bit PIC.  The value
; of CMDOPT determines how many of the 64 will be written in
; each pass of the loop.  Many devices have 8 byte buffers,
; some have 64 byte write buffers.  The value passed must
; be appropriate for the device being programmed.
; 
;
; The target must already have been configured for single
; panel writes:
;    // Direct access to config memory
;   0x8EA6 - bsf EECON1, EEPGD
;   0x8CA6 - bcf EECON1, CFGS
;   // Configure device for single panel writes
;   Set table pointer to: 0x3c0006
;   // Write 0x00 to 0x3C0006 to enable single-panel writes
;   // Direct access to code memory
;   0x8EA6 - bsf EECON1, EEPGD
;   0x9CA6 - bcf EECON1, CFGS

;
; In:   <cmd><DLEN><ADDRL><ADDRH><ADDRU><CMDOPT>
; OUT:  <cmd><DLEN><ADDRL><ADDRH><ADDRU><DATA>...]
;
PIC16WBlock
    ; Set the table pointer to ADDRU:ADDRH:ADDRL.  This routine
    ; conveniently saves the contents of ADDRU:ADDRH:ADDRL so
    ; they can be restored after we use the overlaid locations
    ; for ICSP_CMD:DATA0:DATA1
    call packetInit
    call PIC16SetTablePtr

PIC16WB1
    ; Adjust
    bcf STATUS, C    ; Don't rotate in any leftovers
    rrf CMDOPT, F    ; Divide by 2 for # of words

    ; Total of 32 words (64 bytes) to write
    movlw 0x20
    movwf RETRYCTR

PIC16WBLoop
    ; Move data into the write buffer
    movfw CMDOPT
    movwf TEMP
    decf TEMP, F ; Move all but the last one into place

PIC16WBFillLoop
    ; Write the current word and increment the table pointer
    movlw PIC16_PBLWRT_POSTINC
    movwf ICSP_CMD
    movfw INDF
    movwf DATA0
    incf FSR, F
    movfw INDF
    movwf DATA1
    incf FSR, F
    call PIC16CmdBase

    decfsz TEMP, F
    goto PIC16WBFillLoop

PIC16WBStartProgramming
    ; Move the final byte into place, but don't increment
    ; the table pointer out of the current buffer
    movlw PIC16_STARTPROG
    movwf ICSP_CMD
    movfw INDF
    movwf DATA0
    incf FSR, F
    movfw INDF
    movwf DATA1
    incf FSR, F
    call PIC16CmdBase

    ; At this point we need to emit a NOP command and hold the
    ; CLK line high for the duration of the write.
    clrf ICSP_CMD
    call PIC16WriteBase

    ; Increment the address pointer to the next buffer
    movlw PIC16_PBLWRT_POSTINC
    movwf ICSP_CMD
    clrf DATA0
    clrf DATA1
    call PIC16CmdBase

    ; Subtract the number written in this pass from
    ; the total.  If any left, then continue writing.
    movfw CMDOPT
    subwf RETRYCTR, F
    btfss STATUS, Z   ; All done with writes
    goto PIC16WBLoop  ; More buffers to fill

    ; Read the entire block and send it back.  It is the
    ; host software that is responsible for performing
    ; verification.  At the start of this routine, the
    ; call to PIC16SetTablePtr saved the address values.
    ; Restore the address and then invoke the block read
    ; routine
    movfw PCADDR2
    movwf ADDRU
    movfw PCADDR1
    movwf ADDRH
    movfw PCADDR0
    movwf ADDRL
    ; Invoke the block read routine.  No need for a 'return"
    ; from this routine - the block read will take care of it.
    movlw 0x40
    movwf CMDOPT
    goto PIC16RBlock

;
; Read a block of data from a 16 bit PIC data memory (EEPROM
;
; In:    <cmd><DLEN><ADDRL><ADDRH><N/A><CMDOPT>
; Out:   <cmd><DLEN><ADDRL>N/A><N/A><cmdopt><data>
;
; DLEN on input tells us how many bytes of information are being
; requested.  The maximum is 64 bytes to prevent overflow.
; ADDRH:ADDRU are not used - the maximum data size for the
; PICs currently supported is 256.
;
; Affects: TEMP
;
PIC16REBlock
    call packetInit

PIC16REB1
    movfw CMDOPT            ; Move requested # of bytes into TEMP
    movwf TEMP              ; to use for loop counter

    coreCmd1 0x9ea6         ; bsf EECON1, EEPGD
    coreCmd1 0x9ca6         ; bsf EECON1, CFGS

PIC16REBLoop
    ; Set address of the byte we want to read
    coreCmd2 0x0e, PCADDR0
    coreCmd1 0x6ea9         ; movwf EEADR

    ; Set high byte of address
    coreCmd2 0x0e, PCADDR1
    coreCmd1 0x6eaa         ; movwf EEADRH

    ; Initiate a memory read
    coreCmd1 0x80a6         ; bsf EECON1, RD

    ; Load the data into the serial data holding register
    coreCmd1 0x50a8         ; movf EEDATA, W, 0
    coreCmd1 0x6ef5         ; movwf TABLAT

    coreCmd1 0x0000         ; Required NOP

    ; Shift out the data
    movlw PIC16_SHIFT_TBLAT
    movwf ICSP_CMD
    call PIC16ReadBase

    ; Move response to output buffer (LSB first)
    movfw PICDATA0
    movwf INDF
    incf FSR, F

    ; See if there are any more words to read
    incf PCADDR0, F
    decfsz TEMP, F
    goto PIC16REBLoop

    movfw CMDOPT          ; Output size is CMDOPT + cmd/dlen/address/opt
    addlw 0x6
    goto WritePacket      ; Send back the data

;
; Write a block of data to a 16 bit PIC data memory (EEPROM
;
; In:    <cmd><DLEN><ADDRL><ADDRH><N/A><CMDOPT>
; Out:   <cmd><DLEN><ADDRL>N/A><N/A><cmdopt><data>
;
PIC16WEBlock
    call packetInit

PIC16WEB1
    movfw CMDOPT          ; Move requested # of bytes into TEMP
    movwf TEMP            ; to use for loop counter

    ; use z for data address so PCADDR is preserved
    movfw PCADDR0
    movwf TEMP0
    movfw PCADDR1
    movwf TEMP1

PIC16WEBLoop
    ; Direct access to data EEPROM
    coreCmd1 0x9ea6       ; bcf EECON1, EEPGD
    coreCmd1 0x9ca6       ; bcf EECON1, CFGS

    ; Set the data EEPROM address pointer
    coreCmd2 0x0e, TEMP0
    coreCmd1 0x6ea9       ; movwf EEADR
    coreCmd2 0x0e, TEMP1
    coreCmd1 0x6eaa       ; movwf EEADRH

    ; Load the data to be written
    coreCmd2 0x0e, INDF   ; movlw <Data>
    incf FSR, F           ; Move pointer to next input byte
    coreCmd1 0x6ea8       ; movwf EEDATA

	; Enable memory writes
    coreCmd1 0x84a6       ; bsf EECON, WREN

    ; Perform required sequence
    coreCmd1 0x0e55       ; movlw 0x55
    coreCmd1 0x6ea7       ; movwf EECON2
    coreCmd1 0x0eaa       ; movlw 0xaa
    coreCmd1 0x6ea7       ; movwf EECON2

    ; Initiate write
    coreCmd1 0x82a6       ; bsf EECON1, WR

    ; Poll WR bit, repeat until clear.  Datasheet says
    ; that max programming time for EEPROM locations is
    ; 5 msec.
    ;   Delay count = (18.432 / 4) * 5000 = 23040 (0x5a00)
    ; Each pass of this loop has a 100 instruction delay
    ; after checking the WR bit.  This means we need to
    ; go through the loop at most 230 (0xe6) times.  If
    ; the write completes more quickly than the max, we
    ; can continue writing the rest of the bytes that
    ; much sooner.
    movlw 0xe6
    movwf y
PIC16WEPoll
    coreCmd1 0x50a6       ; movf EECON1, W, 0
    coreCmd1 0x6ef5       ; movwf TABLAT
    movlw PIC16_SHIFT_TBLAT
    movwf ICSP_CMD
    call PIC16ReadBase    ; Shift out data
    btfss PICDATA0, 1     ; Test WR bit
    goto PIC16WEDonePoll  ; WR bit clear - done with write
    ; Write not complete.  Wait a bit and test again
    movlw 0x20
    call delay
    decfsz y, F
    goto PIC16WEPoll
PIC16WEDonePoll

    ; Hold PGC low for time P10 (5 - 100 usec)
    movlw 0xa0
    call delay

    ; Disable writes
    coreCmd1 0x94a6       ; bcf EECON1, WREN

    ; See if there are any more words to read
    incf TEMP0, F
    decfsz TEMP, F
    goto PIC16WEBLoop

    ; Read the entire block and send it back.
    movfw PCADDR2
    movwf ADDRU
    movfw PCADDR1
    movwf ADDRH
    movfw PCADDR0
    movwf ADDRL
    goto PIC16REBlock

;
; Loop reading and writing single characters until we get an
; ESC character.  If we don't see a valid character, then
; send a '!' to show we are alive (and give the PC something
; to synch to).
;
serial_test
    call RdRS232
    btfss ERRORS, 0
    goto gotone

    movlw '!';
    movwf SERIAL_OUT
    call serialTx

    goto serial_test

gotone
    movfw SERIAL_IN
    sublw 0x1b       ; See if this is an ESC character (decimal 27)
    btfss STATUS, Z
    goto programLoop
    movfw SERIAL_IN
    movwf SERIAL_OUT
    call serialTx
    goto serial_test

;
; Send the data buffer back.
;
; <STX><STX>[<DATA>...]<CHKSUM><ETX>
;
; Note that the data length in W must account for everything
; in the packet from command byte to final byte in the
; packet.  This is not the same as the DATALEN value when we
; read a packet - that value excludes the command and itself.
;
sendAck
    movlw 0x01      ; Send acknowledge (only command byte goes back)

WritePacket
    movwf COUNTER

    movlw STX       ; Send start condition
    call WrRS232
    movlw STX
    call WrRS232

    clrf CHECKSUM   ; Reset checksum

    movlw PACKET    ; Setup pointer to buffer area
    movwf FSR
    
SendNext            ; Send DATA
    movfw INDF
    addwf CHECKSUM,F
    incf FSR,F
    call WrData
    decfsz COUNTER,F
    goto SendNext

    comf CHECKSUM,W ; Send checksum
    addlw 0x01
    call WrData

    movlw ETX       ; Send stop condition
    call WrRS232

    goto programLoop

; Write a byte to the serial port (escape with control
; character if necessary).
WrData
    movwf WRTEMP      ; Save the data

    xorlw STX         ; Check for a STX
    btfsc STATUS,Z
    goto WrDLE        ; Send DLE, then the STX

    movfw WRTEMP
    xorlw ETX         ; Check for a ETX
    btfsc STATUS,Z
    goto WrDLE        ; Send DLE, then the ETX

    movfw WRTEMP
    xorlw DLE         ; Check for a DLE
    btfss STATUS,Z    ; Send DLE, then the DLE
    goto WrNext       ; Not a problem byte - just send

WrDLE
    movlw DLE         ; Yes, send DLE first
    call WrRS232

WrNext
    movfw WRTEMP      ; Then send STX

WrRS232
    clrwdt
#if _USE_UART
    bcf STATUS, RP1
    btfss PIR1, TXIF  ; Write only if TXREG is ready
    goto $-1
    
    movwf TXREG       ; Start sending
#else
    movwf SERIAL_OUT
    call serialTx
#endif
    return

; Serial transmit at 38.4 & 115.2 KBaud (8 Data bits,
; No parity, 1 stop bit).
;
; Common entry point for serial transmit routine.  Clock
; and speed specific routines below.
;
serialTx
  if _USE_RTSCTS == 1
    movlw 0x80
    movwf ICTR       ; timeout counter
    clrf ICTR        ; timeout ctr
TxLoop1
    clrf TEMP        ; ctr2
TxLoop2
    btfss RTS
    goto TxStart     ; PC is ready to read a char
    goto serialTx
    decfsz TEMP, F
    goto TxLoop2
    decfsz ICTR, F
    goto TxLoop1
TxError
    movlw 0x07       ; timeout - return error
    movwf ERRORS
    return
  endif ; _USE_RTSCTS == 1

; Serial receive at 38.4 & 115.2 KBaud (8 Data bits,
; No parity, 1 stop bit).
;
; Common entry point for serial receive routine.  Clock
; and speed specific routines below.
;
; Timeout after 32K passes through RxLoop1.  For 18.432 MHz
; crystal, (just over 5 instruction cycles per loop - there
; is overhead from RxLoop1), gives a timeout 28.125 times
; per second (1/28.125 = 35.6 msec).  When using the internal
; 8 MHz clock, the timeout will be close to 100 msec.
RdRS232
#if _USE_UART
    btfsc RCSTA,OERR ; Reset on overun
    goto VReset

    btfss PIR1, RCIF ; Wait for data from RS232
    goto    $-1
    movf    RCREG,W  ; Save the data
    movwf SERIALIN
 #else
serialRx
  if _USE_RTSCTS == 1
    bcf CTS
  endif
    movlw 0x80
    movwf ICTR       ; timeout counter
RxLoop1
    clrf TEMP        ;               6    From outer loop
RxLoop2
#if ECHO_ALL
    ; For test purposes only!  Using this echo
    ; will require that you only send one byte
    ; at a time from the PC (otherwise bytes
    ; will be dropped while we are sending).
    ;
    ; Echo every byte we get back to the PC
    btfsc RX
    goto RXLoop3
    call RXLow
    btfss ERRORS, 0
    call WrRS232
    return
RXLoop3
#else
    ; See if the start bit has begun.
    btfss RX         ; 0..1    5..6  7..8 See if Rx line went low
    goto RXlow       ; 1..2    6..7  8..9 Got start bit - read byte
#endif
    ; Worse case is if line went low right after btfss RX
    decfsz TEMP, F   ; 1..2
    goto RxLoop2     ; 2..3
    decfsz ICTR, F   ; 3..4
    goto RxLoop1     ; 4..5

RxError
    movlw ERROR_TIMEOUT ; timeout - return error
    movwf ERRORS
  if _USE_RTSCTS == 1
    bsf CTS
  endif
#endif ; _USE_UART
    return

; Wait until the end of the start bit (look for falling edge as the
; start bit begins, followed by a rising edge at the end).  This is in
; expectation that we will then get 4 0's and 4 1's right after.
WaitForRise
    btfsc RX  ; Wait for a falling edge
    goto WaitForRise
    clrwdt    ; Do we need this?
WtSR
    btfss RX  ; Wait for starting edge
    goto WtSR
    return

; RS232 Serial IO for 115200 and 38400 Baud.  There are four
; separate routines.  There are separate routines for 115.2K
; and 38.4K Baud rates.  These routines are duplicated for
; PIC operating frequencies of 18.432 MHz and 8 MHz.
; TBD: One routine based on constants/variables.
;
; 115.2K Baud is 1 bit every 8.680555... usec.
; 38.4K Baud is 1 bit every 26.041666... usec.
; 18.432 MHz clock is 4.608 instructions per usec
; 8 MHz clock is 2 instructions per usec
;
; Table of instructions per bit:
;          18.432 MHz         8 MHz
; 115.2K     40                17.361
;  38.3K    120                52.083
;
#if _USE_XTAL_18432
; These routines assume the use of a 18.432 MHz clock.
; The PIC will execute 4.608 instruction cycles per
; microsecond.  For 115.2 KBaud this is 40 instructions
; per bit.  At 38.4K there would be 120 instruction per
; bit.

#if BAUD_RATE_1152
; Serial transmit at 115.2 KBaud (8 Data bits, No parity, 1 stop bit)
TxStart
    movlw 0x09       ; Start bit plus 8 data bits
    movwf TEMP
    goto TxLow       ; Send start bit

TxLoop
    rrf SERIAL_OUT,1 ; 36
    btfsc STATUS,C   ; 37
    goto TxHigh      ; 38..39
    nop              ; 39
TxLow
    bcf TX           ; 0
    goto TxBit       ; 1..2
TxHigh
    bsf TX           ; 0
    goto TxBit       ; 1..2
TxBit
    movlw 0x08       ; 3
    call delay       ; 4..31
    nop              ; 32
    decfsz TEMP, 1   ; 33
    goto TxLoop      ; 34-35
    nop              ; 36
    nop              ; 37
    nop              ; 38
    nop              ; 39
    bsf TX           ; 0   (stop bit)
    movlw 0x0b       ; 1
    call delay       ; 2..38
    return

; Serial receive at 115.2 KBaud (8 Data bits, No parity,
; 1 stop bit).  Entry is from serialRx
;
; If the RX line went low, then we got a start bit.  Wait for
; a half a bit and start sampling.  This added delay puts the
; sampling at about the midpoint of each bit as it comes in.
; The soonest we could get here is 3 instruction cycles after
; the line went low (btfss/goto).  The latest would be 6-7
; cycles.
;
RXlow
  if _USE_RTSCTS == 1
    bsf CTS          ; 2       8
  endif

#if 1
    movlw 0x02       ; 3       9
    call delay       ; 4..13   10..19
#else
    movlw 0x03       ; 3
    call delay       ; 4..16
#endif
    movlw 0x08       ; 14      20
    movwf TEMP       ; 15      21
    btfsc RX         ; 16      22  Make sure start bit is still low
    goto  RxError   ;

    ; movlw 0x04
    movlw 0x07       ; 18      24
    call delay       ; 19..43  49
    nop              ; 44      50
    nop              ; 45      51

; Read the current bit
RxLoop
    rrf SERIAL_IN, 1 ; 38 [46  52]
    bcf SERIAL_IN, 7 ; 39 [47  53]
    btfsc RX         ; 0  [48  54]  20-35% of the way into the next bit
    bsf SERIAL_IN, 7 ; 1
    movlw 0x09       ; 2
    call delay       ; 3..33
    nop              ; 34
    decfsz TEMP, 1   ; 35
    goto RxLoop      ; 36..37

    btfss RX         ; 37..38   ensure stop bit is set
    goto RxError

  if _USE_RTSCTS == 1
    bsf CTS          ; 39
  endif

    clrf ERRORS       ; 40 (40 without added delay)
    return            ; 41..42 (leaves 30 clocks to get back to serialRx)
#else
; TBD: 38.4K routine needs to have delay loops calculated and
; the values folded back into the original routine above.
;

; Serial transmit at 38.4 KBaud (8 Data bits, No parity, 1 stop bit)

; Serial receive at 38.4 KBaud (8 Data bits, No parity, 1 stop bit)
#endif
#else
;
; RS232 Transmit/Receive routines for 8 MHz clock (2 MIPS).
;
#if BAUD_RATE_1152
; Serial transmit at 115.2 KBaud (8 Data bits, No parity, 1 stop bit)
;
; At 8 MHz, the PIC will be able to execute 17.361 instructions
; per bit.  The 0.361 part causes drift in the point within
; each bit that we sample as well as preventing transmission
; of exactly 115.2K Baud.  The receive end can be accommodated
; by ensuring that the drift will never result in sampling
; outside any one bit within the current byte being transmitted
; from the host.
TxStart
    movlw 0x09       ; Start bit plus 8 data bits
    movwf TEMP
    goto TxLow       ; Send start bit

TxLoop
    rrf SERIAL_OUT,1 ; 13
    btfsc STATUS,C   ; 14
    goto TxHigh      ; 15..16
    nop              ; 16
TxLow
    bcf TX           ; 0
    goto TxBit       ; 1..2
TxHigh
    bsf TX           ; 0
    goto TxBit       ; 1..2
TxBit
    nop              ; 3
    nop              ; 4
    nop              ; 5
    nop              ; 6
    nop              ; 7
    nop              ; 8
    nop              ; 9
    decfsz TEMP, 1   ; 10
    goto TxLoop      ; 11..12

    nop              ; 12
    nop              ; 13
    nop              ; 14
    nop              ; 15
    nop              ; 16  End of final bit in transmission
    bsf TX           ; 0   Send stop bit
    movlw 0x03       ; 1
    call delay       ; 2..14
    return           ; 15..16
;
; If the RX line went low, then we got a start bit.  Wait for
; the start bit to finish plus a little more to get into the
; first data bit before we start sampling.  This added delay
; puts the sampling somewere near the midpoint of each data
; bit as it comes in. The soonest we could get here is 3
; instruction cycles after the line went low (btfss/goto).
; The latest would be 10-11 cycles (which is well into a bit).
;
RXlow
  if _USE_RTSCTS == 1
    bsf CTS          ;  3      10
  endif

    ; Set TEMP to read 7 bits in a loop (8th will be read
    ; outside the loop for timing reasons.)
    movlw 0x07       ;  3      11
    movwf TEMP       ;  4      12
    btfsc RX         ;  5      13  Make sure start bit is still low
    goto  RxError    ;

    ; Wait until we are into the first data bit
    movlw 0x03       ;  7      15
    call delay       ;  8..20  16..28

; Read the current bit
RxLoop
    rrf SERIAL_IN, 1 ; 15      21  29
    bcf SERIAL_IN, 7 ; 16      22  30
    ; Sample the value of the bit between 5 and 11 instructions
    ; (out of 17) , putting us around the middle of where the
    ; bit is expected to be.  This should give enough leeway
    ; so that the 0.36 instructions/bit extra won't pull us back
    ; so much that one bit gets measured twice.  Drift will be
    ; 0.36 * 9 = 3.24 - over three instructions change across
    ; the byte.  17 instructions is shorter than the bit length,
    ; so the measurements will occur earlier in each successive bit.
    btfsc RX         ; 0       23  31
    bsf SERIAL_IN, 7 ; 1
    movlw 0x01       ; 2
    call delay       ; 3..9
    nop              ; 10
    nop              ; 11
    decfsz TEMP, 1   ; 12
    goto RxLoop      ; 13..14

; Read the final bit
    rrf SERIAL_IN, 1 ; 14
    bcf SERIAL_IN, 7 ; 15
    nop              ; 16
    btfsc RX         ; 0      23   31
    bsf SERIAL_IN, 7 ; 1

; Skip the stop bit to give more time to process this
; byte and get back to the serial Rx routine.

  if _USE_RTSCTS == 1
    bsf CTS          ; 2
  endif

    clrf ERRORS      ; 3      26   34

    ; At this point we have between 0 and 8 instructions
    ; left in the final bit.  Together with the stop bit,
    ; we have from 17 to 25 instructions to get back to
    ; serialRx.
    return           ; 4..5
#else
;
; Serial transmit at 38400 BPS
; 38400 is 52.08 (52) cycles at 8MHz
TxStart
    movlw 0x09       ; Start bit plus 8 data bits
    movwf TEMP
    goto TxLow       ; Send start bit

TxLoop
    rrf SERIAL_OUT,1 ; 48
    btfsc STATUS,C   ; 49
    goto TxHigh      ; 50..51
    nop              ; 51
TxLow
    bcf TX           ; 0
    goto TxBit       ; 1..2
TxHigh
    bsf TX           ; 0
    goto TxBit       ; 1..2
TxBit
    movlw 0x0c       ; 3
    call delay       ; 4..43
    nop              ; 44
    decfsz TEMP, 1   ; 45
    goto TxLoop      ; 46-47
    nop              ; 47
    nop              ; 48
    nop              ; 49
    nop              ; 50
    nop              ; 51
    bsf TX           ; 0      Send the stop bit
    movlw 0x0f       ; 1
    call delay       ; 2..50  Wait until stop bit is done
    return

; If the RX line went low, then we got a start bit.  Wait for
; the start bit to finish plus a little more to get into the
; first data bit before we start sampling.  This added delay
; puts the sampling somewere near the midpoint of each data
; bit as it comes in.
RxLow
    movlw 0x05       ; 3
    call delay       ; 4..23
    movlw 0x08       ; 24
    movwf TEMP       ; 25
    btfsc RX         ; 26     See if start bit still low
    goto  RxError    ; 27     No - must not have been a start bit

    movlw 0x0d       ; 28
    call delay       ; 29..72
    nop              ; 73
    nop              ; 74

                     ; Receive bit
RxLoop
    rrf SERIAL_IN, 1 ; 50 [75]
    bcf SERIAL_IN, 7 ; 51 [76]
    btfsc RX         ; 0  [77]    52 + 26 is through start and 1/2 first bit
    bsf SERIAL_IN, 7 ; 1
    movlw 0x0d       ; 2
    call delay       ; 3..46
    decfsz TEMP, 1   ; 47
    goto RxLoop      ; 48..49

    btfss RX         ; See if stop bit is set
    goto RxError

    clrf ERRORS
  if _USE_RTSCTS == 1
    bsf CTS
  endif
    return
#endif
#endif


#if _INDUCTOR_BASED
; Change duty cycle of PWM based on analog reading of feedback voltage
PWMAdjust
#if 1
    bsf STATUS, RP0 ; All PWM & analog vars in bank 1
    ; See if the A/D val is larger or smaller than desired
    movlw fixVal
    subwf analogVal, w
    btfsc STATUS, Z
    goto donePWMAdjust ; analogVal matches max, don't increase duty cycle
    btfss STATUS, C
    goto incPWM    ; analogVal too low, increase duty cycle

decPWM
    ; Don't bother lowering the PWM if we are below the maximum
    ; acceptable voltage
    movlw maxAnalog
    subwf analogVal, w
    btfss STATUS, C
    goto donePWMAdjust ; analogVal < maxAnalog - don't change PWM

    ; the analog reading is larger than the value we want.  Reduce
    ; the duty cycle by 1
    movfw PWMVal
    btfsc STATUS, Z
    goto donePWMAdjust ; duty cycle already zero
    decf PWMVal, f
    setPeriod PWMVal
    goto donePWMAdjust

incPWM
    ; Don't bother raising the PWM if we are above the minimum
    ; acceptable voltage
    movlw minAnalog
    subwf analogVal, w
    btfsc STATUS, C
    goto donePWMAdjust ; analogVal > minAnalog - don't change PWM

    ; Check to see if the PWM value is already at the max
    movlw maxPWM
    subwf PWMVal, w
    btfsc STATUS, Z
    goto donePWMAdjust  ; already exactly at the max
    btfsc STATUS, C
    goto resetPWMAdjust ; duty cycle larger than allowed, reset it
    incf PWMVal, f
    setPeriod PWMVal
    goto donePWMAdjust

resetPWMAdjust
    movlw maxPWM
    movwf PWMVal
    setPeriod PWMVal    
#else
    ; Test - don't change PWM from starting value.
#endif

    ; Done fiddling with the PWM duty cycle
donePWMAdjust
    bcf STATUS, RP0 ; Bank 0
    return;
#endif ; _INDUCTOR_BASED

; Setup of interrupts.
; Sample code below uses Timer0 to generate
; an interrupt every millisecond (at 8MHz).
setupInterrupts
#if 0
    bsf STATUS, RP0 ; bank 1
    clrf PIE1
    movlw 0x04
    movwf OPTION_REG ; internal clock, prescaler = 1:32
    bcf STATUS, RP0 ; bank 0
    clrf TMR0
    bsf INTCON, T0IE ; Turn on interrupts for timer 0
    bsf INTCON, GIE
    bsf INTCON, PEIE
#endif
    return

setupClock
#if _USE_XTAL_18432
; No special code needed if we are using a crystal.
#else
; Set clock to 8MHz
    movlw 0x71  ; Clock 8MHz, use speed in OSCCON<6:4>
    bsf STATUS, RP0 ; bank 1
    movwf OSCCON
    bcf STATUS, RP0 ; bank 0
#endif
    return

setupPins
; Set I/O directions
; Use RX an input for RS232 comm (labeled SERIALTX because it is
; the PC that is transmitting), TX for RS232 output (labelled
; SERIALRX because the PC will be listening).  RC0 is set to input
; so it can be used as an A/D converter.  All the rest of the
; PORTC pins start as outputs, with RC2 toggling as appropriate
; to read data from the ICSP DATA line.
;
    bcf STATUS, RP0   ; bank 0
    ; Init PORTA and PORTC
    clrf PORTA
    clrf PORTC
    ; Turn off comparators
    movlw 0x07
    movwf CMCON0

    bsf STATUS, RP0   ; bank 1

    ; All pins digital I/O default
    clrf ANSEL

    ; Set I/O directions for the pins we are using
    bcf TX_MASK         ; Tx is an output
    bsf RX_MASK         ; Rx is an input
#if _USE_RTSCTS
    bcf CTS_MASK        ; CTS is an output
    bsf RTS_MASK        ; RTS is an input
#endif
    bcf VPP1_MASK       ; 13V control is an output
    bcf VDD_MASK        ; 5V control is an output
    bsf ICSP_CLK_MASK   ; CLK output when in use, otherwise tristate
    bsf ICSP_DAT_MASK   ; DAT output when in use, otherwise tristate

    bcf STATUS, RP0 ; bank 0

    ; Set default values for I/O pins
    bcf TX
    bcf RX
#if _USE_RTSCTS
    bsf CTS        ; Start with CTS low - not ready for input
    bcf RTS
#endif
    bsf VPP1        ; 13V line is inverted from logic level, start high
    bsf VDD        ; 5V line is inverted, start high
    bcf ICSP_CLK
    bcf ICSP_DAT

 ; Turn on the Analog to Digital Converter.
; According to the pic12f683 datasheet, the
; A/D conversion clock (1/Tad) must be selected
; to ensure a minimum Tad of 1.6us.  For an 8 MHz
; oscillator, this means using 16 or 32 Tosc
; (2.0 us or 4 us).  The internal RC oscillator
; only recommended if conversion will be
; performed during sleep.
    bsf STATUS, RP0 ; bank 1
    bsf AD_FEEDBACK_ENABLE ; Turn on A/D on appropriate port(s)
#if _USE_XTAL_18432
    movlw 0x10      ; Fosc/8
#else
    movlw 0x10      ; Fosc/8
#endif
    movwf ADCON1

    bcf STATUS, RP0 ; bank 0
    movlw 0x11      ; Left, Vdd ref, AN4, converter on
    movwf ADCON0

   return

; Reset variables used for I/O and programming
setupVars
    bcf STATUS,RP0    ; select bank 0 again
    clrf ICTR
    clrf ERRORS
    clrf COMMAND
    clrf DATALEN
    clrf DATA0
    clrf DATA1
#if _INDUCTOR_BASED
    movlw 0x10
    movwf PWMCtr
#endif
    bcf STATUS, RP0
    return

; Read the analog register
readAN4
	; Delay code to ensure enough time for acquisition to
	; occur.  [TBD: Is it necessary?]
    bsf STATUS, RP0
    movlw sampleTime
    movwf delayVal
ANDelay1
    decfsz delayVal, F
    goto ANDelay1
    ; Acquisition done.  Wait for conversion
    ; to finish - bit GO in ADCON0 will be
    ; cleared by the PIC when it is done.
    bcf STATUS, RP0
    bsf ADCON0, GO
ADCWaitLoop
    btfsc ADCON0, GO
    goto ADCWaitLoop
    movfw ADRESH
    bsf STATUS, RP0
    movwf analogVal
    bcf STATUS, RP0
    return

#if _INDUCTOR_BASED
; Setup PWM
setupPWM
    ; Set RC5 to an output.
    bsf STATUS, RP0 ; bank 1
    bcf PWM_MASK ; Ensure RC5 is an output
    movlw 0x1f   ; Set PWM period to 31.
    movwf PR2
    ; Set two LSB bits of duty cycle to
    ; "11", CCP1 into PWM mode.
    movlw 0x3c
    bcf STATUS, RP0 ; bank 0
    movwf CCP1CON
    bsf STATUS, RP0 ; bank 1
    ; Start with a low value for PWM to reduce problems with
    ; current saturating the inductor
    movlw startPWM
    movwf PWMVal
    setPeriod PWMVal
    bcf STATUS, RP0 ; bank 0
    ; Enable Timer 2
    bcf PIR1, 1
    movlw 0x04
    movwf T2CON
tmrWait
    btfss PIR1, 1
    goto tmrWait
    return
#endif

;
; delay for 3n+4 instructions (including call/return)
;
delay
    movwf DLYTEMP     ; 1
delayLoop
    decfsz DLYTEMP, 1 ; 0
    goto delayLoop    ; 1..2
    return            ;

; Interrupt service routine
intServ
    ; Insert interrupt code here
checkTimer0
    ; See if Timer0 overflowed
    btfss INTCON, T0IF
    goto checkINTF
    ; Process Timer0 interrupt
#if _INDUCTOR_BASED
    ; call readAN4    ; Read analog value from AN4
    movfw ADRESH
    bsf STATUS, RP0
    movwf analogVal
    call PWMAdjust
#endif
    ; clear Timer0
    bcf INTCON, T0IF
    retfie
checkINTF
    ; See if the RA2/INT external interrupt occurred
    btfss INTCON, INTF
    goto checkRAIF
    bcf INTCON, INTF
checkRAIF
    btfss INTCON, RAIF
    goto checkADIF
    ; See if a pin on PORTA changed state
    bcf INTCON, RAIF
checkADIF
    ; See if an A/D conversion completed
    btfsc PIR1, ADIF
    goto checkTMR1IF
    bcf PIR1, ADIF
checkTMR1IF
    btfsc PIR1, TMR1IF
    goto intDone
    bcf PIR1, TMR1IF
intDone
    retfie

    org 0x2100
    de "Xander - EVA 2.9", 9, 0, 2, 0

    org 0x2120
    de "0123456789abcdef"
    de "fedcba9876543210"
    de "00112233445566778899aabbccddeeff"
    
    org 0x21d8
    de "Copyright 2005 - Alexander Enzmann"
    end
