;
; Code for PIC programmer that generates it's own 13V programming
; voltage using the PWM to drive a boost circuit.  Communications
; with the host PC is over RS232 (when compiled for the 16F737) or
; USB if compiled for the 18F2455.  Note that when this firmware
; is compiled for the 2455, it must be loaded at runtime by the
; EvaUSB firmware.

#ifdef __18F2455
  messg "Compile for 18f2455"
#define procID 2455
#define procSize 16
#include "p18f2455.inc"
#else
#ifdef __16F737
  messg "Compile for 16f737"
#define procID 737
#define procSize 14
#include "p16f737.inc"
#else
#ifdef __16F913
  messg "Compile for 16f913"
#define procID 913
#define procSize 14
#include "p16f913.inc"
#else
  error "Unsupported processor for Eva"
#endif
#endif
#endif

; BAUD_RATE is set to the desired serial rate for RS232 communications.
; Acceptable values are: 115 (115.2), 57 (57.6), 38 (38.4)
#define BAUD_RATE 115

#if procSize == 16
  ; For 18F2455 (USB) firmware, this code is loaded into higher memory,
  ; much like a bootloader operates.
    org 0x4000

; CLK_SPD is used by timing routines to ensure we don't exceed
; the capabilities of chips being programmed.  Note that this
; value has to be consistent with the base firmware (see
; the file evadefs.h).  TBD: detect clock speed at runtime.
#define CLK_SPD 24

#else
; Clock 
#define _USE_XTAL_20 1 ; 0 = Internal 8MHz clock, 1 = 20 MHz crystal
#define _DEBUG_CMDS 0 ; 0 = no debug output, 1 = debug write to RS232
#if _USE_XTAL_20
#if (procID == 737)
    ; External high-speed crystal
    __config 0x2007, 0x3f82
    __config 0x2008, 0x3fbc
#else
    ; External high-speed crystal
    __config 0x30d2
#endif
#define CLK_SPD 20
#else
    ; Internal 4MHz Osc with pin used at general IO.  Clock is
    ; increased to 8MHz during setup.
#if (procID == 737)
    __config 0x2007, 0x3f90
    __config 0x2008, 0x3fbc
#else
    ; TBD: Not working on 913 yet - Xtal version is ok.
    __config 0x30d4
#endif
#define CLK_SPD 8
#endif

; Reset vector
    org 0x00

    clrf INTCON   ; Reset vector
    goto startup

; Interrupt vector
    org 0x04
    ; Save important registers
    movwf TMPW
    movf STATUS, W
    movwf TMPSTATUS
    movf PCLATH, W
    movwf TMPPCLATH

    ; Select program memory page 1 and jump to the ISR
    bsf PCLATH, 3
    goto intServ

    org 0x10
#endif

#if procSize == 14
; RS-232 Parameters
#define _USE_RTSCTS 1     ; 1 = Use hardware flow control, 0 = don't use it
#define _USE_UART 1       ; 0 = Don't use/don't have UART
#define ECHO_ALL 0        ; Echo any characters we get
#define BAUD_RATE_1152 1  ; 0 = 38.4K Baud, 1 = 115.2K Baud

; Definitions to provide equivalent instructions/registers on the 16F

; Macro to duplicate the functionality of the movff instruction
; for 14 bit PIC processors.
movff macro a, b
    movf a, w
    movwf b
    endm
#define rrcf rrf
#define rlcf rlf
#define INDF0 INDF
#define FSR0L FSR
#if (procID == 737)
#define CMCON0 CMCON
#endif
#if (procID == 913)
#define GO GO_DONE
#endif
#endif

; Current version: 2.9
VERSION0 equ 0x09 ; Minor version #
VERSION1 equ 0x02 ; Major version #

;
; Pin definitions
;
#define VBOOST_MASK TRISA, 0
#define VBOOST PORTA, 0
#define EXT_PWR_MASK TRISA, 1
#define EXT_PWR PORTA, 1
#define CTS_MASK TRISA, 2
#define CTS PORTA, 2
#define RTS_MASK TRISA, 3
#define RTS PORTA, 3
#define LED2_MASK TRISA, 4
#define LED2 PORTA, 4
#define LED1_MASK TRISA, 5
#define LED1 PORTA, 5
; RA6 - Tied to oscillator
; RA7 - N/A

#define SDA_MASK TRISB, 0
#define SDA PORTB, 0
#define SCL_MASK TRISB, 1
#define SCL PORTB, 1
#define VPP1_MASK TRISB, 2
#define VPP1 PORTB, 2
#define ICSP_CLK_MASK TRISB, 3
#define ICSP_CLK PORTB, 3
#define ICSP_DAT_MASK TRISB, 4
#define ICSP_DAT PORTB, 4
#define VDD_MASK TRISB, 5
#define VDD PORTB, 5
#define VPP2_MASK TRISB, 6
#define VPP2 PORTB, 6
#define LED3_MASK TRISB, 7
#define LED3 PORTB, 7

#define PRG_SEL_MASK TRISC, 0
#define PRG_SEL PORTC, 0
#define USB_PWR_MASK TRISC, 1
#define USB_PWR PORTC, 1
#define VPP_PWM_MASK TRISC, 2
#define VPP_PWM PORTC, 2
; RC3 - VUSB
; RC4 - USB D-
; RC5 - USB D+
#define SERIAL_TX_MASK TRISC, 6
#define SERIAL_TX PORTB, 6
#define SERIAL_RX_MASK TRISC, 7
#define SERIAL_RX PORTB, 7

; The variable SFlags is used to track the status of several
; events including: Entire packet has been received (PKTReadyFlag),
; escape character while reading from RS232 (DLEFlag).
;

; Bank 0 variables
CTRX        equ 0x20 ; 16 bit counter - timing delays during programming

; Registers for I/O and programming variables (bank 0)
PCADDR0     equ 0x22 ; Low byte of current ICSP address
PCADDR1     equ 0x23 ; High byte of current ICSP address
PCADDR2     equ 0x24 ; Upper byte of current ICSP address
ICTR        equ 0x25 ; Temp counter used for ICSP and Serial I/O
TEMP        equ 0x26 ; Temp value
DLYTEMP     equ 0x27 ; Temp counter used in delay routine
PWMCtr      equ 0x28 ; Number of passes to wait before checking PWM
SFlags      equ 0x29 ; Bit flags for various I/O status events (see above)
trxChar     equ 0x2a ; trxChar - last byte input from RS232
packetCtr   equ 0x2b ; packetCtr - # of bytes read via RS232
STXCount    equ 0x2c ; stxCount - # of STX bytes via RS232
TMPSTATUS   equ 0x2d ; Save STATUS during interrupt
TMPW        equ 0x2e ; Save W during interrupt
TMPPCLATH   equ 0x2f ; Save PCLATH during interrupt
CHECKSUM    equ 0x30 ; Checksum for command packet
CMDTEMP     equ 0x31 ; Temp ICSP command value (for block programming)
PROGTEMP    equ 0x32 ; Temp ICSP begin programming command value
ENDPROGTEMP equ 0x33 ; Temp ICSP end programming command value
RETRYCTR    equ 0x34 ; # of times write attempted at one location
PICDATA0    equ 0x35 ; Low byte returned from target PIC
PICDATA1    equ 0x36 ; High byte returned from target PIC
PICDATA2    equ 0x37 ; Upper byte returned from target PIC
analogVal   equ 0x38 ; Last value read from A/D converter
PWMVal      equ 0x39 ; PWM duty cycle value
; Storage for command packet:
;    <cmd><data len><addrL><addrH><addrU><cmdOPT><...data...>
;    <cmd><data len><data0><data1><data2>
;    <cmd><data len><ICSP cmd><data high byte><data low byte>
;
PACKET      equ 0x3a ; Start of packet
COMMAND     equ 0x3a ; Command value is first entry in the packet
DATALEN     equ 0x3b ; Length of data to follow
DATABUF0    equ 0x3c ; Byte 0 of generic data buffer
ADDRL       equ 0x3c ; Low byte of target address
ICSP_CMD    equ 0x3c ; command value to send to PIC
DATABUF1    equ 0x3d ; Byte 1 of generic data buffer
ADDRH       equ 0x3d ; High byte of target address
DATA0       equ 0x3d ; Low byte of data to send to PIC
DATABUF2    equ 0x3e ; Byte 2 of generic data buffer
ADDRU       equ 0x3e ; Upper byte of target address
DATA1       equ 0x3e ; High byte of data to send to PIC
CMDOPT      equ 0x3f ; Optional value (e.g., block length)
DATA2       equ 0x3f ; Upper byte of data to send to PIC
DATA_BUF    equ 0x40 ; Up to 64 bytes of data to send/recieve
;
; Address 0x40..0x7f are used for data payload.
;

; DLEFlag - escape flag during RS232 input
#define DLEFlag SFlags, 0
; An entire serial packet has arrived and is ready for processing
#define PKTReadyFlag SFlags, 1

; Bank 1 variables
TMPFSR1     equ 0xac ; Save FSR during interrupt
TMPSTATUS1  equ 0xad ; Save STATUS during interrupt
TMPW1       equ 0xae ; Save W during interrupt
TMPPCLATH1  equ 0xaf ; Save PCLATH during interrupt

; Definitions
; Constants for PWM management (for 20 MHz clock)
PWMPeriod equ 0x3f ; Setting for PR2 to get 62.5 KHz PWM
maxPWM    equ 0x50 ; Max duty cycle allowed
#if 1
minAnalog equ 0xaa ; 15V - Minimum voltage before we increase duty cycle
fixVal    equ 0xb0 ; TBD V (reading of about 2.9V from A/D converter)
maxAnalog equ 0xb8 ; 16V - Max voltage before we decrease duty cycle
#else
maxPWM    equ 0x40 ; Max duty cycle allowed
minAnalog equ 0x92 ; 15V - Minimum voltage before we increase duty cycle
fixVal    equ 0x94 ; ~15.24V (reading of about 2.9V from A/D converter)
maxAnalog equ 0x9b ; 16V - Max voltage before we decrease duty cycle
#endif
startPWM  equ 0x10 ; Low duty cycle to start with

MAX_RETRIES equ 0x04 ; Max times write tried before failure

MAX_BUFLEN  equ 0x46 ; Limit buffer count to 6 (header) + 64 (data)

; ICSP commands for 14 bit PIC devices
PIC14_LOADCFG       equ 0x00
PIC14_LOADDATA      equ 0x02
PIC14_READDATA      equ 0x04
PIC14_INCADDR       equ 0x06
PIC14_BEGINPROG     equ 0x08
PIC14_ENDPROG       equ 0x0E
PIC14_ENDPROGA      equ 0x17
PIC14_BEGINPROGONLY equ 0x18
PIC14_LOADE2DATA    equ 0x03
PIC14_READE2DATA    equ 0x05
PIC14_BULKERASEPROG equ 0x09
PIC14_BULKERASEE2   equ 0x0B
PIC14_CHIPERASE     equ 0x1F

; ICSP commands for 16 bit PIC devices
PIC16_CORE           equ 0x00 ; 16 bit PIC core Instruction
PIC16_SHIFT_TBLAT    equ 0x02 ; Shift out TABLAT register
PIC16_TBLRD          equ 0x08 ; Table Read
PIC16_TBLRD_POSTINC  equ 0x09 ; Table Read, post-increment
PIC16_TBLRD_POSTDEC  equ 0x0A ; Table Read, post-decrement
PIC16_TBLRD_PREINC   equ 0x0B ; Table Read, pre-increment
PIC16_TBLWRT         equ 0x0C ; Table Write
PIC16_PBLWRT_POSTINC equ 0x0D ; Table Write, post-increment by 2
PIC16_TBLWRT_POSTDEC equ 0x0E ; Table Write, post-decrement by 2
PIC16_STARTPROG      equ 0x0F ; Table Write, start programming

; STDP commands for 24 bit PIC devices
PIC24_SIX            equ 0x00 ; 24 bit PIC Shift In and eXecute
PIC24_REGOUT         equ 0x01 ; 24 bit PIC Shift out VISI register

;
; Incoming data format:
;
;    <STX><STX><DATA><CHKSUM><ETX>
;              /    \
;     ________/      \____________________________
;    /                                         \
;    <COMMAND><DLEN><ADDRL><ADDRH><ADDRU><DATA>...
;
;
; Definitions:
;
;     STX     -  Start of packet indicator
;     ETX     -  End of packet indicator
;     DATA    -  General data up to 64 bytes
;     CHKSUM  -  The 8-bit two's compliment sum of DLEN & DATA
;     COMMAND -  Base command
;     DLEN    -  Length of data buffer
;     ADDR    -  Address up to 24 bits
;     DATA    -  Data buffer (if any)
STX      equ 0x55
ETX      equ 0x04
DLE      equ 0x05
;
;
; Commands:
;
VERSION_CMD      equ 0x01 ; Version of firmware
SETLINES_CMD     equ 0x02 ; Turn all control lines on/off
VPP_CMD          equ 0x03 ; Turn Vpp on/off
; (unused - ETX)
PRGSEL_CMD       equ 0x05 ; Turn PRG_SEL on/off
VDD_CMD          equ 0x06 ; Turn Vdd on/off
CLKDAT_CMD       equ 0x07 ; Set the ICSP CLK and DAT lines
READ_LINES_CMD   equ 0x08 ; Read current value of signal lines
READ_REGS_CMD    equ 0x09 ; Read the value of SFRs
ECHO_CMD         equ 0x0a ; Send back the input
BEGINPROG_CMD    equ 0x0b ; Enter prog mode: lower DAT/CLK, raise Vdd and Vpp
ENDPROG_CMD      equ 0x0c ; Exit prog mode: lower Vdd/Vpp, tristate CLK/DAT
PIC14_CMD        equ 0x0d ; Send command to 14 bit PIC
PIC14R_CMD       equ 0x0e ; Send cmd to 14 bit PIC, read response
; TBD equ 0x0f
READ_ANALOG_CMD  equ 0x10 ; Read the current value of the A/D converter
SET_ANALOG_CMD   equ 0x11 ; Set desired voltage bounds
PIC16ADDR_CMD    equ 0x12 ; Set address of a 16 bit PIC
PIC16_CMD        equ 0x13 ; Send command to 16 bit PIC
PIC16R_CMD       equ 0x14 ; Send cmd to 16 bit PIC, read response
PIC16W_CMD       equ 0x15 ; Send write cmd to 16 bit PIC, hold CLK for time P9 (1-5 msec)
READ_MASKS_CMD   equ 0x16 ; Read current value of TRIS registers
READ_PORTS_CMD   equ 0x17 ; Read current value of PORT registers
READ_PWM_CMD     equ 0x18 ; Read current PWM duty cycle
PIC14RBP_CMD     equ 0x19 ; Read a block from a 14 bit PIC
PIC14WBP_CMD     equ 0x1a ; Write a block to a 14 bit PIC
PIC14RBE_CMD     equ 0x1b ; Read a block from a 14 bit PIC EEPROM
PIC14WBE_CMD     equ 0x1c ; Write a block to a 14 bit PIC EEPROM
PIC16RB_CMD      equ 0x1d ; Read a block from a 16 bit PIC
PIC16WB_CMD      equ 0x1e ; Write a block to a 16 bit PIC
PIC16RBE_CMD     equ 0x1f ; Read a block from a 16 bit PIC EEPROM
PIC16WBE_CMD     equ 0x20 ; Write a block to a 16 bit PIC EEPROM
PIC16WBUF_CMD    equ 0x21 ; Write one buffer of bytes to a 16 bit PIC
PIC16EBUF_CMD    equ 0x22 ; Erase one window of bytes on a 16 bit PIC
PIC16WEBYTE_CMD  equ 0x23 ; Write one byte to a 16 bit PIC EEPROM
PIC24_CMD        equ 0x24 ; Send cmd to 24 bit PIC (dsPIC)
PIC24R_CMD       equ 0x25 ; Send cmd to 24 bit PIC (dsPIC), read response
PIC24RBUF_CMD    equ 0x26 ; Read 16 instructions from a 24 bit PIC
PIC24WBUF_CMD    equ 0x27 ; Write 16 instruction to 24 bit PIC
PIC24WBSETUP     equ 0x28 ; Setup for write to 24 bit PIC
PIC24WBFINISH    equ 0x29 ; Finish buffer write to 24 bit PIC
LAST_CMD         equ 0x29 ; [Adjust if commands added/deleted]

ERROR_OK       equ 0   ; Error codes
ERROR_BADCMD   equ 1
ERROR_TIMEOUT  equ 3

; Macro to send a core command to a 16 bit PIC target.  The
; argument is a 16 bit literal.
; Affects: ICSP_CMD, DATA0, DATA1, ICTR
coreCmd1 macro a
    clrf ICSP_CMD
    movlw (((a) & 0xff00) >> 8)
    movwf DATA1
    movlw ((a) & 0xff)
    movwf DATA0
    call PIC16CmdBase
    endm

; Macro to send a core command to a 16 bit PIC target.  The
; first argument is an 8 bit literal value to be used as the
; high byte of the command.  The second argument is the address
; of an 8 bit value to be used as the low byte of the command
; Affects: ICSP_CMD, DATA0, DATA1, ICTR
coreCmd2 macro h, l
    clrf ICSP_CMD
    movlw (h)
    movwf DATA1
    movf (l), W
    movwf DATA0
    call PIC16CmdBase
    endm

; Macro to send a core command to a 24 bit PIC target.  The
; argument is a 24 bit literal.
; Affects: ICSP_CMD, DATA0, DATA1, DATA2, ICTR
coreCmd24a macro a
    movlw PIC24_SIX
    movwf ICSP_CMD
    movlw (((a) >> 0x10) & 0xff)
    movwf DATA2
    movlw (((a) >> 0x08) & 0xff)
    movwf DATA1
    movlw ((a) & 0xff)
    movwf DATA0
    call PIC24CmdBase
    endm

; Execute REGOUT command on a dsPIC.  Resulting 16 bits will be
; in PICDATA0 and PICDATA1.
; Affects: ICSP_CMD, PICDATA0, PICDATA1, ICTR
coreCmd24b macro
    movlw PIC24_REGOUT
    movwf ICSP_CMD
    call PIC24ReadBase
    endm

; Macro to set PWM period (must be in RAM bank 0)
setPeriod macro a ; a is 1 SFR
    swapf a, w     ; This macro will change w
    andlw 0x30     ; Set PWM duty cycle LSBs
    iorlw 0x0c     ; ECCP mode - use P1A active-high, single output
    movwf CCP1CON
    bcf STATUS, C  ; Ensure no carry goes into the result
    rrf a, w       ; Shift out LSBs of period
    movwf TEMP
    bcf STATUS, C
    rrf TEMP,w
    movwf CCPR1L   ; Top 6 bits of period now in CCPR1L
    endm

; Macro to adjust PWM.  Program addresses must be in bank 0 when invoked.
adjustPWM macro
    call readAN0  ; Read analog value from AN4
    bsf PCLATH, 3 ; Page 1 program memory
    call PWMAdjust
    bcf PCLATH, 3 ; Page 0 program memory
    movlw 0xff
    movwf PWMCtr
    endm

; Macro to subtract 1 from a 16 bit value
DEC16 macro DST
    decf   (DST), F       ; 0   Decrement low byte
    incfsz (DST), W       ; 1   Check for underflow
    incf   (DST)+1, F     ; 2   Update
    decf   (DST)+1, F     ; 3   Fixup
    movf   (DST), W       ; 4  
    iorwf  (DST)+1, W     ; 5   Set Z bit
    endm

;
; Start of code
;
#if procID == 2455
processCommand
    ; Make sure command is in the correct range
    movf COMMAND, W
    sublw LAST_CMD
    btfss STATUS, C
    goto badCommand

    ; Computed goto, based on value of command
    movlw UPPER(cmdJumpTable)
    movwf PCLATU
    movlw HIGH(cmdJumpTable)
    movwf PCLATH
    ; Each goto is 4 bytes (on an 18F PIC).  Take the command value and
    ; multiply by 4 to determine the offset from cmdJumpTable we want
    ; to go to.  TBD: Fix so it will work with commands larger than 64 - the current
    ; code ignores overflows when multiplying COMMAND * 4.
    movf COMMAND, W
    andlw 0x3f      ; Mask off high bits to prevent them from rolling over to low bits
    movwf TEMP
    rlncf TEMP      ; Multiply by 2
    rlncf TEMP      ; Multiply by 2 more (2 * 2 = 4...)
    movlw LOW(cmdJumpTable)
    addwf TEMP
    btfsc STATUS, C
    incf PCLATH ; Increment high part of address if we cross a page boundary
    ;movff TEMP, CMDOPT
    movf TEMP, W

    ; Jump to cmdJumpTable + COMMAND
    movwf PCL
#else
; Main code
startup
    ; Disable interrupts until everything
    ; has been configured.

    ; Initialization is in memory page 1
    bsf PCLATH, 3
    call setup
    ; Ensure subsequent call/goto are in page 0
    bcf PCLATH, 3

programLoop
    ; Every few passes through the loop we measure
    ; the output voltage and adjust PWM duty cycle
    ; as necessary.
    decfsz PWMCtr, F
    goto noPWMCheck
    adjustPWM      ; Adjust PWM/Voltage levels
noPWMCheck
    ; TBD: See if there is a rx/tx error and clear it.

    ; See if an entire packet has been received.
    banksel SFlags
    btfss PKTReadyFlag
    goto programLoop ; No packet ready - keep waiting

#if _USE_RTSCTS == 1
    bsf CTS
#endif

    ; Toggle LED1 to show we got a packet
#if procSize == 14
    bsf PCLATH, 3
#endif
    call toggleLED1
#if procSize == 14
    ; Ensure subsequent call/goto are in page 0
    bcf PCLATH, 3
#endif

#if _DEBUG_CMDS
 movlw 'C'
 call TxRS232
 movlw 'm'
 call TxRS232
 movlw 'd'
 call TxRS232
 movfw COMMAND
 call TxRS232
#endif

    bcf INTCON, GIE

    ; Process the packet
    call processCommand

    bsf INTCON, GIE

    ; See if there is a response packet
    call WritePacket

    bcf PKTReadyFlag    ; Ok to read a new packet
#if _USE_RTSCTS == 1
    bcf CTS
#endif

    ; TBD: RTS/CTS control
    goto programLoop

processCommand   ; Stack depth 1
    ; Make sure command is in the correct range
    movf COMMAND, W
    sublw LAST_CMD
    btfss STATUS, C
    goto badCommand

    ; Computed goto, based on value of command
    movf COMMAND, W
    addwf PCL, F
#endif

; Start of indirect goto table for commands.  Both processors
; jump through this table to execute the input command - the
; difference is how the 16F and 18F perform indirect goto's.
cmdJumpTable:
    goto nullCommand     ; 0x00 - NOP
    goto readVersionCmd  ; 0x01 - Return the version of this firmware
    goto setLinesCmd     ; 0x02 - Set all control/LEDs on/off
    goto vppCmd          ; 0x03 - Turn Vpp on/off
    goto nullCommand     ; 0x04 - (unused - ETX)
    goto prgSelCmd       ; 0x05 - Turn control selection line on/off
    goto vddCmd          ; 0x06 - Turn Vdd on/off
    goto setCLKDATCmd    ; 0x07 - Set the ICSP CLK and DAT lines
    goto readLinesCmd    ; 0x08 - Read current value of CLK/DAT/Vpp/Vdd/PrgSel lines
    goto readRegsCmd     ; 0x09 - Read registers
    goto echoCmd         ; 0x0a - Send back the input
    goto beginProgCmd    ; 0x0b - Enter prog mode: lower DAT/CLK, raise Vdd and Vpp
    goto endProgCmd      ; 0x0c - Exit prog mode: lower Vdd/Vpp, tristate CLK/DAT
    goto PIC14Cmd        ; 0x0d - Send command to 14 bit PIC
    goto PIC14RCmd       ; 0x0e - Send cmd to 14 bit PIC, read response
    goto nullCommand     ; 0x0f - (unused)
    goto readAnalogCmd   ; 0x10 - Read the current value of the A/D converter
    goto setAnalogCmd    ; 0x11 - Set desired voltage bounds
    goto PIC16SetAddrCmd ; 0x12 - Set address of a 16 bit PIC
    goto PIC16Cmd        ; 0x13 - Send command to 16 bit PIC
    goto PIC16ReadCmd    ; 0x14 - Send cmd to 16 bit PIC, read response
    goto PIC16WriteCmd   ; 0x15 - Send write cmd to 16 bit PIC, hold CLK for time P9 (1-5 msec)
    goto readMasksCmd    ; 0x16 - Read current value of TRIS registers
    goto readPortsCmd    ; 0x17 - Read current value of PORT registers
    goto readPWMCmd      ; 0x18 - Read current PWM duty cycle
    goto PIC14RBPgm      ; 0x19 - Read a block of code from 14 bit PIC
    goto PIC14WBPgm      ; 0x1a - Write a block of code to 14 bit PIC
    goto PIC14RBDat      ; 0x1b - Read a block of data from 14 bit PIC EEPROM
    goto PIC14WBDat      ; 0x1c - Write a block of data to 14 bit PIC EEPROM
    goto PIC16RBlock     ; 0x1d - Read a block of bytes from 16 bit PIC
    goto PIC16WBlock     ; 0x1e - Write a block of bytes to 16 bit PIC
    goto PIC16REBlock    ; 0x1f - Read a block of data from a 16 bit PIC EEPROM
    goto PIC16WEBlock    ; 0x20 - Write a block of data to a 16 bit PIC EEPROM
    goto PIC16WBuf       ; 0x21 - Write one buffer of bytes to a 16 bit PIC
    goto PIC16WEByte     ; 0x22 - Write one byte to a 16 bit PIC EEPROM
    goto PIC16EBuf       ; 0x23 - Erase one window of bytes on a 16 bit PIC
    goto PIC24Cmd        ; 0x24 - Send command to 24 bit PIC
    goto PIC24ReadCmd    ; 0x25 - Send cmd to 24 bit PIC, read response
    goto PIC24RdBuf      ; 0x26 - Read one buffer of words from a 24 bit PIC
    goto PIC24WrBuf      ; 0x27 - Write one buffer of words to a 24 bit PIC
    goto PIC24WrBufSetup ; 0x28 - Setup to write a buffer to a 24 bit PIC
    goto PIC24WrBufFinish; 0x29 - Complete write of a buffer to a 24 bit PIC

;
; Set error value to indicate we got a command we can't interpret
;
nullCommand
    retlw 0x00

;
; Set error value to indicate we got a command we can't interpret
;
badCommand
    retlw ERROR_BADCMD

;
; Commands
;

; Send back the current firmware version
; In : <STX><STX>[<0x01><0x01>]<0xfe><CHKSUM><ETX>
; Out: <STX><STX>[<0x01><DLEN><N/A><VERL><VERH>]<CHKSUM><ETX>
readVersionCmd
    movlw 0x03
    movwf DATALEN
    movlw VERSION0
    movwf DATA0
    movlw VERSION1
    movwf DATA1

    retlw 0x05

;
; Send back exactly what we got
;
; In : <STX><STX>[<cmd><DATALEN><DATA...>]<CHKSUM><ETX>
; Out: <STX><STX>[<cmd><DATALEN><DATA...>]<CHKSUM><ETX>
echoCmd
    movf DATALEN, W
    addlw 0x02
    return

;
; Set the value of all signal lines (Vdd, Vpp, PrgSel) and
; LEDs.  Note: This routine is intended for debugging
; purposes only.  For programming, use: vppCmd, vddCmd,
; prgSelCmd, setCLKDATCmd.
;
; In : <STX><STX>[<cmd><0x01><flags><flags>]<CHKSUM><ETX>
; Out: <STX><STX>[<cmd>]<CHKSUM><ETX>
setLinesCmd
#if procSize == 14
    ; Initialization is in memory page 1
    bsf PCLATH, 3
#endif
    call setLineValues
#if procSize == 14
    ; Ensure subsequent call/goto are in page 0
    bcf PCLATH, 3
#endif
    movlw 0x01
    movwf DATALEN
    retlw 0x04

;
; Turn Vpp on/off based on bit 1 of the data word
;
; In : <STX><STX>[<cmd><0x01><flags>]<CHKSUM><ETX>
; Out: <STX><STX>[<cmd>]<CHKSUM><ETX>
vppCmd
    btfsc DATABUF0, 1
    goto Vpp1High
    bcf VPP1      ; Pull Vpp1 low (action is opposite of logic level)
    goto Vpp2Cmd
Vpp1High
    bsf VPP1      ; Pull Vpp1 high

    ; Turn Vpp on/off separately on the two
    ; pins of the programming socket that need it.
    ; Rationale: 40 pin chips are unhappy with the
    ; 13V coming in on one of the data lines - they
    ; lock up instead of allowing programming.  Smaller
    ; chips don't have this problem - only one Vpp
    ; line ever touches them.
Vpp2Cmd
    btfsc DATABUF0, 2
    goto Vpp2High
    bcf VPP2
    goto VppDone
Vpp2High
    bsf VPP2

VppDone
    movlw 0x01
    movwf DATALEN
    movlw 0x04
    return

;
; Turn Vdd on/off based on bit 0 of the data word
;
; In : <STX><STX>[<cmd><0x01><flags>]<CHKSUM><ETX>
; Out: <STX><STX>[<cmd>]<CHKSUM><ETX>
vddCmd
    movlw 0x01
    movwf DATALEN
    btfsc DATABUF0, 1
    goto VddHigh
    bcf VDD
    movlw 0x04
    return
VddHigh
    bsf VDD
    retlw 0x04

;
; Turn PRG_SEL on/off based on bit 0 of the data word
;
; In : <STX><STX>[<cmd><0x01><flags>]<CHKSUM><ETX>
; Out: <STX><STX>[<cmd>]<CHKSUM><ETX>
prgSelCmd
    movlw 0x01
    movwf DATALEN
    btfsc DATABUF0, 1
    goto prgSelHigh
    bcf PRG_SEL
    movlw 0x04
    return
prgSelHigh
    bsf PRG_SEL
    retlw 0x04

;
; Return values of control lines (Vpp, Vdd, PrgSel) and status
; LEDs
;
; In : <STX><STX>[<cmd><0x00><ETX>
; Out: <STX><STX>[<cmd><DLEN><N/A><flags><N/A>]<CHKSUM><ETX>
readLinesCmd
#if procSize == 14
    ; Initialization is in memory page 1
    bsf PCLATH, 3
#endif
    call readLineValues
#if procSize == 14
    ; Ensure subsequent call/goto are in page 0
    bcf PCLATH, 3
#endif

    retlw 0x05

;
; Turn CLK on/off based on bit 0 of data word; turn DAT on/off
; based on bit 1 of data word
;
; In : <STX><STX>[<cmd><0x01><flags><CHKSUM><ETX>
; Out:  <STX><STX>[<cmd>]<CHKSUM><ETX>
setCLKDATCmd
    btfsc DATABUF0, 0
    goto CLKHigh
    bcf ICSP_CLK
    goto checkDAT
CLKHigh
    bsf ICSP_CLK
checkDAT
    btfsc DATABUF0, 1
    goto DATHigh
    bcf ICSP_DAT
    goto sendAck
DATHigh
    bsf ICSP_DAT
    goto sendAck

;
; Return values of a number of registers
;
; In : <STX><STX>[<cmd><0x00><ETX>
; Out: <STX><STX>[<cmd><DLEN><N/A><N/A><N/A><N/A>][list of reg values]<CHKSUM><ETX>
readRegsCmd
#if procSize == 14
    ; Initialization is in memory page 1
    bsf PCLATH, 3
#endif
    call readRegValues
#if procSize == 14
    ; Ensure subsequent call/goto are in page 0
    bcf PCLATH, 3
#endif

    return

;
; Read the current value of the A/D converter
;
; In : <STX><STX>[<cmd><0x00><CHKSUM><ETX>
; Out: <STX><STX>[<cmd><DLEN><N/A><AD Val><N/A>]<CHKSUM><ETX>
readAnalogCmd
    ; Use current value of AN0
    movf analogVal, W
    movwf DATA0

    ; The bit set/clear of CHS0 toggles between AN0 and AN1 -
    ; it is important to leave with AN0 selected so the
    ; voltage boost routine will measure the right line.
    bsf ADCON0, CHS0  ; Select AN1
    bsf ADCON0, GO    ; Set GO bit
waitAN1
    btfsc ADCON0, GO
    goto waitAN1      ; Wait till conversion is done
    bcf ADCON0, CHS0  ; Select AN0

    movf ADRESH, W
    movwf DATA1

    movlw 0x03
    movwf DATALEN

    retlw 0x05

;
; Set the minimum and maximum voltage thresholds used for PWM
; adjustment.
;
; In : <STX><STX>[<cmd><0x02><Min V><Max V><CHKSUM><ETX>
; Out: <STX><STX>[<cmd>]<CHKSUM><ETX>
setAnalogCmd
    movlw 0x00
    goto sendAck
;
; Read the current value of the PWM duty cycle
;
; In : <STX><STX>[<cmd><0x00><CHKSUM><ETX>
; Out: <STX><STX>[<cmd><DLEN><N/A><PWM Val><N/A>]<CHKSUM><ETX>
readPWMCmd
    movf PWMVal, W
    movwf DATA0
    movlw 0x03
    movwf DATALEN
    clrf DATA1

    retlw 0x05

;
; Read the current value of port A and C direction masks
;
; In : <STX><STX>[<cmd><0x00><CHKSUM><ETX>
; Out: <STX><STX>[<cmd><0x02><N/A><mask0><mask1>]<CHKSUM><ETX>
readMasksCmd
#if procSize == 14
    banksel TRISA     ; bank 1
#endif
    movf TRISA, W
#if procSize == 14
    banksel DATA0      ; bank 0
#endif
    movwf DATA0
#if procSize == 14
    banksel TRISB     ; bank 1
#endif
    movf TRISB, W
#if procSize == 14
    banksel DATA1     ; bank 0
#endif
    movwf DATA1
#if procSize == 14
    banksel TRISC     ; bank 1
#endif
    movf TRISC, W
#if procSize == 14
    banksel CMDOPT    ; bank 0
#endif
    movwf CMDOPT

    movlw 0x03
    movwf DATALEN
    retlw 0x05

;
; Read the current value of port A and C
;
; In : <STX><STX>[<cmd><0x00><CHKSUM><ETX>
; Out: <STX><STX>[<cmd><0x03><N/A><val0><val1>]<CHKSUM><ETX>
readPortsCmd
    movlw 0x03
    movwf DATALEN
    movf PORTA, W
    movwf DATA0
    movf PORTC, W
    movwf DATA1
    retlw 0x05

;
; Enter programming mode
;
; Vdd and Vpp must be off when entering this routine
;
; In : <STX><STX>[<cmd><0x00><CHKSUM><ETX>
; Out: <STX><STX>[<cmd>]<CHKSUM><ETX>
beginProgCmd
#if procSize == 14
    banksel 0x80      ; bank 1
#endif
   ; Start in known good state, CLK tristate
    bsf ICSP_CLK_MASK
    bcf ICSP_DAT_MASK
#if procSize == 14
    banksel 0x00      ; bank 0
#endif
    ; Common code - pull CLK and DAT lines low
    ; Which PGC/PGD line should we use?
    bcf PRG_SEL
    btfsc DATABUF0, 2  ; Must be a dsPIC
    bsf PRG_SEL
    btfsc PRG_SEL
    goto dsPICInit
; Set CLK/DAT lines for standard PIC (12F/16F/18F)
#if procSize == 14
    banksel 0x80      ; bank 1
#endif
    bcf ICSP_CLK_MASK
    goto doneCLKDATInit
; Set CLK/DAT lines for dsPIC
dsPICInit
#if procSize == 14
    banksel 0x80      ; bank 1
#endif
    bcf ICSP_CLK_MASK
doneCLKDATInit
#if procSize == 14
    banksel 0x00      ; bank 0
#endif
    bcf ICSP_CLK      ; lower bits
    bcf ICSP_DAT
    movlw 0x02
    call delay        ; 22 instructions - 11 usec @ 8MHz, ~4.8 usec @ 18.432MHz
    ; Generic entry: raise Vdd, wait 5 usec (14 bit)/100 nsec (16 bit);
    bsf VDD
    movlw 0x0c
    call delay
    ; raise Vpp; wait 5 usec (14 bit) 2 usec (16 bit)
    btfsc DATABUF0, 0
    bsf VPP2           ; Pull Vpp2 high
    btfsc DATABUF0, 1
    bsf VPP1           ; Pull Vpp1 high
    movlw 0x06
    call delay        ; 22 instructions - 11 usec @ 8MHz, ~4.8 usec @ 18.432MHz
    ; After entering programming mode, assume PC is at address 0
    clrf PCADDR0
    clrf PCADDR1
    clrf PCADDR2
    ; Now safe to toggle CLK/DAT lines
    goto sendAck      ; done

;
; Tristate CLK and DATA, turn Vdd and Vpp off
;
; In : <STX><STX>[<cmd><0x00><CHKSUM><ETX>
; Out: <STX><STX>[<cmd>]<CHKSUM><ETX>
endProgCmd
    bcf ICSP_CLK      ; ensure bits are low
    bcf ICSP_DAT
#if procSize == 14
    banksel TRISA     ; bank 1
#endif
    bsf ICSP_CLK_MASK  ; tristate ICSP_CLK line
    bsf ICSP_DAT_MASK  ; tristate ICSP_DAT line
#if procSize == 14
    banksel 0x00      ; bank 0
#endif
    bcf VPP1          ; Pull Vpp low
    bcf VPP2          ; Pull Vpp low
    bcf VDD           ; Pull Vdd low
    goto sendAck      ; done

;
;
; 14 bit PIC code
;
;

;
; Transmit command to the PIC - 6 bits.  Command must be in
; ICSP_CMD.  Affects: ICTR, ICSP_CMD.
;
PIC14TxCmd
    ; Ensure ICSP_DAT line is an output and starts low.
#if procSize == 14
    banksel 0x80 ; bank 1
#endif
    bcf ICSP_DAT_MASK
#if procSize == 14
    banksel 0x00 ; bank 0
#endif
    bcf ICSP_DAT

    movlw 0x06
    movwf ICTR

P14CmdLoop1
    bsf ICSP_CLK      ; raise clock
    bcf ICSP_DAT      ; default data low
    rrcf ICSP_CMD, F  ; get next bit in carry
    btfsc STATUS,C    ; is the bit set?
    bsf ICSP_DAT      ; yes - raise ICSP_DAT
    bcf ICSP_CLK      ; lower clock, PIC latches data
    decfsz ICTR, F    ; dec ctr
    goto P14CmdLoop1  ; loop if not zero
    bcf ICSP_DAT      ; Ensure ICSP_DAT low
    return

;
; Transmit a data payload to the target 14 bit PIC and
; read any response data.  Response will be in
; PICDATA1:PICDATA0 (lsb in PICDATA0).  PIC14TxCmd must
; have been called immediately before this routine.
; Affects: ICTR, DATA0, DATA1, PICDATA0, PICDATA1.
PIC14TxRxData
    movlw 0x10
    movwf ICTR

    ; Prepare 14 bit value to send to target.  We want it
    ; to sit in the middle of the 16 bits we send (0 for
    ; bit 0 and bit 15).
    bcf STATUS, C
    rlcf DATA0, F
    rlcf DATA1, F

; Transmit/Receive data to/from the PIC - 16 bits
P14DatLoop1
    bsf ICSP_CLK      ; raise clock
    bcf ICSP_DAT      ; default data low
    rrcf DATA1, F     ; rotate data and read low bit
    rrcf DATA0, F     ; move carry into low byte
    btfsc STATUS,C
    bsf ICSP_DAT      ; pull data high
    nop               ; ensure TSET1 is satisfied
incCtr2
    bcf ICSP_CLK      ; lower clock, PIC latches data
    ; Read any output from the PIC
    rrcf PICDATA1, F
    rrcf PICDATA0, F
    bcf PICDATA1, 7   ; Default is clear bit
#if procSize == 14
    banksel 0x80      ; bank 1
#endif
    bsf ICSP_DAT_MASK ; Ensure ICSP_DAT is an input
#if procSize == 14
    banksel 0x00
#endif
    btfsc ICSP_DAT    ; check the value of output from PIC
    bsf PICDATA1, 7      ; set if PIC is sending a 1
#if procSize == 14
    banksel 0x80
#endif
    bcf ICSP_DAT_MASK ; Set ICSP_DAT as output to send data
#if procSize == 14
    banksel 0x00
#endif
    decfsz ICTR, F    ; dec ctr
    goto P14DatLoop1  ; loop if not zero
    bcf ICSP_DAT      ; pull ICSP_DAT low

    ; Bits 0 and 15 aren't valid.  Rotate the result and mask off
    ; the two high bits of PICDATA1.  Final result sent back will
    ; the actual 14 bit value in code memory.
    rrcf PICDATA1, F
    rrcf PICDATA0, F
    movlw 0x3f
    andwf PICDATA1, F

    return;

;
; Clock and data should both be OUTPUT/LOW at this stage...
; we're going to clock 6 bits of command data from ICSP_CMD, then
; clock 16 bits of data from DATA1/LO - in the case of read commands,
; data will be tristate anyway (handy pullup).
;
; In : <STX><STX>[<cmd><0x03><ICSP Cmd><Data0><Data1>]<CHKSUM><ETX>
; Out: <STX><STX>[<cmd>]<CHKSUM><ETX>
; Out: <STX><STX>[<cmd><0x03><ICSP Cmd><Data0><Data1>]<CHKSUM><ETX>
PIC14Cmd
    call PIC14TxCmd   ; Send the ICSP command
    goto sendAck      ; Send ack response

PIC14RCmd
    call PIC14TxCmd
    call PIC14TxRxData

    ; Send back the command #, the ICSP command, and any data
    ; we got from the PIC (which will be in DATA0 and DATA1)
    movlw 0x03
    movwf DATALEN
    movf PICDATA0, W
    movwf DATA0
    movf PICDATA1, W
    movwf DATA1
    retlw 0x05

;
; Read a block of data from a 14 bit PIC
;
; In:    <STX><STX>[<0x19><DLEN><ADDRL><ADDRH><ADDRU><DATA>...]<CHKSUM><ETX>
; Out:   <STX><STX>[<0x19><DLEN><ADDRL><ADDRH><ADDRU>]<CHKSUM><ETX>
;
; DLEN on input tells us how many words of information are being
; requested.  The maximum is 32 (64 bytes) to prevent overflow.
; TBD: Use the input address to find the data...
;
; Affects: TEMP
;
PIC14RBlock
    ; TBD: Save contents of ADDRU:ADDRH:ADDRL so they can be
    ; restored after we use the overlaid locations for
    ; ICSP_CMD:DATA0:DATA1

    ; TBD: Pull out requested address and compare to the
    ; one saved in PCADDR2:PCADDR1:PCADDR0.  If they don't
    ; match, then either increment until we get there, or
    ; reset the target and increment.  (Note, on 16 bit parts
    ; we can directly change the target address.)

    ; Prepare response packet
    ; Point to the output buffer
#if procSize == 16
    movlw HIGH(DATA_BUF)
    movwf FSR0H
#endif
    movlw LOW(DATA_BUF)
    movwf FSR0L

    movf CMDOPT, W   ; Get # of bytes requested
    andlw 0xc0       ; Is # requested more than 64?
    btfsc STATUS, Z
    goto PIC14RB1    ; Nope, continue on

    ; Too many bytes requested.  Limit to 64
    movlw 0x40
    movwf CMDOPT     ; Send back the reduced #

PIC14RB1
    movf CMDOPT, W   ; Move requested # of bytes into TEMP
    movwf TEMP       ; to use for loop counter
    bcf STATUS, C    ; Don't rotate in any leftovers
    rrcf TEMP, F     ; Divide by 2 for # of words

PIC14RBLoop
    ; Read data from the current memory location
    movf CMDTEMP, W
    movwf ICSP_CMD
    call PIC14TxCmd
    call PIC14TxRxData

    ; Move response to output buffer (LSB first)
#if procSize == 16
    movf PICDATA0, W
    movwf POSTINC0
    movf PICDATA1, W
    movwf POSTINC0
#else
    movf PICDATA0, W
    movwf INDF
    incf FSR, F
    movf PICDATA1, W
    movwf INDF
    incf FSR, F
#endif

    ; Increment the memory address
    movlw PIC14_INCADDR ; Increment address command for 14 bit PICs
    movwf ICSP_CMD
    call PIC14TxCmd

    ; See if there are any more words to read
    decfsz TEMP, F
    goto PIC14RBLoop

    movf CMDOPT, W   ; Output size is CMDOPT + cmd/dlen/address/opt
    addlw 0x6
    return           ; Send back the data
 
PIC14RBPgm
    movlw PIC14_READDATA   ; Read code command for 14 bit PICs
    movwf CMDTEMP
    goto PIC14RBlock

PIC14RBDat
    movlw PIC14_READE2DATA ; Read data command for 14 bit PICs
    movwf CMDTEMP
    goto PIC14RBlock

;
; Write a block of code to a 14 bit PIC
;
; In:    <STX><STX>[<0x01><DLEN><ADDRL><ADDRH><ADDRU><CMDOPT><DATA>]<CHKSUM><ETX>
; In:    <STX><STX>[<0x01><DLEN><PROGCMD><ENDPROGCMD><N/A><CMDOPT><DATA>]<CHKSUM><ETX>
; OUT:    <STX><STX>[<0x01><DLEN><N/A><N/A><N/A><DATA>]<CHKSUM><ETX>
;

PIC14WBPgm
    movf DATABUF0, W
    movwf PROGTEMP
    movf DATABUF1, W
    movwf ENDPROGTEMP
    movf DATABUF2, W
    movwf PCADDR2

#if procSize == 16
    movlw HIGH(DATA_BUF)   ; Point to the output buffer
    movwf FSR0H
#endif
    movlw LOW(DATA_BUF)
    movwf FSR0L

    movf CMDOPT, W   ; Get # of bytes requested
    sublw 0xc0       ; Is # requested more than 64?
    btfsc STATUS, C
    goto PIC14WB1    ; Nope, continue on

    ; Too many bytes requested.  Limit to 64
    movlw 0x40
    movwf CMDOPT     ; Send back the reduced #

PIC14WB1
    movf CMDOPT, W   ; Move requested # of bytes into TEMP
    movwf TEMP       ; to use for loop counter
    bcf STATUS, C    ; Don't rotate in any leftovers
    rrcf TEMP, F     ; Divide by 2 for # of words

PIC14WBLoop
    ; Up to MAX_RETRIES before abandoning write
    movlw MAX_RETRIES
    movwf RETRYCTR
PIC14WBRetryLoop

    ; Read data from the current memory location
    movlw PIC14_READDATA
    movwf ICSP_CMD
    call PIC14TxCmd
    call PIC14TxRxData

    ; Compare to desired value. If same then
    ; go to next one in block.
    ; Move response to output buffer
    movf PICDATA0, W  ; Compare low bytes
    xorwf INDF0, W
    btfss STATUS, Z
    goto PIC14WBRetry1
    incf FSR0L, F
    movf PICDATA1, W  ; Compare high bytes
    xorwf INDF0, W
    btfss STATUS, Z
    goto PIC14WBRetry2
    incf FSR0L, F

    goto PIC14WBOk    ; Write confirmed ok

PIC14WBRetry2
    decf FSR0L, F      ; Back up pointer
PIC14WBRetry1
    ; If not a match then see if we have tried
    ; enough times
    decfsz RETRYCTR, F
    goto PIC14WBWrite
PIC14WBFail
    ; Write failed, send back neg ack
    movlw 0xee
    movwf COMMAND
#if 1
    movlw 0x03
    movwf DATALEN
    movf TEMP, W
    movwf DATA0
    movf RETRYCTR, W
    movwf DATA1
    retlw 0x05
#else
    goto sendAck
#endif

PIC14WBWrite
    ; Move current data into DATA0:DATA1 without moving INDF0
    ; past the current data value.
#if procSize == 16
    movff POSTINC0, DATA0
    movff POSTDEC0, DATA1
#else
    movf INDF, W
    movwf DATA0
    incf FSR, F
    movf INDF, W
    movwf DATA1
    decf FSR, F
#endif

    ; Load the data
    movlw PIC14_LOADDATA
    movwf ICSP_CMD
    call PIC14TxCmd
    call PIC14TxRxData

    ; Initiate programming
    ; movlw PIC14_BEGINPROG
    movf PROGTEMP, w
    movwf ICSP_CMD
    call PIC14TxCmd

    ; Wait for programming to complete (max 2.5 msec according to
    ; datasheets).  Passes needed is (clock / 4) * (# msec / 9)
    ; (9 instructions per pass through the loop, 4 clocks per
    ; instruction.
    ;   At 18.432 MHz: (18.432 / 4) * (2500 / 9) = 1280 (0x500)
    ;   At 8 MHz: (8 / 4) * (2500 / 9) = 555.5 (0x22c)
    ;   At 48 MHz: (48 / 4) * (2500 / 9) = 3333.3 (0xD05)
    ; TBD: Make the duration of this delay based on an input.  It
    ; can change quite a bit
    ; Note: 16F83-84, 16F84A seem to have minimums of more than 4 msec.
    clrf CTRX
    movf PCADDR2, W ; See if there is a delay value in DATABUF2
    btfss STATUS, Z  ; If so, then use it
    goto setDelay2
    ; No input delay value - use default
setDelay1
#if CLK_SPD > 24
    ; 24-48 MHz clock
    movlw 0x0d
#else
#if CLK_SPD > 16
    ; 16-24 MHz clock
    movlw 0x08
#else
#if CLK_SPD > 8
    ; 8-16 MHz clock
    movlw 0x04
#else
    ; 5-8 MHz clock
    movlw 0x02
#endif
#endif
#endif

setDelay2
    movwf CTRX+1
PIC14ProgLoop
    DEC16 CTRX
    btfss STATUS, Z
    goto PIC14ProgLoop

    ; After the programming cycle is done, some parts need to
    ; have an end program command (e.g., 16F88).  If the value
    ; of ENDPROGTEMP is nonzero then send it to the PIC.
    movf ENDPROGTEMP, W
    btfsc STATUS, Z        ; See if ENDPROGTEMP is zero
    goto PIC14EndProgDone  ; It is - don't need end prog command

    ; ENDPROGTEMP is set, so we need to send it to the chip
    movwf ICSP_CMD
    call PIC14TxCmd

PIC14EndProgDone
    ; Programming delay done.  See if the word was successfully written
    goto PIC14WBRetryLoop

PIC14WBOk
    ; Increment the memory address
    movlw PIC14_INCADDR    ; Increment address command for 14 bit PICs
    movwf ICSP_CMD
    call PIC14TxCmd

    ; See if there are any more words to read
    decfsz TEMP, F
    goto PIC14WBLoop

#if 1
    goto sendAck
#else
    movf CMDOPT, W         ; Output size is CMDOPT + cmd/dlen/address/opt
    addlw 0x6
    return       ; Send back the data
#endif
    return;

;
; Write a block of data to a 14 bit PIC
;
; In:    <STX><STX>[<0x01><DLEN><ADDRL><ADDRH><ADDRU><CMDOPT><DATA>]<CHKSUM><ETX>
; OUT:    <STX><STX>[<0x01><DLEN><ADDRL><ADDRH><ADDRU><DATA>]<CHKSUM><ETX>
;
PIC14WBDat
    movff DATABUF0, PROGTEMP
    movff DATABUF1, ENDPROGTEMP

#if procSize == 16
    movlw HIGH(DATA_BUF)   ; Point to the output buffer
    movwf FSR0H
#endif
    movlw LOW(DATA_BUF)
    movwf FSR0L

    movf CMDOPT, W   ; Get # of bytes requested
    andlw 0xc0       ; Is # requested more than 64?
    btfsc STATUS, Z
    goto PIC14WBE1   ; Nope, continue on

    ; Too many bytes requested.  Limit to 64
    movlw 0x40
    movwf CMDOPT     ; Send back the reduced #

PIC14WBE1
    movf CMDOPT, W   ; Move requested # of bytes into TEMP
    movwf TEMP       ; to use for loop counter

PIC14WBELoop
    ; Up to MAX_RETRIES before abandoning write
    movlw MAX_RETRIES
    movwf RETRYCTR
PIC14WBERetryLoop

    ; Read data from the current memory location
    movlw PIC14_READE2DATA
    movwf ICSP_CMD
    call PIC14TxCmd
    call PIC14TxRxData

    ; Compare low byte to desired value. If same then
    ; move response to output buffer
    movf PICDATA0, W   ; Compare low bytes
    xorwf INDF0, W
    btfsc STATUS, Z
    goto PIC14WBEOk   ; Write confirmed ok

PIC14WBERetry1
    ; If not a match then see if we have tried
    ; enough times
    decfsz RETRYCTR, F
    goto PIC14WBEWrite
PIC14WBEFail
    ; Write failed, send back neg ack
    movlw 0xee
    movwf COMMAND
#if 0
    goto sendAck
#else
    movlw 0x01
    movwf DATALEN
    movf TEMP, W
    movwf DATABUF0
    retlw 0x03
#endif

PIC14WBEWrite
    ; Move current data into DATA0 without moving INDF0
    ; past the current data value.  Clear high byte (not
    ; used when writing EEPROM data.
    movf INDF0, W
    movwf DATA0
    clrf DATA1

    ; Load the data
    movlw PIC14_LOADE2DATA
    movwf ICSP_CMD
    call PIC14TxCmd
    call PIC14TxRxData

    ; Initiate programming
    movf PROGTEMP, W
    movwf ICSP_CMD
    call PIC14TxCmd

    ; Wait for programming to complete (5 to 6 msec according to
    ; datasheets).  Passes needed is (clock / 4) * (# msec / 9)
    ; (9 instructions per pass through the loop, 4 clocks per
    ; instruction.
    ;   At 48 MHz: (48 / 4) * (6000 / 9) = 8000 (0x1F40)
    clrf CTRX
#if CLK_SPD > 24
    ; 24-48 MHz clock
    movlw 0x20
#else
#if CLK_SPD > 16
    ; 16-24 MHz clock
    movlw 0x10
#else
#if CLK_SPD > 8
    ; 8-16 MHz clock
    movlw 0x0b
#else
    ; 5-8 MHz clock
    movlw 0x06
#endif
#endif
#endif
    movwf CTRX+1
PIC14ProgELoop
    DEC16 CTRX
    btfss STATUS, Z
    goto PIC14ProgELoop

    ; After the programming cycle is done, some parts need to
    ; have an end program command (e.g., 16F88).  If the value
    ; of ENDPROGTEMP is nonzero then send it to the PIC.
    movf ENDPROGTEMP, W
    btfsc STATUS, Z        ; See if ENDPROGTEMP is zero
    goto PIC14EndProgEDone ; It is - don't need end prog command

    ; ENDPROGTEMP is set, so we need to send it to the chip
    movwf ICSP_CMD
    call PIC14TxCmd

PIC14EndProgEDone
    ; Programming delay done.  See if the word was successfully written
    goto PIC14WBERetryLoop

PIC14WBEOk
    ; Increment the memory address
    incf FSR0L, F           ; Move to next byte in the buffer
    movlw PIC14_INCADDR    ; Increment address command for 14 bit PICs
    movwf ICSP_CMD
    call PIC14TxCmd

    ; See if there are any more bytes to read
    decfsz TEMP, F
    goto PIC14WBELoop

    movf CMDOPT, W         ; Output size is CMDOPT + cmd/dlen/address/opt
    addlw 0x6
    return                 ; Send back the data

;
;
; 16 bit PIC code
;
;

;
; Set the target chips address to ADDRU:ADDRH:ADDRL
PIC16SetTablePtr
    ; Save the address
    movf ADDRU, W
    movwf PCADDR2
    movf ADDRH, W
    movwf PCADDR1
    movf ADDRL, W
    movwf PCADDR0

    coreCmd2 0x0e, PCADDR2
    coreCmd1 0x6ef8
    coreCmd2 0x0e, PCADDR1
    coreCmd1 0x6ef7
    coreCmd2 0x0e, PCADDR0
    coreCmd1 0x6ef6

    ; A final NOP to finish things off
    coreCmd1 0x0000

    return

; Set the target address
; In:    <cmd><DLEN><ADDRL><ADDRH><ADDRU>
PIC16SetAddrCmd
    call PIC16SetTablePtr

    movlw 0x04
    movwf DATALEN
    movf PCADDR2, W
    movwf ADDRU
    movf PCADDR1, W
    movwf ADDRH
    movf PCADDR0, W
    movwf ADDRL
    retlw 0x06

; Setup CLK/DAT lines and send 4 bit command to a 16 bit
; PIC processor.  ICSP_CLK and ICSP_DAT should both be
; low at this point.  Serial clock period P2 ranges from
; 100 nsec to 1 usec, depending on Vdd.  Even at 18MHz,
; the loop below will not be shorter than the minimum
; (loop is 10 instruction cycles long, ~2 usec at 18MHz,
; ~4.6 usec at 8MHz).
FourBitCmdSetup
    ; Ensure ICSP_DAT line is an output and starts low.
#if procSize == 14
    banksel 0x80          ; bank 1
#endif
    bcf ICSP_DAT_MASK
#if procSize == 14
    banksel 0x00          ; bank 0
#endif
    bcf ICSP_DAT

    movlw 0x04            ; 4 bits in the command
    movwf ICTR
FourBitCmdLoop
    bsf ICSP_CLK          ; Pull clock high
    bcf ICSP_DAT          ; Start data low, raise if cmd bit set
    rrcf ICSP_CMD, F      ; Read bit in command
    btfsc STATUS,C        ; If set then raise ICSP_DAT
    bsf ICSP_DAT          ; Pull ICSP_DAT high
    bcf ICSP_CLK          ; PIC latches in data when clock goes low
#if CLK_SPD > 24
    nop
    nop
#endif
    decfsz ICTR, F        ; Fall out of bottom if all 4 bits sent
    goto FourBitCmdLoop     ; Keep sending the command

    ; Need to ensure P5 time (20 nsec) is met here.  Shouldn't
    ; be an issue if this code is running at less than 200 MHz
    bcf ICSP_DAT          ; Ensure DAT is low before continuing
    rrcf ICSP_CMD, F      ; top and bottom of byte should be swapped
    return

; Transmit command to the PIC - 4 bit command for 16 bit devices
; followed by 16 bits of data output.  This routine is for commands
; that do not require a data response from the PIC.  Must be in
; bank 0, ICSP_CLK aand ICSP_DAT must be low.
; Affects: ICTR, DATA0, DATA1.
;
; Total time is approximately 220 instructions
;
PIC16CmdBase ; Stack depth 2 or 3
    ; Send the 4 bit command to the target PIC
    call FourBitCmdSetup

    ; Send 16 bit data payload to the target
    movlw 0x10
    movwf ICTR
pic16DATLoop1
    ; Serial clock period must be > P2 (100 nsec to 1 usec)
    bsf ICSP_CLK      ; raise clock
    bcf ICSP_DAT      ; default data low
    rrcf DATA1, F     ; rotate data and read low bit
    rrcf DATA0, F
    btfsc STATUS,C
    bsf ICSP_DAT      ; pull data high
    bcf ICSP_CLK      ; lower clock, PIC latches data
#if CLK_SPD > 24
    nop
    nop
#endif
    decfsz ICTR, F    ; dec ctr
    goto pic16DATLoop1; loop if not zero
    bcf ICSP_DAT      ; pull DAT low

    return


; Send a command and 16 bit payload to the target.
; In:  <cmd><data len><ICSP cmd><data high byte><data low byte>
; Out: <STX><STX>[<cmd><0x03><ICSP cmd><data0><data1>]<CHKSUM><ETX>
PIC16Cmd
    ; Save off the input data
    movf ICSP_CMD, W
    movwf CMDTEMP
    movf DATA0, W
    movwf PICDATA0
    movf DATA1, W
    movwf PICDATA1

    call PIC16CmdBase

    ; Send back input as an ack
    movlw 0x03
    movwf DATALEN
    movf CMDTEMP, W
    movwf ICSP_CMD
    movf PICDATA0, W
    movwf DATA0
    movf PICDATA1, W
    movwf DATA1

    movlw 0x05        ; 5 - cmd, data len, ICSP cmd, Data0, Data1
    return

; Send 4 bit command, 5 msec hold of clk/dat, then 16 bit payload
PIC16DlyCmd
     ; Save off the input data
    movf ICSP_CMD, W
    movwf CMDTEMP
    movf DATA0, W
    movwf PICDATA0
    movf DATA1, W
    movwf PICDATA1

   ; Send the 4 bit command to the target PIC
    call FourBitCmdSetup

    ; Wait time P11 (5 msec) +P10 (5 usec) and then disable writes.  The
    ; loop below takes 9 instructions per cycle.  For 5 msec, there are
    ; 30,000 instructions (at 24 MHz), giving 0x0d05 loops.

    ; Ensure clk/dat low
    bcf ICSP_DAT
    bcf ICSP_CLK

    ; 
    clrf CTRX
    movlw 0x1b
    movwf CTRX+1
PIC16CmdDlyLoop
    DEC16 CTRX
    btfss STATUS, Z
    goto PIC16CmdDlyLoop

    ; Send 16 bit data payload to the target
    movlw 0x10
    movwf ICTR
pic16CmdDlyLoop1
    ; Serial clock period must be > P2 (100 nsec to 1 usec)
    bsf ICSP_CLK      ; raise clock
    bcf ICSP_DAT      ; default data low
    rrcf DATA1, F     ; rotate data and read low bit
    rrcf DATA0, F
    btfsc STATUS,C
    bsf ICSP_DAT      ; pull data high
    bcf ICSP_CLK      ; lower clock, PIC latches data
#if CLK_SPD > 24
    nop
    nop
#endif
    decfsz ICTR, F    ; dec ctr
    goto pic16CmdDlyLoop1; loop if not zero
    bcf ICSP_DAT      ; pull DAT low

    ; Send back input as an ack
    movlw 0x03
    movwf DATALEN
    movf CMDTEMP, W
    movwf ICSP_CMD
    movf PICDATA0, W
    movwf DATA0
    movf PICDATA1, W
    movwf DATA1

    retlw 0x05        ; 5 - cmd, data len, ICSP cmd, Data0, Data1

; Transmit a command to the PIC, then hold the CLK line high for
; the maximum programming time (P9).  This routine is necessary
; when initiating a memory write on the target.  
PIC16WriteBase
    movlw 0x04            ; 4 bits in the command
    movwf ICTR
pic16WriteLoop
    bsf ICSP_CLK          ; Pull clock high
    bcf ICSP_DAT          ; Start data low, raise if cmd bit set
    rrcf ICSP_CMD, F      ; Read bit in command
    btfsc STATUS,C        ; If set then raise ICSP_DAT
    bsf ICSP_DAT          ; Pull ICSP_DAT high
    decfsz ICTR, F        ; Done writing if all 4 bits sent
    goto continueWrite    ; Not done - pull clock low and continue
    goto cmdWriteDone
continueWrite
    bcf ICSP_CLK
    goto pic16WriteLoop
cmdWriteDone
; Need to delay for time P9 (1 msec) after clocking out the instruction.
;
; Delay loop for 9*N instructions.  The duration of the delay
; will be determined by the clock speed.  4 instruction cycles
; per microsecond for an 48MHz clock.
    ; 12000 (0x2EE0) instructions per msec at 48 MHz.
    ; 9 instructions per loop, giving 1333 (0x535) loops.
#if CLK_SPD > 24
    ; 24-48 MHz clock
    movlw 0x35
    movwf CTRX
    movlw 0x05
    movwf CTRX+1
#else
#if CLK_SPD > 16
    ; 16-24 MHz clock
    movlw 0x9a
    movwf CTRX
    movlw 0x02
    movwf CTRX+1
#else
#if CLK_SPD > 8
    ; 8-16 MHz clock
    movlw 0xbc
    movwf CTRX
    movlw 0x01
    movwf CTRX+1
#else
    ; 5-8 MHz clock
    movlw 0xde
    movwf CTRX
    movlw 0x00
    movwf CTRX+1
#endif
#endif
#endif

cmdWriteDelay
    DEC16 CTRX         ; 0..5
    btfss STATUS, Z    ; 6
    goto cmdWriteDelay ; 7..8

    bcf ICSP_CLK        ; now pull clock low

   ; delay for time P10 (5 usec) for high voltage discharge.
    ; At 48MHz, there are 12 instructions per usec, so
    ; we need to wait 60 instruction cycles for 5 usec.
#if CLK_SPD > 24
    ; 24-48 MHz clock
    movlw 0x12        ; 0
#else
#if CLK_SPD > 16
    ; 16-24 MHz clock
    movlw 0x09
#else
#if CLK_SPD > 8
    ; 8-16 MHz clock
    movlw 0x05
#else
    ; 5-8 MHz clock
    movlw 0x02
#endif
#endif
#endif

    call delay        ; 1..25

    ; Clear data registers and clock out 16 bits.  No data
    ; going either way - all commands have a payload, even
    ; if they are ignored.
    clrf DATA1
    clrf DATA0
    movlw 0x10
    movwf ICTR
    call pic16DATLoop1

    ; All done with programming write
    return

; In:  <cmd><data len><ICSP cmd><data0><data1>
; Out: <cmd><0x03><ICSP cmd><data0><data1>
PIC16WriteCmd
    call PIC16WriteBase
    goto sendAck

; Transmit command to the PIC - 4 bit command for 16 bit devices
; followed by 16 bits of data I/O.  We are looking for 8 bits of
; data in the second set of 8 bits
; Affects; ICTR, ICSP_CMD, PICDATA0
PIC16ReadBase
    call FourBitCmdSetup
    ; Toggle the clock 8 times and ignore anything that comes back
    movlw 0x08
    movwf ICTR
    bcf ICSP_DAT      ; Hold ICSP_DAT low for first 8 bits

    ; Serial clock period must be > P2 (100 nsec to 1 usec)  With the
    ; 48MHz crystal, there are 12 instructions per usec.  For slower
    ; clock rates (24 MHz or below), the 6 instruction loop will be
    ; adequate to meet timing requirements.
pic16RdLoop1
#if CLK_SPD > 24
    bsf ICSP_CLK         ; 0      raise clock
    nop                  ; 1
    nop                  ; 2
    nop                  ; 3
    nop                  ; 4
    nop                  ; 5
    bcf ICSP_CLK         ; 6      lower clock, PIC latches data
    nop                  ; 7
    nop                  ; 8
    decfsz ICTR, F       ; 9
    goto pic16RdLoop1    ; 10..11   loop if not zero
#else
    bsf ICSP_CLK         ; 0      raise clock
    nop                  ; 1
    bcf ICSP_CLK         ; 2      lower clock, PIC latches data
    decfsz ICTR, F       ; 3      dec ctr
    goto pic16RdLoop1    ; 4..5   loop if not zero
#endif

    ; Ensure P6 (20 nsec) between first 8 bits and second

    ; Toggle 8 more times and read the result byte
#if procSize == 14
    banksel 0x80      ; bank 1
#endif
    bsf ICSP_DAT_MASK ; Set DAT line as input
#if procSize == 14
    banksel 0x00      ; bank 0
#endif
    movlw 0x08
    movwf ICTR
    clrf PICDATA0
pic16RdLoop2
    ; Serial clock period must be > P2 (100 nsec to 1 usec)
    bsf ICSP_CLK      ; Raise ICSP_CLK
    rrcf PICDATA0, F  ; rotate previous contents before setting bit
    bcf PICDATA0, 7   ; Start output bit as 0
    bcf ICSP_CLK      ; Lower ICSP_CLK, PIC shifts data out
    btfsc ICSP_DAT    ; Check value of output from PIC
    bsf PICDATA0, 7   ; Set bit if PIC is sending a 1
    decfsz ICTR, F    ; 
    goto pic16RdLoop2 ; loop if not zero
#if procSize == 14
    banksel 0x80      ; bank 1
#endif
    bcf ICSP_DAT_MASK ; Set DAT back to output
#if procSize == 14
    banksel 0x00      ; bank 0
#endif
    bcf ICSP_DAT      ; pull DAT low

    return


; 16 bit PIC table read command has to already be in ICSP_CMD
; In:  <cmd><data len><ICSP cmd><N/A><N/A>
; Out: <cmd><DLEN><N/A><data val><N/A>
PIC16ReadCmd
    movf ICSP_CMD, W
    movwf CMDTEMP

    ; Read a byte from the target
    call PIC16ReadBase

    ; Move value read from PIC to output buffer
    movf PICDATA0, W
    movwf DATA0
    movf PICDATA1, W
    movwf DATA1
    movlw 0x03
    movwf DATALEN

    movlw 0x05
    return

; Routine to prepare a packet for block read/write
packetInit
    ; Prepare response packet
#if procSize == 16
    movlw HIGH(DATA_BUF)
    movwf FSR0H
#endif
    movlw LOW(DATA_BUF) ; Point to the output buffer
    movwf FSR0L         ;

    ; Save the address
    movf ADDRL, W
    movwf PCADDR0
    movf ADDRH, W
    movwf PCADDR1
    movf ADDRU, W
    movwf PCADDR2

    movf CMDOPT, W   ; Get # of bytes requested
    andlw 0xc0       ; Is # requested more than 64?
    btfsc STATUS, Z
    return           ; Size is ok - continue on

    ; Too many bytes requested.  Limit to 64
    movlw 0x40
    movwf CMDOPT     ; Send back the reduced #
    return

;
; Read a block of data from a 16 bit PIC
;
; In:    <cmd><DLEN><ADDRL><ADDRH><ADDRU><CMDOPT>
; Out:   <cmd><DLEN><ADDRL><ADDRH><ADDRU><cmdopt><data>
;
; DLEN on input tells us how many words of information are being
; requested.  The maximum is 32 (64 bytes) to prevent overflow.
; TBD: Use the input address to find the data...
;
; Affects: TEMP
;
PIC16RBlock
    call packetInit
    call PIC16SetTablePtr ; Set the table pointer to ADDRU:ADDRH:ADDRL
PIC16RB1
    movf CMDOPT, W
    movwf TEMP ; Move # of bytes into TEMP for loop counter

PIC16RBLoop
    ; Read data from the current memory location
    movlw PIC16_TBLRD_POSTINC
    movwf ICSP_CMD
    call PIC16ReadBase

    ; Move response to output buffer (LSB first)
#if procSize == 16
    movff PICDATA0, POSTINC0
#else
    movf PICDATA0, W
    movwf INDF
    incf FSR, F
#endif

    ; See if there are any more words to read
    decfsz TEMP, F
    goto PIC16RBLoop

    movf CMDOPT, W  ; Output size is CMDOPT + cmd/dlen/address/opt
    addlw 0x6
    return          ; Send back the data

; Write a single buffer of bytes of code to a 16 bit PIC.  The
; value of CMDOPT gives the size of the buffer.
PIC16WBuf
    call packetInit
    call PIC16SetTablePtr ; Set the table pointer to ADDRU:ADDRH:ADDRL

    bcf STATUS, C    ; Don't rotate in any leftovers
    rrcf CMDOPT, F  ; Divide by 2 for # of words
    movf CMDOPT, W
    movwf TEMP

    decf TEMP, F     ; Move all but the last one into place

PIC16WBufFillLoop
    ; Write the current word and increment the table pointer
    movlw PIC16_PBLWRT_POSTINC
    movwf ICSP_CMD
#if procSize == 16
    movff POSTINC0, DATA0
    movff POSTINC0, DATA1
#else
    movf INDF, W
    movwf DATA0
    incf FSR, F
    movf INDF, W
    movwf DATA1
    incf FSR, F
#endif
    call PIC16CmdBase

    decfsz TEMP, F
    goto PIC16WBufFillLoop

    ; Move the final byte into place, but don't increment
    ; the table pointer out of the current buffer
    movlw PIC16_STARTPROG
    movwf ICSP_CMD
#if procSize == 16
    movff POSTINC0, DATA0
    movff POSTINC0, DATA1
#else
    movf INDF, W
    movwf DATA0
    incf FSR, F
    movf INDF, W
    movwf DATA1
    incf FSR, F
#endif
    call PIC16CmdBase

    ; Externally timed write

    ; At this point we need to emit a NOP command and hold the
    ; CLK line high for the duration of the write.
    clrf ICSP_CMD
    clrf DATA0
    clrf DATA1
    call PIC16WriteBase

    movlw PIC16_PBLWRT_POSTINC
    movwf ICSP_CMD
    clrf DATA0
    clrf DATA1
    call PIC16CmdBase

    ; Read the entire block and send it back.  It is the
    ; host software that is responsible for performing
    ; verification.  At the start of this routine, the
    ; call to PIC16SetTablePtr saved the address values.
    ; Restore the address and then invoke the block read
    ; routine
    movf PCADDR2, W
    movwf ADDRU
    movf PCADDR1, W
    movwf ADDRH
    movf PCADDR0, W
    movwf ADDRL
    ; Invoke the block read routine.  No need for a 'return"
    ; from this routine - the block read will take care of it.
    movlw 0x40
    movwf CMDOPT
    goto PIC16RBlock

PIC16EBuf
    call PIC16SetTablePtr

    ; Direct access to code memory
    coreCmd1 0x8ea6
    coreCmd1 0x9ca6

    ; Enable memory writes and setup erase
    coreCmd1 0x84a6
    coreCmd1 0x88a6

    ; Perform Flash unlock sequency
    coreCmd1 0x0e55
    coreCmd1 0x6ea7
    coreCmd1 0x0eaa
    coreCmd1 0x6ea7

    ; Initiate erase
    coreCmd1 0x82a6
    coreCmd1 0x0000

    ; Wait time P11 (5 msec) +P10 (5 usec) and then disable writes.  The
    ; loop below takes 9 instructions per cycle.  For 5 msec, there are
    ; 60,000 instructions (at 48 MHz), giving 0x1a0a loops.
    ; 
    clrf CTRX
    movlw 0x1b
    movwf CTRX+1
PIC16EBufLoop
    DEC16 CTRX
    btfss STATUS, Z
    goto PIC16EBufLoop

    coreCmd1 0x94a6

    retlw 0x04

;
; Write a block of 64 bytes of code to a 16 bit PIC.  The value
; of CMDOPT determines how many of the 64 will be written in
; each pass of the loop.  Many devices have 8 byte buffers,
; some have 64 byte write buffers.  The value passed must
; be appropriate for the device being programmed.
; 
;
; The target must already have been configured for single
; panel writes:
;    // Direct access to config memory
;   0x8EA6 - bsf EECON1, EEPGD
;   0x8CA6 - bcf EECON1, CFGS
;   // Configure device for single panel writes
;   Set table pointer to: 0x3c0006
;   // Write 0x00 to 0x3C0006 to enable single-panel writes
;   // Direct access to code memory
;   0x8EA6 - bsf EECON1, EEPGD
;   0x9CA6 - bcf EECON1, CFGS

;
; In:   <cmd><DLEN><ADDRL><ADDRH><ADDRU><CMDOPT>
; OUT:  <cmd><DLEN><ADDRL><ADDRH><ADDRU><DATA>...]
;
PIC16WBlock
    ; Set the table pointer to ADDRU:ADDRH:ADDRL.  This routine
    ; conveniently saves the contents of ADDRU:ADDRH:ADDRL so
    ; they can be restored after we use the overlaid locations
    ; for ICSP_CMD:DATA0:DATA1
    call packetInit
    call PIC16SetTablePtr

PIC16WB1
    ; Adjust
    bcf STATUS, C    ; Don't rotate in any leftovers
    rrcf CMDOPT, F  ; Divide by 2 for # of words

    ; Total of 32 words (64 bytes) to write
    movlw 0x20
    movwf RETRYCTR

PIC16WBLoop
    ; Move data into the write buffer
    movf CMDOPT, W
    movwf TEMP
    decf TEMP, F ; Move all but the last one into place

PIC16WBFillLoop
    ; Write the current word and increment the table pointer
    movlw PIC16_PBLWRT_POSTINC
    movwf ICSP_CMD
#if procSize == 16
    movff POSTINC0, DATA0
    movff POSTINC0, DATA1
#else
    movf INDF, W
    movwf DATA0
    incf FSR, F
    movf INDF, W
    movwf DATA1
    incf FSR, F
#endif
    call PIC16CmdBase

    decfsz TEMP, F
    goto PIC16WBFillLoop

PIC16WBStartProgramming
    ; Move the final byte into place, but don't increment
    ; the table pointer out of the current buffer
    movlw PIC16_STARTPROG
    movwf ICSP_CMD
#if procSize == 16
    movff POSTINC0, DATA0
    movff POSTINC0, DATA1
#else
    movf INDF, W
    movwf DATA0
    incf FSR, F
    movf INDF, W
    movwf DATA1
    incf FSR, F
#endif
    call PIC16CmdBase

    ; At this point we need to emit a NOP command and hold the
    ; CLK line high for the duration of the write.
    clrf ICSP_CMD
    call PIC16WriteBase

    ; Increment the address pointer to the next buffer
    movlw PIC16_PBLWRT_POSTINC
    movwf ICSP_CMD
    clrf DATA0
    clrf DATA1
    call PIC16CmdBase

    ; Subtract the number written in this pass from
    ; the total.  If any left, then continue writing.
    movf CMDOPT, W
    subwf RETRYCTR, F
    btfss STATUS, Z   ; All done with writes
    goto PIC16WBLoop  ; More buffers to fill

    ; Read the entire block and send it back.  It is the
    ; host software that is responsible for performing
    ; verification.  At the start of this routine, the
    ; call to PIC16SetTablePtr saved the address values.
    ; Restore the address and then invoke the block read
    ; routine
    movf PCADDR2, W
    movwf ADDRU
    movf PCADDR1, W
    movwf ADDRH
    movf PCADDR0, W
    movwf ADDRL

    ; Invoke the block read routine.  No need for a 'return"
    ; from this routine - the block read will take care of it.
    movlw 0x40
    movwf CMDOPT
    goto PIC16RBlock

;
; Read a block of data from a 16 bit PIC data memory (EEPROM
;
; In:    <cmd><DLEN><ADDRL><ADDRH><N/A><CMDOPT>
; Out:   <cmd><DLEN><ADDRL>N/A><N/A><cmdopt><data>
;
; DLEN on input tells us how many bytes of information are being
; requested.  The maximum is 64 bytes to prevent overflow.
; ADDRH:ADDRU are not used - the maximum data size for the
; PICs currently supported is 256.
;
; Affects: TEMP
;
PIC16REBlock
    call packetInit

PIC16REB1
    movf CMDOPT, W          ; Move requested # of bytes into TEMP
    movwf TEMP              ; to use for loop counter

    coreCmd1 0x9ea6         ; bsf EECON1, EEPGD
    coreCmd1 0x9ca6         ; bsf EECON1, CFGS

PIC16REBLoop
    ; Set address of the byte we want to read
    coreCmd2 0x0e, PCADDR0
    coreCmd1 0x6ea9         ; movwf EEADR

    ; Set high byte of address
    coreCmd2 0x0e, PCADDR1
    coreCmd1 0x6eaa         ; movwf EEADRH

    ; Initiate a memory read
    coreCmd1 0x80a6         ; bsf EECON1, RD

    ; Load the data into the serial data holding register
    coreCmd1 0x50a8         ; movf EEDATA, W, 0
    coreCmd1 0x6ef5         ; movwf TABLAT

    coreCmd1 0x0000         ; Required NOP

    ; Shift out the data
    movlw PIC16_SHIFT_TBLAT
    movwf ICSP_CMD
    call PIC16ReadBase

    ; Move response to output buffer (LSB first)
#if procSize == 16
    movff  PICDATA0, POSTINC0
#else
    movf PICDATA0, W
    movwf INDF
    incf FSR, F
#endif

    ; See if there are any more words to read
    incf PCADDR0, F
    decfsz TEMP, F
    goto PIC16REBLoop

    movf CMDOPT, W          ; Output size is CMDOPT + cmd/dlen/address/opt
    addlw 0x6
    return                  ; Send back the data


; Write a single byte to EEPROM.
PIC16WEByte
    ; Direct access to data EEPROM
    coreCmd1 0x9ea6       ; bcf EECON1, EEPGD
    coreCmd1 0x9ca6       ; bcf EECON1, CFGS

    ; Set the data EEPROM address pointer
    coreCmd2 0x0e, PCADDR0
    coreCmd1 0x6ea9       ; movwf EEADR
    coreCmd2 0x0e, PCADDR1
    coreCmd1 0x6eaa       ; movwf EEADRH

    ; Load the data to be written
    coreCmd2 0x0e, CMDOPT ; movlw <Data>
    coreCmd1 0x6ea8       ; movwf EEDATA

	; Enable memory writes
    coreCmd1 0x84a6       ; bsf EECON, WREN

    ; Perform required sequence
    coreCmd1 0x0e55       ; movlw 0x55
    coreCmd1 0x6ea7       ; movwf EECON2
    coreCmd1 0x0eaa       ; movlw 0xaa
    coreCmd1 0x6ea7       ; movwf EECON2

    ; Initiate write
    coreCmd1 0x82a6       ; bsf EECON1, WR

    ; Poll WR bit, repeat until clear.  Datasheet says
    ; that max programming time for EEPROM locations is
    ; 5 msec.
    ;   Delay count = (48 / 4) * 5000 = 60000 (0xea60) instructions.
    ; Each pass of this loop invokes several hundred instructions (~800)
    ; in order to check the WR bit.  The entire loop will be invoked
    ; a number of times to ensure the write was successful.
#if CLK_SPD > 24
    ; 24-48 MHz clock
    movlw 0x50
#else
#if CLK_SPD > 16
    ; 16-24 MHz clock
    movlw 0x28
#else
#if CLK_SPD > 8
    ; 8-16 MHz clock
    movlw 0x14
#else
    ; 5-8 MHz clock
    movlw 0x08
#endif
#endif
#endif
    movwf CTRX
    clrf CTRX+1
PIC16WEPoll
    coreCmd1 0x50a6       ; movf EECON1, W, 0
    coreCmd1 0x6ef5       ; movwf TABLAT
    coreCmd1 0x0000       ; nop
    movlw PIC16_SHIFT_TBLAT
    movwf ICSP_CMD
    call PIC16ReadBase    ; Shift out data
    btfss PICDATA0, 1     ; Test WR bit
    goto PIC16WEDonePoll  ; WR bit clear - done with write
    ; Write not complete - keep polling
    decfsz CTRX, F
    goto PIC16WEPoll
PIC16WEDonePoll

    ; Hold PGC low for time P10 (5 - 100 usec)
    movlw 0xa0
    call delay

    ; Disable writes
    coreCmd1 0x94a6       ; bcf EECON1, WREN

    retlw 0x01

;
; Write a block of data to a 16 bit PIC data memory (EEPROM
;
; In:    <cmd><DLEN><ADDRL><ADDRH><N/A><CMDOPT>
; Out:   <cmd><DLEN><ADDRL>N/A><N/A><cmdopt><data>
;
PIC16WEBlock
    call packetInit

PIC16WEB1
    movf CMDOPT, W        ; Move requested # of bytes into TEMP
    movwf TEMP            ; to use for loop counter

    ; use z for data address so PCADDR is preserved
    movf PCADDR0, W
    movwf CMDTEMP
    movf PCADDR1, W
    movwf PROGTEMP

PIC16WEBLoop
    movf INDF0, W
    movwf CMDOPT
    call PIC16WEByte
    incf FSR0L, F         ; Move pointer to next input byte

    ; See if there are any more words to read
    incf PCADDR0, F
    btfsc STATUS, C
    incf PCADDR1, F
    decfsz TEMP, F
    goto PIC16WEBLoop

#if 1
    ; Restore data address
    movff CMDTEMP, PCADDR0
    movff PROGTEMP, PCADDR1

    ; Read the entire block and send it back.
    movff PCADDR2, ADDRU
    movff PCADDR1, ADDRH
    movff PCADDR0, ADDRL
    goto PIC16REBlock
#else
    goto sendAck
#endif

oneClkCycle macro
    bsf ICSP_CLK
    bcf ICSP_CLK
    endm

twoClkCycles macro
    oneClkCycle
    oneClkCycle
    endm

fourClkCycles macro
    twoClkCycles
    twoClkCycles
    endm

eightClkCycles macro
    fourClkCycles
    fourClkCycles
    endm

sixteenClkCycles macro
    eightClkCycles
    eightClkCycles
    endm

twentyFourClkCycles macro
    sixteenClkCycles
    eightClkCycles
    endm

; Interactions with a dsPIC require transmitting a lot of NOP
; commands.  This routine provides that one command in as fast
; a form as possible.
PIC24Nop
#if 1
    bcf ICSP_DAT  ; Data line stays low throughout
    ; Send 4 command bits (all 0's), followed by 24 instruction
    ; bits (all 0's).  Total of 28 bits
#if 1
    movlw 0x1c
    movwf ICTR
PIC24NopLoop
    bsf ICSP_CLK
    bcf ICSP_CLK
    decfsz ICTR, F
    goto PIC24NopLoop
#else
    ; Unrolled loop, about twice as fast
    fourClkCycles       ; Four bits for command
    twentyFourClkCycles ; Twenty four bits for instruction
#endif
#else
    coreCmd24a 0x000000 ;  // nop
#endif
    return

; Transmit command to the PIC - 4 bit command (in ICSP_CMD),
; 24 bit payload (in DATA2:DATA1:DATA0) with DATA2 the most
; significant byte and DATA0 the least significant.  The bits
; are clocked out from least significant to most (IAW programming
; specification).
; Affects: ICSP_CMD, DATA0, DATA1, DATA2, ICTR
PIC24CmdBase
    ; Send the 4 bit command to the target PIC
    call FourBitCmdSetup

    ; Send 24 bit data payload to the target
    movlw 0x18
    movwf ICTR
pic24CmdLoop1
    ; Serial clock period must be > P1 (50 nsec) - no problem, even
    ; using 48 MHz clock speed.
    bsf ICSP_CLK      ; raise clock
    bcf ICSP_DAT      ; default data low
    rrcf DATA2, F
    rrcf DATA1, F     ; rotate data and read low bit
    rrcf DATA0, F
    btfsc STATUS,C
    bsf ICSP_DAT      ; pull data high
    bcf ICSP_CLK      ; lower clock, PIC latches data
    decfsz ICTR, F    ; dec ctr
    goto pic24CmdLoop1; loop if not zero
    bcf ICSP_DAT      ; pull DAT low

    return

; Send a command and 24 bit payload to the target.
; In:  <cmd><data len><data upper byte><data high byte><data low byte>
; Out: <STX><STX>[<cmd><0x03><data2><data0><data1>]<CHKSUM><ETX>
PIC24Cmd
    ; Save off the input data
    movf DATABUF1, W
    movwf PICDATA0
    movf DATABUF2, W
    movwf PICDATA1
    movf DATABUF0, W
    movwf PICDATA2

    ; Move input data to DATAx locations to be shifted out
    movff PICDATA0, DATA0
    movff PICDATA1, DATA1
    movff PICDATA2, DATA2

    ; Shift out the values
    movlw PIC24_SIX
    movwf ICSP_CMD
    call PIC24CmdBase

    ; Send back input as an ack
    movlw 0x73
    movwf DATALEN
    movf PICDATA2, W
    movwf ICSP_CMD
    movf PICDATA0, W
    movwf DATA0
    movf PICDATA1, W
    movwf DATA1

    retlw 0x05        ; 5 - cmd, data len, ICSP cmd, Data0, Data1


; Transmit command to the PIC - 4 bit command, 8 empty clocks,
;  followed by 16 bits of data out.
; Affects; ICTR, ICSP_CMD, PICDATA0, PICDATA1
PIC24ReadBase
    call FourBitCmdSetup

    ; Toggle the clock 8 times and ignore anything that comes back
    movlw 0x08
    movwf ICTR
    bcf ICSP_DAT      ; Hold ICSP_DAT low for first 8 bits
pic24RdLoop1
    bsf ICSP_CLK         ; 0      raise clock
    nop                  ; 1
    nop                  ; 2
    bcf ICSP_CLK         ; 3      lower clock, PIC latches data
    decfsz ICTR, F       ; 4      dec ctr
    goto pic24RdLoop1    ; 5..6   loop if not zero

    ; Ensure P5 (20 nsec) between first 8 bits and 16 bit response

    ; Toggle 16 more times and read the result bytes
#if procSize == 14
    banksel 0x80      ; bank 1
#endif
    bsf ICSP_DAT_MASK ; Set DAT line as input
#if procSize == 14
    banksel 0x00      ; bank 0
#endif
    movlw 0x10
    movwf ICTR
    clrf PICDATA0
    clrf PICDATA1
pic24RdLoop2
    ; Serial clock period must be > P2 (100 nsec to 1 usec)
    bsf ICSP_CLK      ; Raise ICSP_CLK
    rrcf PICDATA1, F
    rrcf PICDATA0, F  ; rotate previous contents before reading bit
    bcf PICDATA1, 7   ; Start output bit as 0
    nop
    nop
    bcf ICSP_CLK      ; Lower ICSP_CLK, PIC shifts data out    
    btfsc ICSP_DAT    ; Check value of output from PIC
    bsf PICDATA1, 7   ; Set bit if PIC is sending a 1
    decfsz ICTR, F    ; 
    goto pic24RdLoop2 ; loop if not zero

#if procSize == 14
    banksel 0x80      ; bank 1
#endif
    bcf ICSP_DAT_MASK ; Set DAT back to output
#if procSize == 14
    banksel 0x00      ; bank 0
#endif
    bcf ICSP_DAT      ; pull DAT low

    return


; 24 bit PIC table read command has to already be in ICSP_CMD
; In:  <cmd><data len><ICSP cmd><N/A><N/A>
; Out: <cmd><DLEN><N/A><data val><N/A>
PIC24ReadCmd
    ; Read a word from the target
    coreCmd24b
    ;movlw PIC24_REGOUT
    ;movwf ICSP_CMD
    ;call PIC24ReadBase

    ; Move value read from PIC to output buffer
    movf PICDATA0, W
    movwf DATA0
    movf PICDATA1, W
    movwf DATA1
    movlw 0x37
    movwf DATALEN

    movlw 0x05
    return

pushPICDATA macro
#if procSize == 16
    movff PICDATA0, POSTINC0
    movff PICDATA1, POSTINC0
#else
    movf PICDATA0, W
    movwf INDF
    incf FSR, F
    movf PICDATA1, W
    movwf INDF
    incf FSR, F
#endif
    endm

; After the 24 bit PIC has been put into STDP mode, the following
; routine is used to pull a set of 4 words from code memory (at
; the current address on the chip (as defined by W6) and put them
; into the buffer pointed to by FSR0H:FSR0L.
PIC24Rd4Words

    ; Initialize the write pointer (W7) and store the next four
    ; locations of code memory to W0:W5
    coreCmd24a 0xeb0380 ;  // clr W7
    coreCmd24a 0xba1b96 ;  // TBLRDL [W6], [W7++]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop
    coreCmd24a 0xbadbb6 ;  // TBLRDH, B [W6++], [W7++]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop
    coreCmd24a 0xbadbd6 ;  // TBLRDH, B [++W6], [W7++]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop
    coreCmd24a 0xba1bb6 ;  // TBLRDL [W6++], [W7++]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop
    coreCmd24a 0xba1b96 ;  // TBLRDL [W6], [W7++]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop
    coreCmd24a 0xbadbb6 ;  // TBLRDH, B [W6++], [W7++]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop
    coreCmd24a 0xbadbd6 ;  // TBLRDH, B [++W6], [++W7]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop
    coreCmd24a 0xba0bb6 ;  // TBLRDL [W6++], [W7]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop

	; Output W0:W5 using the VISI register and REGOUT command
    coreCmd24a 0x883c20 ;  // mov W0, VISI
    coreCmd24a 0x000000 ;  // nop
    coreCmd24b          ; Read a word from the target
    coreCmd24a 0x000000 ;  // nop
    pushPICDATA

    coreCmd24a 0x883c21 ;  // mov W1, VISI
    call PIC24Nop       ;  // nop
    coreCmd24b          ; Read a word from the target
    call PIC24Nop       ;  // nop
    pushPICDATA

    coreCmd24a 0x883c22 ;  // mov W2, VISI
    call PIC24Nop       ;  // nop
    coreCmd24b          ; Read a word from the target
    call PIC24Nop       ;  // nop
    pushPICDATA

    coreCmd24a 0x883c23 ;  // mov W3, VISI
    call PIC24Nop       ;  // nop
    coreCmd24b          ; Read a word from the target
    call PIC24Nop       ;  // nop
    pushPICDATA

    coreCmd24a 0x883c24 ;  // mov W4, VISI
    call PIC24Nop       ;  // nop
    coreCmd24b          ; Read a word from the target
    call PIC24Nop       ;  // nop
    pushPICDATA

    coreCmd24a 0x883c25 ;  // mov W5, VISI
    call PIC24Nop       ;  // nop
    coreCmd24b          ; Read a word from the target
    call PIC24Nop       ;  // nop
    pushPICDATA

    ; Reset the device internal PC
    coreCmd24a 0x040100 ;  //goto 0x100
    call PIC24Nop       ;  // nop

    return

; In:  <cmd><data len><ADDRL><ADDRH><ADDRU><N/A>
; Out: <cmd><DLEN><N/A><N/A><N/A><N/A><4 words>
; Note: The 4 instruction words are stored in the packed format
; described in the programming spec for dsPICs (using 12 bytes).
PIC24RdBuf
    ; Save off the address
    movff ADDRL, PCADDR0
    movff ADDRH, PCADDR1
    movff ADDRU, PCADDR2

    ; Use indirect addressing to fill the data buffer
#if procSize == 16
    movlw HIGH(DATA_BUF)
    movwf FSR0H
#endif
    movlw LOW(DATA_BUF)
    movwf FSR0L

    ; Read four blocks of 4 instructions (48 bytes)
    call PIC24Rd4Words
    call PIC24Rd4Words
    call PIC24Rd4Words
    call PIC24Rd4Words

    movlw 0x30
    movwf DATALEN

    retlw 0x36

packCmd macro w
    ; mov <word>, Wx
    swapf INDF0, W
    andlw 0xf0
    iorlw (w & 0x0f)
    movwf DATA0
#if procSize == 16
    swapf POSTINC0, W
#else
    swapf INDF, W
    incf FSR, F
#endif
    andlw 0x0f
    movwf DATA1
    swapf INDF0, W
    andlw 0xf0
    iorwf DATA1, F
#if procSize == 16
    swapf POSTINC0, W
#else
    swapf INDF, W
    incf FSR, F
#endif
    andlw 0x0f
    iorlw 0x20
    movwf DATA2
    endm

; Execute setup for buffer write on a 24 bit PIC
PIC24WrBufSetup
    ; Save off the address
    movff ADDRL, PCADDR0
    movff ADDRH, PCADDR1
    movff ADDRU, PCADDR2

    ; Set NVMCON to program 32 instruction words
    coreCmd24a 0x24001a ;  // mov 0x4001, W10
    coreCmd24a 0x883b0a ;  // mov W10, NVMCON

    ; Initialize the write pointer (W7) for TBLWT instruction
    ; 0x200xx0 - mov <addr 23:16>, W0
    swapf PCADDR2, W
    andlw 0xf0
    movwf DATA0
    swapf PCADDR2, W
    andlw 0x0f
    movwf DATA1
    movlw 0x20
    movwf DATA2
    movlw PIC24_SIX
    movwf ICSP_CMD
    call PIC24CmdBase

    coreCmd24a 0x880190 ;  // mov W0, TBLPAG

    ; 0x2xxxx0 - mov <addr 15:0>, W7
    swapf PCADDR1, W
    andlw 0x0f
    iorlw 0x20
    movwf DATA2
    swapf PCADDR1, W
    andlw 0xf0
    movwf DATA1
    swapf PCADDR0, W
    andlw 0x0f
    iorwf DATA1, F
    swapf PCADDR0, W
    andlw 0xf0
    iorlw 0x07
    movwf DATA0
    movlw PIC24_SIX
    movwf ICSP_CMD
    call PIC24CmdBase

    movff PCADDR0, ADDRL
    movff PCADDR1, ADDRH
    movff PCADDR2, ADDRU

    movlw 0x05
    return

; Take the next 12 bytes from INDF0 and pack them into DATA0:DATA1:DATA2
; in the appropriate manner to send them to the PIC.
PIC24PackForWrite
    ; mov LSW, W0
    packCmd 0x00
    movlw PIC24_SIX
    movwf ICSP_CMD
    call PIC24CmdBase
    packCmd 0x01
    movlw PIC24_SIX
    movwf ICSP_CMD
    call PIC24CmdBase
    packCmd 0x02
    movlw PIC24_SIX
    movwf ICSP_CMD
    call PIC24CmdBase
    packCmd 0x03
    movlw PIC24_SIX
    movwf ICSP_CMD
    call PIC24CmdBase
    packCmd 0x04
    movlw PIC24_SIX
    movwf ICSP_CMD
    call PIC24CmdBase
    packCmd 0x05
    movlw PIC24_SIX
    movwf ICSP_CMD
    call PIC24CmdBase

    return

PICWrCmd
	; Initialize read pointer (W6) and load W0:W5 with the next 4 instruction
	; words to program
    call PIC24PackForWrite

	; Set the read pointer (W6) and load the write latches
    coreCmd24a 0xeb0300 ;  // clr W6
    call PIC24Nop       ;  // nop
    coreCmd24a 0xbb0bb6 ;  // TBLWRT [W6++], [W7]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop
    coreCmd24a 0xbbdbb6 ;  // TBLWTH, B [W6++], [W7++]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop
    coreCmd24a 0xbbebb6 ;  // TBLWTH, B [W6++], [++W7]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop
    coreCmd24a 0xbb1bb6 ;  // TBLWTL [W6++], [W7++]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop
    coreCmd24a 0xbb0bb6 ;  // TBLWTL [W6++], [W7]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop
    coreCmd24a 0xbbdbb6 ;  // TBLWTH, B [W6++], [W7++]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop
    coreCmd24a 0xbbebb6 ;  // TBLWTH, B [W6++], [++W7]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop
    coreCmd24a 0xbb1bb6 ;  // TBLWTL [W6++], [W7++]
    call PIC24Nop       ;  // nop
    call PIC24Nop       ;  // nop
    return

; In:  <cmd><data len><ADDRL><ADDRH><ADDRU><N/A><4 words>
; Out: <cmd><DLEN><N/A><N/A><N/A><N/A><4 words>
; Note: The 17 instruction words are stored in the packed format
; described in the programming spec for dsPICs (using 12 bytes for
; each group of 4 instructions).
PIC24WrBuf
    ; Save off the address
    movff ADDRL, PCADDR0
    movff ADDRH, PCADDR1
    movff ADDRU, PCADDR2

    ; Use indirect addressing to fill the data buffer
#if procSize == 16
    movlw HIGH(DATA_BUF)
    movwf FSR0H
#endif
    movlw LOW(DATA_BUF)
    movwf FSR0L

    call PICWrCmd
    call PICWrCmd
    call PICWrCmd
    call PICWrCmd

    retlw 0x12

PIC24WrBufFinish
	; Unlock the NVMCON for writing
    coreCmd24a 0x200558 ;  // mov 0x55, W8
    coreCmd24a 0x883b38 ;  // mov W8, NVMKEY
    coreCmd24a 0x200aa9 ;  // mov 0xaa, W9
    coreCmd24a 0x883b39 ;  // mov W9, NVMKEY

	; Initiate the write cycle
    coreCmd24a 0xa8e761 ;  // bset NVMCON, WR
    coreCmd24a 0x000000 ;  // nop
    coreCmd24a 0x000000 ;  // nop
    coreCmd24a 0x000000 ;  // nop
    coreCmd24a 0x000000 ;  // nop

    ; // Externally timed 2 msec
    call delay1msec
    call delay1msec

    coreCmd24a 0xa9e761 ;  // bclr NVMCON, WR
    coreCmd24a 0x000000 ;  // nop
    coreCmd24a 0x000000 ;  // nop
    coreCmd24a 0x000000 ;  // nop
    coreCmd24a 0x000000 ;  // nop

	; Reset the device internal PC
    coreCmd24a 0x040100 ;  // goto 0x100
    coreCmd24a 0x000000 ;  // nop

    retlw 0x05

;
; Send the data buffer back.
;
; <STX><STX>[<DATA>...]<CHKSUM><ETX>
;
; Note that the data length in W must account for everything
; in the packet from command byte to final byte in the
; packet.  This is not the same as the DATALEN value when we
; read a packet - that value excludes the command and itself.
;
sendAck
    retlw 0x01      ; Send acknowledge (only command byte goes back)

WritePacket
    movwf ICTR

    movlw STX       ; Send start condition
    call TxRS232
    movlw STX
    call TxRS232

    clrf CHECKSUM   ; Reset checksum

#if procSize == 16
    movlw HIGH(PACKET)
    movwf FSR0H
#endif
    movlw LOW(PACKET)    ; Setup pointer to buffer area
    movwf FSR0L

SendNext            ; Send DATA
    movf INDF0, W
    addwf CHECKSUM, F
    incf FSR0L, F
    call WrData
    decfsz ICTR, F
    goto SendNext

    comf CHECKSUM,W ; Send checksum
    addlw 0x01
    call WrData

    movlw ETX       ; Send stop condition
    call TxRS232

    return

; Write a byte to the serial port (escape with control
; character if necessary).
WrData
    movwf TEMP       ; Save the data

    xorlw STX         ; Check for a STX
    btfsc STATUS,Z
    goto WrDLE        ; Send DLE, then the STX

    movf TEMP, W
    xorlw ETX         ; Check for a ETX
    btfsc STATUS,Z
    goto WrDLE        ; Send DLE, then the ETX

    movf TEMP, W
    xorlw DLE         ; Check for a DLE
    btfss STATUS,Z    ; Send DLE, then the DLE
    goto WrNext       ; Not a problem byte - just send

WrDLE
    movlw DLE         ; Yes, send DLE first
    call TxRS232

WrNext
    movf TEMP, W      ; Then send STX

TxRS232
    banksel TXSTA
    btfss TXSTA, TRMT
    goto TxRS232
    banksel TXREG
    movwf TXREG
    return

;
; delay for 3n+4 instructions (including call/return)
;
delay
    movwf DLYTEMP     ; 1
delayLoop
    decfsz DLYTEMP, 1 ; 0
    goto delayLoop    ; 1..2
    return            ;

delay1msec
; Delay loop for 9*N instructions.  The duration of the delay
; will be determined by the clock speed.  4 instruction cycles
; per microsecond for an 48MHz clock.
    ; 12000 (0x2EE0) instructions per msec at 48 MHz.
    ; 9 instructions per loop, giving 1333 (0x535) loops.
#if CLK_SPD > 24
    ; 24-48 MHz clock
    movlw 0x35
    movwf CTRX
    movlw 0x05
    movwf CTRX+1
#else
#if CLK_SPD > 16
    ; 16-24 MHz clock
    movlw 0x9a
    movwf CTRX
    movlw 0x02
    movwf CTRX+1
#else
#if CLK_SPD > 8
    ; 8-16 MHz clock
    movlw 0xbc
    movwf CTRX
    movlw 0x01
    movwf CTRX+1
#else
    ; 5-8 MHz clock
    movlw 0xde
    movwf CTRX
    movlw 0x00
    movwf CTRX+1
#endif
#endif
#endif

delay1msecLoop
    DEC16 CTRX          ; 0..5
    btfss STATUS, Z     ; 6
    goto delay1msecLoop ; 7..8

    return

#if (procID == 737) || (procID == 913)

#if 0
RdRS232
    ; Check for Overrun Error
    btfsc RCSTA, OERR
    goto readCh1

    ; Overrun error, try clearing CREN get rid of it
    bcf RCSTA, CREN
    bsf RCSTA, CREN

readCh1
    ; Wait until there is an input character.
    ; TBD: timeout if nothing arrives
    clrf CTRX
waitChLoop
    btfsc PIR1, RCIF
    goto gotRS232Char
    incfsz CTRX, F
    goto waitChLoop

    ; Timeout waiting for character
    movlw 0x01
    return

gotRS232Char
    movf RCREG, W
    banksel trxChar
    movwf trxChar

    return
#endif

; Read the analog register
readAN0
	; Start the A/D conversion, then wait for it
    ; to complete (bit GO in ADCON0 will be cleared).
    banksel ADCON0
    bsf ADCON0, GO
readAN0WaitLoop
    btfsc ADCON0, GO
    goto readAN0WaitLoop
    movf ADRESH, W
    movwf analogVal
    return

; The 737 has banked program memory - anything above 0x800 requires a
; separate step before a goto or call instruction.  Everything but
; the initialization code for the 737 fits in the first page, so we
; start the init code right at 0x800
    org 0x800


; Call all the initialization routines
setup
    call setupClock
    call setupInterrupts
    call setupPins
    call setupPWM
    call setupVars
    call setupUSART
    return

; Setup of interrupts.
; Sample code below uses Timer0 to generate
; an interrupt every millisecond (at 8MHz).
setupInterrupts
    clrf INTCON
    banksel PIE1
    clrf PIE1
    clrf PIE2
#if 1
    banksel OPTION_REG
    movlw 0x04
    movwf OPTION_REG ; internal clock, prescaler = 1:32
    ;banksel TMR0
    ;clrf TMR0
    ;bsf INTCON, T0IE ; Turn on interrupts for timer 0
    bsf INTCON, GIE
    bsf INTCON, PEIE
#endif
    return

setupClock
#if _USE_XTAL_20
; No special code needed if we are using a crystal.
#else
; Set clock to 8MHz
    banksel OSCCON   ; bank 1
    movlw 0x72       ; Clock 8MHz, use speed in OSCCON<6:4>
    movwf OSCCON
    banksel 0x00     ; bank 0
    clrf OSCTUNE
#endif
    return

setupPins
; Set I/O directions
    ; Init port values (not strictly necessary, but will ensure
    ; that we have pulled Vdd and Vpp lines low.
    banksel PORTA ; Bank 0
    clrf PORTA
    clrf PORTB
    clrf PORTC

    ; Configure and turn on the Analog to Digital Converter.
#if (procID == 737)
    movlw 0x40
    movwf ADCON0 ; Fosc/32, A/D off, AN0 is selected channel
    ; Configure how analog values will be read
    banksel ADCON1
    movlw 0x0d
    movwf ADCON1 ; Left justified, no divide, VRef Vdd/Vss, only AN0/AN1 analog
    movlw 0x20
    movwf ADCON2 ; 8 Tad
#else
    clrf ADCON0 ; Left justified, Vref Vdd/Vss, A/D off, AN0 selected, A/D off
    ; Configure how analog values will be read
    banksel ADCON1
    movlw 0x20
    movwf ADCON1 ; Fosc/32
    movlw 0x03
    movwf ANSEL  ; only AN0/AN1 analog
#endif

    ; Turn off comparators
    banksel CMCON0        ; Bank 1
    movlw 0x07
    movwf CMCON0

    ; Set I/O directions for the pins we are using
    ; PORTA 
    bsf VBOOST_MASK       ; RA0 - A/D input is an input
    bsf EXT_PWR_MASK      ; RA1 - A/D input for external power
    bcf CTS_MASK          ; RA2 - CTS is an output
    bsf RTS_MASK          ; RA3 - RTS is an input
    bcf LED2_MASK         ; RA4 - RA4 is an output
    bcf LED1_MASK         ; RA5 - RA5 is an output
    ; RA6, RA7 N/A

    ; PORTB
    bsf SDA_MASK          ; RB0 - TBD: I2C data is an input?
    bcf SCL_MASK          ; RB1 - I2C clock is an output
    bcf VPP1_MASK         ; RB2 - 13V control is an output
    bsf ICSP_CLK_MASK     ; RB3 - CLK output when in use, otherwise tristate
    bsf ICSP_DAT_MASK     ; RB4 - DAT output when in use, otherwise tristate
    bcf VDD_MASK          ; RB5 - 5V control is an output
    bcf VPP2_MASK         ; RB6 - 13V control is an output
    bcf LED3_MASK         ; RB7 - LED3

    ; PORTC
    bcf PRG_SEL_MASK      ; RC0 - PRG_SEL is an output
    bsf USB_PWR_MASK      ; RC1 - USB power status
    bcf VPP_PWM_MASK      ; RC2 - CCP1 is an output
    bsf TRISC, 3          ; RC3 - N/A - pin used on 2455 for VUSB
    bsf TRISC, 4          ; RC4 - N/A - pin used on 2455 for USB D-
    bsf TRISC, 5          ; RC5 - N/A - pin used on 2455 for USB D+
    bcf SERIAL_TX_MASK    ; RC6 - Tx is an output
    bsf SERIAL_RX_MASK    ; RC7 - Rx is an input

    ; Configure Timer0 - want a couple of hundred interrupts
    ; per second to check the analog value.  With a prescale
    ; value of 1:128, counting to 256, and using the internal
    ; instruction cycle, Timer0 will overflow every 6.55 msec,
    ; giving 152.6 interrupts per second (with 20 MHz clock).
    movf OPTION_REG, W
    andlw 0xc0 ; Clear all timer 0 flags
    iorlw 0x07 ; use CLK0, 1:128 prescale
    movwf OPTION_REG

    banksel ADCON0   ; bank 0
    bsf ADCON0, ADON ; enable the A/D module

    ; Set default values for I/O pins
    bcf LED1
    bcf LED2
    bsf LED3

    bcf SERIAL_TX
    bcf SERIAL_RX
    bcf CTS        ; Start with CTS on - ready to get characters
    bcf RTS
    bcf VPP_PWM
    bcf VPP1       ; 
    bcf VPP2       ;
    bcf VDD        ; 
    bcf ICSP_CLK
    bcf ICSP_DAT
    bcf PRG_SEL

    return

; Reset variables used for I/O and programming
setupVars
    banksel 0x80
    movlw 0xff
    movwf PWMCtr
	clrf packetCtr
	clrf STXCount
    clrf SFlags

    banksel 0x00
    clrf ICTR
    clrf COMMAND
    clrf DATALEN
    clrf DATA0
    clrf DATA1

    return

; Setup PWM
setupPWM
    banksel TRISA    ; bank 1
    bcf VPP_PWM_MASK ; Ensure CCP pin is an output
    movlw 0x1f       ; Set PWM period.  0x1f -> 156.3 KHz @ 20 MHz.
    movwf PR2
    banksel CCP1CON  ; bank 0
    ; Set two LSB bits of duty cycle to
    ; "11", CCP1 into PWM mode.
    movlw 0x3c
    movwf CCP1CON
    ; Start with a low value for PWM to reduce problems with
    ; current saturating the inductor
    movlw startPWM
    movwf PWMVal
    setPeriod PWMVal
    bcf PIR1, TMR2IF ; Disable timer 2 interrupts till it's configured
    ; Enable Timer 2
    movlw 0x7c ; 1:1 prescale, 1:16 postscale, T2 on
    movwf T2CON
    ; Wait until one cycle has elapsed
tmrWait
    btfss PIR1, TMR2IF
    goto tmrWait
    return

setupUSART
    banksel TXSTA
    movlw 0x24
    movwf TXSTA ; Asynch, TSR empty, BRGH=1
#if _USE_XTAL_20
    ; Baud rate formula for BRG16=1, BRGH=1: Baud Rate = Fosc/(4 (n + 1)),
    ; or n = (Fosc / (4 * BaudRate)) - 1
    ; At 20 MHz, for 115.2K Baud:
    ;     SPBRGH:SPBRG = X = Fosc / (16 * 115200) - 1 = 10 (9.85)
	; Calculated Baud Rate: 20000000 / (16 * (X + 1)) = 113636
	; Error: (115200 - 116279) / 115200 = 1.4%
#if BAUD_RATE == 115
    movlw 0x0a ; At 20 MHz, SPBRGH=1, SPBRG=10 gives ~115.2K Baud, 20 gives 57.6 KB
#else
#if BAUD_RATE == 57
    movlw 0x14
#else
    movlw 0x1e
#endif
#endif
#else
    ; 115.2 KBaud does not work well using the 8 MHz internal
    ; oscillator.  The reason appears to be the significant
    ; error between the desired rate and what is possible with
    ; integer values for SPBRG.  To fix this we use OSCTUNE to
    ; modify the clock to get a more even division of the
    ; baud rate.
    ;
	; Calculated Baud Rate: 8000000 / (16 * (X + 1)) = 125000 (X = 3)
	; Error: (115200 - 125000) / 115200 = 8.5%
    ;
    ; For SPBRG = X = 3, ideal clock freq: 16 * 4 * 115200 = 7372800.
    ; Need to reduce clock speed by -7.8%.  OSCTUNE varies the clock
    ; up to 12.5% over the range from: -32 to 31 (16f737); or -16 to
    ; 15 (16f913)
    ; For 7x7:
    ;    (OSCTUNE / 31) = -7.8% / 12.5%
    ;    OSCTUNE = -31 * (7.8 / 12.5) = -19
    ; For 91x:
    ;    (OSCTUNE / 15) = -7.8% / 12.5%
    ;    OSCTUNE = -15 * (7.8 / 12.5) = -9
#if (procID == 737)
    movlw 0xeb
#else
    movlw 0xf6
#endif
    movwf OSCTUNE
#if BAUD_RATE == 115
    movlw 0x03
#else
#if BAUD_RATE == 57
    ; 57.6, 38.4 KBaud seem reliable at 8 MHz
    movlw 0x06 ; At 8 MHz, SPBRGH=1, SPBRG=8 gives ~57.6K Baud
#else
    movlw 0x09 ; At 8 MHz, SPBRGH=1, SPBRG=12 gives ~38.4K Baud
#endif
#endif
#endif
    movwf SPBRG
    banksel RCSTA
    movlw 0x90
    movwf RCSTA ; Enable serial port, enable receiver

	; interrupts for Rx of bytes
    bcf PIR1, RCIF
    banksel PIE1     ; bank 1
    bsf PIE1, RCIE

	; Disable interrupts for Tx
    bcf PIE1, TXIE

    banksel 0x00     ; bank 0

    return

; Change duty cycle of PWM based on analog reading of feedback voltage
PWMAdjust
#if 1
    ; See if the A/D val is larger or smaller than desired
    movlw fixVal
    subwf analogVal, w
    btfsc STATUS, Z
    goto donePWMAdjust ; analogVal matches max, don't increase duty cycle
    btfss STATUS, C
    goto incPWM    ; analogVal too low, increase duty cycle

decPWM
    ; Don't bother lowering the PWM if we are below the maximum
    ; acceptable voltage
    movlw maxAnalog
    subwf analogVal, w
    btfss STATUS, C
    goto donePWMAdjust ; analogVal < maxAnalog - don't change PWM

    ; the analog reading is larger than the value we want.  Reduce
    ; the duty cycle by 1
    movf PWMVal, W
    btfsc STATUS, Z
    goto donePWMAdjust ; duty cycle already zero
    decf PWMVal, f
    setPeriod PWMVal
    goto donePWMAdjust

incPWM
    ; Don't bother raising the PWM if we are above the minimum
    ; acceptable voltage
    movlw minAnalog
    subwf analogVal, w
    btfsc STATUS, C
    goto donePWMAdjust ; analogVal > minAnalog - don't change PWM

    ; Check to see if the PWM value is already at the max
    movlw maxPWM
    subwf PWMVal, w
    btfsc STATUS, Z
    goto donePWMAdjust  ; already exactly at the max
    btfsc STATUS, C
    goto resetPWMAdjust ; duty cycle larger than allowed, reset it
    incf PWMVal, f
    setPeriod PWMVal
    goto donePWMAdjust

resetPWMAdjust
    movlw maxPWM
    movwf PWMVal
    setPeriod PWMVal    
#else
    ; Test - don't change PWM from starting value.
#endif

    ; Done fiddling with the PWM duty cycle
donePWMAdjust
    return;

toggleLED3
    ; Toggle LED3 to show we got a packet
    btfss LED3
    goto setLED3
    bcf LED3
    goto doneSetLED3
setLED3
    bsf LED3
doneSetLED3
    return

toggleLED2
    btfss LED2
    goto setLED2
    bcf LED2
    goto doneSetLED2
setLED2
    bsf LED2
doneSetLED2
    return

toggleLED1
    ; Toggle LED1 to show we got a byte
    btfss LED1
    goto setLED1
    bcf LED1
    goto doneSetLED1
setLED1
    bsf LED1
doneSetLED1
    return

Tx1RS232
    banksel TXSTA
    btfss TXSTA, TRMT
    goto Tx1RS232
    banksel TXREG
    movwf TXREG
    return

; Add the latest byte from the serial port to the packet
; that is under construction.  Must be in RAM bank 1 when
; called.
addByteToPacket
	; See if this is the start of a packet.  We need to
	; see two STX bytes in a row.
 	movlw 0x02
    subwf STXCount, W
    btfsc STATUS, Z
    goto checkDLEFlag ; Already have 2 STX's

#if _DEBUG_CMDS
 ; Echo the character
 movfw trxChar
 call Tx1RS232
#endif

	movlw STX
    subwf trxChar, W
    btfss STATUS, Z
    goto resetPacket ; Didn't get an STX
	incf STXCount, F ; Got one - increment the count
#if _DEBUG_CMDS
 movlw '1'
 call Tx1RS232
 movfw STXCount
 call Tx1RS232
#endif
    retlw 0x00

checkDLEFlag:
	btfss DLEFlag
	goto checkDLE
	; Already got a DLE.  Clear the flag
	; and add this byte to the packet.
    bcf DLEFlag
	goto addByte   ; Process escaped byte

checkDLE:
    movf trxChar, W
    sublw DLE
    btfss STATUS, Z
	goto checkETX          ; Not a DLE
	bsf DLEFlag
#if _DEBUG_CMDS
 movlw '2'
 call Tx1RS232
#endif
    retlw 0x00

checkETX:
    movf trxChar, W
    sublw ETX
    btfss STATUS, Z
	goto checkSTX          ; Not an ETX

	; Packet is complete - can process it now.  Make sure all counters
	; and flags are clear for the next packet.
	clrf packetCtr
	clrf STXCount
	bcf DLEFlag
    bsf PKTReadyFlag
#if _DEBUG_CMDS
 movlw '3'
 call Tx1RS232
#endif

    retlw 0x01

checkSTX:
    movf trxChar, W
    sublw STX
    btfsc STATUS, Z
	goto resetPacket        ; Unexpected STX in middle of packet

; This is a byte we need to add to the packet
addByte:
    movf trxChar, W
    movwf CHECKSUM          ; If next byte is ETX, then this is CHECKSUM
    ; Check length of packet
    movf packetCtr, W
    xorlw MAX_BUFLEN
    btfss STATUS, Z
    goto addByte1

; We have hit the limit of the storage buffer.  Any more
; and we start to overwrite RAM that isn't allocated for
; the input buffer.  Ideally we only get here if the packet
; is full size (64 bytes of payload) and this byte is the
; checksum.  The next byte should be an ETX (end of packet).
serialOverrun
#if _DEBUG_CMDS
 movlw '4'
 call Tx1RS232
#endif
    ; goto resetPacket    ; packetCtr == MAX_BUFLEN - packet size exceeds storage length
    retlw 0x00

addByte1
	; Add data to packet
    movf FSR, W
    banksel TMPFSR1
    movwf TMPFSR1       ; Save old FSR value
    banksel 0x00
	movlw LOW(PACKET)
	addwf packetCtr, W  ; Point to byte at PACKET[packetCtr]
	movwf FSR
	movf trxChar, W
    movwf INDF          ; Put the byte in the packet
    banksel TMPFSR1
    movf TMPFSR1, W
    banksel 0x00
    movwf FSR           ; Restore old FSR value

	; Increment pointer into packet
	incf packetCtr, F

#if _DEBUG_CMDS
 movlw '5'
 call Tx1RS232
 movf FSR, W
 call Tx1RS232
#endif

	; Not done with packet yet
    retlw 0x00

; An error occured in the packet.  Clear all counters and
; start processing as if we hadn't seen any data to this point
resetPacket:
#if _DEBUG_CMDS
 movlw '6'
 call Tx1RS232
#endif
	clrf packetCtr
	clrf STXCount
	bcf DLEFlag
    retlw 0x00

; Interrupt service routine
;
; Note: stack depth on 14 bit PICs is very limited (8 words)
intServ
    ; Don't know what RAM bank we started in, so will need to
    ; save STATUS in a place we can find it later.  Use bank 0
    ; value to be sure.
#if _DEBUG_CMDS
 movlw 'i'
 call Tx1RS232
#if 0
 movf STATUS, W
 call Tx1RS232
 movf TMPPCLATH, W
 call Tx1RS232
 movf PCLATH, W
 call Tx1RS232
 movlw ' '
 call Tx1RS232
#endif
#endif
    movf TMPSTATUS, W
    movwf TMPSTATUS
    clrf STATUS         ; bank 0

checkTimer0
    ; See if Timer0 overflowed
    btfss INTCON, T0IF 
    goto checkRX
    bcf INTCON, T0IF

checkRX
    ; See if a character has arrived
    btfss PIR1, RCIF
    goto checkADIF

    ; Add new byte to the current packet
    movf RCREG, W
    movwf trxChar
#if _DEBUG_CMDS
 movf trxChar, W
 call Tx1RS232
#endif
    btfss PKTReadyFlag
    goto tryAddByte
    ; Programmer is busy right now, discard this character
    goto doneCheckRX

tryAddByte
    bcf PKTReadyFlag
    call addByteToPacket
doneAddByte
    xorlw 0x01
    btfsc STATUS, Z
    ; Packet is complete
    bsf PKTReadyFlag
doneCheckRX
    goto checkRX ; Keep getting bytes while they are available

checkADIF
    ; See if an A/D conversion completed
    btfsc PIR1, ADIF
    goto checkTMR1IF
    bcf PIR1, ADIF

checkTMR1IF
    btfsc PIR1, TMR1IF
    goto intDone
    bcf PIR1, TMR1IF

intDone
    ; Restore status to put us in the right RAM bank
    clrf STATUS
    movf TMPSTATUS, W
    movwf STATUS

    ; Restore PCLATH to put us in the right code bank
    movf TMPPCLATH, W
    movwf PCLATH

    ; Re-restore STATUS in case PCLATH changed it
    movf TMPSTATUS, W
    movwf STATUS
    ; Finally restore W without changing STATUS
    swapf TMPW, F
    swapf TMPW, W

    retfie

#endif ; End of 16F737 specific routines


;
; Debugging/status oriented routines
;
setLineValues
#if procSize == 14
    banksel TRISA
#endif
    bcf ICSP_DAT_MASK
    bcf ICSP_CLK_MASK
    bcf PRG_SEL_MASK
#if procSize == 14
    banksel DATABUF0
#endif
    btfsc DATABUF0, 0
    goto setLineVddHigh
    bcf VDD
    goto setLineVpp1
setLineVddHigh
    bsf VDD

setLineVpp1
    btfsc DATABUF0, 1
    goto setLineVpp1High
    bcf VPP1      ; Pull Vpp1 low
    goto setLineVpp2
setLineVpp1High
    bsf VPP1      ; Pull Vpp1 high

setLineVpp2
    btfsc DATABUF0, 2
    goto setLineVpp2High
    bcf VPP2
    goto setLineCLK
setLineVpp2High
    bsf VPP2

setLineCLK
    btfsc DATABUF0, 3
    goto setLineCLKHigh
    bcf ICSP_CLK
    goto setLineDAT
setLineCLKHigh
    bsf ICSP_CLK

setLineDAT
    btfsc DATABUF0, 4
    goto setLineDATHigh
    bcf ICSP_DAT
    goto setLinePrgSel
setLineDATHigh
    bsf ICSP_DAT

setLinePrgSel
    btfsc DATABUF0, 5
    goto setLinePrgSelHigh
    bcf PRG_SEL
    goto setLineLEDs
setLinePrgSelHigh
    bsf PRG_SEL

; TBD: Should the LEDs be set the same way as the
; control lines?  The code currently turns them
; off, then turns back on if the bit is set.
setLineLEDs
    bcf LED1
    btfsc DATABUF1, 0
    bsf LED1
    bcf LED2
    btfsc DATABUF1, 1
    bsf LED2
    bcf LED3
    btfsc DATABUF1, 2
    bsf LED3

    return

readLineValues
    movlw 0x02
    movwf DATALEN
    clrf DATA0
    btfss VDD
    bsf DATA0, 0
    btfss VPP1
    bsf DATA0, 1
    btfss VPP2
    bsf DATA0, 2
    btfsc ICSP_CLK
    bsf DATA0, 3
    btfss ICSP_DAT
    bsf DATA0, 4
    btfsc PRG_SEL
    bsf DATA0, 5
    clrf DATA1
    btfsc LED1
    bsf DATA1, 0
    btfsc LED2
    bsf DATA1, 1
    btfsc LED3
    bsf DATA1, 2
    return

readRegValues
    clrf DATABUF0
    clrf DATABUF1
    clrf DATABUF2
    clrf CMDOPT

    ; Use indirect addressing to fill the data buffer
#if procSize == 16
    movlw HIGH(DATA_BUF)
    movwf FSR0H
    movlw LOW(DATA_BUF)
    movwf FSR0L

    movlw 69
    movwf DATALEN

    movff INTCON, POSTINC0   ; 0
    movff INTCON2, POSTINC0
    movff INTCON3, POSTINC0
    movff T0CON, POSTINC0
    movff OSCCON, POSTINC0
    movff HLVDCON, POSTINC0
    movff WDTCON, POSTINC0
    movff RCON, POSTINC0
    movff T1CON, POSTINC0
    movff PR2, POSTINC0
    movff T2CON, POSTINC0   ; 10
    movff ADRESH, POSTINC0
    movff ADRESL, POSTINC0
    movff ADCON0, POSTINC0
    movff ADCON1, POSTINC0
    movff ADCON2, POSTINC0
    movff CCPR1H, POSTINC0
    movff CCPR1L, POSTINC0
    movff CCP1CON, POSTINC0
    movff CCPR2H, POSTINC0
    movff CCPR2L, POSTINC0  ; 20
    movff CCP2CON, POSTINC0
    movff BAUDCON, POSTINC0
    movff CVRCON, POSTINC0
    movff CMCON, POSTINC0
    movff T3CON, POSTINC0
    movff SPBRGH, POSTINC0
    movff SPBRG, POSTINC0
    movff TXSTA, POSTINC0
    movff RCSTA, POSTINC0
    movff IPR2, POSTINC0    ; 30
    movff PIR2, POSTINC0
    movff PIE2, POSTINC0
    movff IPR1, POSTINC0
    movff PIR1, POSTINC0
    movff PIE1, POSTINC0
    movff OSCTUNE, POSTINC0
#ifdef TRISE
    movff TRISE, POSTINC0
#else
    clrf POSTINC0
#endif
#ifdef TRISD
    movff TRISD, POSTINC0
#else
    clrf POSTINC0
#endif
    movff TRISC, POSTINC0
    movff TRISB, POSTINC0  ; 40
    movff TRISA, POSTINC0
    movff PORTE, POSTINC0
#ifdef PORTD
    movff PORTD, POSTINC0
#else
    clrf POSTINC0
#endif
    movff PORTC, POSTINC0
    movff PORTB, POSTINC0
    movff PORTA, POSTINC0
    movff UEP3, POSTINC0
    movff UEP2, POSTINC0
    movff UEP1, POSTINC0
    movff UEP0, POSTINC0  ; 50
    movff UCFG, POSTINC0
    movff UADDR, POSTINC0
    movff UCON, POSTINC0
    movff USTAT, POSTINC0
    movff UEIE, POSTINC0
    movff UEIR, POSTINC0
    movff UIE, POSTINC0
    movff UIR, POSTINC0
    movff UFRMH, POSTINC0
    movff UFRML, POSTINC0 ; 60
    movff 0x0400, POSTINC0 ; BD0STAT
    movff 0x0401, POSTINC0
    movff 0x0402, POSTINC0
    movff 0x0403, POSTINC0
    movff 0x0404, POSTINC0 ; BD1STAT
    movff 0x0405, POSTINC0
    movff 0x0406, POSTINC0
    movff 0x0407, POSTINC0 ; 68
#else
;
; Read registers on a 14 bit PIC.  TBD: ensure all
; the appropriate banksel statements have been
; inserted into the code.
;
    movlw DATA_BUF
    movwf FSR

    movlw 69
    movwf DATALEN

    movff INTCON, INDF   ; 0
    incf FSR, F
#ifdef INTCON2
    movff INTCON2, INDF
#else
    clrf INDF
#endif
    incf FSR, F
#ifdef INTCON3
    movff INTCON3, INDF
#else
    clrf INDF
#endif
    incf FSR, F
#ifdef T0CON
    movff T0CON, INDF
#else
    clrf INDF
#endif
    incf FSR, F
    banksel OSCCON
    movff OSCCON, INDF
    incf FSR, F
#ifdef HLVDCON
    movff HLVDCON, INDF
#else
    clrf INDF
#endif
    incf FSR, F
    banksel WDTCON
    movff WDTCON, INDF
    incf FSR, F
#ifdef RCON
    movff RCON, INDF
#else
    clrf INDF
#endif
    incf FSR, F
    banksel T1CON
    movff T1CON, INDF
    incf FSR, F
    banksel PR2
    movff PR2, INDF
    incf FSR, F
    banksel T2CON
    movff T2CON, INDF   ; 10
    incf FSR, F
    banksel ADRESH
    movff ADRESH, INDF
    incf FSR, F
    banksel ADRESL
    movff ADRESL, INDF
    incf FSR, F
    banksel ADCON0
    movff ADCON0, INDF
    incf FSR, F
    banksel ADCON1
    movff ADCON1, INDF
    incf FSR, F
#ifdef ADCON2
    movff ADCON2, INDF
#else
    clrf INDF
#endif
    incf FSR, F
    banksel CCPR1H
    movff CCPR1H, INDF
    incf FSR, F
    movff CCPR1L, INDF
    incf FSR, F
    movff CCP1CON, INDF
    incf FSR, F
#ifdef CCPR2H
    movff CCPR2H, INDF
#else
    clrf INDF
#endif
    incf FSR, F
#ifdef CCPR2L
    movff CCPR2L, INDF  ; 20
#else
    clrf INDF
#endif
    incf FSR, F
#ifdef CCP2CON
    movff CCP2CON, INDF
#else
    clrf INDF
#endif
    incf FSR, F
#ifdef BAUDCON
    movff BAUDCON, INDF
#else
    clrf INDF
#endif
    incf FSR, F
#ifdef CVRCON
    banksel CVRCON
    movff CVRCON, INDF
#else
    clrf INDF
#endif
    incf FSR, F
#ifdef CMCON
    movff CMCON, INDF
#else
    clrf INDF
#endif
    incf FSR, F
#ifdef T3CON
    banksel T3CON
    movff T3CON, INDF
#else
    clrf INDF
#endif
    incf FSR, F
#ifdef SPBRGH
    banksel SPBRGH
    movff SPBRGH, INDF
#else
    clrf INDF
#endif
    incf FSR, F
    movff SPBRG, INDF
    incf FSR, F
    movff TXSTA, INDF
    incf FSR, F
    banksel RCSTA
    movff RCSTA, INDF
    incf FSR, F
#ifdef IPR2
    movff IPR2, INDF    ; 30
#else
    clrf INDF
#endif
    incf FSR, F
    movff PIR2, INDF
    incf FSR, F
    movff PIE2, INDF
    incf FSR, F
#ifdef IPR1
    movff IPR1,INDF
#else
    clrf INDF
#endif
    incf FSR, F
    movff PIR1, INDF
    incf FSR, F
    banksel PIE1
    movff PIE1, INDF
    incf FSR, F
    movff OSCTUNE, INDF
    incf FSR, F
#ifdef TRISE
    movff TRISE, INDF
#else
    clrf INDF
#endif
    incf FSR, F
#ifdef TRISD
    movff TRISD, INDF
#else
    clrf INDF
#endif
    incf FSR, F
    movff TRISC, INDF
    incf FSR, F
    movff TRISB, INDF ; 40
    incf FSR, F
    movff TRISA, INDF
    incf FSR, F
    banksel PORTE
    movff PORTE, INDF
    incf FSR, F
#ifdef PORTD
    movff PORTD, INDF
#else
    clrf INDF
#endif
    incf FSR, F
    movff PORTC, INDF
    incf FSR, F
    movff PORTB, INDF
    incf FSR, F
    movff PORTA, INDF ; 46
    incf FSR, F
#endif

    movf DATALEN, W
    addlw 0x06
    return

    end
